@@ -28,7 +28,8 @@ let dumb_astcocci_rule_elem = function
  | A.Ty ty -> ()
  | A.TopId id -> ()
  | A.TopInit init -> ()
- | A.FunHeader (bef,allminus, fninfo, ida, _, paramsa, _, _) -> ()
+ | A.FunHeader (bef,allminus, fninfo, ida, _, paramsa, _, _, _) -> ()
+ | A.TemplateDefinitionHeader (tmpkw,lab,params,rab) -> ()
  | A.Decl decl -> ()
  | A.SeqStart _ -> ()
  | A.SeqEnd _ -> ()
@@ -38,7 +39,7 @@ let dumb_astcocci_rule_elem = function
  | A.WhileHeader (_, _, ea, _) -> ()
  | A.DoHeader _ -> ()
  | A.WhileTail (_,_,ea,_,_) -> ()
- | A.ForHeader (_, _, _, ea2opt, _, ea3opt, _) -> ()
+ | A.ForHeader (_, _, _, _) -> ()
  | A.IteratorHeader (ia1, ia2, ea, ia3) -> ()
  | A.SwitchHeader _ -> ()
  | A.Break _ -> ()
@@ -50,21 +51,19 @@ let dumb_astcocci_rule_elem = function
  | A.Exec(_,_,_,_) -> ()
  | A.DefineHeader _ -> ()
  | A.Undef _ -> ()
- | A.Pragma _ -> ()
- | A.Include _ -> ()
- | A.MetaInclude _ -> ()
+ | A.CppTop _ -> ()
  | A.Default _ -> ()
  | A.Case _ -> ()
  | A.AsRe _ -> ()
  | A.DisjRuleElem _ -> failwith "not possible - compiled away in asttoctl"
 
 let dumb_astcocci_decl = function
-   A.UnInit (stg, typa, sa, attr, _)     -> ()
- | A.Init (stg, typa, sa, attr, _, expa, _) -> ()
+   A.UnInit (al, stg, typa, sa, endattr, _)     -> ()
+ | A.Init (al, stg, typa, sa, endattr, _, expa, _) -> ()
  | A.FunProto _ -> ()
- | A.TyDecl (typa, _, _)     -> ()
- | A.MacroDecl(stg, fn, _, eas, _, _, _) -> ()
- | A.MacroDeclInit(stg, fn, _, eas, _, _, _, _) -> ()
+ | A.TyDecl (typa, _)     -> ()
+ | A.MacroDecl(stg, _, fn, _, eas, _, _, _) -> ()
+ | A.MacroDeclInit(stg, _, fn, _, eas, _, _, _, _, _) -> ()
  | A.MetaDecl _ -> ()
  | A.AsDecl _ -> ()
  | A.Typedef(d,ty1,ty2,pv) -> ()
@@ -73,12 +72,12 @@ let dumb_astcocci_decl = function
  | A.OptDecl _ -> ()
 
 let dumb_astcocci_initialiser = function (* seems same as the above *)
-    A.Init(stg,ty,id,attr,eq,ini,sem) -> ()
-  | A.UnInit(stg,ty,id,attr,sem) -> ()
+    A.Init(al,stg,ty,id,endattr,eq,ini,sem) -> ()
+  | A.UnInit(al,stg,ty,id,endattr,sem) -> ()
   | A.FunProto _ -> ()
-  | A.MacroDecl(_, fn, _, eas, _, _, _) -> ()
-  | A.MacroDeclInit(_, fn, _, eas, _, _, _, _) -> ()
-  | A.TyDecl(ty,attr,sem) -> ()
+  | A.MacroDecl(_, _, fn, _, eas, _, _, _) -> ()
+  | A.MacroDeclInit(_, _, fn, _, eas, _, _, _, _, _) -> ()
+  | A.TyDecl(ty,sem) -> ()
   | A.Typedef(d,ty1,ty2,pv) -> ()
   | A.DisjDecl(decls) -> ()
   | A.ConjDecl(decls) -> ()
@@ -87,7 +86,9 @@ let dumb_astcocci_initialiser = function (* seems same as the above *)
   | A.OptDecl(decl) -> ()
 
 let dumb_astcocci_field = function
-    A.Field (typa, sa, _bf, _)     -> ()
+    A.Field (typa, sa, _bf, _endattr, _)     -> ()
+  | A.MacroDeclField(fn, _, eas, _, _, _) -> ()
+  | A.CppField _ -> ()
   | A.MetaField _ -> ()
   | A.MetaFieldList _ -> ()
 
@@ -123,9 +124,14 @@ let dumb_astcocci_expr = function
  | A.ArrayAccess (ea1, _, ea2, _) -> ()
  | A.RecordAccess (ea, _, ida) -> ()
  | A.RecordPtAccess (ea, _, ida) -> ()
- | A.Cast (_, typa, attra, _, ea) -> ()
+ | A.QualifiedAccess (typa,_,ida) -> ()
+ | A.Cast (_, typa, _, ea) -> ()
  | A.SizeOfExpr (_, ea) -> ()
  | A.SizeOfType (_, _, typa, _) -> ()
+ | A.Delete (_, ea) -> ()
+ | A.DeleteArr (_,_,_,ea) -> ()
+ | A.New(_, _, _, _, _, _) -> ()
+ | A.TemplateInst(_,_,_,_) -> ()
  | A.TypeExp (typa) -> ()
  | A.Constructor (_, typa, _, ia) -> ()
  | A.Paren (_, ea, _) -> ()
@@ -137,7 +143,7 @@ let dumb_astcocci_expr = function
  | A.OptExp _ -> ()
 
 let dumb_astcocci_fulltype = function
-    A.Type(_,cv,ty) -> ()
+    A.Type(_,cv,attrs,ty) -> ()
   | A.AsType(_,_) -> ()
   | A.DisjType(types) -> ()
   | A.ConjType(types) -> ()
@@ -152,14 +158,16 @@ let dumb_astcocci_type = function
  | A.FunctionType(ty,lp,params,rp) -> ()
  | A.Array (typa, _, eaopt, _) -> ()
  | A.Decimal(_, _, _, _, _, _) -> ()
- | A.EnumName(en, ena) -> ()
- | A.EnumDef(ty, lb, ids, rb) -> ()
+ | A.EnumName(k, key, ena) -> ()
+ | A.EnumDef(ty, base, lb, ids, rb) -> ()
  | A.StructUnionName(sa, sua) -> ()
  | A.StructUnionDef(ty, lb, decls, rb) -> ()
  | A.TypeOfExpr(tf, lp, exp, rp) -> ()
  | A.TypeOfType(tf, lp, ty, rp) -> ()
  | A.TypeName sa -> ()
+ | A.QualifiedType(ty,_,ida) -> ()
  | A.AutoType _ -> ()
+ | A.TemplateType _ -> ()
 
 
 (* ------------------------------------------------------------------------- *)
