@@ -102,6 +102,8 @@ module type PARAM =
       (Ast_cocci.meta_name Ast_cocci.mcode, Ast_c.assignOp) matcher
     val distrf_binaryOp :
       (Ast_cocci.meta_name Ast_cocci.mcode, Ast_c.binaryOp) matcher
+    val distrf_pragma_info :
+      (Ast_cocci.meta_name Ast_cocci.mcode, Ast_c.info) matcher
     val distrf_args :
       (Ast_cocci.meta_name Ast_cocci.mcode,
       (Ast_c.argument, Ast_c.il) Common.either list)
@@ -117,6 +119,13 @@ module type PARAM =
     val distrf_param :
       (Ast_cocci.meta_name Ast_cocci.mcode, Ast_c.parameterType) matcher
 
+    val distrf_template_params :
+      (Ast_cocci.meta_name Ast_cocci.mcode,
+      (Ast_c.templateParameterType, Ast_c.il) Common.either list)
+      matcher
+    val distrf_template_param :
+      (Ast_cocci.meta_name Ast_cocci.mcode, Ast_c.templateParameterType) matcher
+
     val distrf_ini :
       (Ast_cocci.meta_name Ast_cocci.mcode, Ast_c.initialiser) matcher
     val distrf_inis :
