@@ -318,10 +318,13 @@ module XMATCH = struct
   let distrf_e              = distrf Lib_parsing_c.ii_of_expr
   let distrf_assignOp       = distrf Lib_parsing_c.ii_of_assignOp
   let distrf_binaryOp       = distrf Lib_parsing_c.ii_of_binaryOp
+  let distrf_pragma_info    = distrf (fun x -> [x])
   let distrf_args           = distrf Lib_parsing_c.ii_of_args
   let distrf_type           = distrf Lib_parsing_c.ii_of_type
   let distrf_param          = distrf Lib_parsing_c.ii_of_param
-  let distrf_params         = distrf Lib_parsing_c.ii_of_params
+  let distrf_params         = distrf Lib_parsing_c.ii_of_params 
+  let distrf_template_param = distrf Lib_parsing_c.ii_of_template_param
+  let distrf_template_params= distrf Lib_parsing_c.ii_of_template_params
   let distrf_ini            = distrf Lib_parsing_c.ii_of_ini
   let distrf_inis           = distrf Lib_parsing_c.ii_of_inis
   let distrf_decl           = distrf Lib_parsing_c.ii_of_decl
@@ -336,6 +339,7 @@ module XMATCH = struct
   let distrf_ident_list     = distrf Lib_parsing_c.ii_of_ident_list
   let distrf_exec_code_list = distrf Lib_parsing_c.ii_of_exec_code_list
   let distrf_attr           = distrf Lib_parsing_c.ii_of_attr
+  let distrf_attr_arg       = distrf Lib_parsing_c.ii_of_attr_arg
   let distrf_attrs          = distrf Lib_parsing_c.ii_of_attrs
 
 
@@ -412,14 +416,28 @@ module XMATCH = struct
             Ast_c.MetaIdVal (a)    ->
 	      success(Ast_c.MetaIdVal(a))
           | Ast_c.MetaAssignOpVal op      ->
+	      let op =
+		if strip
+		then Lib_parsing_c.al_assignOp op
+		else Lib_parsing_c.semi_al_assignOp op in
 	      success(Ast_c.MetaAssignOpVal op)
           | Ast_c.MetaBinaryOpVal op      ->
+	      let op =
+		if strip
+		then Lib_parsing_c.al_binaryOp op
+		else Lib_parsing_c.semi_al_binaryOp op in
 	      success(Ast_c.MetaBinaryOpVal op)
+          | Ast_c.MetaPragmaInfoVal pi      ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_info pi
+		else Lib_parsing_c.semi_al_info pi in
+	      success(Ast_c.MetaPragmaInfoVal stripped)
           | Ast_c.MetaFuncVal a      ->
 	      success(Ast_c.MetaFuncVal a)
           | Ast_c.MetaLocalFuncVal a ->
 	      success(Ast_c.MetaLocalFuncVal a) (*more?*)
-          | Ast_c.MetaExprVal (a,c,ty) ->
+          | Ast_c.MetaExprVal (a,original,c,ty) ->
 	      (* c in the value is only to prepare for the future in which
 		 we figure out how to have subterm constraints on unbound
 		 variables.  Now an environment will only contain expression
@@ -431,13 +449,13 @@ module XMATCH = struct
 		else Lib_parsing_c.semi_al_expr a in
 	      let inh_stripped = Lib_parsing_c.al_inh_expr a in
 	      let rec loop = function
-		  [] -> success(Ast_c.MetaExprVal(stripped,[],ty))
+		  [] -> success(Ast_c.MetaExprVal(stripped,original,[],ty))
 		| c::cs ->
 		    let tmp =
 		      Common.optionise
 			(fun () -> tin.binding0 +> List.assoc c) in
 		    (match tmp with
-		      Some (Ast_c.MetaExprVal(v,_,_)) ->
+		      Some (Ast_c.MetaExprVal(v,_,_,_)) ->
 			if C_vs_c.subexpression_of_expression inh_stripped v
 			then loop cs (* forget satisfied constraints *)
 			else None (* failure *)
@@ -446,12 +464,12 @@ module XMATCH = struct
 			 doesn't exist *)
 		    | None -> None) in
 	      loop c
-          | Ast_c.MetaExprListVal a ->
-	      success
-		(Ast_c.MetaExprListVal
-		   (if strip
-		   then Lib_parsing_c.al_arguments a
-		   else Lib_parsing_c.semi_al_arguments a))
+          | Ast_c.MetaExprListVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_arguments a
+		else Lib_parsing_c.semi_al_arguments a in
+	      success(Ast_c.MetaExprListVal(stripped,original))
 
           | Ast_c.MetaDeclVal(a,original) ->
 	      let stripped =
@@ -459,65 +477,78 @@ module XMATCH = struct
 		then Lib_parsing_c.al_declaration a
 		else Lib_parsing_c.semi_al_declaration a in
 	      success(Ast_c.MetaDeclVal(stripped,original))
-          | Ast_c.MetaFieldVal a ->
-	      success
-		(Ast_c.MetaFieldVal
-		   (if strip
-		   then Lib_parsing_c.al_field a
-		   else Lib_parsing_c.semi_al_field a))
-          | Ast_c.MetaFieldListVal a ->
-	      success
-		(Ast_c.MetaFieldListVal
-		   (if strip
-		   then Lib_parsing_c.al_fields a
-		   else Lib_parsing_c.semi_al_fields a))
+          | Ast_c.MetaFieldVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_field a
+		else Lib_parsing_c.semi_al_field a in
+	      success(Ast_c.MetaFieldVal(stripped,original))
+          | Ast_c.MetaFieldListVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_fields a
+		else Lib_parsing_c.semi_al_fields a in
+	      success(Ast_c.MetaFieldListVal(stripped,original))
           | Ast_c.MetaStmtVal(a,original,ty) ->
 	      let stripped =
 		if strip
 		then Lib_parsing_c.al_statement a
 		else Lib_parsing_c.semi_al_statement a in
 	      success(Ast_c.MetaStmtVal(stripped,original,ty))
-          | Ast_c.MetaStmtListVal(a,ty) ->
+          | Ast_c.MetaStmtListVal(a,original,ty) ->
 	      let stripped =
 		if strip
 		then Lib_parsing_c.al_statement_seq_list a
 		else Lib_parsing_c.semi_al_statement_seq_list a in
-	      success(Ast_c.MetaStmtListVal(stripped,ty))
-          | Ast_c.MetaTypeVal a ->
-	      success
-		(Ast_c.MetaTypeVal
-		   (if strip
-		   then Lib_parsing_c.al_type a
-		   else Lib_parsing_c.semi_al_type a))
+	      success(Ast_c.MetaStmtListVal(stripped,original,ty))
+          | Ast_c.MetaTypeVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_type a
+		else Lib_parsing_c.semi_al_type a in
+	      success(Ast_c.MetaTypeVal(stripped,original))
 
-          | Ast_c.MetaInitVal a ->
-	      success
-		(Ast_c.MetaInitVal
-		   (if strip
-		   then Lib_parsing_c.al_init a
-		   else Lib_parsing_c.semi_al_init a))
+          | Ast_c.MetaInitVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_init a
+		else Lib_parsing_c.semi_al_init a in
+	      success(Ast_c.MetaInitVal(stripped,original))
 
-          | Ast_c.MetaInitListVal a ->
-	      success
-		(Ast_c.MetaInitListVal
-		   (if strip
-		   then Lib_parsing_c.al_inits a
-		   else Lib_parsing_c.semi_al_inits a))
+          | Ast_c.MetaInitListVal (newlines,a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_inits a
+		else Lib_parsing_c.semi_al_inits a in
+	      success(Ast_c.MetaInitListVal(newlines,stripped,original))
 
           | Ast_c.MetaListlenVal a -> success(Ast_c.MetaListlenVal a)
 
-          | Ast_c.MetaParamVal a ->
-	      success
-		(Ast_c.MetaParamVal
-		   (if strip
-		   then Lib_parsing_c.al_param a
-		   else Lib_parsing_c.semi_al_param a))
-          | Ast_c.MetaParamListVal a ->
-	      success
-		(Ast_c.MetaParamListVal
-		   (if strip
-		   then Lib_parsing_c.al_params a
-		   else Lib_parsing_c.semi_al_params a))
+          | Ast_c.MetaParamVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_param a
+		else Lib_parsing_c.semi_al_param a in
+	      success(Ast_c.MetaParamVal(stripped,original))
+          | Ast_c.MetaParamListVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_params a
+		else Lib_parsing_c.semi_al_params a in
+	      success(Ast_c.MetaParamListVal(stripped,original))
+
+          | Ast_c.MetaTemplateParamVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_template_param a
+		else Lib_parsing_c.semi_al_template_param a in
+	      success(Ast_c.MetaTemplateParamVal(stripped,original))
+          | Ast_c.MetaTemplateParamListVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_template_params a
+		else Lib_parsing_c.semi_al_template_params a in
+	      success(Ast_c.MetaTemplateParamListVal(stripped,original))
 
           | Ast_c.MetaDParamListVal a ->
 	      success
@@ -538,12 +569,12 @@ module XMATCH = struct
 		   (if strip
 		   then Lib_parsing_c.al_string_format a
 		   else Lib_parsing_c.semi_al_string_format a))
-          | Ast_c.MetaAttributeVal a ->
-	      success
-		(Ast_c.MetaAttributeVal
-		   (if strip
-		   then Lib_parsing_c.al_attribute a
-		   else Lib_parsing_c.semi_al_attribute a))
+          | Ast_c.MetaAttrArgVal(a,original) ->
+	      let stripped =
+		if strip
+		then Lib_parsing_c.al_attr_arg a
+		else Lib_parsing_c.semi_al_attr_arg a in
+	      success(Ast_c.MetaAttrArgVal(stripped,original))
 
           | Ast_c.MetaPosVal (pos1,pos2) ->
 	      success(Ast_c.MetaPosVal (pos1,pos2))
