@@ -7,6 +7,7 @@
 open Common
 
 module F = Control_flow_c
+module V = Visitor_ast
 
 (*****************************************************************************)
 (* The functor argument  *)
@@ -114,36 +115,74 @@ module XTRANS = struct
   (* When env is used in + code, have to strip it more to avoid circular
   references due to local variable information *)
 
+  let drop_exp_types = (* keep position information *)
+    { Visitor_c.default_visitor_c_s with
+      Visitor_c.kinfo_s = (fun (k,_) i ->
+	{ i with Ast_c.cocci_tag = ref Ast_c.emptyAnnot; }
+      );
+      Visitor_c.kexpr_s = (fun (k,_) e ->
+	let (e', ty),ii' = k e in
+	(e', Ast_c.noType()), ii'
+      );
+    }
+
   let clean_env env =
     List.map
       (function (v,vl) ->
 	match vl with
-	| Ast_c.MetaExprVal(e,ml,ty) ->
-	    (v,Ast_c.MetaExprVal(Lib_parsing_c.real_al_expr e,ml,ty))
-	| Ast_c.MetaExprListVal(es) ->
-	    (v,Ast_c.MetaExprListVal(Lib_parsing_c.real_al_arguments es))
-	| Ast_c.MetaTypeVal(ty) ->
-	    (v,Ast_c.MetaTypeVal(Lib_parsing_c.real_al_type ty))
-	| Ast_c.MetaInitVal(i) ->
-	    (v,Ast_c.MetaInitVal(Lib_parsing_c.real_al_init i))
-	| Ast_c.MetaInitListVal(is) ->
-	    (v,Ast_c.MetaInitListVal(Lib_parsing_c.real_al_inits is))
+	| Ast_c.MetaExprVal(e,original,ml,ty) ->
+	    (v,Ast_c.MetaExprVal(Visitor_c.vk_expr_s drop_exp_types e,
+				 Visitor_c.vk_expr_s drop_exp_types original,ml,ty))
+	| Ast_c.MetaExprListVal(es,original) ->
+	    (v,Ast_c.MetaExprListVal(Visitor_c.vk_arguments_s drop_exp_types es,
+				     Visitor_c.vk_arguments_s drop_exp_types original))
+	| Ast_c.MetaTypeVal(ty,original) ->
+	    (v,Ast_c.MetaTypeVal(Visitor_c.vk_type_s drop_exp_types ty,
+				 Visitor_c.vk_type_s drop_exp_types original))
+	| Ast_c.MetaInitVal(i,original) ->
+	    (v,Ast_c.MetaInitVal(Visitor_c.vk_ini_s drop_exp_types i,
+				 Visitor_c.vk_ini_s drop_exp_types original))
+	| Ast_c.MetaInitListVal(newlines,is,original) ->
+	    (v,Ast_c.MetaInitListVal(newlines,Visitor_c.vk_inis_s drop_exp_types is,
+				     Visitor_c.vk_inis_s drop_exp_types original))
 	| Ast_c.MetaDeclVal(d,original) ->
 	    (v,
 	     Ast_c.MetaDeclVal
-	       (Lib_parsing_c.real_al_decl d,
-		Lib_parsing_c.real_al_decl_with_comments original))
+	       (Visitor_c.vk_decl_s drop_exp_types d,
+		Visitor_c.vk_decl_s drop_exp_types original))
 	| Ast_c.MetaStmtVal(s,original,ty) ->
 	    (v,
 	     Ast_c.MetaStmtVal
-	       (Lib_parsing_c.real_al_statement s,
-		Lib_parsing_c.real_al_statement_with_comments original,ty))
+	       (Visitor_c.vk_statement_s drop_exp_types s,
+		Visitor_c.vk_statement_s drop_exp_types original,ty))
+	| Ast_c.MetaStmtListVal(ss,original,ty) ->
+	    (v,Ast_c.MetaStmtListVal(Visitor_c.vk_statement_sequencable_list_s drop_exp_types ss,
+				     Visitor_c.vk_statement_sequencable_list_s drop_exp_types original,
+				     ty))
+
+	| Ast_c.MetaParamVal(p,original) ->
+	    (v,Ast_c.MetaParamVal(Visitor_c.vk_param_s drop_exp_types p,
+				  Visitor_c.vk_param_s drop_exp_types original))
+	| Ast_c.MetaParamListVal(p,original) ->
+	    (v,Ast_c.MetaParamListVal(Visitor_c.vk_params_s drop_exp_types p,
+				      Visitor_c.vk_params_s drop_exp_types original))
+	| Ast_c.MetaFieldVal(fld,original) ->
+	    (v,Ast_c.MetaFieldVal(Visitor_c.vk_struct_field_s drop_exp_types fld,
+				  Visitor_c.vk_struct_field_s drop_exp_types original))
+	| Ast_c.MetaFieldListVal(fld,original) ->
+	    (v,Ast_c.MetaFieldListVal(Visitor_c.vk_struct_fields_s drop_exp_types fld,
+				      Visitor_c.vk_struct_fields_s drop_exp_types original))
+	| Ast_c.MetaAttrArgVal(a,original) ->
+	    (v,Ast_c.MetaAttrArgVal(Visitor_c.vk_attr_arg_s drop_exp_types a,
+				    Visitor_c.vk_attr_arg_s drop_exp_types original))
+
 	(* These don't contain local variables, but the cocci_tag field
 	   causes problems too.  Why is this not needd for other metavars? *)
 	| Ast_c.MetaAssignOpVal(b) ->
-	    (v,Ast_c.MetaAssignOpVal(Lib_parsing_c.real_al_assignop b))
+	    (v,Ast_c.MetaAssignOpVal(Visitor_c.vk_assignOp_s drop_exp_types b))
 	| Ast_c.MetaBinaryOpVal(b) ->
-	    (v,Ast_c.MetaBinaryOpVal(Lib_parsing_c.real_al_binop b))
+	    (v,Ast_c.MetaBinaryOpVal(Visitor_c.vk_binaryOp_s drop_exp_types b))
+
 	| _ -> (v,vl))
       env
 
@@ -284,16 +323,11 @@ module XTRANS = struct
           Ast_cocci.rewrap e
 	    (Ast_cocci.MetaExpr(name,Ast_cocci.CstrTrue,u,ty,form,i,bitfield))
       | _ -> e in
-    let fn = Visitor_ast.rebuilder
-	mcode mcode mcode mcode mcode mcode mcode mcode mcode
-	mcode mcode mcode mcode mcode
-	donothing donothing donothing donothing donothing donothing donothing
-	ident expression donothing donothing donothing donothing
-	donothing donothing donothing donothing donothing donothing
-	donothing donothing donothing donothing donothing donothing
-	donothing donothing donothing donothing donothing in
+    let fn =
+      V.rebuilder {V.rmcode=mcode} {V.rdonothing=donothing}
+	~ident:ident ~expr:expression donothing in
 
-  fn.Visitor_ast.rebuilder_anything anything
+  fn.V.rebuilder_anything anything
 
   let strip_minus_code = function
       Ast_cocci.REPLACEMENT(l,c) ->
@@ -606,6 +640,9 @@ module XTRANS = struct
   let distribute_mck_binaryOp (maxpos, minpos) = fun (lop,mop,rop,bop) -> fun x ->
     Visitor_c.vk_binaryOp_s (mk_bigf (maxpos, minpos) (lop,mop,rop,bop)) x
 
+  let distribute_mck_pragma_info (maxpos, minpos) = fun (lop,mop,rop,bop) -> fun x ->
+    Visitor_c.vk_info_s (mk_bigf (maxpos, minpos) (lop,mop,rop,bop)) x
+
   let distribute_mck_args (maxpos, minpos) = fun (lop,mop,rop,bop) -> fun x ->
     Visitor_c.vk_args_splitted_s (mk_bigf (maxpos, minpos) (lop,mop,rop,bop)) x
 
@@ -637,6 +674,13 @@ module XTRANS = struct
     Visitor_c.vk_params_splitted_s (mk_bigf (maxpos, minpos) (lop,mop,rop,bop))
       x
 
+  let distribute_mck_template_param (maxpos, minpos) = fun (lop,mop,rop,bop) -> fun x ->
+    Visitor_c.vk_template_param_s (mk_bigf (maxpos, minpos) (lop,mop,rop,bop)) x
+
+  let distribute_mck_template_params (maxpos, minpos) = fun (lop,mop,rop,bop) ->fun x ->
+    Visitor_c.vk_template_params_splitted_s (mk_bigf (maxpos, minpos) (lop,mop,rop,bop))
+      x
+
   let distribute_mck_fragments (maxpos, minpos) =
     fun (lop,mop,rop,bop) ->fun x ->
       Visitor_c.vk_string_fragments_splitted_s
@@ -732,10 +776,15 @@ module XTRANS = struct
   let distrf_e    = distrf (Lib_parsing_c.ii_of_expr,  distribute_mck_expr)
   let distrf_assignOp = distrf (Lib_parsing_c.ii_of_assignOp, distribute_mck_assignOp)
   let distrf_binaryOp = distrf (Lib_parsing_c.ii_of_binaryOp, distribute_mck_binaryOp)
+  let distrf_pragma_info = distrf ((fun x -> [x]), distribute_mck_pragma_info)
   let distrf_args = distrf (Lib_parsing_c.ii_of_args,  distribute_mck_args)
   let distrf_type = distrf (Lib_parsing_c.ii_of_type,  distribute_mck_type)
   let distrf_param  = distrf (Lib_parsing_c.ii_of_param, distribute_mck_param)
   let distrf_params = distrf (Lib_parsing_c.ii_of_params,distribute_mck_params)
+  let distrf_template_param =
+    distrf (Lib_parsing_c.ii_of_template_param, distribute_mck_template_param)
+  let distrf_template_params =
+    distrf (Lib_parsing_c.ii_of_template_params,distribute_mck_template_params)
   let distrf_ini = distrf (Lib_parsing_c.ii_of_ini,distribute_mck_ini)
   let distrf_inis = distrf (Lib_parsing_c.ii_of_inis,distribute_mck_inis)
   let distrf_decl = distrf (Lib_parsing_c.ii_of_decl,distribute_mck_decl)
@@ -868,7 +917,7 @@ let (transform2: string (* rule name *) -> string list (* dropped_isos *) ->
       (* assert that have done something. But with metaruleElem sometimes
          don't modify fake nodes. So special case before on Fake nodes. *)
       (match F.unwrap node with
-      | F.Enter | F.Exit | F.ErrorExit
+      | F.Enter | F.PreExit _ | F.Exit | F.ErrorExit
       | F.EndStatement _ | F.CaseNode _
       | F.Fake
       | F.TrueNode _ | F.FalseNode | F.AfterNode _ | F.FallThroughNode
