@@ -18,9 +18,7 @@ module Inc = Includes
  * globals/config.ml, mainly a standard.h and standard.iso file *)
 
 let cocci_file = ref ""
-
 let opt_c_files = ref []
-
 let output_file = ref "" (* resulting code *)
 let tmp_dir = ref "" (* temporary files for parallelism *)
 let aux_file_suffix =
@@ -157,6 +155,50 @@ let quiet_profile = (
 
   ])
 
+let testing_profile = (
+  [
+    Common.print_to_stderr;
+    FC.show_diff;
+  ],
+  [
+
+    Flag.show_misc;
+    Flag.show_trying;
+    Flag.show_transinfo;
+
+    FC.show_c;
+    FC.show_cocci;
+    FC.show_flow;
+    FC.show_before_fixed_flow;
+    FC.show_ctl_tex;
+    FC.show_ctl_text;
+    FC.show_binding_in_out;
+
+    FC.verbose_cocci;
+
+    Flag_parsing_c.show_parsing_error;
+
+    Flag_parsing_c.verbose_lexing;
+    Flag_parsing_c.verbose_parsing;
+    Flag_parsing_c.verbose_type;
+    Flag_parsing_c.verbose_cfg;
+    Flag_parsing_c.verbose_unparsing;
+    Flag_parsing_c.verbose_visit;
+    Flag_parsing_c.verbose_cpp_ast;
+    Flag_parsing_c.verbose_includes;
+
+    Flag_matcher.verbose_matcher;
+    Flag_matcher.debug_engine;
+
+    Flag_parsing_c.debug_unparsing;
+
+    Flag_parsing_cocci.show_SP;
+    Flag_parsing_cocci.show_iso_failures;
+
+    Flag_ctl.verbose_ctl_engine;
+    Flag_ctl.verbose_match;
+  ])
+
 (* some information that is useful in seeing why a semantic patch doesn't
 work properly *)
 let debug_profile = (
@@ -370,6 +412,15 @@ let short_options = [
   "--use-coccigrep",
   Arg.Unit (function _ -> Flag.scanner := Flag.CocciGrep),
   "    find relevant files using cocci grep";
+  "--use-diff",
+  Arg.Unit (function _ -> Flag.scanner := Flag.PatchDiff),
+  "    process files in the diff for a directory";
+  "--use-patch-diff",
+  Arg.String
+    (function s ->
+      let (startid,endid) = Patch_diff.split s in
+      Flag.scanner := Flag.PatchDiffRange(startid,endid)),
+  "    process files in the diff for a directory";
   "--patch",
     Arg.String (function s -> Flag.patch := Some (Cocci.normalize_path s)),
   ("    <dir> path name with respect to which a patch should be created\n"^
@@ -660,7 +711,7 @@ let other_options = [
     "  make a small attempt to parse C++ files";
     (* Special case, for when a user gives a specific C++ version *)
     "--c++=",
-    Arg.Symbol (["98"; "03"; "0x"; "11"; "1y"; "14"; "1z"; "17"; "2a"; "20"],
+    Arg.Symbol (["98"; "03"; "0x"; "11"; "1y"; "14"; "1z"; "17"; "2a"; "20"; "23"],
                 (fun i -> Flag.set_c_plus_plus (Some i))),
     "  <int> same as --c++ but allows to specify a version (e.g. --c++=11)";
     "--ibm", Arg.Set Flag.ibm,
@@ -684,6 +735,10 @@ let other_options = [
     "--external-analysis-file", Arg.String
       (Externalanalysis.load_external_results),
     "  import results from an external analysis";
+    "--ocaml-lib-search-path",
+    Arg.String (function s ->
+      Flag.ocaml_lib_search_path := s :: !Flag.ocaml_lib_search_path),
+    "  add a path when searching for dynamically loadable modules";
   ];
 
   "concurrency",
@@ -756,6 +811,8 @@ let other_options = [
     "   do not update the score file when -testall succeeds";
     "--relax-include-path", Arg.Set Inc.relax_include_path,
     " ";
+    "--batch_mode", Arg.Set _batch_mode,
+    " no interactivity"
   ];
 
   "action mode",
@@ -802,10 +859,10 @@ let arg_align2 xs =
 let check_include_path () =
   let opt = Array.get Sys.argv !Arg.current in
   let is_include_re = Str.regexp "-I\\(.*\\)" in
-  if Str.string_match is_include_re opt 0 then
+  if Str.string_match is_include_re opt 0
+  then
     let path = Str.matched_group 1 opt in
-	Inc.include_path:= path::!Inc.include_path
-  else ()
+    Inc.include_path:= path::!Inc.include_path
 
 let rec arg_parse_no_fail ~current l f msg argv =
   try
@@ -841,16 +898,16 @@ let arg_parse2 ~current l f msg argv argv_location =
 
 let short_usage () =
  begin
-  Common.short_usage usage_msg short_options;
-  pr2 "";
-  pr2 "Example of use:";
-  pr2 "  ./spatch --sp-file foo.cocci foo.c -o /tmp/newfoo.c";
-  pr2 "";
+  Common.short_usage usage_msg ~short_opt:short_options;
+  pr "";
+  pr "Example of use:";
+  pr "  ./spatch --sp-file foo.cocci foo.c -o /tmp/newfoo.c";
+  pr "";
  end
 
 
 let long_usage () =
-  Common.long_usage usage_msg short_options other_options
+  Common.long_usage usage_msg ~short_opt:short_options ~long_opt:other_options
 
 let _ = short_usage_func := short_usage
 let _ = long_usage_func := long_usage
@@ -882,7 +939,7 @@ let glimpse_filter2 (_,query,_,_) dir =
       let rec loop = function
 	  [] -> None (* error, eg due to pattern too big *)
 	| query::queries ->
-	    Printf.fprintf stderr "%s\n" ("glimpse request = " ^ query);
+	    Printf.eprintf "%s\n" ("glimpse request = " ^ query);
 	    let command = spf "glimpse -y -H %s -N -W -w '%s'" dir query in
 	    let (glimpse_res,stat) = Common.cmd_to_list_and_status command in
 	    match stat with
@@ -940,13 +997,24 @@ let idutils_filter (_,_,_,query) dir =
       Some
 	(files +>
 	 List.filter
-	   (fun file -> List.mem (Common.filesuffix file) suffixes))
+     (fun file -> List.mem (Common.filesuffix file) suffixes))
+
+let diff_filter _ dir =
+  let struc = Patch_diff.getdiff dir in
+  Some (List.map (function x -> x.Patch_diff.file_name) struc)
+
+let patchdiff_filter startid endid _ dir =
+  let struc = Patch_diff.getpatchdiff dir startid endid in
+  Some (List.map (function x -> x.Patch_diff.file_name) struc)
 
 let scanner_to_interpreter = function
     Flag.Glimpse -> glimpse_filter
   | Flag.IdUtils -> idutils_filter
   | Flag.CocciGrep -> coccigrep_filter
   | Flag.GitGrep -> gitgrep_filter
+  | Flag.PatchDiff -> diff_filter
+  | Flag.PatchDiffRange(startid,endid) ->
+      patchdiff_filter startid endid
   | _ -> failwith "impossible"
 
 (*****************************************************************************)
@@ -1021,7 +1089,7 @@ let rec main_action xs =
   | _ -> failwith "only one .cocci file allowed");
   Iteration.base_file_list := xs;
   previous_merges := ([], []);
-  let rec toploop = function
+  let rec toploop patching_failed = function
       [] -> failwith "no C files provided"
     | x::xs ->
       (* a more general solution would be to use
@@ -1079,7 +1147,8 @@ let rec main_action xs =
 		     " or multiple files")
               | _, false, _, _, _ -> [List.map (fun x -> (x,None)) (x::xs)]
 	      |	_, true, "",
-		  (Flag.Glimpse|Flag.IdUtils|Flag.CocciGrep|Flag.GitGrep),
+		  (Flag.Glimpse|Flag.IdUtils|Flag.CocciGrep|Flag.GitGrep|
+		  Flag.PatchDiff|Flag.PatchDiffRange _),
 		  [] ->
 		    let interpreter = scanner_to_interpreter !Flag.scanner in
 		    let files =
@@ -1088,7 +1157,8 @@ let rec main_action xs =
 		      | Some files -> files in
                     files +> List.map (fun x -> [(x,None)])
               | _, true, s,
-		  (Flag.Glimpse|Flag.IdUtils|Flag.CocciGrep|Flag.GitGrep), _
+		  (Flag.Glimpse|Flag.IdUtils|Flag.CocciGrep|Flag.GitGrep|
+		  Flag.PatchDiff|Flag.PatchDiffRange _), _
 		when s <> "" ->
                   failwith "--use-xxx filters do not work with --kbuild"
                   (* normal *)
@@ -1238,14 +1308,14 @@ let rec main_action xs =
 	    then seq_fold
 	    else par_fold in
 
-          let (outfiles, merges) =
+          let (outfiles, merges, patching_failed) =
             Common.profile_code "Main.outfiles computation" (fun () ->
 	      let res =
 		match infiles with
 		  [] ->
 		    (* parmap case does a lot of work that is not needed
 		       if there is nothing to do *)
-		    [], []
+		    [], [], false
 		| _ ->
 (* merge_vars associates one merge variable assignment to each core.
 The following code relies on the fact that parfold never
@@ -1253,64 +1323,64 @@ passes the accumulator from one core to another before merging the
 results. Therefore, the fold operator may just keep the last value in a
 singleton list (forgetting the former accumulator value) and these
 singleton lists are then just appended to each other during the merge. *)
-		    let merge (outfiles, merge_vars) (outfiles', merge_vars') =
+		    let merge (outfiles, merge_vars, fld) (outfiles', merge_vars', fld') =
 		      let all_outfiles = List.rev_append outfiles outfiles' in
 		      let all_merge_vars =
 			List.rev_append merge_vars merge_vars' in
-		      all_outfiles, all_merge_vars in
+		      all_outfiles, all_merge_vars, fld || fld' in
 		    infiles +> actual_fold merge
-		      (fun (prev_files, _prev_merges as prev) cfiles ->
-		      if (not !Flag.worth_trying_opt) ||
-		      Cocci.worth_trying cfiles constants
-		      then
-			begin
-			  let all_cfiles =
-			    String.concat " " (List.map fst cfiles) in
-			  pr2 ("HANDLING: " ^ all_cfiles);
-			  flush stderr;
+		      (fun ((prev_files, prev_merges, fld) as prev) cfiles ->
+			if (not !Flag.worth_trying_opt) ||
+			Cocci.worth_trying cfiles constants
+			then
+			  begin
+			    let all_cfiles =
+			      String.concat " " (List.map fst cfiles) in
+			    pr2 ("HANDLING: " ^ all_cfiles);
+			    flush stderr;
 
-			  Common.timeout_function_opt all_cfiles !FC.timeout
-			    (fun () ->
-			      let res =
-				Common.report_if_take_time 10 all_cfiles
-				  (fun () ->
-				    try
-				      let optfile =
-					if !output_file <> "" && !compat_mode
-					then Some !output_file
-					else None in
-				      let files, merges =
-					adjust_stdin cfiles (fun () ->
-					  Common.redirect_stdout_opt optfile
-					    (fun () ->
+			    Common.timeout_function_opt all_cfiles !FC.timeout
+			      (fun () ->
+				let res =
+				  Common.report_if_take_time 10 all_cfiles
+				    (fun () ->
+				      try
+					let optfile =
+					  if !output_file <> "" && !compat_mode
+					  then Some !output_file
+					  else None in
+					let files, merges =
+					  adjust_stdin cfiles (fun () ->
+					    Common.redirect_stdout_opt optfile
+					      (fun () ->
 					      (* this is the main call *)
-					      Cocci.full_engine cocci_infos
-						cfiles
-					 )) in
-				      List.rev_append files prev_files,
-				      [merges]
-				    with
-				    | Common.UnixExit x ->
-					raise (Common.UnixExit x)
-				    | Pycocci.Pycocciexception ->
-					raise Pycocci.Pycocciexception
-				    | e ->
-					if !dir || !file_groups
-					then begin
+						Cocci.full_engine cocci_infos
+						  cfiles
+						  )) in
+					List.rev_append files prev_files,
+					[merges], fld
+				      with
+				      | Common.UnixExit x ->
+					  raise (Common.UnixExit x)
+				      | Pycocci.Pycocciexception ->
+					  raise Pycocci.Pycocciexception
+				      | e ->
+					  if !dir || !file_groups
+					  then begin
 					  (* not hidden by --very-quiet *)
-					  Printf.eprintf "EXN: %s in %s\n"
-					    (Printexc.to_string e)
-					    all_cfiles;
-					  flush stderr;
-					  prev (* *)
-					end
-					else raise e) in
-			      (if !dir && !profile_per_file
-			      then Common.reset_profile());
-			      res)
-			end
-		      else prev)
-		      ([], []) in res) in
+					    Printf.eprintf "EXN: %s in %s\n"
+					      (Printexc.to_string e)
+					      all_cfiles;
+					    flush stderr;
+					    (prev_files, prev_merges, true)
+					  end
+					  else raise e) in
+				(if !dir && !profile_per_file
+				then Common.reset_profile());
+				res)
+			  end
+			else prev)
+		      ([], [], patching_failed) in res) in
 	  let outfiles = List.rev outfiles in
 	  let merges =
             List.fold_left Cocci.union_merge_vars !previous_merges merges in
@@ -1321,14 +1391,14 @@ singleton lists are then just appended to each other during the merge. *)
 		Cocci.post_engine cocci_infos merges;
 		Iteration.get_pending_instance()
 	    | pending_instance ->
-               previous_merges := merges;
-               pending_instance in
+		previous_merges := merges;
+		pending_instance in
 	  (match pending_instance with
 	    None ->
-	      (x,xs,cocci_infos,outfiles)
+	      (x,xs,cocci_infos,outfiles,patching_failed)
 	  | Some (files,virt_rules,virt_ids) ->
-	      if outfiles = [] || outfiles = [] || not !FC.show_diff
-		  || !Flag_cocci.inplace_modif
+	      if outfiles = [] || outfiles = [] || not !FC.show_diff ||
+	      !Flag_cocci.inplace_modif
 	      then
 		begin
 		  (if !Flag_cocci.inplace_modif
@@ -1341,30 +1411,35 @@ singleton lists are then just appended to each other during the merge. *)
 		  distrib_index := None;
 		  distrib_max := None;
 		  dir := List.exists Sys.is_directory files;
-		  toploop files
+		  toploop patching_failed files
 		end
 	      else
 		begin
-		  Common.pr2
-		    "Out of place transformation not compatible with iteration. Aborting.\n consider using -no_show_diff or -in_place";
-		  (x,xs,cocci_infos,outfiles)
+		  Printf.eprintf
+		    "Out of place transformation not compatible with iteration. Aborting.\n consider using --no-show-diff or --in-place";
+		  (x,xs,cocci_infos,outfiles,true)
 		end) in
-      let (x,xs,cocci_infos,outfiles) = toploop xs in
+      let (x,xs,cocci_infos,outfiles,patching_failed) = toploop false xs in
       Common.profile_code "Main.result analysis" (fun () ->
 	Ctlcocci_integration.print_bench();
-	generate_outfiles outfiles x xs;
+	(if not patching_failed
+	then generate_outfiles outfiles x xs);
         match !compare_with_expected with
 	  None -> ()
-        | Some extension -> Testing.compare_with_expected outfiles extension)
+        | Some extension -> Testing.compare_with_expected outfiles extension);
+      if patching_failed
+      then
+        Printf.eprintf
+	  "An error occurred when attempting to transform some files.\n";
 
 and debug_restart virt_rules virt_ids =
   if !Flag_parsing_cocci.debug_parse_cocci
   then
     begin
-      Printf.fprintf stderr
+      Printf.eprintf
 	"Starting a new iteration with:\nVirtual rules: %s\n"
 	(String.concat " " virt_rules);
-      Printf.fprintf stderr
+      Printf.eprintf
 	"Virtual identifiers: %s\n\n"
 	(String.concat ", "
 	   (List.map
@@ -1497,8 +1572,6 @@ let main arglist =
 
     let args = ref [] in
 
-    (* Gc.set {(Gc.get ()) with Gc.stack_limit = 1024 * 1024};*)
-
     let arg_parse =
       let speclist = Arg.align all_options in
       let anon_fun = (fun x -> args := x::!args) in
@@ -1632,7 +1705,7 @@ let main arglist =
 		Testing.testone "" x !compare_with_expected
 	      end
 	      else
-	        Printf.fprintf stderr
+	        Printf.eprintf
 		  "ERROR: File %s does not exist\n" testfile
 	end
 
@@ -1642,7 +1715,13 @@ let main arglist =
         let score_file = if !expected_score_file <> ""
                          then !expected_score_file
                          else "tests/SCORE_expected.sexp" in
-        Testing.testall score_file !allow_update_score_file
+        Testing.testall
+	  (fun file ->
+	    run_profile testing_profile;
+	    let cocci_args =
+	      Cocci_args.read_args [file] +> normalize_args in
+	    arg_parse cocci_args "in the cocci file")
+	  score_file !allow_update_score_file
 
     | []  when !test_spacing ->
         (if !Inc.include_path = []
@@ -1650,7 +1729,13 @@ let main arglist =
         let score_file = if !expected_spacing_score_file <> ""
                          then !expected_spacing_score_file
                          else "tests/SCORE_spacing_expected.sexp" in
-        Testing.test_spacing score_file !allow_update_score_file
+        Testing.test_spacing
+	  (fun file ->
+	    run_profile testing_profile;
+	    let cocci_args =
+	      Cocci_args.read_args [file] +> normalize_args in
+	    arg_parse cocci_args "in the cocci file")
+	   score_file !allow_update_score_file
 
     | [] when !test_regression_okfailed ->
         Testing.test_regression_okfailed ()
@@ -1711,15 +1796,15 @@ let main_with_better_error_report arglist =
       main arglist
     with
     | Unix.Unix_error (e, "stat", filename) ->
-        Printf.fprintf stderr "ERROR: File %s does not exist: %s\n"
+        Printf.eprintf "ERROR: File %s does not exist: %s\n"
 	  filename (Unix.error_message e);
         raise (UnixExit (-1))
     | Parse_cocci.Bad_virt s ->
-	Printf.fprintf stderr "virtual rule %s not supported\n" s;
+	Printf.eprintf "virtual rule %s not supported\n" s;
         raise (UnixExit (-1))
     | Parse_cocci.SMPLParseError error_message
     | Failure error_message ->
-	Printf.fprintf stderr "%s\n" error_message;
+	Printf.eprintf "%s\n" error_message;
 	Printexc.print_backtrace stderr;
 	raise (UnixExit (-1))
 
