@@ -15,6 +15,7 @@ val initialization_stack : init_info list ref
 val base_file_list : string list ref
 val parsed_virtual_rules : string list ref
 val parsed_virtual_identifiers : string list ref
+val reset : unit -> unit (* for testing *)
 
 (* ----------------------------------------------------------------------- *)
 
@@ -30,6 +31,9 @@ val add_pending_instance :
 
 val get_pending_instance : unit -> pending_info option
 
+(* for testing *)
+val clear_pending_instance : unit -> unit
+
 (* ----------------------------------------------------------------------- *)
 
 val check_virtual_rule : string -> unit
