@@ -28,7 +28,7 @@ echo "  rm -rf $WORKING_DIRECTORY"
 echo
 echo The release script will _not_ push the updated repository.
 echo You may want to run the following command after the release:
-echo "  git push"
+echo "  git push --tags"
 echo
 if [ -d "$WEBSITE" ]; then
   echo For the website, you need to manually update those files:
@@ -82,5 +82,5 @@ rm -rf $WORKING_DIRECTORY/$PACKAGE
 if [ -d "$WEBSITE" ]; then
    cp $PACKAGE.tar.gz $WEBSITE
    cp $PACKAGE-bin-x86.tar.gz $WEBSITE
-   ( cd $WEBSITE; svn add $PACKAGE.tar.gz $PACKAGE-bin-x86.tar.gz )
+   ( cd $WEBSITE; git add $PACKAGE.tar.gz $PACKAGE-bin-x86.tar.gz )
 fi
