@@ -55,6 +55,7 @@ type param_type =
 		 Token_c.comment_like_token list) list
   | AssignOp of Ast_c.assignOp
   | BinaryOp of Ast_c.binaryOp
+  | PragmaInfo of Ast_c.info
   | Str of string
   | Type of Ast_c.fullType
   | Init of Ast_c.initialiser
@@ -62,6 +63,8 @@ type param_type =
   | Int of int
   | Param of Ast_c.parameterType
   | ParamList of Ast_c.parameterType Ast_c.wrap2 list
+  | TemplateParam of Ast_c.templateParameterType
+  | TemplateParamList of Ast_c.templateParameterType Ast_c.wrap2 list
   | DParamList of (string Ast_c.wrap) Ast_c.wrap2 list
   | Expr of Ast_c.expression
   | ExprList of Ast_c.argument Ast_c.wrap2 list
@@ -113,7 +116,7 @@ let (cstatement_of_string: string -> string -> Ast_c.statement) =
   fun env s ->
 (*  assert (no_format s);*) (* format strings are not getting parsed *)
   let tmpfile = Common.new_temp_file "cocci_stmt_of_s" "c" in
-  Common.write_file tmpfile (Printf.sprintf "void main() {\n%s\n%s\n}" env s);
+  Common.write_file ~file:tmpfile (Printf.sprintf "void main() {\n%s\n%s\n}" env s);
   let program = Parse_c.parse_c_and_cpp false false tmpfile +> fst in
   let _ =
     Type_annoter_c.annotate_program !Type_annoter_c.initial_env
@@ -131,7 +134,7 @@ let (cexpression_of_string: string -> string -> Ast_c.expression) =
   fun env s ->
 (*  assert (no_format s);*) (* format strings are not getting parsed *)
   let tmpfile = Common.new_temp_file "cocci_expr_of_s" "c" in
-  Common.write_file tmpfile (Printf.sprintf "void main() {\n%s\n%s;\n}" env s);
+  Common.write_file ~file:tmpfile (Printf.sprintf "void main() {\n%s\n%s;\n}" env s);
   let program = Parse_c.parse_c_and_cpp false false tmpfile +> fst in
   let _ =
     Type_annoter_c.annotate_program !Type_annoter_c.initial_env
@@ -161,13 +164,13 @@ let make_ident s = Ast_c.MetaIdVal(s)
 let make_expr s =
   parse_failure "expression" s
     (function s ->
-      Ast_c.MetaExprVal(Lib_parsing_c.al_expr(cexpression_of_string "" s), [],
-			Ast_c.WITHOUT_TYPES))
+      let e = Lib_parsing_c.al_expr(cexpression_of_string "" s) in
+      Ast_c.MetaExprVal(e, e, [], Ast_c.WITHOUT_TYPES))
 let make_expr_with_env env s =
   parse_failure "expression" s
     (function s ->
-      Ast_c.MetaExprVal(Lib_parsing_c.al_expr(cexpression_of_string env s), [],
-			Ast_c.WITH_TYPES))
+      let e = Lib_parsing_c.al_expr(cexpression_of_string env s) in
+      Ast_c.MetaExprVal(e, e, [], Ast_c.WITH_TYPES))
 let make_stmt s =
   parse_failure "statement" s
     (function s ->
@@ -181,7 +184,8 @@ let make_stmt_with_env env s =
 let make_type s =
   parse_failure "type" s
     (function s ->
-      Ast_c.MetaTypeVal(Lib_parsing_c.al_type(Parse_c.type_of_string s)))
+      let t = Lib_parsing_c.al_type(Parse_c.type_of_string s) in
+      Ast_c.MetaTypeVal(t,t))
 let make_listlen i = Ast_c.MetaListlenVal i
 let make_full_position fl fn ce_startl ce_startc ce_endl ce_endc
     startl startc endl endc =
@@ -226,9 +230,15 @@ let cocci_version () = Config.version
 (* org mode *)
 
 let build_link p msg color =
-  Printf.sprintf
-    "[[view:%s::face=%s::linb=%d::colb=%d::cole=%d][%s]]"
-    p.file color p.line p.col p.col_end msg
+  if p.line = p.line_end
+  then
+    Printf.sprintf
+      "[[view:%s::face=%s::linb=%d::colb=%d::cole=%d][%s]]"
+      p.file color p.line p.col p.col_end msg
+  else
+    Printf.sprintf
+      "[[view:%s::face=%s::linb=%d::line=%d::colb=%d::cole=%d][%s]]"
+      p.file color p.line p.line_end p.col p.col_end msg
 
 let print_todo ?color:(color="ovl-face1") ?msg:(msg="") p =
   let msg =
