@@ -67,6 +67,7 @@ let string_rep_binding ctr = function
     (Some nm,Ast.MetaPosDecl _) -> print_match ctr nm "Pos"
   | (Some nm,Ast.MetaComDecl _) -> print_match ctr nm "Com"
   | (Some nm,Ast.MetaListlenDecl _) -> print_match ctr nm "Int"
+  | (Some nm,Ast.MetaFreshIdDecl _) -> ""
   | (Some nm,_) (* strings for everything else *) ->
       print_match ctr nm "Str"
   | (None,_) -> ""
@@ -101,6 +102,7 @@ let ast_rep_binding ctr = function
   | (Some nm,Ast.MetaAssignmentOperatorDecl _) ->
       failwith
 	(Printf.sprintf "%s: No AST representation for operator variables" nm)
+  | (Some nm,Ast.MetaPragmaInfoDecl _) -> print_match ctr nm "PragmaInfo"
   | (Some nm,Ast.MetaConstDecl _) -> print_match ctr nm "Expr"
   | (Some nm,Ast.MetaErrDecl _) -> failwith ("not supported: "^nm)
   | (Some nm,Ast.MetaExpDecl _) -> print_match ctr nm "Expr"
@@ -309,10 +311,7 @@ let prepare coccifile code =
   let add_fresh_id_rules prev
       (self, (script_name, lang, params, pos, body)) =
     if lang = "ocaml"
-    then
-      let kind = Ast_cocci.MetaIdDecl (Ast_cocci.NONE, self) in
-      let self = (self, kind) in
-      (script_name, self::params, body) :: prev
+    then (script_name, params, body) :: prev
     else prev in
   let fresh_id_rules =
     List.fold_left add_fresh_id_rules [] !Data.fresh_id_scripts in
@@ -502,9 +501,25 @@ let filter_dep existing_deps (accld, accinc) dep =
 	  (accld, dep::accinc)
 
 let get_dir p =
-  let inclcmd = !Flag.ocamlfind ^" query "^p in
-  let dir = List.hd (Common.cmd_to_list inclcmd) in
-    (dir, p)
+  (* Search user-provided paths first *)
+  let inc =
+    try
+      let res =
+	List.find
+	  (fun dir ->
+	    let path = Printf.sprintf "%s/%s" dir (p ^ ext) in
+	    Sys.file_exists path)
+	  !Flag.ocaml_lib_search_path in
+      Some res
+    with Not_found -> None in
+  match inc with
+  | Some dir -> (dir,p)
+  | None ->
+    (* If not found in user-provided paths, then use ocamlfind *)
+    let inclcmd = !Flag.ocamlfind ^" query "^p in
+    (match Common.cmd_to_list inclcmd with
+    | dir::_ -> (dir,p)
+    | _ -> raise (LinkFailure ((p ^ "("^(p^ext)^")", "Not found"))))
 
 let parse_dep cmifile mlfile depout =
   let empty_deps = ([], "") in
