@@ -74,12 +74,14 @@ type posl = int * int (* line-col, for MetaPosValList, for position variables *)
 type virtual_position = Common.parse_info * int (* character offset *)
  (* with sexp *)
 
+type befaft = Before | After (* reason for fake tok *)
+
 type parse_info =
   (* Present both in ast and list of tokens *)
   | OriginTok of Common.parse_info
   (* Present only in ast and generated after parsing. Used mainly
    * by Julia, to add stuff at virtual places, beginning of func or decl *)
-  | FakeTok of string * virtual_position
+  | FakeTok of string * virtual_position * befaft
   (* Present both in ast and list of tokens.  *)
   | ExpandedTok of Common.parse_info * virtual_position
 
@@ -144,6 +146,8 @@ and 'a wrap3 = 'a * il (* * evotype*)
  *)
 and name =
    | RegularName of string wrap
+   | Operator of bool wrap
+   | QualName of name wrap2 (* the :: separators *) list
    | CppConcatenatedName of (string wrap) wrap2 (* the ## separators *) list
    (* normally only used inside list of things, as in parameters or arguments
     * in which case, cf cpp-manual, it has a special meaning *)
@@ -177,7 +181,7 @@ and name =
  *)
 
 
-and fullType = typeQualifier * typeC
+and fullType = typeQualifier * attribute list * typeC
  and typeC = typeCbis wrap (* todo reput wrap3 *)
 
   and typeCbis =
@@ -189,14 +193,17 @@ and fullType = typeQualifier * typeC
   | Decimal         of constExpression * constExpression option
   | FunctionType    of functionType
 
-  | Enum            of string option * enumType
-  | StructUnion     of structUnion * string option * structType (* new scope *)
+  | EnumDef         of fullType * fullType (*enum_base*) option * enumType
+  | StructUnion     of structUnion * string option * info option (* C++, final *) *
+	base_class wrap2 list (* C++ *) * structType (* new scope *)
 
-  | EnumName        of string
+  | EnumName        of structUnion option * string option
   | StructUnionName of structUnion * string
 
   | TypeName   of name * fullType option (* semantic: filled later *)
 
+  | QualifiedType of fullType option * name 
+
   | FieldType of fullType * name * constExpression option
 
   | ParenType of fullType (* for unparser: *)
@@ -216,9 +223,9 @@ and fullType = typeQualifier * typeC
   | TypeOfExpr of expression
   | TypeOfType of fullType
   | AutoType (* c++ >= 11 *)
+  | TemplateType of fullType * argument wrap2 (* , *) list
 
   (* cppext: IfdefType TODO *)
-
 (* -------------------------------------- *)
      and  baseType = Void
                    | IntType   of intType
@@ -239,19 +246,29 @@ and fullType = typeQualifier * typeC
             and sign = Signed | UnSigned
 
           and floatType = CFloat | CDouble | CLongDouble |
-	                  CFloatComplex | CDoubleComplex | CLongDoubleComplex
+	                  CFloatComplex | CDoubleComplex | CLongDoubleComplex |
+			  CUnknownComplex (* only for parsing *)
 
 
      (* -------------------------------------- *)
-     and structUnion = Struct | Union
+     and structUnion = Struct | Union | Class
      and structType  = field list
          and field =
            | DeclarationField of field_declaration
            (* gccext: *)
            | EmptyField of info
 
+	   (* C++ *)
+	   | FunctionField of definition
+	   | PublicLabel of info list
+	   | ProtectedLabel of info list
+	   | PrivateLabel of info list
+           | ConstructDestructField of c_plus_plus_constructor
+
             (* cppext: *)
-           | MacroDeclField of (string * argument wrap2 list)
+           | MacroDeclField of (string * argument wrap2 list * attribute list)
+                               wrap (* optional ';'*)
+           | MacroDeclFieldInit of (string * argument wrap2 list * attribute list * initialiser)
                                wrap (* optional ';'*)
 
             (* cppext: *)
@@ -268,7 +285,8 @@ and fullType = typeQualifier * typeC
            * can cast into int so enum too, ...
            *)
            and fieldkind =
-             | Simple   of name option * fullType
+               (* storage, attr, and init are for C++; none are currently supported in SmPL *)
+             | Simple   of storage * attribute list * (name * v_init) option * fullType * attribute list (* endattrs *)
              | BitField of name option * fullType *
                  info (* : *) * constExpression
               (* fullType => BitFieldInt | BitFieldUnsigned *)
@@ -287,21 +305,21 @@ and fullType = typeQualifier * typeC
         { p_namei: name option;
           p_register: bool wrap;
           p_type: fullType;
-          p_attr: attribute list;
-          p_midattr: attribute list;
           p_endattr: attribute list;
         }
         (* => (bool (register) * fullType) list * bool *)
 
 
 and typeQualifier = typeQualifierbis wrap
-and typeQualifierbis = {const: bool; volatile: bool}
+and typeQualifierbis = {const: bool; volatile: bool; restrict: bool}
 
 (* gccext: cppext: *)
 and attribute = attributebis wrap
   and attributebis =
     | Attribute of attr_arg
-    | GccAttribute of attr_arg
+    | GccAttribute of argument wrap2 list
+    | CxxAttribute of argument wrap2 list
+    | CxxAttributeUsing of name * argument wrap2 list
 
 and attr_arg = attr_arg_bis wrap
   and attr_arg_bis =
@@ -342,16 +360,17 @@ and expression = (expressionbis * exp_info ref (* semantic: *)) wrap3
   | Unary          of expression * unaryOp
   | Binary         of expression * binaryOp * expression
 
-  | ArrayAccess    of expression * expression
+  | ArrayAccess    of expression * argument wrap2 (* , *) list
 
   (* field ident access *)
   | RecordAccess   of expression * name
   | RecordPtAccess of expression * name
   (* redundant normally, could replace it by DeRef RecordAcces *)
+  | QualifiedAccess of fullType option * name
 
   | SizeOfExpr     of expression
   | SizeOfType     of fullType
-  | Cast           of fullType * attribute list * expression
+  | Cast           of fullType * expression
 
   (* gccext: *)
   | StatementExpr of compound wrap (* ( )     new scope *)
@@ -361,14 +380,14 @@ and expression = (expressionbis * exp_info ref (* semantic: *)) wrap3
   | ParenExpr of expression
 
   (* for C++: *)
-  | New of (argument wrap2 (* , *) list) option * argument
+  | New of (argument wrap2 (* , *) list) option * fullType * (argument wrap2 (* , *) list) option
   | Delete of bool (* true if [] *) * expression
+  | TemplateInst of expression * argument wrap2 list
 
   (* CPP [defined] operator, e.g. #if defined(A) *)
   | Defined of name
 
   (* cppext: IfdefExpr TODO *)
-
   (* cppext: normally just expression *)
   and argument = (expression, weird_argument) Common.either
    and weird_argument =
@@ -507,7 +526,11 @@ and statement = statementbis wrap3
 
   and exprStatement = expression option
 
-  and declOrExpr = ForDecl of declaration | ForExp of expression option wrap
+  and declOrExpr =
+      ForDecl of declaration * exprStatement wrap * exprStatement wrap
+    | ForExp of
+	expression option wrap * exprStatement wrap * exprStatement wrap
+    | ForRange of declaration * initialiser
 
  (* for Switch, need check that all elements in the compound start
   * with a case:, otherwise unreachable code.
@@ -519,13 +542,13 @@ and statement = statementbis wrap3
    | Ifdef_Ite of expression * statement * statement
    (* #ifdef A if e S1 else #else S2 #endif S3 *)
    | Ifdef_Ite2 of expression * statement * statement * statement
+   | TryCatch of statement * (parameterType * statement) wrap list
 
 
   and iteration     =
     | While   of expression * statement
     | DoWhile of statement * expression
-    | For     of declOrExpr * exprStatement wrap * exprStatement wrap *
-                 statement
+    | For     of declOrExpr * statement
     (* cppext: *)
     | MacroIteration of string * argument wrap2 list * statement
 
@@ -569,10 +592,10 @@ and declaration =
   (* cppext: *)
     (* bool is true if there is a ; at the end *)
   | MacroDecl of
-      (storagebis * string * argument wrap2 list * attribute list * bool)
+      (storagebis * attribute list * string * argument wrap2 list * attribute list * bool)
         wrap (* fakestart *)
   | MacroDeclInit of
-      (storagebis * string * argument wrap2 list * initialiser)
+      (storagebis * attribute list * string * argument wrap2 list * attribute list * initialiser)
 	wrap (* fakestart *)
 
      and onedecl =
@@ -585,15 +608,14 @@ and declaration =
          v_storage: storage;
          v_local: local_decl; (* cocci: *)
          v_attr: attribute list; (* gccext: *)
-         v_midattr: attribute list; (* gccext: *)
          v_endattr: attribute list; (* gccext: *)
        }
      and v_init =
-       NoInit | ValInit of info * initialiser
-     | ConstrInit of argument wrap2 (* , *) list wrap
-     and storage       = storagebis * bool (* gccext: inline or not *)
+       NoInit | ValInit of initialiser wrap
+     and storage       = storagebis * bool (* gccext: inline or not *) * align
      and storagebis    = NoSto | StoTypedef | Sto of storageClass
      and storageClass  = Auto  | Static | Register | Extern
+     and align         = NoAlign | Align of argument
 
      and local_decl = LocalDecl | NotLocalDecl
 
@@ -604,6 +626,7 @@ and declaration =
        and initialiserbis =
           | InitExpr of expression
           | InitList of initialiser wrap2 (* , *) list
+          | InitListNoBrace of initialiser wrap2 (* , *) list (* only for #define *)
           (* gccext: *)
           | InitDesignators of designator list * initialiser
           | InitFieldOld  of string * initialiser
@@ -634,13 +657,38 @@ and definition = definitionbis wrap (* ( ) { } fakestart sto *)
   { f_name: name;
     f_type: functionType; (* less? a functionType2 ? *)
     f_storage: storage;
+    f_constr_inherited: expression wrap2 list;
     f_body: compound;
-    f_attr: attribute list; (* gccext: *)
     f_endattr: attribute list; (* gccext: *)
     f_old_c_style: declaration list option;
   }
   (* cppext: IfdefFunHeader TODO *)
 
+and c_plus_plus_constructor = c_plus_plus_constructorbis wrap
+and c_plus_plus_constructorbis =
+  | ConstructorDecl of bool wrap *
+        string * (parameterType wrap2 list * bool wrap) * bool wrap
+  | DestructorDecl  of bool wrap *
+        string * (parameterType wrap2 list * bool wrap) * bool wrap
+  | ConstructorDef  of bool wrap *
+        string * (parameterType wrap2 list * bool wrap) *
+	(constr_init wrap2 (* , *) list) wrap * bool wrap * compound
+  | DestructorDef   of bool wrap *
+        string * (parameterType wrap2 list * bool wrap) * bool wrap * compound
+
+and constr_init = (name * argument wrap2 (* , *) list) wrap
+
+(* ------------------------------------------------------------------------- *)
+(* C++ base classes *)
+(* ------------------------------------------------------------------------- *)
+
+and base_class = base_class_bis wrap
+  and base_class_bis =
+    ClassName of name
+  | CPublic of name
+  | CProtected of name
+  | CPrivate of name
+
 (* ------------------------------------------------------------------------- *)
 (* cppext: cpp directives, #ifdef, #define and #include body *)
 (* ------------------------------------------------------------------------- *)
@@ -650,6 +698,10 @@ and cpp_directive =
   | Pragma of (name * string wrap list) wrap
   | OtherDirective of il
 (*| Ifdef ? no, ifdefs are handled differently, cf ifdef_directive below *)
+  (* C++ *)
+  | UsingTypename of (name * fullType) wrap
+  | UsingNamespace of name wrap
+  | UsingMember of name wrap
 
 and define = string wrap (* #define s eol *) * (define_kind * define_val)
    and define_kind =
@@ -662,6 +714,7 @@ and define = string wrap (* #define s eol *) * (define_kind * define_val)
 
      | DefineStmt of statement
      | DefineType of fullType
+     | DefineAttr of attribute list
      | DefineDoWhileZero of (statement * expression) wrap (* do { } while(0) *)
 
      | DefineFunction of definition
@@ -797,6 +850,12 @@ and toplevel =
 
   (* c++ *)
   | Namespace of toplevel list * il
+  | TemplateDefinition of templateParameterType wrap2 list * toplevel * il
+
+and templateParameterType =
+    TypenameOrClassParam of (name * fullType option) wrap
+  | VarNameParam of (fullType * name * initialiser option) wrap
+  | TemplateParam of (templateParameterType wrap2 list * templateParameterType) wrap
 
 (* ------------------------------------------------------------------------- *)
 and program = toplevel list
@@ -808,33 +867,40 @@ and program = toplevel list
  * we need to decorate each token with some cocci code AND the environment
  * for this cocci code.
  *)
+and newlines = Keep | Compress
+
 and metavars_binding = (Ast_cocci.meta_name, metavar_binding_kind) assoc
   and metavar_binding_kind =
   | MetaIdVal        of string
   | MetaFuncVal      of string
   | MetaLocalFuncVal of string
 
-  | MetaExprVal      of expression (* a "clean expr" *) *
+  | MetaExprVal      of expression (* a "clean expr" *) * expression (* original expr *) *
 	                (*subterm constraints, currently exprs*)
 	                Ast_cocci.meta_name list * stripped
-  | MetaExprListVal  of argument wrap2 list
-  | MetaParamVal     of parameterType
-  | MetaParamListVal of parameterType wrap2 list
+  | MetaExprListVal  of argument wrap2 list * argument wrap2 list
+  | MetaParamVal     of parameterType * parameterType
+  | MetaParamListVal of parameterType wrap2 list * parameterType wrap2 list
+  | MetaTemplateParamVal
+    of templateParameterType * templateParameterType
+  | MetaTemplateParamListVal
+    of templateParameterType wrap2 list * templateParameterType wrap2 list
 
-  | MetaTypeVal      of fullType
-  | MetaInitVal      of initialiser
-  | MetaInitListVal  of initialiser wrap2 list
+  | MetaTypeVal      of fullType * fullType
+  | MetaInitVal      of initialiser * initialiser
+  | MetaInitListVal  of newlines * initialiser wrap2 list * initialiser wrap2 list
   | MetaDeclVal      of declaration * declaration
-  | MetaFieldVal     of field
-  | MetaFieldListVal of field list
+  | MetaFieldVal     of field * field
+  | MetaFieldListVal of field list * field list
   | MetaStmtVal      of statement * statement * stripped
-  | MetaStmtListVal  of statement_sequencable list * stripped
+  | MetaStmtListVal  of statement_sequencable list * statement_sequencable list * stripped
   | MetaDParamListVal of (string wrap) wrap2 list
   | MetaFmtVal       of string_format
-  | MetaAttrArgVal   of attr_arg
+  | MetaAttrArgVal   of attr_arg * attr_arg
   | MetaFragListVal  of string_fragment list
   | MetaAssignOpVal  of assignOp
   | MetaBinaryOpVal  of binaryOp
+  | MetaPragmaInfoVal of info
   (* Could also be in Lib_engine.metavars_binding2 with the ParenVal,
    * because don't need to have the value for a position in the env of
    * a '+'. But ParenVal or LabelVal are used only by CTL, they are not
@@ -897,7 +963,7 @@ and com = comment list ref
 (*****************************************************************************)
 (* Some constructors *)
 (*****************************************************************************)
-let nullQualif = ({const=false; volatile= false}, [])
+let nullQualif = ({const=false; volatile= false; restrict= false}, [])
 let nQ = nullQualif
 
 let defaultInt = (BaseType (IntType (Si (Signed, CInt))))
@@ -942,17 +1008,41 @@ let noInIfdef () =
  * an existing C element.
  * old: or when don't want 'synchronize' on it in unparse_c.ml
  * (now have other mark for the matter).
+ * Position of fake tokens must be unique within a file so that
+ * things added to fake toks are not added at multiple places.
  *)
-let no_virt_pos = ({str="";charpos=0;line=0;column=0;file=""},-1)
+let nonpos = ref 0
+let reset_nonpos _ = nonpos := !nonpos - 1
+let no_virt_pos _ =
+  nonpos := !nonpos - 1;
+  ({str="";charpos=0;line=0;column=0;file=""},!nonpos)
 
-let fakeInfo pi  =
-  { pinfo = FakeTok ("",no_virt_pos);
+(* for types only *)
+let no_virt_pos_no_pos _ =
+  ({str="";charpos=0;line=0;column=0;file=""},0)
+
+let fakeInfo befaft =
+  { pinfo = FakeTok ("",no_virt_pos(),befaft);
     cocci_tag = ref emptyAnnot;
     annots_tag = Token_annot.empty;
     comments_tag = ref emptyComments;
     danger = ref NoDanger;
   }
 
+let fakeInfoNoPos befaft =
+  { pinfo = FakeTok ("",no_virt_pos_no_pos(),befaft);
+    cocci_tag = ref emptyAnnot;
+    annots_tag = Token_annot.empty;
+    comments_tag = ref emptyComments;
+    danger = ref NoDanger;
+  }
+
+(* before fake tokens can be moved over whitespace in unparser.
+after fake tokens don't move, stay at the end of something. *)
+let fakeBeforeInfo _ = fakeInfo Before
+let fakeAfterInfo _  = fakeInfo After
+let fakeAfterInfoNoPos _  = fakeInfoNoPos After
+
 let noii = []
 let noattr = []
 let noi_content = (None: ((Common.filename * program) option))
@@ -967,8 +1057,8 @@ let unwrap2 = fst
 let unwrap_expr ((unwrap_e, typ), iie) = unwrap_e
 let rewrap_expr ((_old_unwrap_e, typ), iie)  newe = ((newe, typ), iie)
 
-let unwrap_typeC (qu, (typeC, ii)) = typeC
-let rewrap_typeC (qu, (typeC, ii)) newtypeC  = (qu, (newtypeC, ii))
+let unwrap_typeC (qu, attr, (typeC, ii)) = typeC
+let rewrap_typeC (qu, attr, (typeC, ii)) newtypeC  = (qu, attr, (newtypeC, ii))
 
 let unwrap_typeCbis (typeC, ii) = typeC
 
@@ -978,7 +1068,7 @@ let unwrap_st (unwrap_st, ii) = unwrap_st
 let mk_e unwrap_e ii = (unwrap_e, noType()), ii
 let mk_e_bis unwrap_e ty ii = (unwrap_e, ty), ii
 
-let mk_ty typeC ii = nQ, (typeC, ii)
+let mk_ty typeC ii = nQ, [], (typeC, ii)
 let mk_tybis typeC ii = (typeC, ii)
 
 let mk_st unwrap_st ii = (unwrap_st, ii)
@@ -989,7 +1079,7 @@ let get_ii_st_take_care (st, ii) = ii
 let get_ii_expr_take_care (e, ii) = ii
 
 let get_st_and_ii (st, ii) = st, ii
-let get_ty_and_ii (qu, (typeC, ii)) = qu, (typeC, ii)
+let get_ty_and_ii (qu, attrs, (typeC, ii)) = (typeC, ii)
 let get_e_and_ii  (e, ii) = e, ii
 
 
@@ -1015,7 +1105,7 @@ let rewrap_str s ii =
     (match ii.pinfo with
       OriginTok pi -> OriginTok { pi with Common.str = s;}
     | ExpandedTok (pi,vpi) -> ExpandedTok ({ pi with Common.str = s;},vpi)
-    | FakeTok (_,vpi) -> FakeTok (s,vpi)
+    | FakeTok (_,vpi,befaft) -> FakeTok (s,vpi,befaft)
     | AbstractLineTok pi -> OriginTok { pi with Common.str = s;})}
 
 let rewrap_charpos charpos ii =
@@ -1024,7 +1114,7 @@ let rewrap_charpos charpos ii =
       OriginTok pi -> OriginTok { pi with Common.charpos = charpos;}
     | ExpandedTok (pi,vpi) ->
 	ExpandedTok ({ pi with Common.charpos = charpos;},vpi)
-    | FakeTok (s,vpi) -> FakeTok (s,vpi)
+    | FakeTok (s,vpi,befaft) -> FakeTok (s,vpi,befaft)
     | AbstractLineTok pi -> OriginTok { pi with Common.charpos = charpos;})}
 
 let rewrap_col col ii =
@@ -1032,7 +1122,7 @@ let rewrap_col col ii =
     (match ii.pinfo with
       OriginTok pi -> OriginTok { pi with Common.column = col;}
     | ExpandedTok (pi,vpi) -> ExpandedTok ({ pi with Common.column = col;},vpi)
-    | FakeTok (s,vpi) -> FakeTok (s,vpi)
+    | FakeTok (s,vpi,befaft) -> FakeTok (s,vpi,befaft)
     | AbstractLineTok pi -> OriginTok { pi with Common.column = col;})}
 
 let rewrap_pinfo pi ii =
@@ -1044,39 +1134,39 @@ let rewrap_pinfo pi ii =
 let get_pi = function
     OriginTok pi -> pi
   | ExpandedTok (_,(pi,_)) -> pi
-  | FakeTok (_,(pi,_)) -> pi
+  | FakeTok (_,(pi,_),_) -> pi
   | AbstractLineTok pi -> pi
 
 (* original info *)
 let get_opi = function
     OriginTok pi -> pi
   | ExpandedTok (pi,_) -> pi (* diff with get_pi *)
-  | FakeTok (_,_) -> failwith "no position information"
+  | FakeTok (_,_,_) -> failwith "no position information"
   | AbstractLineTok pi -> pi
 
 let str_of_info ii =
   match ii.pinfo with
     OriginTok pi -> pi.Common.str
   | ExpandedTok (pi,_) -> pi.Common.str
-  | FakeTok (s,_) -> s
+  | FakeTok (s,_,_) -> s
   | AbstractLineTok pi -> pi.Common.str
 
 let get_info f ii =
   match ii.pinfo with
     OriginTok pi -> f pi
   | ExpandedTok (_,(pi,_)) -> f pi
-  | FakeTok (_,(pi,_)) -> f pi
+  | FakeTok (_,(pi,_),_) -> f pi
   | AbstractLineTok pi -> f pi
 
 let get_orig_info f ii =
   match ii.pinfo with
     OriginTok pi -> f pi
   | ExpandedTok (pi,_) -> f pi (* diff with get_info *)
-  | FakeTok (_,(pi,_)) -> f pi
+  | FakeTok (_,(pi,_),_) -> f pi
   | AbstractLineTok pi -> f pi
 
 let make_expanded ii =
-  {ii with pinfo = ExpandedTok (get_opi ii.pinfo,no_virt_pos)}
+  {ii with pinfo = ExpandedTok (get_opi ii.pinfo,no_virt_pos())}
 
 let pos_of_info   ii = get_info      (function x -> x.Common.charpos) ii
 let opos_of_info  ii = get_orig_info (function x -> x.Common.charpos) ii
@@ -1092,7 +1182,7 @@ let strloc_of_info ii =
 
 let is_fake ii =
   match ii.pinfo with
-    FakeTok (_,_) -> true
+    FakeTok (_,_,_) -> true
   | _ -> false
 
 let is_origintok ii =
@@ -1106,7 +1196,7 @@ type posrv = Real of Common.parse_info | Virt of virtual_position
 let compare_pos ii1 ii2 =
   let get_pos = function
       OriginTok pi -> Real pi
-    | FakeTok (s,vpi) -> Virt vpi
+    | FakeTok (s,vpi,befaft) -> Virt vpi
     | ExpandedTok (pi,vpi) -> Virt vpi
     | AbstractLineTok pi -> Real pi in (* used for printing *)
   let pos1 = get_pos (pinfo_of_info ii1) in
@@ -1143,7 +1233,7 @@ let info_to_fixpos ii =
     OriginTok pi -> Ast_cocci.Real pi.Common.charpos
   | ExpandedTok (_,(pi,offset)) ->
       Ast_cocci.Virt (pi.Common.charpos,offset)
-  | FakeTok (_,(pi,offset)) ->
+  | FakeTok (_,(pi,offset),_) ->
       Ast_cocci.Virt (pi.Common.charpos,offset)
   | AbstractLineTok pi -> failwith "unexpected abstract"
 
@@ -1332,7 +1422,10 @@ let s_of_inc_file_bis inc_file =
 
 let fieldname_of_fieldkind fieldkind =
   match fieldkind with
-  | Simple (sopt, ft) -> sopt
+  | Simple (storage, attrs, sopt, ft, endattrs) ->
+      (match sopt with
+	None -> None
+      | Some(nm,v) -> Some nm)
   | BitField (sopt, ft, info, expr) -> sopt
 
 
@@ -1347,16 +1440,25 @@ let s_of_attr attr =
       get_attr_name a
   | (GccAttribute a, ii) ->
       let (attr_,_,_,_,_) = Common.tuple_of_list5 ii in
-      str_of_info attr_ ^ "((" ^ (get_attr_name a) ^ "))" in
+      str_of_info attr_ ^ "((" ^ "args... (not supported to show)" ^ "))"
+  | (CxxAttribute a, ii) ->
+      "[[ args... (not supported to show) ]]"
+  | (CxxAttributeUsing (atnm,args), ii) ->
+      "[[ args... (not supported to show) ]]" in
   attr
   +> List.map get_attr
   +> String.concat ","
 
 
 (* ------------------------------------------------------------------------- *)
-let str_of_name ident =
+let rec str_of_name ident =
   match ident with
   | RegularName (s,ii) -> s
+  | Operator(space_needed,ii) ->
+      "operator" ^ (if space_needed then " " else "") ^
+      (String.concat "" (List.map str_of_info (List.tl ii)))
+  | QualName xs ->
+      xs +> List.map (fun (nm,ii) -> str_of_name nm) +> String.concat "##"
   | CppConcatenatedName xs ->
       xs +> List.map (fun (x,iiop) -> unwrap x) +> String.concat "##"
   | CppVariadicName (s, ii) -> "##" ^ s
@@ -1365,13 +1467,18 @@ let str_of_name ident =
         (xs +> List.map (fun ((x,iix), iicomma) -> x) +> String.concat ",") ^
         ")"
 
-let get_s_and_ii_of_name name =
+let rec get_s_and_ii_of_name name =
   match name with
   | RegularName (s, iis) -> s, iis
+  | Operator (space_needed,iis) -> str_of_name name, iis
   | CppIdentBuilder ((s, iis), xs) -> s, iis
   | CppVariadicName (s,iis)  ->
       let (iop, iis) = Common.tuple_of_list2 iis in
       s, [iis]
+  | QualName xs ->
+      (match xs with
+      | [] -> raise (Impossible 60)
+      | (nm,noiiop)::xs -> get_s_and_ii_of_name nm)
   | CppConcatenatedName xs ->
       (match xs with
       | [] -> raise (Impossible 60)
@@ -1401,6 +1508,8 @@ let get_local_ii_of_expr_inlining_ii_of_name e =
       ii @ ii_of_name name
   | RecordPtAccess (e, name), ii ->
       ii @ ii_of_name name
+  | QualifiedAccess (_typ, name), ii -> 
+      ii @ ii_of_name name
   | _, ii -> ii
 
 
@@ -1411,7 +1520,7 @@ let get_local_ii_of_tybis_inlining_ii_of_name ty =
 
 (* the following is used to obtain the argument to LocalVar *)
 let info_of_type ft =
-  let (qu, ty) = ft in
+  let (qu, attr, ty) = ft in
   (* bugfix: because of string->name, the ii can be deeper *)
   let ii = get_local_ii_of_tybis_inlining_ii_of_name ty in
   match ii with
@@ -1459,6 +1568,7 @@ let string_of_toplevel = function
   | NotParsedCorrectly _ -> "NotParsedCorrectly"
   | FinalDef _ -> "FinalDef"
   | Namespace _ -> "Namespace"
+  | TemplateDefinition _ -> "TemplateDefinition"
 
 let string_of_inc_file = function
   | Local lst -> "local://" ^ (String.concat "/" lst)
