@@ -1,8 +1,9 @@
 type posl = int * int
 type virtual_position = Common.parse_info * int
+type befaft = Before | After (* reason for fake tok *)
 type parse_info =
     OriginTok of Common.parse_info
-  | FakeTok of string * virtual_position
+  | FakeTok of string * virtual_position * befaft
   | ExpandedTok of Common.parse_info * virtual_position
   | AbstractLineTok of Common.parse_info
 type danger = DangerStart | DangerEnd | Danger | NoDanger
@@ -19,10 +20,12 @@ and 'a wrap2 = 'a * il
 and 'a wrap3 = 'a * il
 and name =
     RegularName of string wrap
+  | Operator of bool wrap
+  | QualName of name wrap2 (* the :: separators *) list
   | CppConcatenatedName of string wrap wrap2 list
   | CppVariadicName of string wrap
   | CppIdentBuilder of string wrap * string wrap wrap2 list
-and fullType = typeQualifier * typeC
+and fullType = typeQualifier * attribute list * typeC
 and typeC = typeCbis wrap
 and typeCbis =
     NoType
@@ -31,15 +34,19 @@ and typeCbis =
   | Array of constExpression option * fullType
   | Decimal of constExpression * constExpression option
   | FunctionType of functionType
-  | Enum of string option * enumType
-  | StructUnion of structUnion * string option * structType
-  | EnumName of string
+  | EnumDef of fullType * fullType option * enumType
+  | StructUnion of structUnion * string option * info option (* C++, final *) *
+	base_class wrap2 list (* C++ *) * structType (* new scope *)
+  | EnumName of structUnion option * string option
   | StructUnionName of structUnion * string
   | TypeName of name * fullType option
+  | QualifiedType of fullType option * name
   | FieldType of fullType * name * constExpression option
   | ParenType of fullType
   | TypeOfExpr of expression
   | TypeOfType of fullType
+  | AutoType (* c++ >= 11 *)
+  | TemplateType of fullType * argument wrap2 (* , *) list (* C++ *)
 and baseType =
     Void
   | IntType of intType
@@ -52,18 +59,26 @@ and signed = sign * base
 and base = CChar2 | CShort | CInt | CLong | CLongLong
 and sign = Signed | UnSigned
 and floatType =
-    CFloat | CDouble | CLongDouble | CFloatComplex | CDoubleComplex | CLongDoubleComplex
-and structUnion = Struct | Union
+    CFloat | CDouble | CLongDouble | CFloatComplex | CDoubleComplex
+  | CLongDoubleComplex | CUnknownComplex (* only for parsing *)
+and structUnion = Struct | Union | Class
 and structType = field list
 and field =
     DeclarationField of field_declaration
   | EmptyField of info
-  | MacroDeclField of (string * argument wrap2 list) wrap
+  | FunctionField of definition (* C++ *)
+  | PublicLabel of info list (* C++ *)
+  | ProtectedLabel of info list (* C++ *)
+  | PrivateLabel of info list (* C++ *)
+  | ConstructDestructField of c_plus_plus_constructor
+  | MacroDeclField of (string * argument wrap2 list * attribute list) wrap
+  | MacroDeclFieldInit of (string * argument wrap2 list * attribute list * initialiser) wrap
   | CppDirectiveStruct of cpp_directive
   | IfdefStruct of ifdef_directive
 and field_declaration = FieldDeclList of fieldkind wrap2 list wrap
 and fieldkind =
-    Simple of name option * fullType
+    (* storage, attr, and init are for C++; none are currently supported in SmPL *)
+    Simple   of storage * attribute list * (name * v_init) option * fullType * attribute list (* endattrs *)
   | BitField of name option * fullType * info * constExpression
 and enumType = oneEnumType wrap2 list
 and oneEnumType = name * (info * constExpression) option
@@ -72,11 +87,18 @@ and parameterType = {
   p_namei : name option;
   p_register : bool wrap;
   p_type : fullType;
+  p_endattr : attribute list;
 }
 and typeQualifier = typeQualifierbis wrap
-and typeQualifierbis = { const : bool; volatile : bool; }
+and typeQualifierbis = { const : bool; volatile : bool; restrict : bool; }
 and attribute = attributebis wrap
-and attributebis = Attribute of string
+and attributebis =
+    Attribute of attr_arg
+  | GccAttribute of argument wrap2 list
+  | CxxAttribute of argument wrap2 list
+  | CxxAttributeUsing of name * argument wrap2 list
+and attr_arg = attr_arg_bis wrap
+and attr_arg_bis = MacroAttr of string | MacroAttrArgs of string * argument wrap2 (* , *) list
 and expression = (expressionbis * exp_info ref) wrap3
 and exp_info = exp_type option * test
 and exp_type = fullType * local
@@ -97,17 +119,19 @@ and expressionbis =
   | Infix of expression * fixOp
   | Unary of expression * unaryOp
   | Binary of expression * binaryOp * expression
-  | ArrayAccess of expression * expression
+  | ArrayAccess of expression * argument wrap2 (* , *) list
   | RecordAccess of expression * name
   | RecordPtAccess of expression * name
+  | QualifiedAccess of fullType option * name
   | SizeOfExpr of expression
   | SizeOfType of fullType
   | Cast of fullType * expression
   | StatementExpr of compound wrap
   | Constructor of fullType * initialiser
   | ParenExpr of expression
-  | New of argument wrap2 list option * argument
+  | New of (argument wrap2 list) option * fullType * (argument wrap2 list) option
   | Delete of bool (* true if [] *) * expression
+  | TemplateInst of expression * argument wrap2 list
   | Defined of name
 and argument = (expression, weird_argument) Common.either
 and weird_argument = ArgType of parameterType | ArgAction of action_macro
@@ -119,7 +143,7 @@ and constant =
   | Int of (string * intType)
   | Float of (string * floatType)
   | DecimalConst of (string * string * string)
-and isWchar = IsWchar | IsChar
+and isWchar = IsWchar | IsUchar | Isuchar | Isu8char | IsChar
 and unaryOp = GetRef | DeRef | UnPlus | UnMinus | Tilde | Not | GetRefLabel
 and assignOpbis = SimpleAssign | OpAssign of arithOp
 and assignOp = assignOpbis wrap
@@ -173,16 +197,21 @@ and statement_sequencable =
   | IfdefStmt of ifdef_directive
   | IfdefStmt2 of ifdef_directive list * statement_sequencable list list
 and exprStatement = expression option
-and declOrExpr = ForDecl of declaration | ForExp of expression option wrap
+and declOrExpr =
+    ForDecl of declaration * exprStatement wrap * exprStatement wrap
+  | ForExp of
+      expression option wrap * exprStatement wrap * exprStatement wrap
+  | ForRange of declaration * initialiser
 and selection =
     If of expression * statement * statement
   | Switch of expression * statement
   | Ifdef_Ite of expression * statement * statement
   | Ifdef_Ite2 of expression * statement * statement * statement
+  | TryCatch of statement * (parameterType * statement) wrap list
 and iteration =
     While of expression * statement
   | DoWhile of statement * expression
-  | For of declOrExpr * exprStatement wrap * exprStatement wrap * statement
+  | For of declOrExpr * statement
   | MacroIteration of string * argument wrap2 list * statement
 and jump =
     Goto of name
@@ -199,9 +228,13 @@ and exec_code_bis = ExecEval of expression | ExecToken
 and exec_code = exec_code_bis wrap
 and declaration =
     DeclList of onedecl wrap2 list wrap
-  | MacroDecl of (storagebis * string * argument wrap2 list * bool) wrap
+  | MacroDecl of
+      (storagebis * attribute list * string *
+	 argument wrap2 list * attribute list * bool)
+        wrap
   | MacroDeclInit of
-      (storagebis * string * argument wrap2 list * initialiser) wrap
+      (storagebis * attribute list * string *
+	 argument wrap2 list * attribute list * initialiser) wrap
 and onedecl = {
   v_namei : (name * v_init) option;
   v_type : fullType;
@@ -213,16 +246,17 @@ and onedecl = {
 }
 and v_init =
     NoInit
-  | ValInit of info * initialiser
-  | ConstrInit of argument wrap2 list wrap
-and storage = storagebis * bool
+  | ValInit of initialiser wrap
+and storage = storagebis * bool * align
 and storagebis = NoSto | StoTypedef | Sto of storageClass
 and storageClass = Auto | Static | Register | Extern
+and align = NoAlign | Align of argument
 and local_decl = LocalDecl | NotLocalDecl
 and initialiser = initialiserbis wrap
 and initialiserbis =
     InitExpr of expression
   | InitList of initialiser wrap2 list
+  | InitListNoBrace of initialiser wrap2 (* , *) list (* only for #define *)
   | InitDesignators of designator list * initialiser
   | InitFieldOld of string * initialiser
   | InitIndexOld of expression * initialiser
@@ -236,15 +270,42 @@ and definitionbis = {
   f_name : name;
   f_type : functionType;
   f_storage : storage;
+  f_constr_inherited: expression wrap2 list;
   f_body : compound;
-  f_attr : attribute list;
+  f_endattr : attribute list;
   f_old_c_style : declaration list option;
 }
+
+and c_plus_plus_constructor = c_plus_plus_constructorbis wrap
+and c_plus_plus_constructorbis =
+  | ConstructorDecl of bool wrap *
+        string * (parameterType wrap2 list * bool wrap) * bool wrap
+  | DestructorDecl  of bool wrap *
+        string * (parameterType wrap2 list * bool wrap) * bool wrap
+  | ConstructorDef  of bool wrap *
+        string * (parameterType wrap2 list * bool wrap) *
+	(constr_init wrap2 (* , *) list) wrap * bool wrap * compound
+  | DestructorDef   of bool wrap *
+        string * (parameterType wrap2 list * bool wrap) * bool wrap * compound
+
+and constr_init = (name * argument wrap2 (* , *) list) wrap
+
+and base_class = base_class_bis wrap
+  and base_class_bis =
+    ClassName of name
+  | CPublic of name
+  | CProtected of name
+  | CPrivate of name
+
 and cpp_directive =
     Define of define
   | Include of includ
   | Pragma of (name * string wrap list) wrap
   | OtherDirective of il
+  | UsingTypename of (name * fullType) wrap
+  | UsingNamespace of name wrap
+  | UsingMember of name wrap
+
 and define = string wrap * (define_kind * define_val)
 and define_kind =
     DefineVar
@@ -254,6 +315,7 @@ and define_val =
     DefineExpr of expression
   | DefineStmt of statement
   | DefineType of fullType
+  | DefineAttr of attribute list
   | DefineDoWhileZero of (statement * expression) wrap
   | DefineFunction of definition
   | DefineInit of initialiser
@@ -286,7 +348,7 @@ and ifdefkind =
 and ifdef_guard =
     Gifdef of macro_symbol
   | Gifndef of macro_symbol
-  | Gif_str of string
+  | Gif_str of Lexing.position * string
   | Gif of expression
   | Gnone
 and macro_symbol = string
@@ -301,30 +363,42 @@ and toplevel =
   | NotParsedCorrectly of il
   | FinalDef of info
   | Namespace of toplevel list * il
+  | TemplateDefinition of templateParameterType wrap2 list * toplevel * il
+and templateParameterType =
+    TypenameOrClassParam of (name * fullType option) wrap
+  | VarNameParam of (fullType * name * initialiser option) wrap
+  | TemplateParam of (templateParameterType wrap2 list * templateParameterType) wrap
 and program = toplevel list
 and metavars_binding =
     (Ast_cocci.meta_name, metavar_binding_kind) Common.assoc
+and newlines = Keep | Compress
 and metavar_binding_kind =
     MetaIdVal of string
   | MetaFuncVal of string
   | MetaLocalFuncVal of string
-  | MetaExprVal of expression * Ast_cocci.meta_name list * stripped
-  | MetaExprListVal of argument wrap2 list
-  | MetaParamVal of parameterType
-  | MetaParamListVal of parameterType wrap2 list
-  | MetaTypeVal of fullType
-  | MetaInitVal of initialiser
-  | MetaInitListVal of initialiser wrap2 list
+  | MetaExprVal of expression * expression * Ast_cocci.meta_name list * stripped
+  | MetaExprListVal of argument wrap2 list * argument wrap2 list
+  | MetaParamVal of parameterType * parameterType
+  | MetaParamListVal of parameterType wrap2 list * parameterType wrap2 list
+  | MetaTemplateParamVal
+    of templateParameterType * templateParameterType
+  | MetaTemplateParamListVal
+    of templateParameterType wrap2 list * templateParameterType wrap2 list
+  | MetaTypeVal of fullType * fullType
+  | MetaInitVal of initialiser * initialiser
+  | MetaInitListVal of newlines * initialiser wrap2 list * initialiser wrap2 list
   | MetaDeclVal of declaration * declaration
-  | MetaFieldVal of field
-  | MetaFieldListVal of field list
+  | MetaFieldVal of field * field
+  | MetaFieldListVal of field list * field list
   | MetaStmtVal of statement * statement * stripped
-  | MetaStmtListVal of statement_sequencable list * stripped
+  | MetaStmtListVal of statement_sequencable list * statement_sequencable list * stripped
   | MetaDParamListVal of string wrap wrap2 list
   | MetaFmtVal of string_format
+  | MetaAttrArgVal of attr_arg * attr_arg
   | MetaFragListVal of string_fragment list
   | MetaAssignOpVal of assignOp
   | MetaBinaryOpVal of binaryOp
+  | MetaPragmaInfoVal of info
   | MetaPosVal of (Ast_cocci.fixpos * Ast_cocci.fixpos)
   | MetaPosValList of
       (Common.filename * string * (posl * posl) option * posl * posl) list
@@ -358,8 +432,11 @@ val mcode_and_env_of_cocciref :
 val emptyComments : comments_around
 val noRelPos : unit -> include_rel_pos option ref
 val noInIfdef : unit -> bool ref
-val no_virt_pos : Common.parse_info * int
-val fakeInfo : 'a -> info
+val reset_nonpos : unit -> unit
+val fakeInfo : befaft -> info
+val fakeBeforeInfo : unit -> info
+val fakeAfterInfo : unit -> info
+val fakeAfterInfoNoPos : unit -> info
 val noii : 'a list
 val noattr : 'a list
 val noi_content : (Common.filename * program) option
@@ -367,20 +444,20 @@ val unwrap : 'a * 'b -> 'a
 val unwrap2 : 'a * 'b -> 'a
 val unwrap_expr : ('a * 'b) * 'c -> 'a
 val rewrap_expr : ('a * 'b) * 'c -> 'd -> ('d * 'b) * 'c
-val unwrap_typeC : 'a * ('b * 'c) -> 'b
-val rewrap_typeC : 'a * ('b * 'c) -> 'd -> 'a * ('d * 'c)
-val unwrap_typeCbis : 'a * 'b -> 'a
+val unwrap_typeC : fullType -> typeCbis
+val rewrap_typeC : fullType -> typeCbis -> fullType
+val unwrap_typeCbis : typeC -> typeCbis
 val unwrap_st : 'a * 'b -> 'a
 val mk_e : 'a -> 'b -> ('a * ('c option * test) ref) * 'b
 val mk_e_bis : 'a -> 'b -> 'c -> ('a * 'b) * 'c
-val mk_ty : 'a -> 'b -> (typeQualifierbis * 'c list) * ('a * 'b)
+val mk_ty : 'a -> 'b -> (typeQualifierbis * 'c list) * attribute list * ('a * 'b)
 val mk_tybis : 'a -> 'b -> 'a * 'b
 val mk_st : 'a -> 'b -> 'a * 'b
 val get_ii_typeC_take_care : 'a * 'b -> 'b
 val get_ii_st_take_care : 'a * 'b -> 'b
 val get_ii_expr_take_care : 'a * 'b -> 'b
 val get_st_and_ii : 'a * 'b -> 'a * 'b
-val get_ty_and_ii : 'a * ('b * 'c) -> 'a * ('b * 'c)
+val get_ty_and_ii : 'a * attribute list * ('b * 'c) -> 'b * 'c
 val get_e_and_ii : 'a * 'b -> 'a * 'b
 val get_type_expr : ('a * 'b ref) * 'c -> 'b
 val set_type_expr : ('a * 'b ref) * 'c -> 'b -> unit
@@ -428,7 +505,7 @@ val unsplit_nocomma : ('a, il) Common.either list -> 'a list
 val s_of_inc_file : inc_file -> string
 val s_of_inc_file_bis : inc_file -> string
 val fieldname_of_fieldkind : fieldkind -> name option
-val s_of_attr : (attributebis * 'a) list -> string
+val s_of_attr : (attributebis * info list) list -> string
 val str_of_name : name -> string
 val get_s_and_ii_of_name : name -> string * il
 val get_s_and_info_of_name : name -> string * info
@@ -437,7 +514,7 @@ val ii_of_name : name -> il
 val get_local_ii_of_expr_inlining_ii_of_name :
   (expressionbis * 'a) * il -> il
 val get_local_ii_of_tybis_inlining_ii_of_name : typeCbis * il -> il
-val info_of_type : 'a * (typeCbis * il) -> parse_info option
+val info_of_type : 'a * attribute list * (typeCbis * il) -> parse_info option
 val get_local_ii_of_st_inlining_ii_of_name :
   statementbis * info list -> info list
 val name_of_parameter : parameterType -> string option
