@@ -22,12 +22,12 @@ type node = node1 * string (* For debugging. Used by print_graph *)
 
   | IfHeader      of statement * expression wrap
   | Else          of info
+  | TryHeader     of statement * info
+  | CatchHeader   of parameterType wrap
   | WhileHeader   of statement * expression wrap
   | DoHeader      of statement * info
   | DoWhileTail   of expression wrap
-  | ForHeader     of statement *
-                 (declOrExpr * exprStatement wrap * exprStatement wrap)
-                 wrap
+  | ForHeader     of statement * declOrExpr wrap
   | SwitchHeader  of statement * expression wrap
   | MacroIterHeader of statement * (string * argument wrap2 list) wrap
 
@@ -52,13 +52,16 @@ type node = node1 * string (* For debugging. Used by print_graph *)
 
   | DefineExpr of expression
   | DefineType of fullType
+  | DefineAttr of attribute list
   | DefineDoWhileZeroHeader of unit wrap
+  | DefineInit of initialiser
 
   | DefineTodo
 
   | Include of includ
 
-  | PragmaHeader of string wrap * pragmainfo
+  | TemplateHeader of templateParameterType wrap2 list wrap
+  | CppTop of cpp_directive
 
   | MacroTop of string * argument wrap2 list * il
 
@@ -77,11 +80,13 @@ type node = node1 * string (* For debugging. Used by print_graph *)
 
   | Asm of statement * asmbody wrap
   | MacroStmt of statement * unit wrap
+  | NestedFunc of statement * definition wrap
 
   | Exec of statement * exec_code list wrap
 
   (* ------------------------ *)
   | Enter
+  | PreExit of info
   | Exit
   | Fake
   | CaseNode of int
@@ -94,6 +99,7 @@ type node = node1 * string (* For debugging. Used by print_graph *)
   | AfterNode of after_type
   | FallThroughNode
   | LoopFallThroughNode
+  | CatchExit
   | ErrorExit
 
 and after_type =
