@@ -68,6 +68,8 @@ let strip_info_visitor _ =
   }
 
 let al_expr      x = Visitor_c.vk_expr_s      (strip_info_visitor()) x
+let al_assignOp  x = Visitor_c.vk_assignOp_s  (strip_info_visitor()) x
+let al_binaryOp  x = Visitor_c.vk_binaryOp_s  (strip_info_visitor()) x
 let al_declaration x = Visitor_c.vk_decl_s    (strip_info_visitor()) x
 let al_field     x = Visitor_c.vk_struct_field_s (strip_info_visitor()) x
 let al_statement x = Visitor_c.vk_statement_s (strip_info_visitor()) x
@@ -78,6 +80,8 @@ let al_init      x = Visitor_c.vk_ini_s       (strip_info_visitor()) x
 let al_inits     x = Visitor_c.vk_inis_s      (strip_info_visitor()) x
 let al_param     x = Visitor_c.vk_param_s     (strip_info_visitor()) x
 let al_params    x = Visitor_c.vk_params_s    (strip_info_visitor()) x
+let al_template_param  x = Visitor_c.vk_template_param_s     (strip_info_visitor()) x
+let al_template_params x = Visitor_c.vk_template_params_s    (strip_info_visitor()) x
 let al_define_params x =
   Visitor_c.vk_define_params_s (strip_info_visitor()) x
 let al_arguments x = Visitor_c.vk_arguments_s (strip_info_visitor()) x
@@ -92,6 +96,7 @@ let al_attr_arg  x = Visitor_c.vk_attr_arg_s  (strip_info_visitor()) x
 let al_node      x = Visitor_c.vk_node_s      (strip_info_visitor()) x
 
 let al_program  x = List.map (Visitor_c.vk_toplevel_s (strip_info_visitor())) x
+let al_info  x = Visitor_c.vk_info_s (strip_info_visitor()) x
 let al_ii    x = Visitor_c.vk_ii_s (strip_info_visitor()) x
 
 
@@ -99,14 +104,7 @@ let al_ii    x = Visitor_c.vk_ii_s (strip_info_visitor()) x
 
 let strip_inh_info_visitor _ =  (* for inherited metavariables *)
   let drop_test_lv ty bigf =
-    let (ty,_) = !ty in
-    let ty =
-      match ty with
-	None -> None
-      |	Some (ty,_) ->
-	  let ty = Visitor_c.vk_type_s bigf ty in
-	  Some (ty,Ast_c.NotLocalVar) in
-    ref ((ty,Ast_c.NotTest) : Ast_c.exp_info) in
+    ref ((None,Ast_c.NotTest) : Ast_c.exp_info) in
 
   { Visitor_c.default_visitor_c_s with
     Visitor_c.kinfo_s =
@@ -117,7 +115,8 @@ let strip_inh_info_visitor _ =  (* for inherited metavariables *)
 
     Visitor_c.kexpr_s = (fun (k,bigf) e ->
       let (e', ty), ii' = k e in
-      (e', drop_test_lv ty bigf), ii' (* keep type, but process it - jll *)
+      (* drop type, which may be affected by a previous change *)
+      (e', drop_test_lv ty bigf), ii'
     );
 
 (*
@@ -151,6 +150,8 @@ let al_inh_string_fragments x =
   Visitor_c.vk_string_fragments_s (strip_inh_info_visitor()) x
 let al_inh_attribute x = Visitor_c.vk_attribute_s (strip_inh_info_visitor()) x
 let al_inh_attr_arg  x = Visitor_c.vk_attr_arg_s  (strip_inh_info_visitor()) x
+let al_inh_info      x = Visitor_c.vk_info_s (strip_inh_info_visitor()) x
+let al_inh_ii        x = Visitor_c.vk_ii_s (strip_inh_info_visitor()) x
 
 
 
@@ -170,6 +171,8 @@ let semi_strip_info_visitor = (* keep position information *)
   }
 
 let semi_al_expr      = Visitor_c.vk_expr_s      semi_strip_info_visitor
+let semi_al_assignOp  = Visitor_c.vk_assignOp_s  semi_strip_info_visitor
+let semi_al_binaryOp  = Visitor_c.vk_binaryOp_s  semi_strip_info_visitor
 let semi_al_declaration = Visitor_c.vk_decl_s    semi_strip_info_visitor
 let semi_al_field = Visitor_c.vk_struct_field_s  semi_strip_info_visitor
 let semi_al_fields = Visitor_c.vk_struct_fields_s semi_strip_info_visitor
@@ -181,6 +184,8 @@ let semi_al_init      = Visitor_c.vk_ini_s       semi_strip_info_visitor
 let semi_al_inits     = Visitor_c.vk_inis_s      semi_strip_info_visitor
 let semi_al_param     = Visitor_c.vk_param_s     semi_strip_info_visitor
 let semi_al_params    = Visitor_c.vk_params_s    semi_strip_info_visitor
+let semi_al_template_param  = Visitor_c.vk_template_param_s     semi_strip_info_visitor
+let semi_al_template_params = Visitor_c.vk_template_params_s    semi_strip_info_visitor
 let semi_al_define_params =
   Visitor_c.vk_define_params_s semi_strip_info_visitor
 let semi_al_arguments = Visitor_c.vk_arguments_s semi_strip_info_visitor
@@ -190,6 +195,7 @@ let semi_al_string_fragments =
   Visitor_c.vk_string_fragments_s semi_strip_info_visitor
 let semi_al_attribute = Visitor_c.vk_attribute_s semi_strip_info_visitor
 let semi_al_attr_arg  = Visitor_c.vk_attr_arg_s semi_strip_info_visitor
+let semi_al_info  = Visitor_c.vk_info_s semi_strip_info_visitor
 
 let semi_al_program =
   List.map (Visitor_c.vk_toplevel_s semi_strip_info_visitor)
@@ -307,6 +313,8 @@ let ii_of_ini  = extract_info_visitor Visitor_c.vk_ini
 let ii_of_inis  = extract_info_visitor Visitor_c.vk_inis_splitted
 let ii_of_param = extract_info_visitor Visitor_c.vk_param
 let ii_of_params = extract_info_visitor Visitor_c.vk_params_splitted
+let ii_of_template_param = extract_info_visitor Visitor_c.vk_template_param
+let ii_of_template_params = extract_info_visitor Visitor_c.vk_template_params_splitted
 let ii_of_enum_fields = extract_info_visitor Visitor_c.vk_enum_fields_splitted
 let ii_of_struct_fields = extract_info_visitor Visitor_c.vk_struct_fields
 (*let ii_of_struct_field = extract_info_visitor Visitor_c.vk_struct_field*)
@@ -363,7 +371,7 @@ let info_to_fixpos ii =
     Ast_c.OriginTok pi -> Ast_cocci.Real pi.Common.charpos
   | Ast_c.ExpandedTok (_,(pi,offset)) ->
       Ast_cocci.Virt (pi.Common.charpos,offset)
-  | Ast_c.FakeTok (_,(pi,offset)) ->
+  | Ast_c.FakeTok (_,(pi,offset),_) ->
       Ast_cocci.Virt (pi.Common.charpos,offset)
   | Ast_c.AbstractLineTok pi ->
       failwith ("unexpected abstract: "^(Dumper.dump pi))
