@@ -2,6 +2,8 @@ val pr2 : string -> unit
 val pr2_once : string -> unit
 val strip_info_visitor : 'a -> Visitor_c.visitor_c_s
 val al_expr : Ast_c.expression -> Ast_c.expression
+val al_assignOp : Ast_c.assignOp -> Ast_c.assignOp
+val al_binaryOp : Ast_c.binaryOp -> Ast_c.binaryOp
 val al_declaration : Ast_c.declaration -> Ast_c.declaration
 val al_field : Ast_c.field -> Ast_c.field
 val al_statement : Ast_c.statement -> Ast_c.statement
@@ -15,6 +17,11 @@ val al_param : Ast_c.parameterType -> Ast_c.parameterType
 val al_params :
   Ast_c.parameterType Ast_c.wrap2 list ->
   Ast_c.parameterType Ast_c.wrap2 list
+val al_template_param :
+    Ast_c.templateParameterType -> Ast_c.templateParameterType
+val al_template_params :
+  Ast_c.templateParameterType Ast_c.wrap2 list ->
+  Ast_c.templateParameterType Ast_c.wrap2 list
 val al_define_params :
   string Ast_c.wrap Ast_c.wrap2 list -> string Ast_c.wrap Ast_c.wrap2 list
 val al_arguments :
@@ -28,6 +35,7 @@ val al_string_fragments :
   Ast_c.string_fragment list -> Ast_c.string_fragment list
 val al_node : Control_flow_c.node -> Control_flow_c.node
 val al_program : Ast_c.toplevel list -> Ast_c.toplevel list
+val al_info : Ast_c.info -> Ast_c.info
 val al_ii : Ast_c.info list -> Ast_c.info list
 val strip_inh_info_visitor : 'a -> Visitor_c.visitor_c_s
 val al_inh_expr : Ast_c.expression -> Ast_c.expression
@@ -48,8 +56,12 @@ val al_inh_string_fragments :
   Ast_c.string_fragment list -> Ast_c.string_fragment list
 val al_inh_attribute : Ast_c.attribute -> Ast_c.attribute
 val al_inh_attr_arg : Ast_c.attr_arg -> Ast_c.attr_arg
+val al_inh_info : Ast_c.info -> Ast_c.info
+val al_inh_ii : Ast_c.info list -> Ast_c.info list
 val semi_strip_info_visitor : Visitor_c.visitor_c_s
 val semi_al_expr : Ast_c.expression -> Ast_c.expression
+val semi_al_assignOp : Ast_c.assignOp -> Ast_c.assignOp
+val semi_al_binaryOp : Ast_c.binaryOp -> Ast_c.binaryOp
 val semi_al_declaration : Ast_c.declaration -> Ast_c.declaration
 val semi_al_field : Ast_c.field -> Ast_c.field
 val semi_al_fields : Ast_c.field list -> Ast_c.field list
@@ -64,6 +76,11 @@ val semi_al_param : Ast_c.parameterType -> Ast_c.parameterType
 val semi_al_params :
   Ast_c.parameterType Ast_c.wrap2 list ->
   Ast_c.parameterType Ast_c.wrap2 list
+val semi_al_template_param :
+    Ast_c.templateParameterType -> Ast_c.templateParameterType
+val semi_al_template_params :
+  Ast_c.templateParameterType Ast_c.wrap2 list ->
+  Ast_c.templateParameterType Ast_c.wrap2 list
 val semi_al_define_params :
   string Ast_c.wrap Ast_c.wrap2 list -> string Ast_c.wrap Ast_c.wrap2 list
 val semi_al_arguments :
@@ -73,6 +90,7 @@ val semi_al_string_fragments :
   Ast_c.string_fragment list -> Ast_c.string_fragment list
 val semi_al_attribute : Ast_c.attribute -> Ast_c.attribute
 val semi_al_attr_arg  : Ast_c.attr_arg -> Ast_c.attr_arg
+val semi_al_info  : Ast_c.info -> Ast_c.info
 val semi_al_program : Ast_c.toplevel list -> Ast_c.toplevel list
 val real_strip_info_visitor : 'a -> Visitor_c.visitor_c_s
 val real_al_expr : Ast_c.expression -> Ast_c.expression
@@ -115,6 +133,9 @@ val ii_of_inis :
 val ii_of_param : Ast_c.parameterType -> Ast_c.info list
 val ii_of_params :
   (Ast_c.parameterType, Ast_c.il) Common.either list -> Ast_c.info list
+val ii_of_template_param : Ast_c.templateParameterType -> Ast_c.info list
+val ii_of_template_params :
+  (Ast_c.templateParameterType, Ast_c.il) Common.either list -> Ast_c.info list
 val ii_of_enum_fields :
   (Ast_c.oneEnumType, Ast_c.il) Common.either list -> Ast_c.info list
 val ii_of_struct_fields : Ast_c.field list -> Ast_c.info list
