@@ -215,9 +215,12 @@ let print_commentized xs =
 	    if newline = !line
 	    then pr2_no_nl (s ^ " ")
 	    else begin
+	      let str =
+		Printf.sprintf "%s:%d: passed:"
+		  pinfo.Common.file newline in
               if !line = -1
-              then pr2_no_nl "passed:"
-              else pr2_no_nl "\npassed:";
+              then pr2_no_nl str
+              else pr2_no_nl ("\n"^str);
               line := newline;
               pr2_no_nl (s ^ " ");
 	    end
@@ -375,9 +378,7 @@ let parse_gen ~cpp ~tos parsefunc infos s =
   result
 
 (* Please DO NOT remove this code, even though most of it is not used *)
-let type_of_string s     =
-  let typname = parse_gen ~cpp:false ~tos:true Parser_c.type_name None s in
-  Common.snd typname
+let type_of_string s     = parse_gen ~cpp:false ~tos:true Parser_c.type_name None s
 let statement_of_string  = parse_gen ~cpp:false ~tos:false Parser_c.statement None
 let expression_of_string = parse_gen ~cpp:false ~tos:false Parser_c.expr None
 let cpp_expression_of_string = parse_gen ~cpp:true ~tos:false Parser_c.expr
@@ -606,10 +607,9 @@ let rec lexer_function ~pass tr = fun lexbuf ->
           (* typedef_fix1 *)
           let v = match v with
             | Parser_c.TIdent (s, ii) ->
-                if
-                  LP.is_typedef s &&
-                    not (!Flag_parsing_c.disable_add_typedef) &&
-                    pass = 1
+                if LP.is_typedef s &&
+                  not (!Flag_parsing_c.disable_add_typedef) &&
+                  pass = 1
                 then Parser_c.TypedefIdent (s, ii)
                 else Parser_c.TIdent (s, ii)
             | x -> x
@@ -621,9 +621,26 @@ let rec lexer_function ~pass tr = fun lexbuf ->
 	    if !in_exec
 	    then v
 	    else
-	      Parsing_hacks.lookahead ~pass
-		(clean_for_lookahead (v::tr.rest_clean))
-		passed_before in
+	      let other _ =
+		Parsing_hacks.lookahead ~pass
+		  (clean_for_lookahead (v::tr.rest_clean))
+		  passed_before in
+	      match v with
+		Parser_c.TIdent(s,ii) ->
+		  let res =
+		    try Some(Data.get_special_name s)
+		    with _ -> None in
+		  (match res with
+		    Some Data.Attr -> Parser_c.TMacroAttr(s,ii)
+		  | Some Data.AttrArgs -> Parser_c.TMacroAttrArgs(s,ii)
+		  | Some Data.Declarer -> Parser_c.TMacroDecl(s, ii)
+		  | Some Data.Iterator -> Parser_c.TMacroIterator(s, ii)
+		  | Some Data.CommaInit ->
+		      tr.rest <- Parsing_hacks.fix_comma_init ii tr.rest;
+		      tr.rest_clean <- Parsing_hacks.fix_comma_init ii tr.rest_clean;
+		      v
+		  | _ -> other())
+	      | _ -> other() in
 
           tr.passed <- v::tr.passed;
 
@@ -677,7 +694,7 @@ let get_one_elem ~pass tr =
       let current = tr.current in
       (*  error recovery, go to next synchro point *)
       let (passed', rest') =
-	Parsing_recovery_c.find_next_synchro tr.rest tr.passed in
+	Parsing_recovery_c.find_next_synchro ~next:tr.rest ~already_passed:tr.passed in
       tr.rest <- rest';
       tr.passed <- passed';
 
@@ -965,6 +982,8 @@ let tree_stack = ref []
 let seen_files = ref []
 
 let normalize file =
+  let file =
+    String.concat "/" (Str.split (Str.regexp_string "/./") file) in
   let pieces = Str.split_delim (Str.regexp "/") file in
   let rec loop prev = function
       [] -> String.concat "/" (List.rev prev)
@@ -1037,7 +1056,8 @@ and handle_include file wrapped_incl k =
     then
       match f with
       | Some header_filename when Common.lfile_exists header_filename ->
-          if not (IC.has_been_parsed header_filename)
+          if not (!Includes.include_headers_for_types) ||
+	     not (IC.has_been_parsed header_filename)
           then
             begin
               IC.add_to_parsed_files header_filename;
@@ -1102,11 +1122,18 @@ and _parse_print_error_heuristic2bis saved_typedefs saved_macros
     if parse_strings
     then Parsing_hacks.fix_tokens_strings toks
     else toks in
+  let toks =
+    if !Flag.c_plus_plus <> Flag.Off
+    then Parsing_hacks.choose_qualtype(Parsing_hacks.convert_templates toks)
+    else toks in
 
+  (* List.iter
+       (fun t -> Printf.eprintf "tok: %s --- %s\n" (TH.str_of_tok t) (TH.string_of_token t))
+       toks; *)
 
   let tr = mk_tokens_state toks in
 
-  let rec loop tr =
+  let rec loop acc tr =
 
     (* todo?: I am not sure that it represents current_line, cos maybe
      * tr.current partipated in the previous parsing phase, so maybe tr.current
@@ -1120,6 +1147,7 @@ and _parse_print_error_heuristic2bis saved_typedefs saved_macros
      *)
     let checkpoint = TH.line_of_tok tr.current in
     let checkpoint_file = TH.file_of_tok tr.current in
+    Ast_c.reset_nonpos();
 
     (* call the parser *)
     let elem =
@@ -1291,11 +1319,11 @@ and _parse_print_error_heuristic2bis saved_typedefs saved_macros
     in
 
     (match elem with
-    | Ast_c.FinalDef x -> [(Ast_c.FinalDef x, info)]
-    | xs -> (xs, info):: loop tr (* recurse *)
+    | Ast_c.FinalDef x -> List.rev ((Ast_c.FinalDef x, info)::acc)
+    | xs -> loop ((xs, info) :: acc) tr (* recurse *)
     )
   in
-  let v = loop tr in
+  let v = loop [] tr in
   let v = with_program2 Parsing_consistency_c.consistency_checking v in
   with_program2_unit Danger.add_danger v;
   let v =
@@ -1426,7 +1454,7 @@ let no_format s =
 let (cstatement_of_string: string -> Ast_c.statement) = fun s ->
   assert (no_format s);
   let tmpfile = Common.new_temp_file "cocci_stmt_of_s" "c" in
-  Common.write_file tmpfile ("void main() { \n" ^ s ^ "\n}");
+  Common.write_file ~file:tmpfile ("void main() { \n" ^ s ^ "\n}");
   let program = fst (parse_c_and_cpp false false tmpfile) in
   program +> Common.find_some (fun (e,_) ->
     match e with
@@ -1437,7 +1465,7 @@ let (cstatement_of_string: string -> Ast_c.statement) = fun s ->
 let (cexpression_of_string: string -> Ast_c.expression) = fun s ->
   assert (no_format s);
   let tmpfile = Common.new_temp_file "cocci_expr_of_s" "c" in
-  Common.write_file tmpfile ("void main() { \n" ^ s ^ ";\n}");
+  Common.write_file ~file:tmpfile ("void main() { \n" ^ s ^ ";\n}");
   let program = fst (parse_c_and_cpp false false tmpfile) in
   program +> Common.find_some (fun (e,_) ->
     match e with
