@@ -1,4 +1,4 @@
-type type_with_ident = Ast_c.fullType -> (unit -> unit) -> (unit -> unit) -> unit
+type type_with_ident = Ast_c.fullType -> (unit -> unit) -> unit
 type type_with_ident_rest = Ast_c.fullType -> (unit -> unit) -> unit
 
 type 'a printer = 'a -> unit
@@ -15,9 +15,11 @@ type pretty_printers = {
   field           : Ast_c.field printer;
   field_list      : Ast_c.field list printer;
   init            : Ast_c.initialiser printer;
-  init_list       : (Ast_c.initialiser Ast_c.wrap2 list) printer;
+  init_list       : (Ast_c.newlines * Ast_c.initialiser Ast_c.wrap2 list) printer;
   param           : Ast_c.parameterType printer;
   paramlist       : (Ast_c.parameterType Ast_c.wrap2 list) printer;
+  template_param  : Ast_c.templateParameterType printer;
+  template_paramlist : (Ast_c.templateParameterType Ast_c.wrap2 list) printer;
   dparamlist      : ((string Ast_c.wrap) Ast_c.wrap2 list) printer;
   ty              : Ast_c.fullType printer;
   type_with_ident : type_with_ident;
@@ -68,6 +70,10 @@ val pp_param_gen:  pr_elem:Ast_c.info printer -> pr_space: unit printer ->
   Ast_c.parameterType printer
 val pp_param_list_gen:  pr_elem:Ast_c.info printer -> pr_space: unit printer ->
   (Ast_c.parameterType Ast_c.wrap2 list) printer
+val pp_template_param_gen:  pr_elem:Ast_c.info printer -> pr_space: unit printer ->
+  Ast_c.templateParameterType printer
+val pp_template_param_list_gen:  pr_elem:Ast_c.info printer -> pr_space: unit printer ->
+  (Ast_c.templateParameterType Ast_c.wrap2 list) printer
 val pp_define_param_list_gen:
     pr_elem:Ast_c.info printer -> pr_space: unit printer ->
       ((string Ast_c.wrap) Ast_c.wrap2 list) printer
@@ -76,7 +82,7 @@ val pp_type_gen:  pr_elem:Ast_c.info printer -> pr_space:unit printer ->
 val pp_init_gen:  pr_elem:Ast_c.info printer -> pr_space:unit printer ->
   Ast_c.initialiser printer
 val pp_init_list_gen:  pr_elem:Ast_c.info printer -> pr_space:unit printer ->
-  (Ast_c.initialiser Ast_c.wrap2 list) printer
+  (Ast_c.newlines * Ast_c.initialiser Ast_c.wrap2 list) printer
 val pp_string_fragment_list_gen:
     pr_elem:Ast_c.info printer -> pr_space:unit printer ->
       Ast_c.string_fragment list printer
@@ -122,6 +128,7 @@ val string_of_expression: Ast_c.expression -> string
  * @author Iago Abal
  *)
 val string_of_ifdef_guard: Ast_c.ifdef_guard -> string
+val string_of_flow: Control_flow_c.node -> string
 val string_of_toplevel: Ast_c.toplevel -> string
 val string_of_fullType: Ast_c.fullType -> string
 val string_of_name: Ast_c.name -> string
