@@ -28,16 +28,16 @@ let test_tokens_c file =
 
 (* Was in main, but using it in test_parsing_c *)
 let get_files path =
-  if !Flag.c_plus_plus
+  if !Flag.c_plus_plus <> Flag.Off
   then
     (* only C++ files, but contains .h files as that extension is ambiguous *)
     cmd_to_list
       (if !Flag.include_headers
       then
-        "find "^ path ^" -name \"*.cpp\" -o -name \"*.cxx\" -o -name \"*.cc\""
-        ^"-o name \"*.h\" -o -name \"*.hpp\" -o -name \"*.hxx\""
+        "find "^ path ^" -type f -name \"*.cpp\" -o -name \"*.cxx\" -o -name \"*.cc\""
+        ^"-o -name \"*.h\" -o -name \"*.hpp\" -o -name \"*.hxx\""
       else
-        "find "^ path ^" -name \"*.cpp\" -o -name \"*.cxx\" -o -name \"*.cc\"")
+        "find "^ path ^" -type f -name \"*.cpp\" -o -name \"*.cxx\" -o -name \"*.cc\"")
   else
     (* only .c files and .h files *)
     cmd_to_list (* same as "true, "", _" case *)
@@ -45,11 +45,11 @@ let get_files path =
 	  (* FIXME : Could we remove xs ?
 	     -use_glimpse requires a singleton.
 	     This is checked some lines before.
-	     then ("find "^(join " " (x::xs))^" -name \"*.[ch]\"")
-	     else ("find "^(join " " (x::xs))^" -name \"*.c\"")
+	     then ("find "^(join " " (x::xs))^" -type f -name \"*.[ch]\"")
+	     else ("find "^(join " " (x::xs))^" -type f -name \"*.c\"")
 	  *)
-      then ("find "^ path ^" -name \"*.[ch]\"")
-      else ("find "^ path ^" -name \"*.c\""))
+      then ("find "^ path ^" -type f -name \"*.[ch]\"")
+      else ("find "^ path ^" -type f -name \"*.c\""))
 
 let new_test_parse_gen xs =
 
@@ -57,6 +57,7 @@ let new_test_parse_gen xs =
   Flag_parsing_c.debug_cpp := true;
   Flag_parsing_c.debug_etdt := false;
   Flag_parsing_c.filter_msg := true;
+  (if !Flag.c_plus_plus = Flag.Off then Flag.set_c_plus_plus None);
 
   (*let dirname_opt =
     match xs with
@@ -66,7 +67,7 @@ let new_test_parse_gen xs =
 
   (* old:
      let xs = if !Flag.dir then
-     process_output_to_list ("find " ^ x ^" -name \"*.c\"") else x::xs in
+     process_output_to_list ("find " ^ x ^" -type f -name \"*.c\"") else x::xs in
   *)
   let fullxs = xs +> List.map get_files +> List.concat in
 
@@ -136,7 +137,7 @@ let test_parse_gen xs ext =
 
   (* old:
      let xs = if !Flag.dir then
-     process_output_to_list ("find " ^ x ^" -name \"*.c\"") else x::xs in
+     process_output_to_list ("find " ^ x ^" -type f -name \"*.c\"") else x::xs in
   *)
   let fullxs = Common.files_of_dir_or_files_no_vcs ext xs in
 
@@ -218,7 +219,7 @@ let test_parse_i xs =
 (* local function that is parametrized by whether to launch gv *)
 let local_test_cfg launchgv file =
   let (file, specific_func) =
-    if file =~ "\\(.*\\.c\\):\\(.*\\)"
+    if file =~ "\\(.*\\.c\\):\\(.*\\)" || file =~ "\\(.*\\.cpp\\):\\(.*\\)"
     then
       let (a,b) = matched2 file in
       a, Some b
@@ -226,7 +227,8 @@ let local_test_cfg launchgv file =
       file, None
   in
 
-  if not (file =~ ".*\\.c")
+  if not (file =~ ".*\\.c" ||
+          (!Flag.c_plus_plus = Flag.Off && file =~ ".*\\.cpp"))
   then pr2 "warning: seems not a .c file";
 
   (* no point to parse format strings *)
@@ -247,8 +249,8 @@ let local_test_cfg launchgv file =
       None -> ()
     | Some fn -> (* old: Flow_to_ast.test !Flag.show_flow def *)
 	try
-          let flow = Ast_to_flow.ast_to_control_flow e in
-          flow +> do_option (fun flow ->
+          let flows = Ast_to_flow.ast_to_control_flow e in
+	  let do_flow mkstring flow =
             Ast_to_flow.deadcode_detection flow;
             let flow = Ast_to_flow.annotate_loop_nodes flow in
 
@@ -258,16 +260,21 @@ let local_test_cfg launchgv file =
               then flow
               else Ctlcocci_integration.fix_flow_ctl flow
 *)
-              flow
-            in
+              flow in
 	    let filename =
 	      if launchgv
-	      then Filename.temp_file "output" ".dot"
+	      then Filename.temp_file (mkstring "output") ".dot"
 	      else
 		let fl = Filename.chop_extension (Filename.basename file) in
-		fl^":"^fn^".dot" in
-            Control_flow_c.G.print_ograph_mutable flow' (filename) launchgv
-          )
+		(mkstring(fl^":"^fn))^".dot" in
+            Control_flow_c.G.print_ograph_mutable flow' (filename) launchgv in
+	  let mkstring i x =
+	    if i > 0
+	    then Printf.sprintf "%s_%d" x i
+	    else x in
+	  Stdcompat.List.iteri
+	    (fun i (_,flow) -> flow +> do_option (do_flow (mkstring i)))
+	    flows
         with Ast_to_flow.Error (x) -> Ast_to_flow.report_error x
       )
 
@@ -283,12 +290,19 @@ let test_cfg_ifdef file =
 
   ast +> List.iter (fun e ->
     (try
-        let flow = Ast_to_flow.ast_to_control_flow e in
-        flow +> do_option (fun flow ->
+        let flows = Ast_to_flow.ast_to_control_flow e in
+	let do_flow mkstring flow =
           Ast_to_flow.deadcode_detection flow;
           let flow = Ast_to_flow.annotate_loop_nodes flow in
-          Control_flow_c.G.print_ograph_mutable flow ("/tmp/output.dot") true
-        )
+          Control_flow_c.G.print_ograph_mutable flow
+	    ((mkstring "/tmp/output")^".dot") true in
+	let mkstring i x =
+	  if i > 0
+	  then Printf.sprintf "%s_%d" x i
+	  else x in
+	Stdcompat.List.iteri
+	  (fun i (_,flow) -> flow +> do_option (do_flow (mkstring i)))
+	  flows
       with Ast_to_flow.Error (x) -> Ast_to_flow.report_error x
     )
   )
@@ -417,9 +431,9 @@ let test_attributes file =
   let ast = Parse_c.program_of_program2 ast2 in
 
   Visitor_c.vk_program { Visitor_c.default_visitor_c with
-    Visitor_c.kdef = (fun (k, bigf) (defbis, ii) ->
-      let sattr  = Ast_c.s_of_attr defbis.f_attr in
-      pr2 (spf "%-30s: %s" (Ast_c.str_of_name (defbis.f_name)) sattr);
+    Visitor_c.ktype = (fun (k, bigf) (q, a, t) ->
+      let sattr  = Ast_c.s_of_attr a in
+      pr2 (spf "%s" sattr);
     );
     Visitor_c.kdecl = (fun (k, bigf) decl ->
       match decl with
