@@ -150,6 +150,7 @@ let is_if_or_else = function
 
 let is_statement = function
   | Tfor _ | Tdo _ | Tif _ | Twhile _ | Treturn _
+  | Ttry _ | Tcatch _
   | Tbreak _ | Telse _ | Tswitch _ | Tcase _ | Tcontinue _
   | Tgoto _
   | TPtVirg _
@@ -178,10 +179,10 @@ let is_start_of_something = function
   | Tchar _  | Tshort _ | Tint _ | Tdouble _ |  Tfloat _ | Tlong _
   | Tunsigned _ | Tsigned _ | Tvoid _ | Tsize_t _ | Tssize_t _ | Tptrdiff_t _
   | TautoType _
-  | Tauto _ | Tregister _ | Textern _ | Tstatic _
+  | Tauto _ | Tregister _ | Textern _ | Tstatic _ | Talignas _
   | Tconst _ | Tvolatile _
   | Ttypedef _
-  | Tstruct _ | Tunion _ | Tenum _ | Tdecimal _
+  | Tstruct _ | Tunion _ | Tcpp_struct _ | Tcpp_union _ | Tclass _ | Tenum _ | Tdecimal _
     -> true
   | _ -> false
 
@@ -201,6 +202,7 @@ let is_stuff_taking_parenthized = function
   | Tswitch _
   | Ttypeof _
   | TMacroIterator _
+  | TMacroAttrArgs _
     -> true
   | _ -> false
 
@@ -211,15 +213,11 @@ let is_ident_like = function
   | TKRParam _
   | TypedefIdent _
   | TIdentDefine  _
-  | TDefParamVariadic _
 
   | TUnknown _
 
   | TMacroAttr _
   | TMacroAttrArgs _
-  | TMacroEndAttr _
-  | TMacroEndAttrArgs _
-  | TMacroGccEndAttr _
   | TMacroStmt _
   | TMacroIdStmt _
   | TMacroString _
@@ -329,7 +327,6 @@ let info_of_tok = function
   | TOParDefine (i1) ->     i1
   | TIdentDefine  (s, i) -> i
   | TCppEscapedNewline (ii) -> ii
-  | TDefParamVariadic (s, i1) ->     i1
 
   | TCppConcatOp (ii) -> ii
 
@@ -340,9 +337,6 @@ let info_of_tok = function
   | TMacroIdentBuilder     (s, i) -> i
   | TMacroAttr             (s, i) -> i
   | TMacroAttrArgs         (s, i) -> i
-  | TMacroEndAttr          (s, i) -> i
-  | TMacroEndAttrArgs      (s, i) -> i
-  | TMacroGccEndAttr       (i) -> i
   | TMacroStmt             (s, i) -> i
   | TMacroIdStmt           (s, i) -> i
   | TMacroString           (s, i) -> i
@@ -381,8 +375,11 @@ let info_of_tok = function
   | TCBrace              (i) -> i
   | TOCro                (i) -> i
   | TCCro                (i) -> i
+  | TOCroCro             (i) -> i
   | TDot                 (i) -> i
+  | TColonColon          (i) -> i
   | TComma               (i) -> i
+  | TNoComma             (i) -> i
   | TPtrOp               (i) -> i
   | TInc                 (i) -> i
   | TDec                 (i) -> i
@@ -402,6 +399,8 @@ let info_of_tok = function
   | TNotEq               (i) -> i
   | TInf                 (i) -> i
   | TSup                 (i) -> i
+  | TInf3                (i) -> i
+  | TSup3                (i) -> i
   | TInfEq               (i) -> i
   | TSupEq               (i) -> i
   | TShl                 (i) -> i
@@ -431,6 +430,7 @@ let info_of_tok = function
   | Tregister            (i) -> i
   | Textern              (i) -> i
   | Tstatic              (i) -> i
+  | Talignas             (i) -> i
   | Tconst               (i) -> i
   | Tvolatile            (i) -> i
 
@@ -440,8 +440,12 @@ let info_of_tok = function
   | Tenum                (i) -> i
   | Tdecimal             (i) -> i
   | Texec                (i) -> i
+  | Ttemplate            (i) -> i
   | Ttypedef             (i) -> i
   | Tunion               (i) -> i
+  | Tcpp_struct          (i) -> i
+  | Tcpp_union           (i) -> i
+  | Tclass               (i) -> i
   | Tbreak               (i) -> i
   | Telse                (i) -> i
   | Tswitch              (i) -> i
@@ -449,6 +453,8 @@ let info_of_tok = function
   | Tcontinue            (i) -> i
   | Tfor                 (i) -> i
   | Tdo                  (i) -> i
+  | Ttry                 (i) -> i
+  | Tcatch               (i) -> i
   | Tif                  (i) -> i
   | Twhile               (i) -> i
   | Treturn              (i) -> i
@@ -462,11 +468,27 @@ let info_of_tok = function
   | Ttypeof              (i) -> i
   | Tnew                 (i) -> i
   | Tdelete              (i) -> i
+  | Tusing               (i) -> i
+  | Tprivate             (i) -> i
+  | Tprotected           (i) -> i
+  | Tpublic              (i) -> i
+  | Toperator            (i) -> i
+  | TTemplateStart       (i) -> i
+  | TTemplateEnd         (i) -> i
+  | TTemplateEndSup      (i) -> i
+  | TTemplateEndTemplateEnd (i) -> i
+  | TTemplateEndTemplateEndTemplateEnd (i) -> i
+  | Tfinal               (i) -> i
+  | Tvirtual             (i) -> i
+  | Ttypename            (i) -> i
   | Tdefined             (i) -> i
   | TOParCplusplusInit   (i) -> i
 
-  | EOF                  (i) -> i
+  | TQualType            (i) -> i
+  | TQualExp             (i) -> i
+  | TQualId              (i) -> i
   | Tnamespace           (i) -> i
+  | EOF                  (i) -> i
   | TTODO _ -> failwith "fake token, should not occur"
 
 
@@ -514,8 +536,6 @@ let visitor_info_of_tok f = function
   | TOParDefine (i1) -> TOParDefine (f i1)
   | TIdentDefine  (s, i) -> TIdentDefine (s, f i)
 
-  | TDefParamVariadic (s, i1) -> TDefParamVariadic (s, f i1)
-
   | TOBraceDefineInit (i1) -> TOBraceDefineInit (f i1)
 
 
@@ -524,16 +544,13 @@ let visitor_info_of_tok f = function
   | TMacroIdentBuilder             (s, i) -> TMacroIdentBuilder (s, f i)
   | TMacroAttr           (s, i)   -> TMacroAttr            (s, f i)
   | TMacroAttrArgs       (s, i)   -> TMacroAttrArgs        (s, f i)
-  | TMacroEndAttr        (s, i)   -> TMacroEndAttr         (s, f i)
-  | TMacroEndAttrArgs    (s, i)   -> TMacroEndAttrArgs     (s, f i)
-  | TMacroGccEndAttr     (i)      -> TMacroGccEndAttr      (f i)
   | TMacroStmt           (s, i)   -> TMacroStmt            (s, f i)
   | TMacroIdStmt         (s, i)   -> TMacroIdStmt          (s, f i)
   | TMacroString         (s, i)   -> TMacroString          (s, f i)
-  | TMacroDecl           (s, i) -> TMacroDecl            (s, f i)
-  | TMacroDeclConst      (i)   -> TMacroDeclConst       (f i)
-  | TMacroIterator       (s, i) -> TMacroIterator        (s, f i)
-(*  | TMacroTop          (s,i) -> TMacroTop             (s,f i) *)
+  | TMacroDecl           (s, i)   -> TMacroDecl            (s, f i)
+  | TMacroDeclConst      (i)      -> TMacroDeclConst       (f i)
+  | TMacroIterator       (s, i)   -> TMacroIterator        (s, f i)
+(*  | TMacroTop          (s,i)    -> TMacroTop             (s,f i) *)
   | TCParEOL (i) ->     TCParEOL (f i)
 
 
@@ -566,8 +583,11 @@ let visitor_info_of_tok f = function
   | TCBrace              (i) -> TCBrace              (f i)
   | TOCro                (i) -> TOCro                (f i)
   | TCCro                (i) -> TCCro                (f i)
+  | TOCroCro             (i) -> TOCroCro             (f i)
   | TDot                 (i) -> TDot                 (f i)
+  | TColonColon          (i) -> TColonColon          (f i)
   | TComma               (i) -> TComma               (f i)
+  | TNoComma             (i) -> TNoComma             (f i)
   | TPtrOp               (i) -> TPtrOp               (f i)
   | TInc                 (i) -> TInc                 (f i)
   | TDec                 (i) -> TDec                 (f i)
@@ -587,6 +607,8 @@ let visitor_info_of_tok f = function
   | TNotEq               (i) -> TNotEq               (f i)
   | TInf                 (i) -> TInf                 (f i)
   | TSup                 (i) -> TSup                 (f i)
+  | TInf3                (i) -> TInf3                (f i)
+  | TSup3                (i) -> TSup3                (f i)
   | TInfEq               (i) -> TInfEq               (f i)
   | TSupEq               (i) -> TSupEq               (f i)
   | TShl                 (i) -> TShl                 (f i)
@@ -616,6 +638,7 @@ let visitor_info_of_tok f = function
   | Tregister            (i) -> Tregister            (f i)
   | Textern              (i) -> Textern              (f i)
   | Tstatic              (i) -> Tstatic              (f i)
+  | Talignas             (i) -> Talignas             (f i)
   | Tconst               (i) -> Tconst               (f i)
   | Tvolatile            (i) -> Tvolatile            (f i)
 
@@ -625,8 +648,12 @@ let visitor_info_of_tok f = function
   | Tenum                (i) -> Tenum                (f i)
   | Tdecimal             (i) -> Tdecimal             (f i)
   | Texec                (i) -> Texec                (f i)
+  | Ttemplate            (i) -> Ttemplate            (f i)
   | Ttypedef             (i) -> Ttypedef             (f i)
   | Tunion               (i) -> Tunion               (f i)
+  | Tcpp_struct          (i) -> Tcpp_struct          (f i)
+  | Tcpp_union           (i) -> Tcpp_union           (f i)
+  | Tclass               (i) -> Tclass               (f i)
   | Tbreak               (i) -> Tbreak               (f i)
   | Telse                (i) -> Telse                (f i)
   | Tswitch              (i) -> Tswitch              (f i)
@@ -634,6 +661,8 @@ let visitor_info_of_tok f = function
   | Tcontinue            (i) -> Tcontinue            (f i)
   | Tfor                 (i) -> Tfor                 (f i)
   | Tdo                  (i) -> Tdo                  (f i)
+  | Ttry                 (i) -> Ttry                 (f i)
+  | Tcatch               (i) -> Tcatch               (f i)
   | Tif                  (i) -> Tif                  (f i)
   | Twhile               (i) -> Twhile               (f i)
   | Treturn              (i) -> Treturn              (f i)
@@ -647,10 +676,26 @@ let visitor_info_of_tok f = function
   | Ttypeof              (i) -> Ttypeof              (f i)
   | Tnew                 (i) -> Tnew                 (f i)
   | Tdelete              (i) -> Tdelete              (f i)
+  | Tusing               (i) -> Tusing               (f i)
+  | Tprivate             (i) -> Tprivate             (f i)
+  | Tprotected           (i) -> Tprotected           (f i)
+  | Tpublic              (i) -> Tpublic              (f i)
+  | Toperator            (i) -> Toperator            (f i)
+  | TTemplateStart       (i) -> TTemplateStart       (f i)
+  | TTemplateEnd         (i) -> TTemplateEnd         (f i)
+  | TTemplateEndSup      (i) -> TTemplateEndSup      (f i)
+  | TTemplateEndTemplateEnd (i) -> TTemplateEndTemplateEnd (f i)
+  | TTemplateEndTemplateEndTemplateEnd (i) -> TTemplateEndTemplateEndTemplateEnd (f i)
+  | Tfinal               (i) -> Tfinal               (f i)
+  | Tvirtual             (i) -> Tvirtual             (f i)
+  | Ttypename            (i) -> Ttypename            (f i)
   | Tdefined             (i) -> Tdefined             (f i)
   | TOParCplusplusInit   (i) -> TOParCplusplusInit   (f i)
-  | EOF                  (i) -> EOF                  (f i)
+  | TQualType            (i) -> TQualType            (f i)
+  | TQualExp             (i) -> TQualExp             (f i)
+  | TQualId              (i) -> TQualId              (f i)
   | Tnamespace           (i) -> Tnamespace           (f i)
+  | EOF                  (i) -> EOF                  (f i)
   | TTODO _ -> failwith "fake token, should not occur"
 
 (*****************************************************************************)
@@ -690,8 +735,11 @@ let string_of_token = function
   | TCBrace _ -> "TCBrace"
   | TOCro _ -> "TOCro"
   | TCCro _ -> "TCCro"
+  | TOCroCro _ -> "TOCroCro"
   | TDot _ -> "TDot"
   | TComma _ -> "TComma"
+  | TColonColon _ -> "TColonColon"
+  | TNoComma _ -> "TNoComma"
   | TPtrOp _ -> "TPtrOp"
   | TInc _ -> "TInc"
   | TDec _ -> "TDec"
@@ -712,6 +760,8 @@ let string_of_token = function
   | TNotEq _ -> "TNotEq"
   | TInf _ -> "TInf"
   | TSup _ -> "TSup"
+  | TInf3 _ -> "TInf3"
+  | TSup3 _ -> "TSup3"
   | TInfEq _ -> "TInfEq"
   | TSupEq _ -> "TSupEq"
   | TShl _ -> "TShl"
@@ -741,14 +791,19 @@ let string_of_token = function
   | Tregister _ -> "Tregister"
   | Textern _ -> "Textern"
   | Tstatic _ -> "Tstatic"
+  | Talignas _ -> "Talignas"
   | Ttypedef _ -> "Ttypedef"
   | Tconst _ -> "Tconst"
   | Tvolatile _ -> "Tvolatile"
   | Tstruct _ -> "Tstruct"
   | Tunion _ -> "Tunion"
+  | Tcpp_struct _ -> "Tcpp_struct"
+  | Tcpp_union _ -> "Tcpp_union"
+  | Tclass _ -> "Tclass"
   | Tenum _ -> "Tenum"
   | Tdecimal _ -> "Tdecimal"
   | Texec _ -> "Texec"
+  | Ttemplate _ -> "Ttemplate"
   | Tbreak _ -> "Tbreak"
   | Telse _ -> "Telse"
   | Tswitch _ -> "Tswitch"
@@ -756,6 +811,8 @@ let string_of_token = function
   | Tcontinue _ -> "Tcontinue"
   | Tfor _ -> "Tfor"
   | Tdo _ -> "Tdo"
+  | Ttry _ -> "Ttry"
+  | Tcatch _ -> "Tcatch"
   | Tif _ -> "Tif"
   | Twhile _ -> "Twhile"
   | Treturn _ -> "Treturn"
@@ -764,6 +821,19 @@ let string_of_token = function
   | Tsizeof _ -> "Tsizeof"
   | Tnew _ -> "Tnew"
   | Tdelete _ -> "Tdelete"
+  | Tusing  _ -> "Tusing"
+  | Tprivate _ -> "Tprivate"
+  | Tprotected _ -> "Tprotected"
+  | Tpublic _ -> "Tpublic"
+  | Toperator _ -> "Toperator"
+  | TTemplateStart _ -> "TTemplateStart"
+  | TTemplateEnd _ -> "TTemplateEnd"
+  | TTemplateEndSup _ -> "TTemplateEndSup"
+  | TTemplateEndTemplateEnd _ -> "TTemplateEndTemplateEnd"
+  | TTemplateEndTemplateEndTemplateEnd _ -> "TTemplateEndTemplateEndTemplateEnd"
+  | Tfinal _ -> "Tfinal"
+  | Tvirtual _ -> "Tvirtual"
+  | Ttypename _ -> "Ttypename"
   | Tdefined _ -> "Tdefined"
   | TOParCplusplusInit _ -> "TOParCplusplusInit"
   | Tnamespace _ -> "Tnamespace"
@@ -774,7 +844,6 @@ let string_of_token = function
   | Tinline _ -> "Tinline"
   | Ttypeof _ -> "Ttypeof"
   | TDefine _ -> "TDefine"
-  | TDefParamVariadic _ -> "TDefParamVariadic"
   | TCppEscapedNewline _ -> "TCppEscapedNewline"
   | TCppConcatOp _ -> "TCppConcatOp"
   | TOParDefine _ -> "TOParDefine"
@@ -801,9 +870,6 @@ let string_of_token = function
   | TCppDirectiveOther _ -> "TCppDirectiveOther"
   | TMacroAttr _ -> "TMacroAttr"
   | TMacroAttrArgs _ -> "TMacroAttrArgs"
-  | TMacroEndAttr _ -> "TMacroEndAttr"
-  | TMacroEndAttrArgs _ -> "TMacroEndAttrArgs"
-  | TMacroGccEndAttr _ -> "TMacroGccEndAttr"
   | TMacroStmt _ -> "TMacroStmt"
   | TMacroIdStmt _ -> "TMacroIdStmt"
   | TMacroIdentBuilder _ -> "TMacroIdentBuilder"
@@ -817,6 +883,9 @@ let string_of_token = function
   | TAction _ -> "TAction"
   | TCommentMisc _ -> "TCommentMisc"
   | TCommentCpp _ -> "TCommentCpp"
+  | TQualType _ -> "TQualType"
+  | TQualExp _ -> "TQualExp"
+  | TQualId _ -> "TQualId"
   | EOF _ -> "EOF"
   | TTODO _ -> failwith "fake token, should not occur"
 
