@@ -109,7 +109,7 @@ let rec is_completed_and_simplified ty =
   | Pointer t -> is_completed_and_simplified t
   | Array (e, t) -> is_completed_and_simplified t
   | Decimal (len, prec_opt) -> true
-  | StructUnion (su, sopt, fields) ->
+  | StructUnion (su, sopt, optfinal, base_classes, fields) ->
       (* recurse fields ? Normally actually don't want,
        * prefer to have a StructUnionName when it's possible *)
       (match sopt with
@@ -119,9 +119,9 @@ let rec is_completed_and_simplified ty =
   | FunctionType ft ->
       (* todo? return type is completed ? params completed ? *)
       true
-  | Enum  (s, enumt) ->
+  | EnumDef  (ename, s, enumt) ->
       true
-  | EnumName s ->
+  | EnumName (key, id) ->
       true
 
   (* we prefer StructUnionName to StructUnion when it comes to typed metavar *)
@@ -135,6 +135,12 @@ let rec is_completed_and_simplified ty =
           (* recurse cos what if it's an alias of an alias ? *)
           is_completed_and_simplified t
       )
+  | QualifiedType (typ, _name) ->
+      (match typ with
+      | None -> false
+      | Some t ->
+              is_completed_and_simplified t
+      )
 
   | FieldType (t, _, _) ->
       is_completed_and_simplified t
@@ -151,6 +157,7 @@ let rec is_completed_and_simplified ty =
       true (* well we don't handle it, so can't really say it's completed *)
   | AutoType -> true (* this version is the most complete until we have
                         inference over auto *)
+  | TemplateType _ -> true
 
 
 let is_completed_typedef_fullType x = raise Todo
@@ -188,10 +195,12 @@ let noTypeHere =
 let do_with_type f (t,_test) =
   match t with
   | None -> noTypeHere
+  | Some ((_,_,(Ast_c.NoType,_)),_) -> noTypeHere
   | Some (t,_local) -> f t
 
 let get_opt_type e =
   match Ast_c.get_type_expr e with
+  | Some ((_,_,(Ast_c.NoType,_)),_), _test -> None
   | Some (t,_), _test -> Some t
   | None, _test -> None
 
@@ -203,14 +212,14 @@ let get_opt_type e =
 
 
 let structdef_to_struct_name ty =
-  let (qu, tybis) = ty in
+  let (qu, attr, tybis) = ty in
   match Ast_c.unwrap_typeC ty with
-  | (StructUnion (su, sopt, fields)) ->
+  | (StructUnion (su, sopt, optfinal, base_classes, fields)) ->
       let iis = Ast_c.get_ii_typeC_take_care tybis in
       (match sopt, iis with
       (* todo? but what if correspond to a nested struct def ? *)
       | Some s , [i1;i2;i3;i4] ->
-          qu, Ast_c.mk_tybis (StructUnionName (su, s)) [i1;i2]
+          qu, attr, Ast_c.mk_tybis (StructUnionName (su, s)) [i1;i2]
       | None, _ ->
           ty
       | x -> raise (Impossible 126)
@@ -228,9 +237,9 @@ let type_of_function (def,ii) =
 
   (* could use the info in the 'ii' ? *)
 
-  let fake = Ast_c.fakeInfo (Common.fake_parse_info) in
+  let fake = Ast_c.fakeAfterInfo() in
   let fake_oparen = Ast_c.rewrap_str "(" fake in
-  let fake = Ast_c.fakeInfo (Common.fake_parse_info) in
+  let fake = Ast_c.fakeAfterInfo() in
   let fake_cparen = Ast_c.rewrap_str ")" fake in
 
   Ast_c.mk_ty (FunctionType ftyp) [fake_oparen;fake_cparen]
@@ -246,7 +255,7 @@ let type_of_decl decl =
       (* todo? for other xs ? *)
       | (x,ii2)::xs ->
           let {v_namei = _var; v_type = v_type;
-               v_storage = (_storage,_inline)} = x in
+               v_storage = (_storage,_inline,_align)} = x in
 
           (* TODO normalize ? what if nested structure definition ? *)
           v_type
@@ -268,10 +277,10 @@ let structdef_of_decl decl =
       (* todo? for other xs ? *)
       | (x,ii2)::xs ->
           let {v_namei = var; v_type = v_type;
-               v_storage = (storage,inline)} = x in
+               v_storage = (storage,inline,align)} = x in
 
           (match Ast_c.unwrap_typeC v_type with
-          | Ast_c.StructUnion (su, _must_be_some, fields) ->
+          | Ast_c.StructUnion (su, _must_be_some, optfinal, base_classes, fields) ->
               (su, fields)
           | _ -> raise (Impossible 130)
           )
@@ -289,9 +298,9 @@ let (fake_function_type:
    fullType option -> argument wrap2 list -> fullType option) =
  fun rettype args ->
 
-  let fake = Ast_c.fakeInfo (Common.fake_parse_info) in
+  let fake = Ast_c.fakeAfterInfo() in
   let fake_oparen = Ast_c.rewrap_str "(" fake in
-  let fake = Ast_c.fakeInfo (Common.fake_parse_info) in
+  let fake = Ast_c.fakeAfterInfo() in
   let fake_cparen = Ast_c.rewrap_str ")" fake in
 
   let (tyargs: parameterType wrap2 list) =
@@ -304,8 +313,6 @@ let (fake_function_type:
                 { Ast_c.p_namei = None;
                   p_register = false, Ast_c.noii;
                   p_type = ft;
-                  p_attr = [];
-                  p_midattr = [];
                   p_endattr = [];
                 }
               in
@@ -385,6 +392,9 @@ let lub op t1 t2 =
 	    | _,Ast_c.FloatType(Ast_c.CDoubleComplex) -> Some t2
 	    | Ast_c.FloatType(Ast_c.CFloatComplex),_ -> Some t1
 	    | _,Ast_c.FloatType(Ast_c.CFloatComplex) -> Some t2
+	    | Ast_c.FloatType(Ast_c.CUnknownComplex),_
+	    | _,Ast_c.FloatType(Ast_c.CUnknownComplex) ->
+		failwith "complex should be float or double"
 
 	    | Ast_c.PtrDiffType,_ -> Some t1
 	    | _,Ast_c.PtrDiffType -> Some t2
@@ -455,7 +465,7 @@ let (type_field:
       | DeclarationField (FieldDeclList (onefield_multivars, iiptvirg)) ->
           onefield_multivars +> List.iter (fun (fieldkind, iicomma) ->
             match fieldkind with
-            | Simple (Some name, t) ->
+            | Simple (storage, attrs, Some (name,_), t, endattrs) ->
                 let s = Ast_c.str_of_name name in
                 if s = fld
                 then Common.push2 t res
@@ -468,11 +478,11 @@ let (type_field:
 		    (rewrap_typeC t (FieldType (t, name, Some align))) res
                 else ()
 
-            | Simple (None, t) ->
+            | Simple (storage, attrs, None, t, endattrs) ->
                 (match Ast_c.unwrap_typeC t with
 
                 (* union *)
-                | StructUnion (Union, _, fields) ->
+                | StructUnion (Union, _, _, _, fields) ->
                     aux_fields fields
 
                 (* Special case of nested structure definition inside
@@ -481,7 +491,7 @@ let (type_field:
                  * cf sparse source, where can access subfields directly.
                  * It can also be used in conjunction with union.
                  *)
-                | StructUnion (Struct, _, fields) ->
+                | StructUnion ((Class|Struct), _, _, _, fields) ->
                     aux_fields fields
 
                 | _ -> ()
@@ -490,7 +500,11 @@ let (type_field:
           )
 
       | EmptyField info -> ()
-      | MacroDeclField _ -> pr2_once "DeclTodo"; ()
+      | FunctionField def -> ()
+      | PublicLabel info | ProtectedLabel info | PrivateLabel info -> ()
+      | ConstructDestructField _ -> pr2_once "ConstructorDestructorTodo"; ()
+      | MacroDeclField _ -> pr2_once "MacroDeclField Todo"; ()
+      | MacroDeclFieldInit _ -> pr2_once "MacroDeclFieldInit Todo"; ()
 
       | CppDirectiveStruct _
       | IfdefStruct _ -> pr2_once "StructCpp";
