@@ -67,9 +67,9 @@ let end_block = function [] -> () | _ -> (*unindent true;*) force_newline () in
 let print_string_box s = print_string s in
 
 let print_option = Common.do_option in
-let print_option_space fn = function
+(*let print_option_space fn = function
     None -> ()
-  | Some x -> fn x; pr_space() in
+  | Some x -> fn x; pr_space() in*)
 let print_option_prespace fn = function
     None -> ()
   | Some x -> pr_space(); fn x in
@@ -88,8 +88,8 @@ let rec param_print_before_and_after before fn = function
 let outdent _ = () (* should go to leftmost col, does nothing now *) in
 
 let pretty_print_c =
-  Pretty_print_c.mk_pretty_printers pr_celem pr_cspace
-    force_newline (fun _ -> ()) outdent (function _ -> ()) in
+  Pretty_print_c.mk_pretty_printers ~pr_elem:pr_celem ~pr_space:pr_cspace
+    ~pr_nl:force_newline ~pr_indent:(fun _ -> ()) ~pr_outdent:outdent ~pr_unindent:(function _ -> ()) in
 
 (* --------------------------------------------------------------------- *)
 (* Only for make_hrule, print plus code, unbound metavariables *)
@@ -370,15 +370,15 @@ let rec expression e =
       | Ast.Mul | Ast.Div | Ast.Mod -> mulit
       | Ast.Min | Ast.Max -> relat
       | Ast.DecLeft | Ast.DecRight -> shift
-      | Ast.And -> bit_and
-      | Ast.Or -> bit_or
-      | Ast.Xor -> bit_xor in
+      | Ast.And _ -> bit_and
+      | Ast.Or _ -> bit_or
+      | Ast.Xor _ -> bit_xor in
     let left_prec_of_logical op =
       match Ast.unwrap_mcode op with
 	Ast.Inf | Ast.Sup | Ast.InfEq | Ast.SupEq -> relat
-      | Ast.Eq | Ast.NotEq -> equal
-      | Ast.AndLog -> log_and
-      | Ast.OrLog -> log_or in
+      | Ast.Eq | Ast.NotEq _ -> equal
+      | Ast.AndLog _ -> log_and
+      | Ast.OrLog _ -> log_or in
     match Ast.unwrap op with
       Ast.Arith op' -> left_prec_of_arith op'
     | Ast.Logical op' -> left_prec_of_logical op'
@@ -396,9 +396,9 @@ let rec expression e =
       | Ast.Mul | Ast.Div | Ast.Mod -> cast
       | Ast.Min | Ast.Max -> shift
       | Ast.DecLeft | Ast.DecRight -> addit
-      | Ast.And -> equal
-      | Ast.Or -> bit_xor
-      | Ast.Xor -> bit_and in
+      | Ast.And _ -> equal
+      | Ast.Or _ -> bit_xor
+      | Ast.Xor _ -> bit_and in
     let right_prec_of_logical op =
       match Ast.unwrap_mcode op with
 	Ast.Inf -> shift
@@ -406,9 +406,9 @@ let rec expression e =
       | Ast.InfEq -> shift
       | Ast.SupEq -> shift
       | Ast.Eq -> relat
-      | Ast.NotEq -> relat
-      | Ast.AndLog -> bit_or
-      | Ast.OrLog -> log_and in
+      | Ast.NotEq _ -> relat
+      | Ast.AndLog _ -> bit_or
+      | Ast.OrLog _ -> log_and in
     match Ast.unwrap op with
       Ast.Arith op' -> right_prec_of_arith op'
     | Ast.Logical op' -> right_prec_of_logical op'
@@ -453,14 +453,16 @@ let rec expression e =
     | Ast_c.ArrayAccess (e1, e2) -> postfix
     | Ast_c.RecordAccess (e, name) -> postfix
     | Ast_c.RecordPtAccess (e, name) -> postfix
+    | Ast_c.QualifiedAccess (t, name) -> postfix
     | Ast_c.SizeOfExpr (e) -> unary
     | Ast_c.SizeOfType (t) -> unary
-    | Ast_c.Cast (t, a, e) -> cast
+    | Ast_c.Cast (t, e) -> cast
     | Ast_c.StatementExpr (statxs, _) -> top
     | Ast_c.Constructor (t, init) -> unary
     | Ast_c.ParenExpr (e) -> primary
-    | Ast_c.New (_, t) -> unary
+    | Ast_c.New (_, t, _) -> unary
     | Ast_c.Delete(_,t) -> unary
+    | Ast_c.TemplateInst(tn,args) -> unary
     | Ast_c.Defined _ -> primary
   in
 
@@ -496,16 +498,18 @@ let rec expression e =
   | Ast.Paren(lp,exp,rp) ->
       mcode print_string_box lp; loop exp top; close_box();
       mcode print_string rp
-  | Ast.ArrayAccess(exp1,lb,exp2,rb) ->
-      loop exp1 postfix; mcode print_string_box lb; loop exp2 top; close_box();
-      mcode print_string rb
+  | Ast.ArrayAccess(fn,lb,args,rb) ->
+      loop fn postfix; mcode (print_string_with_hint StartBox) lb;
+      dots (function _ -> ()) arg_expression args;
+      mcode (print_string_with_hint EndBox) rb
   | Ast.RecordAccess(exp,pt,field) ->
       loop exp postfix; mcode print_string pt; ident field
   | Ast.RecordPtAccess(exp,ar,field) ->
       loop exp postfix; mcode print_string ar; ident field
-  | Ast.Cast(lp,ty,attr,rp,exp) ->
+  | Ast.QualifiedAccess(ty,coloncolon,field) ->
+     Common.do_option fullType ty; mcode print_string coloncolon; ident field 
+  | Ast.Cast(lp,ty,rp,exp) ->
       mcode print_string_box lp; fullType ty; close_box();
-      print_attribute_list attr;
       mcode print_string rp; loop exp cast
   | Ast.SizeOfExpr(sizeof,exp) ->
       mcode print_string sizeof; loop exp unary
@@ -513,6 +517,21 @@ let rec expression e =
       mcode print_string sizeof;
       mcode print_string_box lp; fullType ty; close_box();
       mcode print_string rp
+  | Ast.Delete(dlt,exp) ->
+    mcode print_string dlt; expression exp
+  | Ast.DeleteArr(dlt,lb,rb,exp) ->
+      mcode print_string dlt; mcode print_string lb; mcode print_string rb; expression exp
+  | Ast.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) ->
+      mcode print_string nw;
+      print_option print_args pp_opt;
+      print_option (function e -> mcode print_string e) lp_opt;
+      fullType ty;
+      print_option (function e -> mcode print_string e) rp_opt;
+      print_option print_args args_opt;
+  | Ast.TemplateInst(tn,lp,args,rp) ->
+      loop tn postfix; mcode (print_string_with_hint StartBox) lp;
+      dots (function _ -> ()) arg_expression args;
+      mcode (print_string_with_hint EndBox) rp
   | Ast.TypeExp(ty) -> fullType ty
   | Ast.Constructor(lp,ty,rp,init) ->
       mcode print_string_box lp; fullType ty; close_box();
@@ -523,7 +542,7 @@ let rec expression e =
 
   | Ast.MetaExpr (name,_,_,_typedontcare,_formdontcare,_,_bitfield) ->
       handle_metavar name (function
-        | Ast_c.MetaExprVal ((((e, _), _) as exp),_,_) ->
+        | Ast_c.MetaExprVal (_,(((e, _), _) as exp),_,_) ->
 	    if prec_of_c e < prec then
 	      begin
 		print_text "(";
@@ -537,7 +556,7 @@ let rec expression e =
 
   | Ast.MetaExprList (name,_,_,_,_) ->
       handle_metavar name (function
-        | Ast_c.MetaExprListVal args ->
+        | Ast_c.MetaExprListVal(_,args) ->
             pretty_print_c.Pretty_print_c.arg_list args
 	| Ast_c.MetaParamListVal _ ->
 	    failwith "have meta param list matching meta exp list\n";
@@ -581,6 +600,11 @@ let rec expression e =
   | Ast.OptExp(exp) -> raise CantBeInPlus in
   loop e top
 
+and print_args (lp,args,rp) =
+  mcode print_string lp;
+  dots (function _ -> ()) expression args;
+  mcode print_string rp
+
 and arg_expression e =
   match Ast.unwrap e with
     Ast.EComma(cm) ->
@@ -618,8 +642,8 @@ and unaryOp = function
   | Ast.DeRef -> print_string "*"
   | Ast.UnPlus -> print_string "+"
   | Ast.UnMinus -> print_string "-"
-  | Ast.Tilde -> print_string "~"
-  | Ast.Not -> print_string "!"
+  | Ast.Tilde s -> print_string s
+  | Ast.Not s -> print_string s
 
 and assignOp op =
   match Ast.unwrap op with
@@ -646,20 +670,20 @@ and binaryOp op =
 	| _ -> error name op "binary operator value expected")
 
 and arithOp eq op =
-  let print_string s = if eq then print_string (s^"=") else print_string s in
+  let aprint_string s = if eq then print_string (s^"=") else print_string s in
   match op with
-    Ast.Plus -> print_string "+"
-  | Ast.Minus -> print_string "-"
-  | Ast.Mul -> print_string "*"
-  | Ast.Div -> print_string "/"
-  | Ast.Min -> print_string "<?"
-  | Ast.Max -> print_string ">?"
-  | Ast.Mod -> print_string "%"
-  | Ast.DecLeft -> print_string "<<"
-  | Ast.DecRight -> print_string ">>"
-  | Ast.And -> print_string "&"
-  | Ast.Or -> print_string "|"
-  | Ast.Xor -> print_string "^"
+    Ast.Plus -> aprint_string "+"
+  | Ast.Minus -> aprint_string "-"
+  | Ast.Mul -> aprint_string "*"
+  | Ast.Div -> aprint_string "/"
+  | Ast.Min -> aprint_string "<?"
+  | Ast.Max -> aprint_string ">?"
+  | Ast.Mod -> aprint_string "%"
+  | Ast.DecLeft -> aprint_string "<<"
+  | Ast.DecRight -> aprint_string ">>"
+  | Ast.And s -> print_string s
+  | Ast.Or s -> print_string s
+  | Ast.Xor s -> print_string s
 
 and logicalOp = function
     Ast.Inf -> print_string "<"
@@ -667,9 +691,9 @@ and logicalOp = function
   | Ast.InfEq -> print_string "<="
   | Ast.SupEq -> print_string ">="
   | Ast.Eq -> print_string "=="
-  | Ast.NotEq -> print_string "!="
-  | Ast.AndLog -> print_string "&&"
-  | Ast.OrLog -> print_string "||"
+  | Ast.NotEq s -> print_string s
+  | Ast.AndLog s -> print_string s
+  | Ast.OrLog s -> print_string s
 
 and constant = function
     Ast.String(s,sz) -> print_text (sz2c sz); print_string ("\""^s^"\"")
@@ -691,25 +715,23 @@ and sz2c = function
 
 and fullType ft =
   match Ast.unwrap ft with
-    Ast.Type(_,cv,ty) ->
-      (match Ast.unwrap ty with
-	Ast.Pointer(_,_) ->
-          typeC ty;
-          param_print_between (function _ -> ()) (fun c -> pr_space();
-          mcode const_vol c) cv
-      | _ ->
-          param_print_between (function _ -> pr_space()) (mcode const_vol) cv;
-          typeC ty)
-
+    Ast.Type(_,cvbefore,ty,cvafter) ->
+      let do_cvattr = function
+	  Ast.CV cv -> mcode const_vol cv
+	| Ast.Attr attr -> print_attribute attr in
+      print_between pr_space do_cvattr cvbefore;
+      (if cvbefore <> [] then print_text " ");
+      typeC (cvafter <> []) ty;
+      (if cvafter <> [] then print_text " ");
+      print_between pr_space do_cvattr cvafter
   | Ast.AsType(ty, asty) -> fullType ty
   | Ast.DisjType _ | Ast.ConjType _ -> raise CantBeInPlus
   | Ast.OptType(_) -> raise CantBeInPlus
 
 and print_fninfo = function
     Ast.FStorage(stg) -> mcode storage stg
-  | Ast.FType(ty) -> fullType ty
+  | Ast.FType(ty) -> fullType ty; pr_space()
   | Ast.FInline(inline) -> mcode print_string inline; pr_space()
-  | Ast.FAttr(attr) -> print_attribute attr; pr_space()
 
 and print_attribute_list ?(befspace=true) ?(aftspace=false) attrs =
   if befspace && not (attrs = []) then pr_space();
@@ -719,12 +741,22 @@ and print_attribute_list ?(befspace=true) ?(aftspace=false) attrs =
 and print_attribute attr =
   match Ast.unwrap attr with
     Ast.Attribute(a) -> print_attr_arg a
-  | Ast.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) ->
+  | Ast.GccAttribute(attr_,lp1,lp2,args,rp1,rp2) ->
       mcode print_string attr_;
       mcode print_string_box lp1; mcode print_string_box lp2;
-      print_attr_arg arg; close_box();
-      mcode print_string_box rp1; close_box();
-      mcode print_string_box rp2;
+      dots (function _ -> ()) arg_expression args;
+      close_box(); mcode print_string_box rp1;
+      close_box(); mcode print_string_box rp2
+  | Ast.CxxAttribute(lb1,args,rb1,rb2) ->
+      mcode print_string_box lb1;
+      dots (function _ -> ()) arg_expression args;
+      close_box();
+      mcode print_string_box rb1; mcode print_string_box rb2
+  | Ast.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2) ->
+      mcode print_string_box lb1; mcode print_string usng;
+      ident atnm; mcode print_string dotdot; pr_space();
+      dots (function _ -> ()) expression args; close_box();
+      mcode print_string_box rb1; mcode print_string_box rb2
 
 and print_attr_arg arg =
   match Ast.unwrap arg with
@@ -736,19 +768,20 @@ and print_attr_arg arg =
   | Ast.MetaAttr(name,_,_,_) ->
       handle_metavar name
 	(function
-	    Ast_c.MetaAttrArgVal a ->
+	    Ast_c.MetaAttrArgVal(_,a) ->
               pretty_print_c.Pretty_print_c.attr_arg a
           | _ -> error name arg "attr_arg value expected")
 
-and typeC ty =
+and typeC endattrs ty =
   match Ast.unwrap ty with
     Ast.BaseType(ty,strings) ->
       print_between pr_space (mcode print_string) strings
-  | Ast.SignedT(sgn,ty) -> mcode sign sgn; print_option_prespace typeC ty
+  | Ast.SignedT(sgn,ty) -> mcode sign sgn; print_option_prespace (typeC endattrs) ty
   | Ast.Pointer(ty,star) ->
-      fullType ty; ft_space ty; mcode print_string star; eatspace()
+      fullType ty; ft_space ty; mcode print_string star;
+      if not endattrs then eatspace()
   | Ast.ParenType(lp,ty,rp) ->
-      print_parentype (lp,[],ty,rp) (function _ -> ())
+      print_parentype (lp,ty,rp) (function _ -> ())
   | Ast.FunctionType(ty,lp,params,rp) ->
       fullType ty;
       mcode print_string lp;
@@ -764,14 +797,16 @@ and typeC ty =
       print_option (mcode print_string) comma;
       print_option expression precision_opt;
       mcode print_string rp
-  | Ast.EnumName(kind,name) ->
+  | Ast.EnumName(kind,key,name) ->
       mcode print_string kind;
-      print_option_prespace ident name
-  | Ast.EnumDef(ty,lb,ids,rb) ->
-      fullType ty; ft_space ty;
-      mcode print_string lb;
-      dots force_newline enum_decl ids;
-      mcode print_string rb
+      print_option (mcode structUnion) key;
+      print_option (function x -> ident x) name
+  | Ast.EnumDef(ty,base,lb,ids,rb) ->
+    fullType ty;
+    print_option enum_base base;
+    mcode print_string lb;
+    (dots (function _ -> ()) enum_decl) ids;
+    mcode print_string rb
   | Ast.StructUnionName(kind,name) ->
       mcode structUnion kind; print_option_prespace ident name
   | Ast.StructUnionDef(ty,lb,decls,rb) ->
@@ -787,19 +822,32 @@ and typeC ty =
       mcode print_string typeof;
       mcode print_string_box lp; fullType ty; close_box();
       mcode print_string rp
+  | Ast.QualifiedType(ty,coloncolon,name) ->
+      print_option fullType ty; mcode print_string coloncolon;
+      ident name 
   | Ast.TypeName(name)-> mcode print_string name
+  | Ast.TemplateType(tn,lp,args,rp) ->
+      fullType tn; mcode (print_string_with_hint StartBox) lp;
+      dots (function _ -> ()) arg_expression args;
+      mcode (print_string_with_hint EndBox) rp
   | Ast.AutoType(auto) -> mcode print_string auto
   | Ast.MetaType(name,_,_,_) ->
       handle_metavar name  (function
-          Ast_c.MetaTypeVal exp ->
-            pretty_print_c.Pretty_print_c.ty exp
+          Ast_c.MetaTypeVal(_,ty) ->
+            pretty_print_c.Pretty_print_c.ty ty
         | _ -> error name ty "type value expected")
 
 and baseType ty = print_string (Ast.string_of_baseType ty ^ " ")
 
+and enum_base (td, ty) =
+  mcode print_string td;
+  pr_space();
+  fullType ty
+
 and structUnion = function
     Ast.Struct -> print_string "struct"
   | Ast.Union -> print_string "union"
+  | Ast.Class -> print_string "class"
 
 and sign = function
     Ast.Signed -> print_string "signed"
@@ -822,27 +870,26 @@ and storage = function
 (* --------------------------------------------------------------------- *)
 (* ParenType *)
 
-and print_parentype (lp,midattr,ty,rp) fn =
+and print_parentype (lp,ty,rp) fn =
   let function_pointer ty1 array_dec =
     match Ast.unwrap ty1 with
-     Ast.Type(_,_,fty1) ->
+     Ast.Type(_,_,fty1,_) ->
       (match Ast.unwrap fty1 with
         Ast.Pointer(ty2,star) ->
          (match Ast.unwrap ty2 with
-           Ast.Type(_,_,fty3) ->
+           Ast.Type(_,_,fty3,_) ->
             (match Ast.unwrap fty3 with
               Ast.FunctionType(ty3,lp3,params,rp3) ->
                fullType ty3;
-               print_attribute_list midattr ~befspace:true ~aftspace:true;
                mcode print_string lp;
                mcode print_string star;
                fn();
                let _ =
                  match array_dec with
                    Some(lb1,size,rb1) ->
-                    mcode print_string lb1;
-                    print_option expression size;
-                    mcode print_string rb1
+                     mcode print_string lb1;
+                     print_option expression size;
+                     mcode print_string rb1
                  | None -> () in
                mcode print_string rp;
                mcode print_string lp3;
@@ -853,7 +900,7 @@ and print_parentype (lp,midattr,ty,rp) fn =
        | _ -> failwith "ParenType Unparse_cocci")
      | _ -> failwith "ParenType Unparse_cocci" in
   match Ast.unwrap ty with
-    Ast.Type(_,_,fty1) ->
+    Ast.Type(_,_,fty1,_) ->
       (match Ast.unwrap fty1 with
         Ast.Array(ty1,lb1,size,rb1) ->
           function_pointer ty1 (Some(lb1,size,rb1))
@@ -865,25 +912,31 @@ and print_parentype (lp,midattr,ty,rp) fn =
 (* --------------------------------------------------------------------- *)
 (* Variable declaration *)
 
-and print_named_type ty midattr id =
+and print_named_type ty id =
   match Ast.unwrap ty with
-    Ast.Type(_,[],ty1) ->
+    Ast.Type(_,[],ty1,[]) ->
       (match Ast.unwrap ty1 with
         Ast.Array(_,_,_,_) ->
 	  let rec loop ty k =
 	    match Ast.unwrap ty with
 	      Ast.Array(ty,lb,size,rb) ->
 		(match Ast.unwrap ty with
-		  Ast.Type(_,cv,ty) ->
-		    param_print_between (function _ -> pr_space()) (mcode const_vol) cv;
+		  Ast.Type(_,cvbefore,ty,cvafter) ->
+		    let do_cvattr = function
+			Ast.CV cv -> mcode const_vol cv
+		      | Ast.Attr attr -> print_attribute attr in
+		    print_between (function _ -> pr_space()) do_cvattr cvbefore;
+		    (if cvbefore <> [] then pr_space());
 		    loop ty
 		      (function _ ->
 			k ();
 			mcode print_string lb;
 			print_option expression size;
-			mcode print_string rb)
+			mcode print_string rb);
+		    (if cvafter <> [] then pr_space());
+		    print_between (function _ -> pr_space()) do_cvattr cvafter;
 		| _ -> failwith "complex array types not supported")
-	    | _ -> typeC ty; ty_space ty; print_attribute_list midattr ~befspace:false ~aftspace:true; id(); k () in
+	    | _ -> typeC false ty; ty_space ty; id(); k () in
 	  loop ty1 (function _ -> ())
       | Ast.MetaType(name,_,_,_) ->
           (* MetaType with an array and a pointer have to be treated specially.
@@ -904,24 +957,24 @@ and print_named_type ty midattr id =
            *)
           let (a,info,b,c) = name in
           let (res,name_string,line,lcol,rcol) = lookup_metavar name in
-          let (info_new,straft) = match res with
-              Some (Ast_c.MetaTypeVal ty) ->
-                let (qu, iiqu), (tyy, iity) = ty in
+          let (info_new,straft) =
+	    match res with
+              Some (Ast_c.MetaTypeVal(_,ty)) ->
+                let (qu, iiqu), attrs, (tyy, iity) = ty in
                 (match tyy with
                     Ast_c.Pointer _
-                  | Ast_c.Array _ -> {info with Ast.straft = [] },info.Ast.straft
+                  | Ast_c.Array _ ->
+		      {info with Ast.straft = [] },info.Ast.straft
                   | _ -> info,[])
             | _ -> error name ty "type value expected" in
           handle_metavar (a,info_new,b,c) (function
-              Ast_c.MetaTypeVal ty ->
+              Ast_c.MetaTypeVal(_,ty) ->
 		pretty_print_c.Pretty_print_c.type_with_ident ty
-                  (function _ -> print_attribute_list midattr ~befspace:false ~aftspace:false)
 		  (function _ -> id())
             | _ -> error name ty "type value expected")
       | Ast.ParenType(lp,ty,rp) ->
-          print_parentype (lp,midattr,ty,rp) (function _ -> id())
+          print_parentype (lp,ty,rp) (function _ -> id())
       | Ast.FunctionType(ty,lp,params,rp) ->
-          assert (midattr = []);
           fullType ty;
           id();
           mcode print_string lp;
@@ -930,18 +983,26 @@ and print_named_type ty midattr id =
     (*| should have a case here for pointer to array or function type
         that would put ( * ) around the variable.  This makes one wonder
         why we really need a special case for function pointer *)
-      | _ -> fullType ty; ft_space ty; print_attribute_list midattr ~befspace:false ~aftspace:true; id())
-  | _ -> fullType ty; ft_space ty; print_attribute_list midattr ~befspace:false ~aftspace:true; id()
+      | _ -> fullType ty; ft_space ty; id())
+  | _ -> fullType ty; ft_space ty; id()
 
 and ty_space ty =
   match Ast.unwrap ty with
     Ast.Pointer(_,_) -> ()
+  | Ast.MetaType(name,_,_,_) ->
+      handle_metavar name (function
+          Ast_c.MetaTypeVal(_,(_,_,ty)) ->
+	    (match Ast_c.unwrap ty with
+	      Ast_c.Pointer _ -> ()
+	    | _ -> pr_space())
+        | _ -> error name ty "type value expected")
   | _ -> pr_space()
 
 and ft_space ty =
   match Ast.unwrap ty with
-    Ast.Type(_,cv,ty) ->
+    Ast.Type(_,cvbefore,ty,cvafter) ->
       let isptr =
+	cvafter = [] &&
 	match Ast.unwrap ty with
 	  Ast.Pointer(_,_) -> true
 	| Ast.MetaType(name,_,_,_) ->
@@ -951,15 +1012,21 @@ and ft_space ty =
 		failwith
 		  (Printf.sprintf "variable %s not known on SP line %d\n"
 		     name_string line)
-	    | Some (Ast_c.MetaTypeVal (tq,ty)) ->
+	    | Some (Ast_c.MetaTypeVal(_,(tq,attrs,ty))) ->
 		(match Ast_c.unwrap ty with
-		  Ast_c.Pointer(_,_) ->  true
+		  Ast_c.Pointer(_,_,_) ->  true
 		| _ -> false)
 	    | _ -> false)
 	| _ -> false in
       if isptr then () else pr_space()
   | _ -> pr_space()
 
+and alignas (Ast.Align(align,lpar,expr,rpar)) =
+  mcode print_string align;
+  mcode print_string lpar;
+  expression expr;
+  mcode print_string rpar
+
 and declaration d =
   match Ast.unwrap d with
     Ast.MetaDecl(name,_,_,_) ->
@@ -971,22 +1038,23 @@ and declaration d =
 
   | Ast.AsDecl(decl,asdecl) -> declaration decl
 
-  | Ast.Init(stg,ty,midattr,id,endattr,eq,ini,sem) ->
+  | Ast.Init(al,stg,ty,id,endattr,eq,ini,sem) ->
+      print_option alignas al;
       print_option (mcode storage) stg;
       print_option (function _ -> pr_space()) stg;
-      print_named_type ty midattr (fun _ -> ident id);
+      print_named_type ty (fun _ -> ident id);
       print_attribute_list endattr;
       pr_space(); mcode print_string eq;
       pr_space(); initialiser true ini; mcode print_string sem
-  | Ast.UnInit(stg,ty,midattr,id,endattr,sem) ->
+  | Ast.UnInit(al,stg,ty,id,endattr,sem) ->
+      print_option alignas al;
       print_option (mcode storage) stg;
       print_option (function _ -> pr_space()) stg;
-      print_named_type ty midattr (fun _ -> ident id);
+      print_named_type ty (fun _ -> ident id);
       print_attribute_list endattr;
       mcode print_string sem
-  | Ast.FunProto (fninfo,attr,name,lp1,params,va,rp1,sem) ->
+  | Ast.FunProto (fninfo,name,lp1,params,va,rp1,sem) ->
       List.iter print_fninfo fninfo;
-      print_attribute_list attr ~befspace:false ~aftspace:true;
       ident name; mcode print_string_box lp1;
       parameter_list params;
       begin match va with
@@ -997,29 +1065,31 @@ and declaration d =
       end;
       close_box(); mcode print_string rp1;
       mcode print_string sem
-  | Ast.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
+  | Ast.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
       print_option (mcode storage) stg;
       print_option (function _ -> pr_space()) stg;
+      print_attribute_list preattr;
       ident name; mcode print_string_box lp;
       dots (function _ -> ()) arg_expression args;
       close_box(); mcode print_string rp;
       print_attribute_list attr;
       mcode print_string sem
-  | Ast.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
+  | Ast.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
       print_option (mcode storage) stg;
       print_option (function _ -> pr_space()) stg;
+      print_attribute_list preattr;
       ident name; mcode print_string_box lp;
       dots (function _ -> ()) arg_expression args;
       close_box(); mcode print_string rp;
+      print_attribute_list attr;
       pr_space(); mcode print_string eq;
       pr_space(); initialiser true ini; mcode print_string sem
-  | Ast.TyDecl(ty,attr,sem) ->
+  | Ast.TyDecl(ty,sem) ->
       fullType ty;
-      print_attribute_list attr;
       mcode print_string sem
   | Ast.Typedef(stg,ty,id,sem) ->
       mcode print_string stg; pr_space();
-      print_named_type ty [] (fun _ -> typeC id);
+      print_named_type ty (fun _ -> typeC false id);
       mcode print_string sem
   | Ast.DisjDecl(_) | Ast.ConjDecl(_) -> raise CantBeInPlus
   | Ast.OptDecl(decl) -> raise CantBeInPlus
@@ -1036,26 +1106,69 @@ and field d =
     Ast.MetaField(name,_,_,_) ->
       handle_metavar name
 	(function
-	    Ast_c.MetaFieldVal f ->
+	    Ast_c.MetaFieldVal(_,f) ->
 	      pretty_print_c.Pretty_print_c.field f
 	  | _ -> error name d "field value expected")
 
   | Ast.MetaFieldList(name,_,_,_,_) ->
       handle_metavar name
 	(function
-	    Ast_c.MetaFieldListVal f ->
+	    Ast_c.MetaFieldListVal(_,f) ->
 	      print_between force_newline pretty_print_c.Pretty_print_c.field f
 	  | _ -> error name d "field list value expected")
-  | Ast.Field(ty,id,bf,sem) ->
+  | Ast.Field(ty,id,bf,endattr,sem) ->
       begin
 	match id with
 	  None -> fullType ty
-        | Some id -> print_named_type ty [] (fun _ -> ident id)
+        | Some id -> print_named_type ty (fun _ -> ident id)
       end;
       let bitfield (c, e) =
 	mcode print_string c;
 	expression e in
       Common.do_option bitfield bf;
+      print_attribute_list endattr;
+      mcode print_string sem
+  | Ast.MacroDeclField(name,lp,args,rp,attr,sem) ->
+      ident name; mcode print_string_box lp;
+      dots (function _ -> ()) arg_expression args;
+      close_box(); mcode print_string rp;
+      print_attribute_list attr;
+      mcode print_string sem
+  | Ast.CppField(di) -> directive di
+
+and pragmainfo pi =
+  match Ast.unwrap pi with
+    Ast.PragmaString(s) -> mcode print_string s
+  | Ast.PragmaDots (dots) -> mcode print_string dots
+  | Ast.MetaPragmaInfo(name,_,_,_) ->
+      handle_metavar name (function
+        | Ast_c.MetaPragmaInfoVal(rest) -> print_text (Ast_c.str_of_info rest)
+        | _ -> error name pi "pragma info value expected")
+
+and directive di =
+  match Ast.unwrap di with
+    Ast.Include(inc,s) ->
+      mcode print_string inc; print_text " "; mcode inc_file s
+  | Ast.MetaInclude(inc,s) ->
+      mcode print_string inc; print_text " "; expression s
+  | Ast.Pragma(prg,id,body) ->
+      mcode print_string prg; pr_space(); ident id; pr_space();
+      pragmainfo body
+  | Ast.UsingNamespace(usng,nmspc,name,sem) ->
+      mcode print_string usng; pr_space();
+      mcode print_string nmspc; pr_space();
+      ident name; pr_space();
+      mcode print_string sem
+  | Ast.UsingTypename(usng,name,eq,tn,ty,sem) ->
+      mcode print_string usng; pr_space();
+      ident name; pr_space();
+      mcode print_string eq  ; pr_space();
+      print_option (fun x -> mcode print_string x; pr_space()) tn;
+      fullType ty; pr_space();
+      mcode print_string sem
+  | Ast.UsingMember(usng,name,sem) ->
+      mcode print_string usng; pr_space();
+      ident name; pr_space();
       mcode print_string sem
 
 and annotated_field d =
@@ -1086,23 +1199,37 @@ and enum_decl d =
 (* --------------------------------------------------------------------- *)
 (* Initialiser *)
 
+and oneline i =
+  match Ast.unwrap i with
+    Ast.MetaInitList(name,_,_,_,_) ->
+      let (res,name_string,line,lcol,rcol) = lookup_metavar name in
+      (match res with
+	Some (Ast_c.MetaInitListVal (newlines,_,ini)) ->
+	  newlines = Ast_c.Compress
+      | _ -> false)
+  | _ -> true
+
 and initialiser nlcomma i =
   match Ast.unwrap i with
     Ast.MetaInit(name,_,_,_) ->
       handle_metavar name  (function
-          Ast_c.MetaInitVal ini ->
+          Ast_c.MetaInitVal(_,ini) ->
             pretty_print_c.Pretty_print_c.init ini
         | _ -> error name i "initialiser value expected")
   | Ast.MetaInitList(name,_,_,_,_) ->
       handle_metavar name  (function
-          Ast_c.MetaInitListVal ini ->
-	    pretty_print_c.Pretty_print_c.init_list ini
+          Ast_c.MetaInitListVal (newlines,_,ini) ->
+	    pretty_print_c.Pretty_print_c.init_list (newlines,ini)
         | _ -> error name i "initialiser list value expected")
   | Ast.AsInit(init,asinit) -> initialiser nlcomma init
   | Ast.InitExpr(exp) -> expression exp
   | Ast.ArInitList(lb,initlist,rb) ->
       (match Ast.unwrap initlist with
 	[] -> mcode print_string lb; mcode print_string rb
+      | ([x] as lst) when oneline x -> (* { 0 } is a common idiom *)
+	  mcode print_string lb; pr_space();
+	  initialiser_list nlcomma lst;
+	  pr_space(); mcode print_string rb
       |	lst ->
 	  mcode print_string lb; start_block();
 	  initialiser_list nlcomma lst;
@@ -1155,26 +1282,22 @@ and designator = function
 
 and parameterTypeDef p =
   match Ast.unwrap p with
-    Ast.VoidParam(ty,attr) ->
-      fullType ty;
+    Ast.Param(ty,Some id,attr) ->
+      print_named_type ty (fun _ -> ident id);
       print_attribute_list attr;
-  | Ast.Param(ty,midattr,Some id,attr) ->
-      print_named_type ty midattr (fun _ -> ident id);
-      print_attribute_list attr;
-  | Ast.Param(ty,midattr,None,attr) ->
+  | Ast.Param(ty,None,attr) ->
       fullType ty;
-      assert (midattr = []);
       print_attribute_list attr;
   | Ast.MetaParam(name,_,_,_) ->
       handle_metavar name
 	(function
-	    Ast_c.MetaParamVal p ->
+	    Ast_c.MetaParamVal(_,p) ->
               pretty_print_c.Pretty_print_c.param p
           | _ -> error name p "parameter value expected")
   | Ast.MetaParamList(name,_,_,_,_) ->
       handle_metavar name
 	(function
-	    Ast_c.MetaParamListVal p ->
+	    Ast_c.MetaParamListVal(_,p) ->
               pretty_print_c.Pretty_print_c.paramlist p
           | _ -> error name p "parameter list value expected")
 
@@ -1187,14 +1310,26 @@ and parameterTypeDef p =
   | Ast.Pdots(_) -> raise CantBeInPlus
   | Ast.OptParam(param) -> raise CantBeInPlus
 
-and parameter_list l = dots (function _ -> ()) parameterTypeDef l
-in
+and templateParameterTypeDef p =
+  match Ast.unwrap p with
+    Ast.TypenameOrClassParam(tyorcl,id,eqtyopt) ->
+      mcode print_string tyorcl;
+      ident id;
+      print_option (fun (eq,ty) -> mcode print_string eq; fullType ty) eqtyopt
+  | Ast.VarNameParam(ty,id,eqexpopt) ->
+      print_named_type ty (fun _ -> ident id);
+      print_option (fun (eq,exp) -> mcode print_string eq; initialiser false exp) eqexpopt
+  | Ast.TPComma(comma) -> mcode print_string comma; force_newline()
+  | Ast.TPDots(dots) -> mcode print_string dots; force_newline()
 
+and parameter_list l = dots (function _ -> ()) parameterTypeDef l
+
+and template_parameter_list l = dots (function _ -> ()) templateParameterTypeDef l
 
 (* --------------------------------------------------------------------- *)
 (* CPP code *)
 
-let rec inc_file = function
+and inc_file = function
     Ast.Local(elems) ->
       print_string ("\""^(String.concat "/" (List.map inc_elem elems))^"\"")
   | Ast.NonLocal(elems) ->
@@ -1223,6 +1358,13 @@ and rule_elem arity re =
       close_box(); mcode print_string rp;
       print_attribute_list attrs;
       force_newline()
+  | Ast.TemplateDefinitionHeader(tmpkw,lab,params,rab) ->
+      mcode print_string tmpkw;
+      pr_space();
+      mcode print_string lab;
+      pr_space();
+      template_parameter_list params;
+      mcode print_string rab
   | Ast.Decl decl -> pr_arity arity; annotated_decl decl
 
   | Ast.SeqStart(brace) ->
@@ -1251,12 +1393,10 @@ and rule_elem arity re =
       mcode print_string whl; pr_space(); mcode print_string_box lp;
       expression exp; close_box(); mcode print_string rp;
       mcode print_string sem
-  | Ast.ForHeader(fr,lp,first,e2,sem2,e3,rp) ->
+  | Ast.ForHeader(fr,lp,first,rp) ->
       pr_arity arity;
       mcode print_string fr; mcode print_string_box lp; forinfo first;
-      print_option expression e2; mcode print_string sem2;
-      print_option expression e3; close_box();
-      mcode print_string rp
+      close_box(); mcode print_string rp
   | Ast.IteratorHeader(nm,lp,args,rp) ->
       pr_arity arity;
       ident nm; pr_space(); mcode print_string_box lp;
@@ -1291,18 +1431,12 @@ and rule_elem arity re =
   | Ast.Ty(ty) -> pr_arity arity; fullType ty
   | Ast.TopId(id) -> pr_arity arity; ident id
   | Ast.TopInit(init) -> initialiser false init
-  | Ast.Include(inc,s) ->
-      mcode print_string inc; print_text " "; mcode inc_file s
-  | Ast.MetaInclude(inc,s) ->
-      mcode print_string inc; print_text " "; expression s
+  | Ast.CppTop(di) -> directive di
   | Ast.Undef(def,id) ->
       mcode print_string def; pr_space(); ident id
   | Ast.DefineHeader(def,id,params) ->
       mcode print_string def; pr_space(); ident id;
       print_define_parameters params
-  | Ast.Pragma(prg,id,body) ->
-      mcode print_string prg; print_text " "; ident id; print_text " ";
-      pragmainfo body
   | Ast.Default(def,colon) ->
       mcode print_string def; mcode print_string colon; pr_space()
   | Ast.Case(case,exp,colon) ->
@@ -1328,20 +1462,22 @@ and rule_elem arity re =
 
   | Ast.MetaStmtList(name,_,_,_,_) ->
       handle_metavar name (function
-        | Ast_c.MetaStmtListVal(statxs,_) ->
+        | Ast_c.MetaStmtListVal(_,statxs,_) ->
             pretty_print_c.Pretty_print_c.statement_seq_list statxs
         | _ -> error name re "statement list value expected")
   | Ast.AsRe(re,asre) -> rule_elem arity re
 
-and pragmainfo pi =
-  match Ast.unwrap pi with
-      Ast.PragmaString(s) -> mcode print_string s
-    | Ast.PragmaDots (dots) -> mcode print_string dots
-
 and forinfo = function
-    Ast.ForExp(e1,sem1) ->
-      print_option expression e1; mcode print_string sem1
-  | Ast.ForDecl decl -> annotated_decl decl
+    Ast.ForExp(e1,sem1,e2,sem2,e3) ->
+      print_option expression e1; mcode print_string sem1;
+      print_option expression e2; mcode print_string sem2;
+      print_option expression e3
+  | Ast.ForDecl(ann_decl,e2,sem2,e3) ->
+      annotated_decl ann_decl;
+      print_option expression e2; mcode print_string sem2;
+      print_option expression e3
+  | Ast.ForRange(ann_decl, ini) ->
+      annotated_decl ann_decl; pr_space(); initialiser false ini
 
 and print_define_parameters params =
   match Ast.unwrap params with
@@ -1353,6 +1489,7 @@ and print_define_parameters params =
 and print_define_param param =
   match Ast.unwrap param with
     Ast.DParam(id) -> ident id
+  | Ast.DParamEll(id,dots) -> ident id; mcode print_string dots
   | Ast.MetaDParamList(name,_,_,_,_) ->
       handle_metavar name
 	(function
@@ -1397,6 +1534,11 @@ let indent_if_needed s f =
       indent(); start_block(); f(); unindent true
     end in
 
+(*let rec toplevel arity t =
+  match Ast.unwrap t with
+       Ast.Declaration (d) -> declaration arity d;
+    |  Ast.Definition (d) -> definition arity d;*)
+
 let rec statement arity s =
   match Ast.unwrap s with
     Ast.Seq(lbrace,body,rbrace) ->
@@ -1442,7 +1584,9 @@ let rec statement arity s =
       dots force_newline (statement arity) body;
       end_block (Ast.unwrap body);
       rule_elem arity rbrace
-
+  | Ast.TemplateDefinition(header,stmt) ->
+      rule_elem arity header;
+      statement arity stmt
   | Ast.Define(header,body) ->
       rule_elem arity header; pr_space();
       dots force_newline (statement arity) body
@@ -1533,7 +1677,7 @@ in
 
 let if_open_brace  = function "{" | "else" -> true | _ -> false in
 
-(* boolean result indicates whether an indent is needed *)
+(* boolean result indicates whether a newline is needed *)
 let pp_any = function
   (* assert: normally there is only CONTEXT NOTHING tokens in any *)
     Ast.FullTypeTag(x) -> fullType x; false
@@ -1556,7 +1700,7 @@ let pp_any = function
   | Ast.LogicalOpTag(x) -> logicalOp x unknown unknown; false
 
   | Ast.InitTag(x) -> initialiser false x; false
-  | Ast.DeclarationTag(x) -> declaration x; false
+  | Ast.DeclarationTag(x) -> declaration x; true
   | Ast.FieldTag(x) -> field x; false
   | Ast.EnumDeclTag(x) -> enum_decl x; false
 
@@ -1564,7 +1708,7 @@ let pp_any = function
   | Ast.IncFileTag(x) -> inc_file x unknown unknown; false
 
   | Ast.Rule_elemTag(x) -> rule_elem "" x; false
-  | Ast.StatementTag(x) -> statement "" x; false
+  | Ast.StatementTag(x) -> statement "" x; true
   | Ast.ForInfoTag(x) -> forinfo x; false
   | Ast.CaseLineTag(x) -> case_line "" x; false
   | Ast.StringFragmentTag(x) -> string_fragment x; false
@@ -1611,14 +1755,15 @@ let pp_any = function
      normally there should be no '...' inside them *)
   | Ast.ExprDotsTag(x) -> dots (fun _ -> ()) expression x; false
   | Ast.ParamDotsTag(x) -> parameter_list x; false
+  | Ast.TemplateParamDotsTag(x) -> template_parameter_list x; false
   | Ast.StmtDotsTag(x) -> dots force_newline (statement "") x; false
   | Ast.AnnDeclDotsTag(x) -> dots force_newline annotated_decl x; false
   | Ast.AnnFieldDotsTag(x) -> dots force_newline annotated_field x; false
   | Ast.EnumDeclDotsTag(x) -> dots force_newline enum_decl x; false
   | Ast.DefParDotsTag(x) -> dots (fun _ -> ()) print_define_param x; false
-
-  | Ast.TypeCTag(x) -> typeC x; false
+  | Ast.TypeCTag(x) -> typeC false x; false
   | Ast.ParamTag(x) -> parameterTypeDef x; false
+  | Ast.TemplateParamTag(x) -> templateParameterTypeDef x; false
   | Ast.SgrepStartTag(x) -> failwith "unexpected start tag"
   | Ast.SgrepEndTag(x) -> failwith "unexpected end tag"
 in
@@ -1666,7 +1811,7 @@ in
       (* print a newline at the beginning, if needed *)
       newline_before();
       (* print a newline before each of the rest *)
-      let rec loop leading_newline indent_needed = function
+      let rec loop leading_newline space_after = function
 	  [] -> ()
 	| x::xs ->
 	    (if leading_newline then force_newline());
@@ -1702,18 +1847,18 @@ in
 		    Ast.EComma _ -> false (* due to hint *)
 		  | _ -> true)
 	      |	t -> true in
-	    let indent_needed =
-	      let rec loop space_after indent_needed = function
-		  [] -> indent_needed
+	    let space_after =
+	      let rec loop space_after inner_newline_needed = function
+		  [] -> space_after
 		| x::xs ->
-		    (if indent_needed (* for open brace *)
+		    (if inner_newline_needed
 		    then force_newline()
 		    else if space_after && space_needed_before x
 		    then pr_space());
-		    let indent_needed = pp_any x in
+		    let inner_newline_needed = pp_any x in
 		    let space_after = space_needed_after x in
-		    loop space_after indent_needed xs in
-	      loop false false x in
+		    loop space_after inner_newline_needed xs in
+	      loop space_after false x in
 	    let newline_needed =
 	      (* tokens don't group into multiline terms, so may need to avoid
 		 adding newline *)
@@ -1724,7 +1869,7 @@ in
 		  (* should never be followed by a newline *)
 		  false
 	      |	_ -> true in
-	    loop newline_needed indent_needed xs in
+	    loop newline_needed (not newline_needed && space_after) xs in
       loop false false (x::xs);
       (* print a newline at the end, if needed *)
       newline_after()
@@ -1748,4 +1893,3 @@ let pp_list_list_any (envs, pr, pr_celem, pr_cspace, pr_space, pr_arity,
 		  indent, unindent, eatspace)
 	    generating xxs before)
 	rest
-      
