@@ -78,7 +78,7 @@ let get_paramst env =
     then (argname := name; true)
     else false in
   match env_lookup fn env with
-    Ast_c.MetaParamListVal(paramst) -> (paramst,!argname)
+    Ast_c.MetaParamListVal(_,paramst) -> (paramst,!argname)
   | _ -> failwith "not possible"
 
 let get_function_name rule env =
@@ -130,14 +130,8 @@ let get_function_name rule env =
 	| _ -> [])
     | _ -> []) in
   let names =
-    (V.combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode mcode mcode mcode
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing expression donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing).V.combiner_top_level
+    (V.combiner bind option_default {V.cmcode=mcode} {V.cdonothing=donothing}
+       ~expr:expression donothing).V.combiner_top_level
       rule in
   match names with
     [name] ->
@@ -161,7 +155,7 @@ let rec print_typedef pr = function
 	  td in
       if not (List.mem s !typedefs)
       then (typedefs := s::!typedefs; pr "typedef "; pr s; pr ";\n")
-  | (Ast_c.Pointer(_,ty),_) -> print_typedef pr ty
+  | (Ast_c.Pointer(_,_,ty),_) -> print_typedef pr ty
   | _ -> ()
 
 let rewrap_str s ii =
@@ -171,7 +165,7 @@ let rewrap_str s ii =
 	Ast_c.OriginTok { pi with Common.str = s;}
     | Ast_c.ExpandedTok (pi,vpi) ->
 	Ast_c.ExpandedTok ({ pi with Common.str = s;},vpi)
-    | Ast_c.FakeTok (_,vpi) -> Ast_c.FakeTok (s,vpi)
+    | Ast_c.FakeTok (_,vpi,befaft) -> Ast_c.FakeTok (s,vpi,befaft)
     | Ast_c.AbstractLineTok pi ->
 	Ast_c.AbstractLineTok { pi with Common.str = s;})}
 
@@ -181,35 +175,36 @@ let rewrap_prefix_name prefix name =
       let iis = Common.tuple_of_list1 iiname in
       let iis' = rewrap_str (prefix^s) iis in
       Ast_c.RegularName (prefix ^ s, [iis'])
+  | Ast_c.QualName _ 
   | Ast_c.CppConcatenatedName _ | Ast_c.CppVariadicName _
-  | Ast_c.CppIdentBuilder _
+  | Ast_c.CppIdentBuilder _ | Ast_c.Operator _
       -> raise Common.Todo
 
 
 let print_metavar pr = function
   | {Ast_c.p_namei = Some name;
-     p_type = (_,(Ast_c.Pointer(_,(Ast_c.BaseType(Ast_c.Void),_)),_));
+     p_type = (_,_,(Ast_c.Pointer(_,_,(Ast_c.BaseType(Ast_c.Void),_)),_));
     }
     ->
       let param = Ast_c.str_of_name name in
       pr ("expression "^prefix); pr param
-  | ({Ast_c.p_namei = Some name; p_type = (_,ty)} : Ast_c.parameterType) ->
+  | ({Ast_c.p_namei = Some name; p_type = (_,_,ty)} : Ast_c.parameterType) ->
 
       let name' = rewrap_prefix_name prefix name in
 
       print_typedef pr ty;
 
       Pretty_print_c.pp_param_gen
-	(function x ->
+	~pr_elem:(function x ->
 	  let str = Ast_c.str_of_info x in
-	  if not (List.mem str ["const";"volatile"])
+	  if not (List.mem str ["const";"volatile";"__restrict"])
 	  then pr str)
-	(function _ -> pr " ")
+	~pr_space:(function _ -> pr " ")
         {Ast_c.p_register = (false,[]);
          p_namei = Some name';
-         p_type = (({Ast_c.const = false; Ast_c.volatile = false},[]),ty);
-         p_attr = [];
-         p_midattr = [];
+         p_type =
+	  (({Ast_c.const = false; Ast_c.volatile = false; Ast_c.restrict = false},[]),
+	   [],ty);
          p_endattr = [];
         }
   | _ -> failwith "function must have named parameters"
@@ -230,34 +225,38 @@ let print_extra_typedefs pr env =
     { Visitor_c.default_visitor_c with
       Visitor_c.ktype = (fun (k, bigf) ty ->
 	match ty with
-	  (_,((Ast_c.TypeName(_,_),_) as ty)) -> print_typedef pr ty
+	  (_,_,((Ast_c.TypeName(_,_),_) as ty)) -> print_typedef pr ty
 	| _ -> k ty) } in
   List.iter
     (function (_,vl) ->
       match vl with
 	Ast_c.MetaIdVal(_) | Ast_c.MetaFuncVal(_)
       | Ast_c.MetaAssignOpVal(_) | Ast_c.MetaBinaryOpVal(_)
-      | Ast_c.MetaLocalFuncVal(_) -> ()
-      | Ast_c.MetaExprVal(exp,_,_) -> Visitor_c.vk_expr bigf exp
-      | Ast_c.MetaExprListVal(args) -> Visitor_c.vk_argument_list bigf args
-      | Ast_c.MetaParamVal(param) -> Visitor_c.vk_param bigf param
-      | Ast_c.MetaParamListVal(params) -> Visitor_c.vk_param_list bigf params
-      | Ast_c.MetaDParamListVal(params) ->
+      | Ast_c.MetaPragmaInfoVal(_) | Ast_c.MetaLocalFuncVal(_) -> ()
+      | Ast_c.MetaExprVal(exp,_,_,_) -> Visitor_c.vk_expr bigf exp
+      | Ast_c.MetaExprListVal(args,_) -> Visitor_c.vk_argument_list bigf args
+      | Ast_c.MetaParamVal(param,_) -> Visitor_c.vk_param bigf param
+      | Ast_c.MetaParamListVal(params,_) -> Visitor_c.vk_param_list bigf params
+      | Ast_c.MetaTemplateParamVal(param,_) ->
+	  Visitor_c.vk_template_param bigf param
+      | Ast_c.MetaTemplateParamListVal(params,_) ->
+	  Visitor_c.vk_template_param_list bigf params
+      | Ast_c.MetaDParamListVal params ->
 	  Visitor_c.vk_define_params bigf params
 
-      | Ast_c.MetaTypeVal(ty) -> Visitor_c.vk_type bigf ty
-      | Ast_c.MetaInitVal(ty) -> Visitor_c.vk_ini bigf ty
-      | Ast_c.MetaInitListVal(ty) -> Visitor_c.vk_ini_list bigf ty
+      | Ast_c.MetaTypeVal(ty,_) -> Visitor_c.vk_type bigf ty
+      | Ast_c.MetaInitVal(ty,_) -> Visitor_c.vk_ini bigf ty
+      | Ast_c.MetaInitListVal(_,ty,_) -> Visitor_c.vk_ini_list bigf ty
       | Ast_c.MetaDeclVal(decl,_) -> Visitor_c.vk_decl bigf decl
-      | Ast_c.MetaFieldVal(field) -> Visitor_c.vk_struct_field bigf field
-      | Ast_c.MetaFieldListVal(fields) ->
+      | Ast_c.MetaFieldVal(field,_) -> Visitor_c.vk_struct_field bigf field
+      | Ast_c.MetaFieldListVal(fields,_) ->
 	  Visitor_c.vk_struct_fields bigf fields
       | Ast_c.MetaFmtVal(fmt) -> Visitor_c.vk_string_format bigf fmt
-      | Ast_c.MetaAttrArgVal(name) -> Visitor_c.vk_attr_arg bigf name
+      | Ast_c.MetaAttrArgVal(name,_) -> Visitor_c.vk_attr_arg bigf name
       | Ast_c.MetaFragListVal(frags) ->
 	  Visitor_c.vk_string_fragments bigf frags
       | Ast_c.MetaStmtVal(stm,_,_) -> Visitor_c.vk_statement bigf stm
-      | Ast_c.MetaStmtListVal(stms,_) ->
+      | Ast_c.MetaStmtListVal(stms,_,_) ->
 	  Visitor_c.vk_statement_sequencable_list bigf stms
       | Ast_c.MetaPosVal _ | Ast_c.MetaPosValList _ | Ast_c.MetaComValList _
       | Ast_c.MetaListlenVal _ -> ()
@@ -289,43 +288,60 @@ let rename argids env =
        match vl with
 	 Ast_c.MetaIdVal(_) | Ast_c.MetaFuncVal(_)
        | Ast_c.MetaAssignOpVal(_) | Ast_c.MetaBinaryOpVal(_)
-       | Ast_c.MetaLocalFuncVal(_) -> vl
-       | Ast_c.MetaExprVal(exp,c,ty) ->
-	   Ast_c.MetaExprVal(Visitor_c.vk_expr_s bigf exp,c,ty)
-       | Ast_c.MetaExprListVal(args) ->
-	   Ast_c.MetaExprListVal(Visitor_c.vk_arguments_s bigf args)
-       | Ast_c.MetaParamVal(param) ->
-	   Ast_c.MetaParamVal(Visitor_c.vk_param_s bigf param)
-       | Ast_c.MetaParamListVal(params) ->
-	   Ast_c.MetaParamListVal(Visitor_c.vk_params_s bigf params)
+       | Ast_c.MetaPragmaInfoVal(_) | Ast_c.MetaLocalFuncVal(_) -> vl
+       | Ast_c.MetaExprVal(exp,original,c,ty) ->
+	   Ast_c.MetaExprVal(Visitor_c.vk_expr_s bigf exp,
+			     Visitor_c.vk_expr_s bigf original,c,ty)
+       | Ast_c.MetaExprListVal(args,original) ->
+	   Ast_c.MetaExprListVal(Visitor_c.vk_arguments_s bigf args,
+				 Visitor_c.vk_arguments_s bigf original)
+       | Ast_c.MetaParamVal(param,original) ->
+	   Ast_c.MetaParamVal(Visitor_c.vk_param_s bigf param,
+			      Visitor_c.vk_param_s bigf original)
+       | Ast_c.MetaParamListVal(params,original) ->
+	   Ast_c.MetaParamListVal(Visitor_c.vk_params_s bigf params,
+				  Visitor_c.vk_params_s bigf original)
+       | Ast_c.MetaTemplateParamVal(param,original) ->
+	   Ast_c.MetaTemplateParamVal(Visitor_c.vk_template_param_s bigf param,
+				      Visitor_c.vk_template_param_s bigf original)
+       | Ast_c.MetaTemplateParamListVal(params,original) ->
+	   Ast_c.MetaTemplateParamListVal(Visitor_c.vk_template_params_s bigf params,
+					  Visitor_c.vk_template_params_s bigf original)
        | Ast_c.MetaDParamListVal(params) ->
 	   Ast_c.MetaDParamListVal(Visitor_c.vk_define_params_s bigf params)
 
-       | Ast_c.MetaTypeVal(ty) ->
-	   Ast_c.MetaTypeVal(Visitor_c.vk_type_s bigf ty)
-       | Ast_c.MetaInitVal(ini) ->
-	   Ast_c.MetaInitVal(Visitor_c.vk_ini_s bigf ini)
-       | Ast_c.MetaInitListVal(ini) ->
-	   Ast_c.MetaInitListVal(Visitor_c.vk_inis_s bigf ini)
+       | Ast_c.MetaTypeVal(ty,original) ->
+	   Ast_c.MetaTypeVal(Visitor_c.vk_type_s bigf ty,
+			     Visitor_c.vk_type_s bigf original)
+       | Ast_c.MetaInitVal(ini,original) ->
+	   Ast_c.MetaInitVal(Visitor_c.vk_ini_s bigf ini,
+			     Visitor_c.vk_ini_s bigf original)
+       | Ast_c.MetaInitListVal(newlines,ini,original) ->
+	   Ast_c.MetaInitListVal(newlines,Visitor_c.vk_inis_s bigf ini,
+				 Visitor_c.vk_inis_s bigf original)
        | Ast_c.MetaDeclVal(stm,original) ->
 	   Ast_c.MetaDeclVal(Visitor_c.vk_decl_s bigf stm,
 			     Visitor_c.vk_decl_s bigf original)
-       | Ast_c.MetaFieldVal(stm) ->
-	   Ast_c.MetaFieldVal(Visitor_c.vk_struct_field_s bigf stm)
-       | Ast_c.MetaFieldListVal(stm) ->
-	   Ast_c.MetaFieldListVal(Visitor_c.vk_struct_fields_s bigf stm)
+       | Ast_c.MetaFieldVal(fld,original) ->
+	   Ast_c.MetaFieldVal(Visitor_c.vk_struct_field_s bigf fld,
+			      Visitor_c.vk_struct_field_s bigf original)
+       | Ast_c.MetaFieldListVal(fld,original) ->
+	   Ast_c.MetaFieldListVal(Visitor_c.vk_struct_fields_s bigf fld,
+				  Visitor_c.vk_struct_fields_s bigf original)
        | Ast_c.MetaFmtVal(fmt) ->
 	   Ast_c.MetaFmtVal(Visitor_c.vk_string_format_s bigf fmt)
-       | Ast_c.MetaAttrArgVal(name) ->
-	   Ast_c.MetaAttrArgVal(Visitor_c.vk_attr_arg_s bigf name)
+       | Ast_c.MetaAttrArgVal(name,original) ->
+	   Ast_c.MetaAttrArgVal(Visitor_c.vk_attr_arg_s bigf name,
+				Visitor_c.vk_attr_arg_s bigf original)
        | Ast_c.MetaFragListVal(frags) ->
 	   Ast_c.MetaFragListVal(Visitor_c.vk_string_fragments_s bigf frags)
        | Ast_c.MetaStmtVal(stm,original,ty) ->
 	   Ast_c.MetaStmtVal(Visitor_c.vk_statement_s bigf stm,
 			     Visitor_c.vk_statement_s bigf original,ty)
-       | Ast_c.MetaStmtListVal(stm,ty) ->
+       | Ast_c.MetaStmtListVal(stm,original,ty) ->
 	   Ast_c.MetaStmtListVal
-	     (Visitor_c.vk_statement_sequencable_list_s bigf stm,ty)
+	     (Visitor_c.vk_statement_sequencable_list_s bigf stm,
+	      Visitor_c.vk_statement_sequencable_list_s bigf original,ty)
        | Ast_c.MetaPosVal _ | Ast_c.MetaPosValList _ | Ast_c.MetaComValList _
        | Ast_c.MetaListlenVal _ -> vl
        | Ast_c.MetaNoVal -> failwith "referencing a metavar with no value"))
@@ -336,10 +352,10 @@ let print_one_type pr env ty =
     Some (Ast_cocci.MetaType(name,cstr,keep,inherited)) ->
       (try
 	match List.assoc name env with
-	  Ast_c.MetaTypeVal ty ->
+	  Ast_c.MetaTypeVal(ty,_) ->
 	    Pretty_print_c.pp_type_gen
-	      (function x -> pr (Ast_c.str_of_info x))
-	      (function _ -> pr " ")
+	      ~pr_elem:(function x -> pr (Ast_c.str_of_info x))
+	      ~pr_space:(function _ -> pr " ")
 	      ty
         | _ -> failwith "impossible"
       with Not_found -> pr (Ast_cocci.string_of_fullType ty))
@@ -397,6 +413,8 @@ let pp_meta_decl pr env decl =
       no_arity ar; pr "binary operator "; pp_name name; pr ";\n"
   | Ast.MetaAssignmentOperatorDecl(ar, name) ->
       no_arity ar; pr "assignment operator "; pp_name name; pr ";\n"
+  | Ast.MetaPragmaInfoDecl(ar, name) ->
+      no_arity ar; pr "pragmainfo "; pp_name name; pr ";\n"
   | Ast.MetaConstDecl(ar, name, types) ->
       no_arity ar; pr "constant "; print_types pr env types;
       pp_name name; pr ";\n"
@@ -456,7 +474,7 @@ let print_metavariables pr local_metas paramst env header_req function_name =
   pr (Printf.sprintf "position _p!=same_%s.p;\n" function_name);
   pr "identifier _f;\n";
   let rec loop = function
-      [] | [{Ast_c.p_type =(_,(Ast_c.BaseType(Ast_c.Void),_))},_] -> []
+      [] | [{Ast_c.p_type =(_,_,(Ast_c.BaseType(Ast_c.Void),_))},_] -> []
     | ((first,_) as f)::rest ->
 	print_metavar pr first; pr ";\n";
 	(make_exp f) :: loop rest in
@@ -487,7 +505,7 @@ let pp_def_gen pr defn isexp =
   pr (Ast_c.str_of_name name); pr "(";
   (if b then failwith "not handling variable argument functions");
   (match paramst with
-    [] | [{Ast_c.p_type = (_,(Ast_c.BaseType(Ast_c.Void),_))},_] -> ()
+    [] | [{Ast_c.p_type = (_,_,(Ast_c.BaseType(Ast_c.Void),_))},_] -> ()
   | (first,_)::rest ->
       print_param_name pr first;
       List.iter (function (x,_) -> pr ", "; print_param_name pr x) rest);
@@ -551,13 +569,13 @@ let pp_rule local_metas ast env srcfile =
 	function_name_count in
     let (argids,args) = List.split args in
     let env = rename argids env in
-    let env = (args_name,Ast_c.MetaExprListVal args)::env in
+    let env = (args_name,Ast_c.MetaExprListVal(args,args))::env in
     print_start pr;
     (* for printing C tokens *)
     let pr_c info =
       match Ast_c.pinfo_of_info info with
 	Ast_c.AbstractLineTok _ -> pr (Ast_c.str_of_info info)
-      | Ast_c.FakeTok (s,_) -> pr s
+      | Ast_c.FakeTok (s,_,_) -> pr s
       |	_ ->
 	  Printf.printf "line: %s\n" (Dumper.dump info);
 	  error rule "not an abstract line" in
