@@ -45,24 +45,24 @@ let rec right_decl d =
       call_right right_mcode name d
 	(function name -> Ast0.MetaDecl(name,cstr,pure))
   | Ast0.AsDecl(decl,asdecl) -> failwith "not possible"
-  | Ast0.Init(stg,ty,midattr,id,endattr,eq,ini,sem) ->
+  | Ast0.Init(align,stg,ty,id,endattr,eq,ini,sem) ->
       call_right right_mcode sem d
-	(function sem -> Ast0.Init(stg,ty,midattr,id,endattr,eq,ini,sem))
-  | Ast0.UnInit(stg,ty,midattr,id,endattr,sem) ->
+	(function sem -> Ast0.Init(align,stg,ty,id,endattr,eq,ini,sem))
+  | Ast0.UnInit(align,stg,ty,id,endattr,sem) ->
       call_right right_mcode sem d
-	(function sem -> Ast0.UnInit(stg,ty,midattr,id,endattr,sem))
-  | Ast0.FunProto(fninfo,attr,id,lp1,params,va,rp1,sem) ->
+	(function sem -> Ast0.UnInit(align,stg,ty,id,endattr,sem))
+  | Ast0.FunProto(fninfo,id,lp1,params,va,rp1,sem) ->
       call_right right_mcode sem d
-	(function ty -> Ast0.FunProto(fninfo,attr,id,lp1,params,va,rp1,sem))
-  | Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
+	(function ty -> Ast0.FunProto(fninfo,id,lp1,params,va,rp1,sem))
+  | Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
       call_right right_mcode sem d
-	(function sem -> Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem))
-  | Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
+	(function sem -> Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem))
+  | Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
       call_right right_mcode sem d
-	(function sem -> Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem))
-  | Ast0.TyDecl(ty,attr,sem) ->
+	(function sem -> Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem))
+  | Ast0.TyDecl(ty,sem) ->
       call_right right_mcode sem d
-	(function sem -> Ast0.TyDecl(ty,attr,sem))
+	(function sem -> Ast0.TyDecl(ty,sem))
   | Ast0.Typedef(stg,ty,id,sem) ->
       call_right right_mcode sem d
 	(function sem -> Ast0.Typedef(stg,ty,id,sem))
@@ -74,6 +74,7 @@ let rec right_decl d =
 let rec right_statement s =
   match Ast0.unwrap s with
     Ast0.FunDecl(bef,fi,name,lp,params,va,rp,attrs,lbrace,body,rbrace,aft) -> None
+  | Ast0.TemplateDefinition(_,_,_,_,_) -> None
   | Ast0.Decl(bef,decl) ->
       call_right right_decl decl s
 	(function decl -> Ast0.Decl(bef,decl))
@@ -89,7 +90,7 @@ let rec right_statement s =
   | Ast0.Do(d,body,whl,lp,exp,rp,sem) ->
       call_right right_mcode sem s
 	(function sem -> Ast0.Do(d,body,whl,lp,exp,rp,sem))
-  | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,aft) -> None
+  | Ast0.For(fr,lp,first,rp,body,aft) -> None
   | Ast0.Iterator(nm,lp,args,rp,body,aft) -> None
   | Ast0.Switch(switch,lp,exp,rp,lb,decls,cases,rb) ->
       call_right right_mcode rb s
@@ -132,11 +133,7 @@ let rec right_statement s =
   | Ast0.TopId(id) -> None
   | Ast0.TopInit(init) -> None
   | Ast0.Dots(d,whn) -> None
-  | Ast0.Include(inc,name) ->
-      call_right right_mcode name s
-	(function name -> Ast0.Include(inc,name))
-  | Ast0.MetaInclude(inc,name) ->
-      None (* likely not correct, wait for pb in practice *)
+  | Ast0.CppTop(di) -> None
   | Ast0.Undef(def,id) ->
       (* nothing available for ident, and not sure code can appear
 	 here anyway *)
@@ -144,9 +141,6 @@ let rec right_statement s =
   | Ast0.Define(def,id,params,body) ->
       call_right right_statement_dots body s
 	(function body -> Ast0.Define(def,id,params,body))
-  | Ast0.Pragma(prg,id,body) ->
-      (* seems safe; let the code go with the enclosing statement, if any *)
-      None
   | Ast0.OptStm(re) ->
       call_right right_statement re s (function re -> Ast0.OptStm(re))
 
@@ -157,11 +151,16 @@ and right_statement_dots sd =
 
 let rec left_ty t =
   match Ast0.unwrap t with
-    Ast0.ConstVol(cv,ty) ->
+    Ast0.ConstVol([],ty,cvafter) ->
+      call_right left_ty ty t (function ty -> Ast0.ConstVol([],ty,cvafter))
+  | Ast0.ConstVol(cv::cvbefore,ty,cvafter) ->
       (match cv with
-	[] -> failwith "empty const/vol list in type"
-      | s::r ->
-          call_right left_mcode s t (function s -> Ast0.ConstVol(s::r,ty)))
+	Ast0.CV cv ->
+	  call_right left_mcode cv t
+	    (function cv -> Ast0.ConstVol((Ast0.CV cv)::cvbefore,ty,cvafter))
+      | Ast0.Attr attr ->
+	  call_right left_attribute attr t
+	    (function attr -> Ast0.ConstVol((Ast0.Attr attr)::cvbefore,ty,cvafter)))
   | Ast0.BaseType(ty,strings) ->
       (match strings with
 	[] -> failwith "empty strings in type"
@@ -180,11 +179,11 @@ let rec left_ty t =
   | Ast0.Decimal(dec,lp,length,comma,precision_opt,rp) ->
       call_right left_mcode dec t
 	(function dec -> Ast0.Decimal(dec,lp,length,comma,precision_opt,rp))
-  | Ast0.EnumName(kind,name) ->
-      call_right left_mcode kind t (function kind -> Ast0.EnumName(kind,name))
-  | Ast0.EnumDef(ty,lb,ids,rb) ->
+  | Ast0.EnumName(kind,key,name) ->
+      call_right left_mcode kind t (function kind -> Ast0.EnumName(kind,key,name))
+  | Ast0.EnumDef(ty,base,lb,ids,rb) ->
       call_right left_ty ty t
-	(function ty -> Ast0.EnumDef(ty,lb,ids,rb))
+	(function ty -> Ast0.EnumDef(ty,base,lb,ids,rb))
   | Ast0.StructUnionName(kind,name) ->
       call_right left_mcode kind t
 	(function kind -> Ast0.StructUnionName(kind,name))
@@ -197,6 +196,15 @@ let rec left_ty t =
       call_right left_mcode tf t (function tf -> Ast0.TypeOfType(tf,lb,ty,rb))
   | Ast0.TypeName(name) ->
       call_right left_mcode name t (function name -> Ast0.TypeName(name))
+  | Ast0.QualifiedType(Some ty,coloncolon,name) ->
+      call_right left_ty ty t
+    (function ty -> Ast0.QualifiedType(Some ty,coloncolon,name))
+  | Ast0.QualifiedType(None,coloncolon,name) ->
+      call_right left_mcode coloncolon t
+    (function coloncolon -> Ast0.QualifiedType(None,coloncolon,name))
+  | Ast0.TemplateType(name,lp,args,rp) ->
+      call_right left_ty name t (function name ->
+          Ast0.TemplateType(name,lp,args,rp))
   | Ast0.AutoType(auto) ->
       call_right left_mcode auto t (function auto -> Ast0.AutoType(auto))
   | Ast0.MetaType(name,cstr,x) ->
@@ -207,7 +215,7 @@ let rec left_ty t =
   | Ast0.OptType(ty) ->
       call_right left_ty ty t (function ty -> Ast0.OptType(ty))
 
-let rec left_ident i =
+and left_ident i =
   match Ast0.unwrap i with
     Ast0.Id(name) ->
       call_right left_mcode name i (function name -> Ast0.Id(name))
@@ -224,7 +232,7 @@ let rec left_ident i =
       call_right left_ident id i (function id -> Ast0.OptIdent(id))
   | Ast0.AsIdent(id,asid) -> failwith "not possible"
 
-let left_attr_arg arg =
+and left_attr_arg arg =
   match Ast0.unwrap arg with
     Ast0.MacroAttr(a) ->
       call_right left_mcode a arg (function a -> Ast0.MacroAttr(a))
@@ -234,13 +242,19 @@ let left_attr_arg arg =
   | Ast0.MacroAttrArgs(attr,lp,args,rp) ->
       call_right left_mcode attr arg (function attr -> Ast0.MacroAttrArgs(attr,lp,args,rp))
 
-let left_attribute attr =
+and left_attribute attr =
   match Ast0.unwrap attr with
     Ast0.Attribute(a) ->
       call_right left_attr_arg a attr (function a -> Ast0.Attribute(a))
   | Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) ->
       call_right left_mcode attr_ attr
         (function attr_ -> Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2))
+  | Ast0.CxxAttribute(lb1,arg,rb1,rb2) ->
+      call_right left_mcode lb1 attr
+        (function lb1 -> Ast0.CxxAttribute(lb1,arg,rb1,rb2))
+  | Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,arg,rb1,rb2) ->
+      call_right left_mcode lb1 attr
+        (function lb1 -> Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,arg,rb1,rb2))
 
 let left_fundecl name fninfo =
   let fncall_right processor data cont =
@@ -258,8 +272,6 @@ let left_fundecl name fninfo =
       fncall_right left_ty ty (function ty -> (Ast0.FType ty)::x)
   | (Ast0.FInline inl)::x ->
       fncall_right left_mcode inl (function inl -> (Ast0.FInline inl)::x)
-  | (Ast0.FAttr atr)::x ->
-      fncall_right left_attribute atr (function atr -> (Ast0.FAttr atr)::x)
 
 let rec left_decl decl =
   match Ast0.unwrap decl with
@@ -267,54 +279,64 @@ let rec left_decl decl =
       call_right right_mcode name decl
 	(function name -> Ast0.MetaDecl(name,cstr,pure))
   | Ast0.AsDecl(decl,asdecl) -> failwith "not possible"
-  | Ast0.Init(Some stg,ty,midattr,id,endattr,eq,ini,sem) ->
+  | Ast0.Init(Some(Ast0.Align(align,lp,e,rp)),stg,ty,id,endattr,eq,ini,sem) ->
+      call_right left_mcode align decl
+	(function align ->
+	  Ast0.Init(Some(Ast0.Align(align,lp,e,rp)),stg,ty,id,endattr,eq,ini,sem))
+  | Ast0.Init(None,Some stg,ty,id,endattr,eq,ini,sem) ->
       call_right left_mcode stg decl
-	(function stg -> Ast0.Init(Some stg,ty,midattr,id,endattr,eq,ini,sem))
-  | Ast0.Init(None,ty,midattr,id,endattr,eq,ini,sem) ->
+	(function stg -> Ast0.Init(None,Some stg,ty,id,endattr,eq,ini,sem))
+  | Ast0.Init(None,None,ty,id,endattr,eq,ini,sem) ->
       call_right left_ty ty decl
-	(function ty -> Ast0.Init(None,ty,midattr,id,endattr,eq,ini,sem))
-  | Ast0.UnInit(Some stg,ty,midattr,id,endattr,sem) ->
+	(function ty -> Ast0.Init(None,None,ty,id,endattr,eq,ini,sem))
+  | Ast0.UnInit(Some(Ast0.Align(align,lp,e,rp)),stg,ty,id,endattr,sem) ->
+      call_right left_mcode align decl
+	(function align ->
+	  Ast0.UnInit(Some(Ast0.Align(align,lp,e,rp)),stg,ty,id,endattr,sem))
+  | Ast0.UnInit(None,Some stg,ty,id,endattr,sem) ->
       call_right left_mcode stg decl
-	(function stg -> Ast0.UnInit(Some stg,ty,midattr,id,endattr,sem))
-  | Ast0.UnInit(None,ty,midattr,id,endattr,sem) ->
+	(function stg -> Ast0.UnInit(None,Some stg,ty,id,endattr,sem))
+  | Ast0.UnInit(None,None,ty,id,endattr,sem) ->
       call_right left_ty ty decl
-	(function ty -> Ast0.UnInit(None,ty,midattr,id,endattr,sem))
-  | Ast0.FunProto(fi,attr,name,lp1,params,va,rp1,sem) ->
+	(function ty -> Ast0.UnInit(None,None,ty,id,endattr,sem))
+  | Ast0.FunProto(fi,name,lp1,params,va,rp1,sem) ->
       (match fi with
 	[] ->
           call_right left_ident name decl
-	    (function name -> Ast0.FunProto(fi,attr,name,lp1,params,va,rp1,sem))
+	    (function name -> Ast0.FunProto(fi,name,lp1,params,va,rp1,sem))
       | (Ast0.FStorage sto)::x ->
 	  call_right left_mcode sto decl
 	    (function sto ->
-	      Ast0.FunProto((Ast0.FStorage sto)::x,attr,name,lp1,params,va,rp1,sem))
+	      Ast0.FunProto((Ast0.FStorage sto)::x,name,lp1,params,va,rp1,sem))
       |	(Ast0.FType ty)::x ->
 	  call_right left_ty ty decl
 	    (function ty ->
-	      Ast0.FunProto((Ast0.FType ty)::x,attr,name,lp1,params,va,rp1,sem))
+	      Ast0.FunProto((Ast0.FType ty)::x,name,lp1,params,va,rp1,sem))
       | (Ast0.FInline inl)::x ->
 	  call_right left_mcode inl decl
 	    (function inl ->
-	      Ast0.FunProto((Ast0.FInline inl)::x,attr,name,lp1,params,va,rp1,sem))
-      | (Ast0.FAttr attr0)::x ->
-	  call_right left_attribute attr0 decl
-	    (function attr0 ->
-	      Ast0.FunProto((Ast0.FAttr attr0)::x,attr,name,lp1,params,va,rp1,sem)))
-  | Ast0.MacroDecl(Some stg,name,lp,args,rp,attr,sem) ->
+	      Ast0.FunProto((Ast0.FInline inl)::x,name,lp1,params,va,rp1,sem)))
+  | Ast0.MacroDecl(Some stg,preattr,name,lp,args,rp,attr,sem) ->
       call_right left_mcode stg decl
-	(function stg -> Ast0.MacroDecl(Some stg,name,lp,args,rp,attr,sem))
-  | Ast0.MacroDecl(None,name,lp,args,rp,attr,sem) ->
+	(function stg -> Ast0.MacroDecl(Some stg,preattr,name,lp,args,rp,attr,sem))
+  | Ast0.MacroDecl(None,[],name,lp,args,rp,attr,sem) ->
       call_right left_ident name decl
-	(function name -> Ast0.MacroDecl(None,name,lp,args,rp,attr,sem))
-  | Ast0.MacroDeclInit(Some stg,name,lp,args,rp,eq,ini,sem) ->
+	(function name -> Ast0.MacroDecl(None,[],name,lp,args,rp,attr,sem))
+  | Ast0.MacroDecl(None,preattr::l,name,lp,args,rp,attr,sem) ->
+      call_right left_attribute preattr decl
+	(function preattr -> Ast0.MacroDecl(None,preattr::l,name,lp,args,rp,attr,sem))
+  | Ast0.MacroDeclInit(Some stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
       call_right left_mcode stg decl
 	(function stg ->
-	  Ast0.MacroDeclInit(Some stg,name,lp,args,rp,eq,ini,sem))
-  | Ast0.MacroDeclInit(None,name,lp,args,rp,eq,ini,sem) ->
+	  Ast0.MacroDeclInit(Some stg,preattr,name,lp,args,rp,attr,eq,ini,sem))
+  | Ast0.MacroDeclInit(None,[],name,lp,args,rp,attr,eq,ini,sem) ->
       call_right left_ident name decl
-	(function name -> Ast0.MacroDeclInit(None,name,lp,args,rp,eq,ini,sem))
-  | Ast0.TyDecl(ty,attr,sem) ->
-      call_right left_ty ty decl (function ty -> Ast0.TyDecl(ty,attr,sem))
+	(function name -> Ast0.MacroDeclInit(None,[],name,lp,args,rp,attr,eq,ini,sem))
+  | Ast0.MacroDeclInit(None,preattr::l,name,lp,args,rp,attr,eq,ini,sem) ->
+      call_right left_attribute preattr decl
+	(function preattr -> Ast0.MacroDeclInit(None,preattr::l,name,lp,args,rp,attr,eq,ini,sem))
+  | Ast0.TyDecl(ty,sem) ->
+      call_right left_ty ty decl (function ty -> Ast0.TyDecl(ty,sem))
   | Ast0.Typedef(stg,ty,id,sem) ->
       call_right left_mcode stg decl
 	(function stg -> Ast0.Typedef(stg,ty,id,sem))
@@ -363,13 +385,11 @@ let process =
 	    None -> Ast0.unwrap s
 	  | Some (pragmas,body) ->
 	      Ast0.While(whl,lp,exp,rp,body,update_before pragmas aft))
-      | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,aft) ->
+      | Ast0.For(fr,lp,first,rp,body,aft) ->
 	  (match right_statement body with
 	    None -> Ast0.unwrap s
 	  | Some (pragmas,body) ->
-	      Ast0.For
-		(fr,lp,first,e2,sem2,e3,rp,body,
-		 update_before pragmas aft))
+	      Ast0.For(fr,lp,first,rp,body,update_before pragmas aft))
       | Ast0.Iterator(nm,lp,args,rp,body,aft) ->
 	  (match right_statement body with
 	    None -> Ast0.unwrap s
@@ -377,7 +397,6 @@ let process =
 	      Ast0.Iterator(nm,lp,args,rp,body,update_before pragmas aft))
       | _ -> Ast0.unwrap s) in
 
-  let res = V0.rebuilder
-      {V0.rebuilder_functions with VT0.rebuilder_stmtfn = statement} in
+  let res = V0.rebuilder_default ~stmt:statement () in
 
   List.map res.VT0.rebuilder_rec_top_level
