@@ -90,6 +90,7 @@ let mcodebinaryOp2arity op = match Ast0.unwrap op with
 
 let mcode x = x (* nothing to do ... *)
 
+
 (* --------------------------------------------------------------------- *)
 (* Dots *)
 
@@ -179,13 +180,10 @@ let rec top_expression opt_allowed tgt expr =
       let str = dots (string_fragment arity) str in
       let rq = mcode rq in
       make_exp expr tgt arity (Ast0.StringConstant(lq,str,rq,isWchar))
-  | Ast0.FunCall(fn,lp,args,rp) ->
-      let arity = exp_same (mcode2line lp) [mcode2arity lp;mcode2arity rp] in
-      let fn = expression arity fn in
-      let lp = mcode lp in
-      let args = dots (expression arity) args in
-      let rp = mcode rp in
-      make_exp expr tgt arity (Ast0.FunCall(fn,lp,args,rp))
+  | Ast0.FunCall(fn,lb,args,rb) -> (* TODO FunCall(fn,args) *)
+      let fn = expression tgt fn in
+      let (lb,args,rb) = arg_list tgt (lb,args,rb) in
+      make_exp expr tgt tgt (Ast0.FunCall(fn,lb,args,rb))
   | Ast0.Assignment(left,op,right,simple) ->
       let arity = exp_same (mcodeassignOp2line op) [mcodeassignOp2arity op] in
       let left = expression arity left in
@@ -235,13 +233,10 @@ let rec top_expression opt_allowed tgt expr =
       let exp = expression arity exp in
       let rp = mcode rp in
       make_exp expr tgt arity (Ast0.Paren(lp,exp,rp))
-  | Ast0.ArrayAccess(exp1,lb,exp2,rb) ->
-      let arity = exp_same (mcode2line lb) [mcode2arity lb; mcode2arity rb] in
-      let exp1 = expression arity exp1 in
-      let lb = mcode lb in
-      let exp2 = expression arity exp2 in
-      let rb = mcode rb in
-      make_exp expr tgt arity (Ast0.ArrayAccess(exp1,lb,exp2,rb))
+  | Ast0.ArrayAccess(fn,lb,args,rb) -> (* TODO ArrayAccess(fn,args) *)
+      let fn = expression tgt fn in
+      let (lb,args,rb) = arg_list tgt (lb,args,rb) in
+      make_exp expr tgt tgt (Ast0.ArrayAccess(fn,lb,args,rb))
   | Ast0.RecordAccess(exp,pt,field) ->
       let arity = exp_same (mcode2line pt) [mcode2arity pt] in
       let exp = expression arity exp in
@@ -254,15 +249,20 @@ let rec top_expression opt_allowed tgt expr =
       let ar = mcode ar in
       let field = ident false arity field in
       make_exp expr tgt arity (Ast0.RecordPtAccess(exp,ar,field))
-  | Ast0.Cast(lp,ty,attr,rp,exp) ->
+  | Ast0.QualifiedAccess(ty,coloncolon,field) ->
+      let arity = exp_same (mcode2line coloncolon) [mcode2arity coloncolon] in
+      let ty = get_option (typeC arity) ty in
+      let coloncolon = mcode coloncolon in 
+      let field = ident false arity field in
+      make_exp expr tgt arity (Ast0.QualifiedAccess(ty,coloncolon,field))
+  | Ast0.Cast(lp,ty,rp,exp) ->
       let arity =
 	exp_same (mcode2line lp) (List.map mcode2arity ([lp;rp])) in
       let lp = mcode lp in
       let ty = typeC arity ty in
-      let attr = List.map (attribute arity) attr in
       let rp = mcode rp in
       let exp = expression arity exp in
-      make_exp expr tgt arity (Ast0.Cast(lp,ty,attr,rp,exp))
+      make_exp expr tgt arity (Ast0.Cast(lp,ty,rp,exp))
   | Ast0.SizeOfExpr(szf,exp) ->
       let arity = exp_same (mcode2line szf) [mcode2arity szf] in
       let szf = mcode szf in
@@ -276,6 +276,36 @@ let rec top_expression opt_allowed tgt expr =
       let ty = typeC arity ty in
       let rp = mcode rp in
       make_exp expr tgt arity (Ast0.SizeOfType(szf,lp,ty,rp))
+  | Ast0.Delete(dlt,exp) ->
+      let arity = exp_same (mcode2line dlt) [mcode2arity dlt] in
+      let dlt = mcode dlt in
+      let exp = expression arity exp in
+      make_exp expr tgt arity (Ast0.Delete(dlt,exp))
+  | Ast0.DeleteArr(dlt,lb,rb,exp) ->
+      let arity = exp_same (mcode2line dlt) (List.map mcode2arity [dlt;lb;rb]) in
+      let dlt = mcode dlt in
+      let lb = mcode lb in
+      let rb = mcode rb in
+      let exp = expression arity exp in
+      make_exp expr tgt arity (Ast0.DeleteArr(dlt,lb,rb,exp))
+  | Ast0.New(nw,pp_opt,lp2_opt,ty,rp2_opt,args_opt) ->
+      let arity =
+      match (lp2_opt,rp2_opt) with
+        | (Some lp, Some rp) -> exp_same (mcode2line nw) (List.map mcode2arity [nw;lp;rp])
+        | (None, None) -> exp_same (mcode2line nw) [mcode2arity nw]
+        | _ -> failwith "impossible" in
+      let nw = mcode nw in
+      let pp_opt = get_option (arg_list arity) pp_opt in
+      let lp2_opt = get_option mcode lp2_opt in
+      let ty = typeC arity ty in
+      let rp2_opt = get_option mcode rp2_opt in
+      let args_opt = get_option (arg_list arity) args_opt in
+      make_exp expr tgt arity (Ast0.New(nw,pp_opt,lp2_opt,ty,rp2_opt,args_opt))
+  | Ast0.TemplateInst(tn,lb,args,rb) ->
+      let arity =  all_same false tgt (mcode2line lb) [mcode2arity lb;mcode2arity rb] in
+      let tn = expression arity tn in
+      let (lb,args,rb) = arg_list tgt (lb,args,rb) in
+      make_exp expr tgt arity (Ast0.TemplateInst(tn,lb,args,rb))
   | Ast0.TypeExp(ty) -> Ast0.rewrap expr (Ast0.TypeExp(typeC tgt ty))
   | Ast0.MetaErr(name,constraints,pure)  ->
       let arity = exp_same (mcode2line name) [mcode2arity name] in
@@ -338,6 +368,13 @@ let rec top_expression opt_allowed tgt expr =
 
 and expression tgt exp = top_expression false tgt exp
 
+and arg_list tgt (lp,exp,rp) =
+  let arity = all_same false tgt (mcode2line lp) [mcode2arity lp; mcode2arity rp] in
+  let lp = mcode lp in
+  let exp = dots (expression arity) exp in
+  let rp = mcode rp in
+  (lp,exp,rp)
+
 and make_fragment =
   make_opt
     (function x -> failwith "opt not allowed for string fragment")
@@ -389,12 +426,27 @@ and make_typeC =
 
 and top_typeC tgt opt_allowed typ =
   match Ast0.unwrap typ with
-    Ast0.ConstVol(cv,ty) ->
-      let arity = all_same opt_allowed tgt (mcode2line (List.hd cv))
-	  (List.map mcode2arity cv) in
-      let cv = List.map mcode cv in
+    Ast0.ConstVol(cvbefore,ty,cvafter) ->
+      let cvs =
+	List.fold_left
+	  (fun prev ->
+	    function
+		Ast0.CV cv -> cv::prev
+	      | _ -> prev)
+	  [] (cvbefore@cvafter) in
+      let arity =
+	match cvs with
+	  [] -> tgt (* not sure *)
+	| cv::_ ->
+	    all_same opt_allowed tgt (mcode2line cv)
+	      (List.map mcode2arity cvs) in
+      let do_cvattr = function
+	  Ast0.CV cv -> Ast0.CV(mcode cv)
+	| Ast0.Attr attr -> Ast0.Attr(attribute arity attr) in
+      let cvbefore = List.map do_cvattr cvbefore in
+      let cvafter = List.map do_cvattr cvafter in
       let ty = typeC arity ty in
-      make_typeC typ tgt arity (Ast0.ConstVol(cv,ty))
+      make_typeC typ tgt arity (Ast0.ConstVol(cvbefore,ty,cvafter))
   | Ast0.BaseType(ty,strings) ->
       let arity =
 	all_same opt_allowed tgt (mcode2line (List.hd strings))
@@ -451,21 +503,26 @@ and top_typeC tgt opt_allowed typ =
       let rp = mcode rp in
       make_typeC typ tgt arity
 	(Ast0.Decimal(dec,lp,length,comma,precision_opt,rp))
-  | Ast0.EnumName(kind,name) ->
+  | Ast0.EnumName(kind,key,name) ->
       let arity =
-	all_same opt_allowed tgt (mcode2line kind) [mcode2arity kind] in
+	all_same opt_allowed tgt (mcode2line kind)
+	  ((mcode2arity kind)::(Common.default [] (function x -> [mcode2arity x]) key)) in
       let kind = mcode kind in
+      let key = get_option mcode key in
       let name = get_option (ident false arity) name in
-      make_typeC typ tgt arity (Ast0.EnumName(kind,name))
-  | Ast0.EnumDef(ty,lb,decls,rb) ->
+      make_typeC typ tgt arity (Ast0.EnumName(kind,key,name))
+  | Ast0.EnumDef(ty,base,lb,decls,rb) ->
       let arity =
 	all_same opt_allowed tgt (mcode2line lb)
-	  (List.map mcode2arity [lb;rb]) in
+	  (List.map mcode2arity ((Common.default [] (function (td,ty) -> [td]) base)@[lb;rb])) in
       let ty = typeC arity ty in
+      let base = get_option
+	  (function (td, ty) -> (mcode td, typeC arity ty))
+	  base in
       let lb = mcode lb in
-      let ids = dots (enum_decl tgt) decls in
+      let ids = (dots (enum_decl tgt)) decls in
       let rb = mcode rb in
-      make_typeC typ tgt arity (Ast0.EnumDef(ty,lb,ids,rb))
+      make_typeC typ tgt arity (Ast0.EnumDef(ty,base,lb,ids,rb))
   | Ast0.StructUnionName(kind,name) ->
       let arity =
 	all_same opt_allowed tgt (mcode2line kind)
@@ -500,11 +557,22 @@ and top_typeC tgt opt_allowed typ =
       let ty = typeC arity ty in
       let rp = mcode rp in
       make_typeC typ tgt arity (Ast0.TypeOfType(tf,lp,ty,rp))
+  | Ast0.QualifiedType(ty,coloncolon,name) ->
+      let arity = all_same opt_allowed tgt (mcode2line coloncolon) [mcode2arity coloncolon] in
+      let ty = get_option (typeC arity) ty in
+      let coloncolon = mcode coloncolon in
+      let name = ident false arity name in
+      make_typeC typ tgt arity (Ast0.QualifiedType(ty,coloncolon,name))
   | Ast0.TypeName(name) ->
       let arity =
 	all_same opt_allowed tgt (mcode2line name) [mcode2arity name] in
       let name = mcode name in
       make_typeC typ tgt arity (Ast0.TypeName(name))
+  | Ast0.TemplateType(tn,lb,args,rb) ->
+      let arity =  all_same false tgt (mcode2line lb) [mcode2arity lb;mcode2arity rb] in
+      let tn = typeC arity tn in
+      let (lb,args,rb) = arg_list tgt (lb,args,rb) in
+      make_typeC typ tgt arity (Ast0.TemplateType(tn,lb,args,rb))
   | Ast0.AutoType(auto) ->
       let arity =
 	all_same opt_allowed tgt (mcode2line auto) [mcode2arity auto] in
@@ -556,33 +624,33 @@ and declaration tgt decl =
       let arity = all_same true tgt (mcode2line name) [mcode2arity name] in
       let name = mcode name in
       make_decl decl tgt arity (Ast0.MetaDecl(name,cstr,pure))
-  | Ast0.Init(stg,ty,midattr,id,endattr,eq,exp,sem) ->
+  | Ast0.Init(al,stg,ty,id,endattr,eq,exp,sem) ->
       let arity =
 	all_same true tgt (mcode2line eq)
 	  ((match stg with None -> [] | Some x -> [mcode2arity x]) @
 	   (List.map mcode2arity [eq;sem])) in
+      let al = get_option (alignas tgt) al in
       let stg = get_option mcode stg in
       let ty = typeC arity ty in
-      let midattr = List.map (attribute arity) midattr in
       let id = ident false arity id in
       let endattr = List.map (attribute arity) endattr in
       let eq = mcode eq in
       let exp = initialiser arity exp in
       let sem = mcode sem in
-      make_decl decl tgt arity (Ast0.Init(stg,ty,midattr,id,endattr,eq,exp,sem))
-  | Ast0.UnInit(stg,ty,midattr,id,endattr,sem) ->
+      make_decl decl tgt arity (Ast0.Init(al,stg,ty,id,endattr,eq,exp,sem))
+  | Ast0.UnInit(al,stg,ty,id,endattr,sem) ->
       let arity =
 	all_same true tgt (mcode2line sem)
 	  ((match stg with None -> [] | Some x -> [mcode2arity x]) @
 	   [mcode2arity sem]) in
+      let al = get_option (alignas tgt) al in
       let stg = get_option mcode stg in
       let ty = typeC arity ty in
-      let midattr = List.map (attribute arity) midattr in
       let id = ident false arity id in
       let endattr = List.map (attribute arity) endattr in
       let sem = mcode sem in
-      make_decl decl tgt arity (Ast0.UnInit(stg,ty,midattr,id,endattr,sem))
-  | Ast0.FunProto(fi,attr,name,lp1,params,va,rp1,sem) ->
+      make_decl decl tgt arity (Ast0.UnInit(al,stg,ty,id,endattr,sem))
+  | Ast0.FunProto(fi,name,lp1,params,va,rp1,sem) ->
     let tokens = match va with
       | None -> [lp1;rp1;sem]
       | Some (c1,e1) -> [lp1;c1;e1;rp1;sem] in
@@ -590,7 +658,6 @@ and declaration tgt decl =
 	all_same true tgt (mcode2line lp1)
 	  (List.map mcode2arity tokens) in
       let fi = List.map (fninfo arity) fi in
-      let attr = List.map (attribute arity) attr in
       let name = ident false arity name in
       let lp1 = mcode lp1 in
       let params = parameter_list tgt params in
@@ -599,42 +666,45 @@ and declaration tgt decl =
         | Some (c1, e1) -> Some (mcode c1, mcode e1) in
       let rp1 = mcode rp1 in
       let sem = mcode sem in
-      make_decl decl tgt arity (Ast0.FunProto(fi,attr,name,lp1,params,va,rp1,sem))
-  | Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
+      make_decl decl tgt arity (Ast0.FunProto(fi,name,lp1,params,va,rp1,sem))
+  | Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
       let arity =
 	all_same true tgt (mcode2line lp)
 	  ((match stg with None -> [] | Some x -> [mcode2arity x]) @
 	   (List.map mcode2arity ([lp;rp;sem]))) in
       let stg = get_option mcode stg in
+      let preattr = List.map (attribute arity) preattr in
       let name = ident false arity name in
       let lp = mcode lp in
       let args = dots (expression arity) args in
       let rp = mcode rp in
       let attr = List.map (attribute arity) attr in
       let sem = mcode sem in
-      make_decl decl tgt arity (Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem))
-  | Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
+      make_decl decl tgt arity
+	(Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem))
+  | Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
       let arity =
 	all_same true tgt (mcode2line lp)
 	  ((match stg with None -> [] | Some x -> [mcode2arity x]) @
 	   (List.map mcode2arity [lp;rp;eq;sem])) in
       let stg = get_option mcode stg in
+      let preattr = List.map (attribute arity) preattr in
       let name = ident false arity name in
       let lp = mcode lp in
       let args = dots (expression arity) args in
       let rp = mcode rp in
+      let attr = List.map (attribute arity) attr in
       let ini = initialiser arity ini in
       let sem = mcode sem in
       make_decl decl tgt arity
-	(Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem))
-  | Ast0.TyDecl(ty,attr,sem) ->
+	(Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem))
+  | Ast0.TyDecl(ty,sem) ->
       let arity =
 	all_same true tgt
 	  (mcode2line sem) [mcode2arity sem] in
       let ty = typeC arity ty in
-      let attr = List.map (attribute arity) attr in
       let sem = mcode sem in
-      make_decl decl tgt arity (Ast0.TyDecl(ty,attr,sem))
+      make_decl decl tgt arity (Ast0.TyDecl(ty,sem))
   | Ast0.Typedef(stg,ty,id,sem) ->
       let arity =
 	all_same true tgt (mcode2line sem)
@@ -662,6 +732,18 @@ and declaration tgt decl =
   | Ast0.OptDecl(_) | Ast0.AsDecl _ ->
       failwith "unexpected code"
 
+and alignas tgt (Ast0.Align(align,lp,expr,rp)) =
+      let exp_same = all_same false tgt in
+      let arity = exp_same (mcode2line align)
+        [mcode2arity align; mcode2arity lp; mcode2arity rp] in
+      let align = mcode align in
+      let lp = mcode lp in
+      let expr = expression tgt expr in
+      let rp = mcode rp in
+      if tgt = arity
+      then Ast0.Align(align,lp,expr,rp)
+      else failwith "tgt must be None"
+
 (* --------------------------------------------------------------------- *)
 (* Field declaration *)
 
@@ -679,14 +761,30 @@ and field tgt decl =
       let arity = all_same true tgt (mcode2line name) [mcode2arity name] in
       let name = mcode name in
       make_field decl tgt arity (Ast0.MetaFieldList(name,lenname,cstr,pure))
-  | Ast0.Field(ty,id,bf,sem) ->
+  | Ast0.Field(ty,id,bf,endattr,sem) ->
       let arity = all_same true tgt (mcode2line sem) [mcode2arity sem] in
       let ty = typeC arity ty in
       let id = Common.map_option (ident false arity) id in
       let bitfield (c, e) = (mcode c, expression arity e) in
       let bf = Common.map_option bitfield bf in
+      let endattr = List.map (attribute arity) endattr in
       let sem = mcode sem in
-      make_field decl tgt arity (Ast0.Field(ty,id,bf,sem))
+      make_field decl tgt arity (Ast0.Field(ty,id,bf,endattr,sem))
+  | Ast0.MacroDeclField(name,lp,args,rp,attr,sem) ->
+      let arity =
+	all_same true tgt (mcode2line lp)
+	  (List.map mcode2arity ([lp;rp;sem])) in
+      let name = ident false arity name in
+      let lp = mcode lp in
+      let args = dots (expression arity) args in
+      let rp = mcode rp in
+      let attr = List.map (attribute arity) attr in
+      let sem = mcode sem in
+      make_field decl tgt arity
+	(Ast0.MacroDeclField(name,lp,args,rp,attr,sem))
+  | Ast0.CppField(di) ->
+      let di = directive tgt di in
+      Ast0.rewrap decl (Ast0.CppField(di))
   | Ast0.DisjField(starter,decls,mids,ender) ->
       let decls = List.map (field tgt) decls in
       (match List.rev decls with
@@ -817,35 +915,34 @@ and make_param =
   make_opt
     (function x -> Ast0.OptParam x)
 
+and make_template_param =
+  make_opt
+    (function x -> failwith "opt not allowed here")
+
 and parameterTypeDef tgt param =
   let param_same = all_same true tgt in
   match Ast0.unwrap param with
-    Ast0.VoidParam(ty,attr) ->
-      Ast0.rewrap param
-        (Ast0.VoidParam(typeC tgt ty,List.map (attribute tgt) attr))
-  | Ast0.Param(ty,midattr,Some id,attr) ->
+    Ast0.Param(ty,Some id,attr) ->
       let ty = top_typeC tgt true ty in
-      let midattr = List.map (attribute tgt) midattr in
       let id = ident true tgt id in
       let attr = List.map (attribute tgt) attr in
       Ast0.rewrap param
 	(match (Ast0.unwrap ty,Ast0.unwrap id) with
 	  (Ast0.OptType(ty),Ast0.OptIdent(id)) ->
-	    Ast0.OptParam(Ast0.rewrap param (Ast0.Param(ty,midattr,Some id,attr)))
+	    Ast0.OptParam(Ast0.rewrap param (Ast0.Param(ty,Some id,attr)))
 	| (Ast0.OptType(ty),_) ->
 	    fail param "arity mismatch in param declaration"
 	| (_,Ast0.OptIdent(id)) ->
 	    fail param "arity mismatch in param declaration"
-	| _ -> Ast0.Param(ty,midattr,Some id,attr))
-  | Ast0.Param(ty,midattr,None,attr) ->
+	| _ -> Ast0.Param(ty,Some id,attr))
+  | Ast0.Param(ty,None,attr) ->
       let ty = top_typeC tgt true ty in
       let attr = List.map (attribute tgt) attr in
-      assert (midattr = []);
       Ast0.rewrap param
 	(match Ast0.unwrap ty with
 	  Ast0.OptType(ty) ->
-            Ast0.OptParam(Ast0.rewrap param (Ast0.Param(ty,[],None,attr)))
-        | _ -> Ast0.Param(ty,[],None,attr))
+            Ast0.OptParam(Ast0.rewrap param (Ast0.Param(ty,None,attr)))
+        | _ -> Ast0.Param(ty,None,attr))
   | Ast0.MetaParam(name,cstr,pure) ->
       let arity = param_same (mcode2line name) [mcode2arity name] in
       let name = mcode name in
@@ -865,8 +962,36 @@ and parameterTypeDef tgt param =
   | Ast0.OptParam(_) | Ast0.AsParam _ ->
       failwith "unexpected code"
 
+and templateParameterTypeDef tgt param =
+  let param_same = all_same true tgt in
+  match Ast0.unwrap param with
+    Ast0.TypenameOrClassParam(tyorcl,id,Some (eq,ty)) ->
+      let arity = param_same (mcode2line tyorcl) [mcode2arity tyorcl; mcode2arity eq] in
+      let tyorcl = mcode tyorcl in
+      let id = ident true tgt id in
+      let eq = mcode eq in
+      let ty = top_typeC tgt true ty in
+      make_template_param param tgt arity (Ast0.TypenameOrClassParam(tyorcl,id,Some (eq,ty)))
+  | Ast0.TypenameOrClassParam(tyorcl,id,None) ->
+      let arity = param_same (mcode2line tyorcl) [mcode2arity tyorcl] in
+      let tyorcl = mcode tyorcl in
+      let id = ident true tgt id in
+      make_template_param param tgt arity (Ast0.TypenameOrClassParam(tyorcl,id,None))
+  | Ast0.VarNameParam(tyorcl,id,_) ->
+      failwith "unexpected code"
+  | Ast0.TPComma(cm) ->
+      let arity = param_same (mcode2line cm) [mcode2arity cm] in
+      let cm = mcode cm in
+      make_template_param param tgt arity (Ast0.TPComma(cm))
+  | Ast0.TPDots(dots) ->
+      let arity = param_same (mcode2line dots) [mcode2arity dots] in
+      let dots = mcode dots in
+      make_template_param param tgt arity (Ast0.TPComma(dots))
+
 and parameter_list tgt = dots (parameterTypeDef tgt)
 
+and template_parameter_list tgt = dots (templateParameterTypeDef tgt)
+
 (* --------------------------------------------------------------------- *)
 (* Top-level code *)
 
@@ -940,32 +1065,41 @@ and statement tgt stm =
       let rp = mcode rp in
       let sem = mcode sem in
       make_rule_elem stm tgt arity (Ast0.Do(d,body,wh,lp,exp,rp,sem))
-  | Ast0.For(fr,lp,first,exp2,sem2,exp3,rp,body,aft) ->
+  | Ast0.For(fr,lp,first,rp,body,aft) ->
       let arity =
-	let mcodes = [fr;lp;sem2;rp] in
+	let mcodes = [fr;lp;rp] in
 	let mcodes =
 	  match Ast0.unwrap first with
-	    Ast0.ForExp(exp1,sem1) -> sem1::mcodes
-	  | Ast0.ForDecl _ -> mcodes in
+	    Ast0.ForExp(exp1,sem1,exp2,sem2,exp3) -> sem1::sem2::mcodes
+	  | Ast0.ForDecl(bef,decl,exp2,sem2,exp3) -> sem2::mcodes
+	  | Ast0.ForRange(bef,decl,_) -> mcodes in
 	stm_same (mcode2line fr) (List.map mcode2arity mcodes) in
       let fr = mcode fr in
       let lp = mcode lp in
       let first =
 	match Ast0.unwrap first with
-	  Ast0.ForExp(exp1,sem1) ->
+	  Ast0.ForExp(exp1,sem1,exp2,sem2,exp3) ->
 	    let exp1 = get_option (expression arity) exp1 in
 	    let sem1 = mcode sem1 in
-	    Ast0.rewrap first (Ast0.ForExp(exp1,sem1))
-	| Ast0.ForDecl (bef,decl) ->
+	    let exp2 = get_option (expression arity) exp2 in
+	    let sem2 = mcode sem2 in
+	    let exp3 = get_option (expression arity) exp3 in
+	    Ast0.rewrap first (Ast0.ForExp(exp1,sem1,exp2,sem2,exp3))
+	| Ast0.ForDecl (bef,decl,exp2,sem2,exp3) ->
+	    let decl = declaration arity decl in
+	    let exp2 = get_option (expression arity) exp2 in
+	    let sem2 = mcode sem2 in
+	    let exp3 = get_option (expression arity) exp3 in
 	    Ast0.rewrap first
-	      (Ast0.ForDecl(bef,declaration arity decl)) in
-      let exp2 = get_option (expression arity) exp2 in
-      let sem2 = mcode sem2 in
-      let exp3 = get_option (expression arity) exp3 in
+	      (Ast0.ForDecl(bef,decl,exp2,sem2,exp3))
+	| Ast0.ForRange(bef,decl,ini) ->
+	    let decl = declaration arity decl in
+	    let ini = initialiser arity ini in
+	    Ast0.rewrap first (Ast0.ForRange(bef,decl,ini)) in
       let rp = mcode rp in
       let body = statement arity body in
       make_rule_elem stm tgt arity
-	(Ast0.For(fr,lp,first,exp2,sem2,exp3,rp,body,aft))
+	(Ast0.For(fr,lp,first,rp,body,aft))
   | Ast0.Iterator(nm,lp,args,rp,body,aft) ->
       let arity = stm_same (mcode2line lp) (List.map mcode2arity [lp;rp]) in
       let nm = ident false arity nm in
@@ -1175,17 +1309,13 @@ and statement tgt stm =
       let rbrace = mcode rbrace in
       make_rule_elem stm tgt arity
 	(Ast0.FunDecl(bef,fi,name,lp,params,newva,rp,attrs,lbrace,body,rbrace,aft))
-  | Ast0.Include(inc,s) ->
+  | Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt) ->
       let arity =
-	all_same true tgt (mcode2line inc) [mcode2arity inc; mcode2arity s] in
-      let inc = mcode inc in
-      let s = mcode s in
-      make_rule_elem stm tgt arity (Ast0.Include(inc,s))
-  | Ast0.MetaInclude(inc,s) ->
-      let arity = all_same true tgt (mcode2line inc) [mcode2arity inc] in
-      let inc = mcode inc in
-      let s = expression arity s in
-      make_rule_elem stm tgt arity (Ast0.MetaInclude(inc,s))
+	all_same true tgt (mcode2line lab)
+          (List.map mcode2arity [tmpkw;lab;rab]) in
+      let stmt = statement arity stmt in
+      let params = template_parameter_list arity params in
+      make_rule_elem stm tgt arity (Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt))
   | Ast0.Undef(def,id) ->
       let arity = all_same true tgt (mcode2line def) [mcode2arity def] in
       let def = mcode def in
@@ -1198,12 +1328,9 @@ and statement tgt stm =
       let params = define_parameters arity params in
       let body = dots (statement arity) body in
       make_rule_elem stm tgt arity (Ast0.Define(def,id,params,body))
-  | Ast0.Pragma(prg,id,body) ->
-      let arity = all_same true tgt (mcode2line prg) [mcode2arity prg] in
-      let prg = mcode prg in
-      let id = ident false arity id in
-      let body = pragmainfo arity body in
-      make_rule_elem stm tgt arity (Ast0.Pragma(prg,id,body))
+  | Ast0.CppTop(di) ->
+      let di = directive tgt di in
+      Ast0.rewrap stm (Ast0.CppTop(di))
   | Ast0.OptStm(_) | Ast0.AsStmt _ ->
       failwith "unexpected code"
 
@@ -1221,6 +1348,65 @@ and pragmainfo tgt pi =
       let arity = all_same false tgt (mcode2line dots) [mcode2arity dots] in
       let dots = mcode dots in
       make_pragma pi tgt arity (Ast0.PragmaDots (dots))
+  | Ast0.MetaPragmaInfo(name,constraints,pure)  ->
+      let arity = all_same false tgt (mcode2line name) [mcode2arity name] in
+      let name = mcode name in
+      make_pragma pi tgt arity (Ast0.MetaPragmaInfo(name,constraints,pure))
+
+and make_directive =
+  make_opt
+    (function x -> failwith "opt not allowed for pragma")
+
+and directive tgt di =
+  match Ast0.unwrap di with
+    Ast0.Include(inc,s) ->
+      let arity =
+	all_same true tgt (mcode2line inc) [mcode2arity inc; mcode2arity s] in
+      let inc = mcode inc in
+      let s = mcode s in
+      make_directive di tgt arity (Ast0.Include(inc,s))
+  | Ast0.MetaInclude(inc,s) ->
+      let arity = all_same true tgt (mcode2line inc) [mcode2arity inc] in
+      let inc = mcode inc in
+      let s = expression arity s in
+      make_directive di tgt arity (Ast0.MetaInclude(inc,s))
+  | Ast0.Pragma(prg,id,body) ->
+      let arity = all_same true tgt (mcode2line prg) [mcode2arity prg] in
+      let prg = mcode prg in
+      let id = ident false arity id in
+      let body = pragmainfo arity body in
+      make_directive di tgt arity (Ast0.Pragma(prg,id,body))
+  | Ast0.UsingNamespace(usng,nmspc,name,sem) ->
+      let arity =
+	all_same true tgt (mcode2line usng) [mcode2arity usng; mcode2arity nmspc; mcode2arity sem] in
+      let usng = mcode usng in
+      let nmspc = mcode nmspc in
+      let name = ident false arity name in
+      let sem = mcode sem in
+      make_directive di tgt arity (Ast0.UsingNamespace(usng,nmspc,name,sem))
+  | Ast0.UsingTypename(usng,name,eq,tn,ty,sem) ->
+      let mclist =
+        [mcode2arity usng; mcode2arity eq; mcode2arity sem] in
+      let mclist =
+        match tn with
+         Some tn -> (mcode2arity tn) :: mclist
+       | None -> mclist in
+      let arity =
+        all_same true tgt (mcode2line usng) mclist in
+      let usng = mcode usng in
+      let name = ident false arity name in
+      let eq = mcode eq in
+      let tn = get_option mcode tn in
+      let ty = typeC arity ty in
+      let sem = mcode sem in
+      make_directive di tgt arity (Ast0.UsingTypename(usng,name,eq,tn,ty,sem))
+  | Ast0.UsingMember(usng,name,sem) ->
+      let arity =
+        all_same true tgt (mcode2line usng) [mcode2arity usng; mcode2arity sem] in
+      let usng = mcode usng in
+      let name = ident false arity name in
+      let sem = mcode sem in
+      make_directive di tgt arity (Ast0.UsingMember(usng,name,sem))
 
 and define_parameters tgt params =
   match Ast0.unwrap params with
@@ -1247,6 +1433,11 @@ and define_param tgt param =
 	  Ast0.OptIdent(id) ->
 	    Ast0.OptDParam(Ast0.rewrap param (Ast0.DParam(id)))
 	| _ -> Ast0.DParam(new_id))
+  | Ast0.DParamEll(id,dots) ->
+      let arity = all_same true tgt (mcode2line dots) [mcode2arity dots] in
+      let new_id = ident true tgt id in
+      let new_dots = mcode dots in
+      make_define_param param tgt arity (Ast0.DParamEll(new_id,new_dots))
   | Ast0.MetaDParamList(name,lenname,cstr,pure) ->
       let arity = all_same true tgt (mcode2line name) [mcode2arity name] in
       let name = mcode name in
@@ -1269,7 +1460,6 @@ and fninfo arity = function
     Ast0.FStorage(stg) -> Ast0.FStorage(mcode stg)
   | Ast0.FType(ty) -> Ast0.FType(typeC arity ty)
   | Ast0.FInline(inline) -> Ast0.FInline(mcode inline)
-  | Ast0.FAttr(attr) -> Ast0.FAttr(attribute arity attr)
 
 and fninfo2arity fninfo =
   List.concat
@@ -1277,8 +1467,7 @@ and fninfo2arity fninfo =
        (function
 	   Ast0.FStorage(stg) -> [mcode2arity stg]
 	 | Ast0.FType(ty) -> []
-	 | Ast0.FInline(inline) -> [mcode2arity inline]
-	 | Ast0.FAttr(attr) -> [])
+	 | Ast0.FInline(inline) -> [mcode2arity inline])
        fninfo)
 
 and make_attribute =
@@ -1290,15 +1479,39 @@ and attribute tgt attr =
     Ast0.Attribute(arg) ->
       let arg = attr_arg tgt arg in
       Ast0.rewrap attr (Ast0.Attribute(arg))
-  | Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) ->
+  | Ast0.GccAttribute(attr_,lp1,lp2,args,rp1,rp2) ->
       let arity = all_same false tgt (mcode2line lp1) [mcode2arity lp1;mcode2arity rp2] in
       let attr_ = mcode attr_ in
       let lp1 = mcode lp1 in
       let lp2 = mcode lp2 in
-      let arg = attr_arg tgt arg in
+      let args = dots (expression arity) args in
       let rp1 = mcode rp1 in
       let rp2 = mcode rp2 in
-      make_attribute attr tgt arity (Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2))
+      make_attribute attr tgt arity (Ast0.GccAttribute(attr_,lp1,lp2,args,rp1,rp2))
+  | Ast0.CxxAttribute(lb1,args,rb1,rb2) ->
+      let arity =
+        all_same false tgt (mcode2line lb1)
+	  [mcode2arity lb1;mcode2arity rb1;mcode2arity rb2] in
+      let lb1 = mcode lb1 in
+      let args = dots (expression arity) args in
+      let rb1 = mcode rb1 in
+      let rb2 = mcode rb2 in
+      make_attribute attr tgt arity
+	(Ast0.CxxAttribute(lb1,args,rb1,rb2))
+
+  | Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2) ->
+      let arity =
+        all_same false tgt (mcode2line lb1)
+	  [mcode2arity lb1;mcode2arity rb1;mcode2arity rb2] in
+      let lb1 = mcode lb1 in
+      let usng = mcode usng in
+      let atnm = ident false arity atnm in
+      let dotdot = mcode dotdot in
+      let args = dots (expression arity) args in
+      let rb1 = mcode rb1 in
+      let rb2 = mcode rb2 in
+      make_attribute attr tgt arity
+	(Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2))
 
 and make_attr_arg =
   make_opt
