@@ -34,13 +34,11 @@ type info = { pos_info : position_info;  whitespace : string;
 	      strings_after : (Ast_cocci.added_string * position_info) list;
 	      isSymbolIdent : bool; (* is the token a symbol identifier or not *) }
 
-type adjacency = int
-
-type fake_mcode = info * mcodekind * adjacency
+type fake_mcode = info * mcodekind * Ast_cocci.adj
 
 type 'a mcode =
     'a * arity * info * mcodekind * anything list ref (* pos, - only *) *
-      adjacency (* adjacency_index *)
+      Ast_cocci.adj (* adjacency_index *)
 
 and 'a wrap =
     { node : 'a;
@@ -110,15 +108,24 @@ and base_expression =
   | Nested         of expression * binaryOp * expression
   | Paren          of string mcode (* ( *) * expression *
                       string mcode (* ) *)
-  | ArrayAccess    of expression * string mcode (* [ *) * expression *
+  | ArrayAccess    of expression * string mcode (* [ *) * expression dots *
 	              string mcode (* ] *)
   | RecordAccess   of expression * string mcode (* . *) * ident
   | RecordPtAccess of expression * string mcode (* -> *) * ident
-  | Cast           of string mcode (* ( *) * typeC * attr list *
+  | QualifiedAccess of typeC option * string mcode (* :: *) * ident
+  | Cast           of string mcode (* ( *) * typeC *
                       string mcode (* ) *) * expression
   | SizeOfExpr     of string mcode (* sizeof *) * expression
   | SizeOfType     of string mcode (* sizeof *) * string mcode (* ( *) *
-                      typeC * string mcode (* ) *)
+                   typeC * string mcode (* ) *)
+  | Delete         of string mcode (* delete *) * expression
+  | DeleteArr      of string mcode (* delete *) * string mcode (* [ *) * string mcode (* ] *) *
+	           expression
+  | New            of string mcode (* new *) * arguments option *
+	           string mcode (* ( *) option * typeC *
+	           string mcode (* ) *) option * arguments option
+  | TemplateInst of expression (* name *) * string mcode (* < *) *
+        expression dots * string mcode (* > *)
   | TypeExp        of typeC
   | Constructor    of string mcode (* ( *) * typeC * string mcode (* ) *) *
  	              initialiser
@@ -144,6 +151,8 @@ and base_expression =
 
 and expression = base_expression wrap
 
+and arguments = string mcode (* ( *) * expression dots * string mcode (* ) *)
+
 and constraints = expression Ast_cocci.generic_constraints
 
 and listlen =
@@ -187,7 +196,7 @@ and binaryOp = base_binaryOp wrap
 (* Types *)
 
 and base_typeC =
-    ConstVol        of Ast_cocci.const_vol mcode list * typeC
+    ConstVol        of cvattr list * typeC * cvattr list
   | BaseType        of Ast_cocci.baseType * string mcode list
   | Signed          of Ast_cocci.sign mcode * typeC option
   | Pointer         of typeC * string mcode (* * *)
@@ -200,9 +209,9 @@ and base_typeC =
 	               expression *
 	               string mcode option (* , *) * expression option *
 	               string mcode (* ) *) (* IBM C only *)
-  | EnumName        of string mcode (*enum*) * ident option (* name *)
-  | EnumDef  of typeC (* either StructUnionName or metavar *) *
-	string mcode (* { *) * enum_decl dots * string mcode (* } *)
+  | EnumName        of string mcode (*enum*) * Ast_cocci.structUnion mcode option (* struct/class/None *) * ident option (* name *)
+  | EnumDef         of typeC (* EnumName *) *
+	            enum_base option * string mcode (* { *) * enum_decl dots * string mcode (* } *)
   | StructUnionName of Ast_cocci.structUnion mcode * ident option (* name *)
   | StructUnionDef  of typeC (* either StructUnionName or metavar *) *
 	string mcode (* { *) * field dots * string mcode (* } *)
@@ -211,7 +220,10 @@ and base_typeC =
   | TypeOfType      of string mcode (* typeof *) * string mcode (* ( *) *
                        typeC * string mcode (* ) *)
   | TypeName        of string mcode
+  | QualifiedType   of typeC option * string mcode (* :: *) * ident (* name *)
   | AutoType        of string mcode (* auto *) (* c++ >= 11 *)
+  | TemplateType    of typeC (* name *) * string mcode (* < *) *
+                       expression dots * string mcode (* > *)
   | MetaType        of Ast_cocci.meta_name mcode * constraints
 	* pure
   | AsType          of typeC * typeC (* as type, always metavar *)
@@ -223,6 +235,8 @@ and base_typeC =
 
 and typeC = base_typeC wrap
 
+and enum_base = string mcode (* : *) * typeC
+
 (* --------------------------------------------------------------------- *)
 (* Variable declaration *)
 (* Even if the Cocci program specifies a list of declarations, they are
@@ -231,23 +245,23 @@ and typeC = base_typeC wrap
 and base_declaration =
     MetaDecl   of Ast_cocci.meta_name mcode * constraints * pure (* variables *)
   | AsDecl        of declaration * declaration
-  | Init of Ast_cocci.storage mcode option * typeC * attr list * ident *
+  | Init of alignas option * Ast_cocci.storage mcode option * typeC * ident *
         attr list * string mcode (*=*) * initialiser * string mcode (*;*)
-  | UnInit of Ast_cocci.storage mcode option * typeC * attr list * ident *
+  | UnInit of alignas option * Ast_cocci.storage mcode option * typeC * ident *
 	attr list * string mcode (* ; *)
   | FunProto of
-	fninfo list * attr list * ident (* name *) *
+	fninfo list * ident (* name *) *
 	string mcode (* ( *) * parameter_list *
         (string mcode (* , *) * string mcode (* ...... *) ) option *
 	string mcode (* ) *) * string mcode (* ; *)
-  | TyDecl of typeC * attr list * string mcode (* ; *)
-  | MacroDecl of Ast_cocci.storage mcode option *
+  | TyDecl of typeC * string mcode (* ; *)
+  | MacroDecl of Ast_cocci.storage mcode option * attr list *
 	ident (* name *) * string mcode (* ( *) *
         expression dots * string mcode (* ) *) *
         attr list * string mcode (* ; *)
-  | MacroDeclInit of Ast_cocci.storage mcode option *
+  | MacroDeclInit of Ast_cocci.storage mcode option * attr list *
 	ident (* name *) * string mcode (* ( *) *
-        expression dots * string mcode (* ) *) * string mcode (*=*) *
+        expression dots * string mcode (* ) *) * attr list * string mcode (*=*) *
         initialiser * string mcode (* ; *)
   | Typedef of string mcode (* typedef *) * typeC * typeC * string mcode (*;*)
   | DisjDecl   of string mcode * declaration list * string mcode list *
@@ -258,6 +272,9 @@ and base_declaration =
 
 and declaration = base_declaration wrap
 
+and alignas = Align of string mcode (* alignas *) * string mcode (* ( *)
+					* expression * string mcode (* ) *)
+
 (* --------------------------------------------------------------------- *)
 (* Field declaration *)
 
@@ -265,7 +282,11 @@ and base_field =
   | MetaField  of Ast_cocci.meta_name mcode * constraints *
 	pure (* structure fields *)
   | MetaFieldList of Ast_cocci.meta_name mcode * listlen * constraints * pure
-  | Field     of typeC * ident option * bitfield option * string mcode (* ; *)
+  | Field     of typeC * ident option * bitfield option * attr list * string mcode (* ; *)
+  | MacroDeclField of ident (* name *) * string mcode (* ( *) *
+        expression dots * string mcode (* ) *) *
+        attr list * string mcode (* ; *)
+  | CppField    of directive
   | DisjField   of string mcode * field list * string mcode list *
 	          string mcode
   | ConjField   of string mcode * field list * string mcode list *
@@ -321,8 +342,7 @@ and initialiser_list = initialiser dots
 (* Parameter *)
 
 and base_parameterTypeDef =
-    VoidParam     of typeC * attr list
-  | Param         of typeC * attr list * ident option * attr list
+    Param         of typeC * ident option * attr list
   | MetaParam     of Ast_cocci.meta_name mcode * constraints * pure
   | MetaParamList of Ast_cocci.meta_name mcode * listlen * constraints * pure
   | AsParam       of parameterTypeDef * expression (* expr, always metavar *)
@@ -339,6 +359,7 @@ and parameter_list = parameterTypeDef dots
 
 and base_define_param =
     DParam        of ident
+  | DParamEll     of ident * string mcode (* ... *)
   | MetaDParamList of Ast_cocci.meta_name mcode * listlen * constraints * pure
   | DPComma       of string mcode
   | DPdots        of string mcode (* ... *)
@@ -375,8 +396,7 @@ and base_statement =
 	             expression * string mcode (* ) *) *
                      string mcode (* ; *)
   | For           of string mcode (* for *) * string mcode (* ( *) * forinfo *
-	             expression option * string mcode (*;*) *
-                     expression option * string mcode (* ) *) * statement *
+	             string mcode (* ) *) * statement *
 	             fake_mcode (* after info *)
   | Iterator      of ident (* name *) * string mcode (* ( *) *
 	             expression dots * string mcode (* ) *) *
@@ -403,6 +423,7 @@ and base_statement =
   | Ty            of typeC (* only at top level *)
   | TopId         of ident (* only at top level *)
   | TopInit       of initialiser (* only at top level *)
+  | CppTop        of directive
   | Disj          of string mcode * statement dots list * string mcode list *
 	             string mcode
   | Conj          of string mcode * statement dots list * string mcode list *
@@ -419,23 +440,34 @@ and base_statement =
 	string mcode (* { *) * statement dots *
 	string mcode (* } *) *
 	(info * mcodekind) (* after the function decl *)
-  | Include of string mcode (* #include *) * Ast_cocci.inc_file mcode(* file *)
-  | MetaInclude of string mcode (* #include *) * expression (* file *)
+  | TemplateDefinition of
+	string mcode (* template *) * string mcode (* < *) *
+        template_parameter_list * string mcode (* > *) *
+        statement
   | Undef of string mcode (* #define *) * ident (* name *)
   | Define of string mcode (* #define *) * ident (* name *) *
 	define_parameters (*params*) * statement dots
-  | Pragma of string mcode (* #pragma *) * ident * pragmainfo
   | OptStm   of statement
 
-and base_pragmainfo =
-    PragmaString of string mcode
-  | PragmaDots of string mcode
-
-and pragmainfo = base_pragmainfo wrap
+and base_templateParameterTypeDef =
+    TypenameOrClassParam of string mcode (* typename|class *) * ident (* name *) *
+	(string mcode (* = *) * typeC) option
+  | VarNameParam of typeC * ident (* name *) * (string mcode (* = *) * initialiser) option
+  | TPComma of string mcode
+  | TPDots of string mcode (* ... *)
+  (* Note: TemplateParam not supported yet. *)
+and templateParameterTypeDef = base_templateParameterTypeDef wrap
+and template_parameter_list = templateParameterTypeDef dots
 
 and base_forinfo =
-    ForExp of expression option * string mcode (*;*)
-  | ForDecl of (info * mcodekind) (* before the decl *) * declaration
+    ForExp of expression option * string mcode (*;*) *
+	expression option * string mcode (*;*) *
+        expression option
+  | ForDecl of (info * mcodekind) (* before the decl *) * declaration *
+	expression option * string mcode (*;*) *
+        expression option
+  | ForRange of (info * mcodekind) (* before the decl *) * declaration *
+	initialiser
 
 and forinfo = base_forinfo wrap
 
@@ -443,13 +475,21 @@ and fninfo =
     FStorage of Ast_cocci.storage mcode
   | FType of typeC
   | FInline of string mcode
-  | FAttr of attr
+
+and cvattr =
+    CV of Ast_cocci.const_vol mcode
+  | Attr of attr
 
 and base_attr =
     Attribute of attr_arg
   | GccAttribute of string mcode (* __attribute__ *) *
                     string mcode (* ( *) * string mcode (* ( *) *
-                    attr_arg * string mcode (* ) *) * string mcode (* ) *)
+                    expression dots * string mcode (* ) *) * string mcode (* ) *)
+  | CxxAttribute of string mcode (* [[ *) *
+                    expression dots * string mcode (* ] *) * string mcode (* ] *)
+  | CxxAttributeUsing of string mcode (* [[ *) * string mcode (* using *) *
+                    ident * string mcode (* : *) *
+                    expression dots * string mcode (* ] *) * string mcode (* ] *)
 
 and attr = base_attr wrap
 
@@ -471,6 +511,26 @@ and ('a,'b) whencode =
 
 and statement = base_statement wrap
 
+and base_directive = 
+  | Include of string mcode (* #include *) * Ast_cocci.inc_file mcode (* file *)
+  | MetaInclude of string mcode (* #include *) * expression (* file *)
+  | Pragma of string mcode (* #pragma *) * ident * pragmainfo
+  | UsingNamespace of string mcode (*using*) * string mcode (*namespace*) *
+      ident (*name*) * string mcode (*;*)
+  | UsingTypename of string mcode (*using*) * ident (*name*) *
+      string mcode (*=*) * string mcode option (*typename*) *
+      typeC (*full_type*) * string mcode (*;*)
+  | UsingMember of string mcode (*using*) * ident (*name*) * string mcode (*;*)
+
+and directive = base_directive wrap
+
+and base_pragmainfo =
+    PragmaString of string mcode
+  | PragmaDots of string mcode
+  | MetaPragmaInfo of Ast_cocci.meta_name mcode * constraints * pure
+
+and pragmainfo = base_pragmainfo wrap
+
 and base_case_line =
     Default of string mcode (* default *) * string mcode (*:*) * statement dots
   | Case of string mcode (* case *) * expression * string mcode (*:*) *
@@ -555,6 +615,7 @@ and anything =
     DotsExprTag of expression dots
   | DotsInitTag of initialiser dots
   | DotsParamTag of parameterTypeDef dots
+  | DotsTemplateParamTag of templateParameterTypeDef dots
   | DotsStmtTag of statement dots
   | DotsDeclTag of declaration dots
   | DotsFieldTag of field dots
@@ -569,6 +630,7 @@ and anything =
   | TestExprTag of expression (* for isos *)
   | TypeCTag of typeC
   | ParamTag of parameterTypeDef
+  | TemplateParamTag of templateParameterTypeDef
   | InitTag of initialiser
   | DeclTag of declaration
   | FieldTag of field
@@ -591,6 +653,7 @@ and anything =
 val dotsExpr : expression dots -> anything
 val dotsInit : initialiser dots -> anything
 val dotsParam : parameterTypeDef dots -> anything
+val dotsTemplateParam : templateParameterTypeDef dots -> anything
 val dotsStmt : statement dots -> anything
 val dotsDecl : declaration dots -> anything
 val dotsField : field dots -> anything
@@ -603,6 +666,7 @@ val assignOp : assignOp -> anything
 val binaryOp : binaryOp -> anything
 val typeC : typeC -> anything
 val param : parameterTypeDef -> anything
+val template_param : templateParameterTypeDef -> anything
 val ini : initialiser -> anything
 val decl : declaration -> anything
 val field : field -> anything
