@@ -183,14 +183,8 @@ let inline_mcodes =
     | Ast0.PLUS _ -> () in
   let do_nothing r k e = do_nothing_dir true r k e in
   let do_nothing_end r k e = do_nothing_dir false r k e in
-  V0.flat_combiner bind option_default
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing do_nothing_end do_nothing_end do_nothing do_nothing
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+  V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=do_nothing}
+    ~decl:do_nothing_end ~field:do_nothing_end ()
 
 (* --------------------------------------------------------------------- *)
 (* For function declarations.  Can't use the mcode at the root, because that
@@ -267,14 +261,9 @@ let check_allminus =
 	List.for_all r.VT0.combiner_rec_case_line case_lines
     | _ -> k e in
 
-  V0.flat_combiner bind option_default
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode mcode mcode
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing
-    donothing ident expression donothing donothing typeC initialiser donothing
-    declaration field donothing statement donothing case_line donothing
-    donothing donothing donothing
+  V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=donothing}
+    ~ident:ident ~expr:expression ~ty:typeC ~init:initialiser
+    ~decl:declaration ~field:field ~stmt:statement ~case:case_line ()
 
 (* --------------------------------------------------------------------- *)
 (* --------------------------------------------------------------------- *)
@@ -433,21 +422,46 @@ and expression e =
 	Ast.Nested(expression left,binaryOp op,expression right)
     | Ast0.Paren(lp,exp,rp) ->
 	Ast.Paren(mcode lp,expression exp,mcode rp)
-    | Ast0.ArrayAccess(exp1,lb,exp2,rb) ->
-	Ast.ArrayAccess(expression exp1,mcode lb,expression exp2,mcode rb)
+    | Ast0.ArrayAccess(fn,lb,args,rb) ->
+	let fn = expression fn in
+	let lb = mcode lb in
+	let args = dots expression args in
+	let rb = mcode rb in
+	Ast.ArrayAccess(fn,lb,args,rb)
     | Ast0.RecordAccess(exp,pt,field) ->
 	Ast.RecordAccess(expression exp,mcode pt,ident field)
     | Ast0.RecordPtAccess(exp,ar,field) ->
 	Ast.RecordPtAccess(expression exp,mcode ar,ident field)
-    | Ast0.Cast(lp,ty,attr,rp,exp) ->
+    | Ast0.QualifiedAccess(ty,coloncolon,field) ->
+        let allminus = check_allminus.VT0.combiner_rec_expression e in
+        Ast.QualifiedAccess(get_option (typeC allminus) ty,mcode coloncolon,ident field)
+    | Ast0.Cast(lp,ty,rp,exp) ->
 	let allminus = check_allminus.VT0.combiner_rec_expression e in
-	let attr = List.map attribute attr in
-	Ast.Cast(mcode lp,typeC allminus ty,attr,mcode rp,expression exp)
+	Ast.Cast(mcode lp,typeC allminus ty,mcode rp,expression exp)
     | Ast0.SizeOfExpr(szf,exp) ->
 	Ast.SizeOfExpr(mcode szf,expression exp)
     | Ast0.SizeOfType(szf,lp,ty,rp) ->
 	let allminus = check_allminus.VT0.combiner_rec_expression e in
 	Ast.SizeOfType(mcode szf, mcode lp,typeC allminus ty,mcode rp)
+    | Ast0.Delete(dlt,exp) ->
+	Ast.Delete(mcode dlt,expression exp)
+    | Ast0.DeleteArr(dlt,lb,rb,exp) ->
+	Ast.DeleteArr(mcode dlt, mcode lb, mcode rb, expression exp)
+    | Ast0.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) ->
+	let allminus = check_allminus.VT0.combiner_rec_expression e in
+	let nw = mcode nw in
+	let pp_opt = get_option argslist pp_opt in
+	let lp2 = get_option mcode lp_opt in
+	let ty = typeC allminus ty in
+	let rp2 = get_option mcode rp_opt in
+	let args_opt = get_option argslist args_opt in
+	Ast.New(nw,pp_opt,lp2,ty,rp2,args_opt)
+    | Ast0.TemplateInst(tn,lp,args,rp) ->
+	let tn = expression tn in
+	let lp = mcode lp in
+	let args = dots expression args in
+	let rp = mcode rp in
+	Ast.TemplateInst(tn,lp,args,rp)
     | Ast0.TypeExp(ty) ->
 	let allminus = check_allminus.VT0.combiner_rec_expression e in
 	Ast.TypeExp(typeC allminus ty)
@@ -489,6 +503,12 @@ and expression e =
     | Ast0.OptExp(exp) -> Ast.OptExp(expression exp)) in
   if Ast0.get_test_exp e then Ast.set_test_exp e1 else e1
 
+and argslist (lp,args,rp) =
+  let lp = mcode lp in
+  let args = dots expression args in
+  let rp = mcode rp in
+  (lp,args,rp)
+
 and assignOp op =
   rewrap op no_isos
     (match Ast0.unwrap op with
@@ -542,7 +562,10 @@ and rewrap_iso t t1 = rewrap t (do_isos (Ast0.get_iso t)) t1
 and typeC allminus t =
   rewrap t (do_isos (Ast0.get_iso t))
     (match Ast0.unwrap t with
-      Ast0.ConstVol(cv,ty) ->
+      Ast0.ConstVol(cvbefore,ty,cvafter) ->
+	let do_cvattr = function
+	    Ast0.CV cv -> Ast.CV (mcode cv)
+	  | Ast0.Attr attr -> Ast.Attr(attribute attr) in
 	let rec collect_disjs t =
 	  match Ast0.unwrap t with
 	    Ast0.DisjType(_,types,_,_) ->
@@ -554,8 +577,9 @@ and typeC allminus t =
 	  List.map
 	    (function ty ->
 	      Ast.Type
-		(allminus, List.map mcode cv,
-		 rewrap_iso ty (base_typeC allminus ty)))
+		(allminus, List.map do_cvattr cvbefore,
+		 rewrap_iso ty (base_typeC allminus ty),
+		 List.map do_cvattr cvafter))
 	    (collect_disjs ty) in
 	(* one could worry that isos are lost because we flatten the
 	   disjunctions.  but there should not be isos on the disjunctions
@@ -566,11 +590,12 @@ and typeC allminus t =
     | Ast0.BaseType(_) | Ast0.Signed(_,_) | Ast0.Pointer(_,_)
     | Ast0.ParenType(_,_,_) | Ast0.FunctionType(_,_,_,_)
     | Ast0.Array(_,_,_,_) | Ast0.Decimal(_,_,_,_,_,_)
-    | Ast0.EnumName(_,_) | Ast0.StructUnionName(_,_)
-    | Ast0.StructUnionDef(_,_,_,_) | Ast0.EnumDef(_,_,_,_)
+    | Ast0.TemplateType(_,_,_,_)
+    | Ast0.EnumName(_,_,_) | Ast0.StructUnionName(_,_)
+    | Ast0.StructUnionDef(_,_,_,_) | Ast0.EnumDef(_,_,_,_,_)
     | Ast0.TypeOfExpr(_,_,_,_) | Ast0.TypeOfType(_,_,_,_)
-    | Ast0.TypeName(_) | Ast0.AutoType(_) | Ast0.MetaType(_,_,_) ->
-        Ast.Type(allminus,[],rewrap t no_isos (base_typeC allminus t))
+    | Ast0.TypeName(_) | Ast0.AutoType(_) | Ast0.MetaType(_,_,_) | Ast0.QualifiedType(_,_,_) ->
+        Ast.Type(allminus,[],rewrap t no_isos (base_typeC allminus t),[])
     | Ast0.DisjType(_,types,_,_) ->
 	Ast.DisjType(List.map (typeC allminus) types)
     | Ast0.ConjType(_,types,_,_) ->
@@ -599,10 +624,11 @@ and base_typeC allminus t =
       Ast.Decimal(mcode dec,mcode lp,expression length,
 		  get_option mcode comma,get_option expression precision_opt,
 		  mcode rp)
-  | Ast0.EnumName(kind,name) ->
-      Ast.EnumName(mcode kind,get_option ident name)
-  | Ast0.EnumDef(ty,lb,ids,rb) ->
-      Ast.EnumDef(typeC allminus ty,mcode lb,enum_decl_dots ids,mcode rb)
+  | Ast0.EnumName(kind,key,name) ->
+      Ast.EnumName(mcode kind,get_option mcode key,get_option ident name)
+  | Ast0.EnumDef(ty,base,lb,ids,rb) ->
+      Ast.EnumDef(typeC allminus ty,get_option (enum_base allminus) base,
+		  mcode lb, enum_decl_dots ids, mcode rb)
   | Ast0.StructUnionName(kind,name) ->
       Ast.StructUnionName(mcode kind,get_option ident name)
   | Ast0.StructUnionDef(ty,lb,decls,rb) ->
@@ -613,11 +639,19 @@ and base_typeC allminus t =
   | Ast0.TypeOfType(tf,lp,ty,rp) ->
       let allminus = check_allminus.VT0.combiner_rec_typeC t in
       Ast.TypeOfType(mcode tf, mcode lp,typeC allminus ty,mcode rp)
+  | Ast0.QualifiedType(ty,coloncolon,name) ->
+      Ast.QualifiedType(get_option (typeC allminus) ty,mcode coloncolon,ident name)
   | Ast0.TypeName(name) -> Ast.TypeName(mcode name)
   | Ast0.AutoType(auto) -> Ast.AutoType(mcode auto)
   | Ast0.MetaType(name,cstr,_) ->
       let cstr' = constraints cstr in
       Ast.MetaType(mcode name,cstr',unitary,false)
+  | Ast0.TemplateType(tn,lp,args,rp) ->
+      let tn = typeC allminus tn in
+      let lp = mcode lp in
+      let args = dots expression args in
+      let rp = mcode rp in
+      Ast.TemplateType(tn,lp,args,rp)
   | _ -> failwith "ast0toast: unexpected type"
 
 (* --------------------------------------------------------------------- *)
@@ -625,6 +659,13 @@ and base_typeC allminus t =
 (* Even if the Cocci program specifies a list of declarations, they are
    split out into multiple declarations of a single variable each. *)
 
+and alignas (Ast0.Align(align,lpar,expr,rpar)) =
+  let align = mcode align in
+  let lpar = mcode lpar in
+  let expr = expression expr in
+  let rpar = mcode rpar in
+  Ast.Align(align,lpar,expr,rpar)
+
 and declaration d =
   rewrap d (do_isos (Ast0.get_iso d))
     (match Ast0.unwrap d with
@@ -632,26 +673,25 @@ and declaration d =
 	Ast.MetaDecl(mcode name,constraints cstr,unitary,false)
     | Ast0.AsDecl(decl,asdecl) ->
 	Ast.AsDecl(declaration decl,declaration asdecl)
-    | Ast0.Init(stg,ty,midattr,id,endattr,eq,ini,sem) ->
+    | Ast0.Init(al,stg,ty,id,endattr,eq,ini,sem) ->
 	let allminus = check_allminus.VT0.combiner_rec_declaration d in
+	let al = get_option alignas al in
 	let stg = get_option mcode stg in
 	let ty = typeC allminus ty in
 	let id = ident id in
-	let midattr = List.map attribute midattr in
 	let endattr = List.map attribute endattr in
 	let eq = mcode eq in
 	let ini = initialiser ini in
 	let sem = mcode sem in
-	Ast.Init(stg,ty,midattr,id,endattr,eq,ini,sem)
-    | Ast0.UnInit(stg,ty,midattr,id,endattr,sem) ->
+	Ast.Init(al,stg,ty,id,endattr,eq,ini,sem)
+    | Ast0.UnInit(al,stg,ty,id,endattr,sem) ->
 	let allminus = check_allminus.VT0.combiner_rec_declaration d in
-	let midattr = List.map attribute midattr in
+	let al = get_option alignas al in
 	let endattr = List.map attribute endattr in
-	Ast.UnInit(get_option mcode stg,typeC allminus ty,midattr,ident id,endattr,
+	Ast.UnInit(al,get_option mcode stg,typeC allminus ty,ident id,endattr,
 		   mcode sem)
-    | Ast0.FunProto(fi,attr,name,lp,params,va,rp,sem) ->
+    | Ast0.FunProto(fi,name,lp,params,va,rp,sem) ->
 	  let fi = List.map fninfo fi in
-          let attr = List.map attribute attr in
 	  let name = ident name in
 	  let lp = mcode lp in
 	  let params = parameter_list params in
@@ -660,37 +700,39 @@ and declaration d =
             | Some (comma,ellipsis) -> Some(mcode comma,mcode ellipsis) in
 	  let rp = mcode rp in
 	  let sem = mcode sem in
-	  Ast.FunProto(fi,attr,name,lp,params,va,rp,sem)
-    | Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
+	  Ast.FunProto(fi,name,lp,params,va,rp,sem)
+    | Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
 	(* this would seem to need allminus... *)
 	let stg = get_option mcode stg in
+	let preattr = List.map attribute preattr in
 	let name = ident name in
 	let lp = mcode lp in
 	let args = dots expression args in
 	let rp = mcode rp in
 	let attr = List.map attribute attr in
 	let sem = mcode sem in
-	Ast.MacroDecl(stg,name,lp,args,rp,attr,sem)
-    | Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
+	Ast.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem)
+    | Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
 	(* this would seem to need allminus... *)
 	let stg = get_option mcode stg in
+	let preattr = List.map attribute preattr in
 	let name = ident name in
 	let lp = mcode lp in
 	let args = dots expression args in
 	let rp = mcode rp in
+	let attr = List.map attribute attr in
 	let eq = mcode eq in
 	let ini = initialiser ini in
 	let sem = mcode sem in
-	Ast.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem)
-    | Ast0.TyDecl(ty,attr,sem) ->
+	Ast.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem)
+    | Ast0.TyDecl(ty,sem) ->
 	let allminus = check_allminus.VT0.combiner_rec_declaration d in
-	let attr = List.map attribute attr in
-	Ast.TyDecl(typeC allminus ty,attr,mcode sem)
+	Ast.TyDecl(typeC allminus ty,mcode sem)
     | Ast0.Typedef(stg,ty,id,sem) ->
 	let allminus = check_allminus.VT0.combiner_rec_declaration d in
 	let id = typeC allminus id in
 	(match Ast.unwrap id with
-          Ast.Type(_,[],id) -> (* only MetaType or Id *)
+          Ast.Type(_,[],id,[]) -> (* only MetaType or Id *)
 	    Ast.Typedef(mcode stg,typeC allminus ty,id,mcode sem)
 	| _ -> failwith "bad typedef")
     | Ast0.DisjDecl(_,decls,_,_) -> Ast.DisjDecl(List.map declaration decls)
@@ -724,11 +766,22 @@ and field d =
     | Ast0.MetaFieldList(name,lenname,cstr,_) ->
 	let cstr' = constraints cstr in
 	Ast.MetaFieldList(mcode name,do_lenname lenname,cstr',unitary,false)
-    | Ast0.Field(ty,id,bf,sem) ->
+    | Ast0.Field(ty,id,bf,endattr,sem) ->
 	let allminus = check_allminus.VT0.combiner_rec_field d in
 	let bitfield (c, e) = (mcode c, expression e) in
 	let bf = Common.map_option bitfield bf in
-	Ast.Field(typeC allminus ty,Common.map_option ident id, bf, mcode sem)
+	let endattr = List.map attribute endattr in
+	Ast.Field(typeC allminus ty,Common.map_option ident id, bf, endattr, mcode sem)
+    | Ast0.MacroDeclField(name,lp,args,rp,attr,sem) ->
+	(* this would seem to need allminus... *)
+	let name = ident name in
+	let lp = mcode lp in
+	let args = dots expression args in
+	let rp = mcode rp in
+	let attr = List.map attribute attr in
+	let sem = mcode sem in
+	Ast.MacroDeclField(name,lp,args,rp,attr,sem)
+    | Ast0.CppField(di) -> Ast.CppField(directive di)
     | Ast0.ConjField(_,_,_,_)
     | Ast0.DisjField(_,_,_,_)
     | Ast0.OptField(_)
@@ -758,6 +811,11 @@ and annotated_field bef d =
 
 and field_dots l = dots (annotated_field None) l
 
+and enum_base allminus (tdd, typ) =
+  let tdd = mcode tdd in
+  let typ = typeC allminus typ in
+  (tdd, typ)
+
 and enum_decl d =
   rewrap d (do_isos (Ast0.get_iso d))
     (match Ast0.unwrap d with
@@ -863,11 +921,9 @@ and designator = function
 and parameterTypeDef p =
   rewrap p no_isos
     (match Ast0.unwrap p with
-      Ast0.VoidParam(ty,attr) ->
-        Ast.VoidParam(typeC false ty,List.map attribute attr)
-    | Ast0.Param(ty,midattr,id,attr) ->
+      Ast0.Param(ty,id,attr) ->
 	let allminus = check_allminus.VT0.combiner_rec_parameter p in
-	Ast.Param(typeC allminus ty,List.map attribute midattr,get_option ident id,List.map attribute attr)
+	Ast.Param(typeC allminus ty,get_option ident id,List.map attribute attr)
     | Ast0.MetaParam(name,cstr,_) ->
 	Ast.MetaParam(mcode name,constraints cstr,unitary,false)
     | Ast0.MetaParamList(name,lenname,cstr,_) ->
@@ -879,8 +935,45 @@ and parameterTypeDef p =
     | Ast0.Pdots(dots) -> Ast.Pdots(mcode dots)
     | Ast0.OptParam(param) -> Ast.OptParam(parameterTypeDef param))
 
+and templateParameterTypeDef p =
+  let allminus = check_allminus.VT0.combiner_rec_template_parameter p in
+  rewrap p no_isos
+    (match Ast0.unwrap p with
+       Ast0.TypenameOrClassParam(tyorcl,id,None) ->
+	 Ast.TypenameOrClassParam(mcode tyorcl,ident id,None)
+    |  Ast0.TypenameOrClassParam(tyorcl,id,Some (eq,ty)) ->
+	Ast.TypenameOrClassParam
+	  (mcode tyorcl,ident id,Some (mcode eq,typeC allminus ty))
+    |  Ast0.VarNameParam(ty,id,None) ->
+       Ast.VarNameParam(typeC allminus ty,ident id,None)
+    |  Ast0.VarNameParam(ty,id,Some (eq,ini)) ->
+       Ast.VarNameParam
+	  (typeC allminus ty,ident id,Some (mcode eq,initialiser ini))
+    |  Ast0.TPComma(cm) -> Ast.TPComma(mcode cm)
+    |  Ast0.TPDots(dots) -> Ast.TPDots(mcode dots)
+    (* Note: TemplateParam not supported yet. *)
+    )
+
 and parameter_list l = dots parameterTypeDef l
 
+and template_parameter_list l = dots templateParameterTypeDef l
+
+
+and directive d =
+  rewrap d no_isos
+    (match Ast0.unwrap d with
+      Ast0.Include(inc,str) -> Ast.Include(mcode inc,mcode str)
+    | Ast0.MetaInclude(inc,str) -> Ast.MetaInclude(mcode inc,expression str)
+    | Ast0.Pragma(prg,id,body) ->
+	Ast.Pragma(mcode prg,ident id,pragmainfo body)
+    | Ast0.UsingNamespace(nmspc,usng,name,sem) ->
+        Ast.UsingNamespace(mcode nmspc,mcode usng,ident name,mcode sem)
+    | Ast0.UsingTypename(usng,name,eq,tn,ty,sem) ->
+        Ast.UsingTypename(mcode usng,ident name,mcode eq,get_option mcode tn,
+                          typeC false ty,mcode sem)
+    | Ast0.UsingMember(usng,name,sem) ->
+        Ast.UsingMember(mcode usng,ident name,mcode sem))
+
 (* --------------------------------------------------------------------- *)
 (* Top-level code *)
 
@@ -942,17 +1035,14 @@ and statement s =
 		 tokenwrap wh s
 		   (Ast.WhileTail(wh,mcode lp,expression exp,mcode rp,
 				  mcode sem)))
-      | Ast0.For(fr,lp,first,exp2,sem2,exp3,rp,body,aft) ->
+      | Ast0.For(fr,lp,first,rp,body,aft) ->
 	  let fr = mcode fr in
 	  let lp = mcode lp in
 	  let first = forinfo first in
-	  let exp2 = get_option expression exp2 in
-	  let sem2= mcode sem2 in
-	  let exp3 = get_option expression exp3 in
 	  let rp = mcode rp in
 	  let body = statement Ast.NotSequencible body in
 	  Ast.For(rewrap_rule_elem s
-		    (Ast.ForHeader(fr,lp,first,exp2,sem2,exp3,rp)),
+		    (Ast.ForHeader(fr,lp,first,rp)),
 		  body,([],[],[],convert_fake_mcode aft))
       | Ast0.Iterator(nm,lp,args,rp,body,aft) ->
 	  Ast.Iterator(rewrap_rule_elem s
@@ -1062,11 +1152,16 @@ and statement s =
                       body,
                       tokenwrap rbrace s (Ast.SeqEnd(rbrace)),
                       ([],[],[],convert_allminus_mcodekind allminus aft))
-      |	Ast0.Include(inc,str) ->
-	  Ast.Atomic(rewrap_rule_elem s (Ast.Include(mcode inc,mcode str)))
-      |	Ast0.MetaInclude(inc,str) ->
-	  Ast.Atomic
-	    (rewrap_rule_elem s (Ast.MetaInclude(mcode inc,expression str)))
+      | Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt) ->
+	  let tmpkw = mcode tmpkw in
+	  let lab = mcode lab in
+	  let params = template_parameter_list params in
+	  let rab = mcode rab in
+          let stmt = statement seqible stmt in
+          Ast.TemplateDefinition
+	    (rewrap_rule_elem s (Ast.TemplateDefinitionHeader(tmpkw,lab,params,rab)),
+	     stmt)
+      | Ast0.CppTop(di) -> Ast.Atomic(rewrap_rule_elem s (Ast.CppTop(directive di)))
       |	Ast0.Undef(def,id) ->
 	  Ast.Atomic(rewrap_rule_elem s (Ast.Undef(mcode def,ident id)))
       | Ast0.Define(def,id,params,body) ->
@@ -1075,9 +1170,6 @@ and statement s =
 	       (Ast.DefineHeader
 		  (mcode def,ident id, define_parameters params)),
 	     statement_dots Ast.NotSequencible (*not sure*) body)
-      |	Ast0.Pragma(prg,id,body) ->
-	  Ast.Atomic(rewrap_rule_elem s
-		       (Ast.Pragma(mcode prg,ident id,pragmainfo body)))
       | Ast0.OptStm(stm) -> Ast.OptStm(statement seqible stm))
 
   and whencode notfn alwaysfn = function
@@ -1165,7 +1257,10 @@ and pragmainfo pi =
   rewrap pi no_isos
     (match Ast0.unwrap pi with
       Ast0.PragmaString(s) -> Ast.PragmaString(mcode s)
-    | Ast0.PragmaDots (dots) -> Ast.PragmaDots (mcode dots))
+    | Ast0.PragmaDots (dots) -> Ast.PragmaDots (mcode dots)
+    | Ast0.MetaPragmaInfo(mv, cstr, _) ->
+	let cstr' = constraints cstr in
+	Ast.MetaPragmaInfo(mcode mv, cstr', unitary, false))
 
 and define_parameters p =
   rewrap p no_isos
@@ -1178,6 +1273,7 @@ and define_param p =
   rewrap p no_isos
     (match Ast0.unwrap p with
       Ast0.DParam(id) -> Ast.DParam(ident id)
+    | Ast0.DParamEll(id,dots) -> Ast.DParamEll(ident id,mcode dots)
     | Ast0.MetaDParamList(name,lenname,cstr,_) ->
 	let cstr' = constraints cstr in
 	Ast.MetaDParamList(mcode name,do_lenname lenname,cstr',unitary,false)
@@ -1189,24 +1285,56 @@ and define_param_dots l = dots define_param l
 
 and forinfo fi =
   match Ast0.unwrap fi with
-    Ast0.ForExp(exp1,sem1) ->
+    Ast0.ForExp(exp1,sem1,exp2,sem2,exp3) ->
       let exp1 = get_option expression exp1 in
       let sem1 = mcode sem1 in
-      Ast.ForExp(exp1,sem1)
-  | Ast0.ForDecl ((_,bef),decl) -> Ast.ForDecl(annotated_decl (Some bef) decl)
+      let exp2 = get_option expression exp2 in
+      let sem2 = mcode sem2 in
+      let exp3 = get_option expression exp3 in
+      Ast.ForExp(exp1,sem1,exp2,sem2,exp3)
+  | Ast0.ForDecl ((_,bef),decl,exp2,sem2,exp3) ->
+      let decl = annotated_decl (Some bef) decl in
+      let exp2 = get_option expression exp2 in
+      let sem2 = mcode sem2 in
+      let exp3 = get_option expression exp3 in
+      Ast.ForDecl(decl,exp2,sem2,exp3)
+  | Ast0.ForRange ((_,bef),decl,ini) ->
+      let decl = annotated_decl (Some bef) decl in
+      let ini = initialiser ini in
+      Ast.ForRange(decl,ini)
 
 and fninfo = function
     Ast0.FStorage(stg) -> Ast.FStorage(mcode stg)
   | Ast0.FType(ty) -> Ast.FType(typeC false ty)
   | Ast0.FInline(inline) -> Ast.FInline(mcode inline)
-  | Ast0.FAttr(attr) -> Ast.FAttr(attribute attr)
 
 and attribute a =
   rewrap a no_isos
     (match Ast0.unwrap a with
       Ast0.Attribute(attr) -> Ast.Attribute(attr_arg attr)
-    | Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) ->
-	Ast.GccAttribute(mcode attr_,mcode lp1,mcode lp2,attr_arg arg,mcode rp1,mcode rp2))
+    | Ast0.GccAttribute(attr_,lp1,lp2,args,rp1,rp2) ->
+        let attr_ = mcode attr_ in
+	let lp1 = mcode lp1 in
+	let lp2 = mcode lp2 in
+	let args = dots expression args in
+	let rp1 = mcode rp1 in
+	let rp2 = mcode rp2 in
+	Ast.GccAttribute(attr_,lp1,lp2,args,rp1,rp2)
+    | Ast0.CxxAttribute(lb1,args,rb1,rb2) ->
+	let lb1 = mcode lb1 in
+	let args = dots expression args in
+	let rb1 = mcode rb1 in
+	let rb2 = mcode rb2 in
+	Ast.CxxAttribute(lb1,args,rb1,rb2)
+    | Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2) ->
+	let lb1 = mcode lb1 in
+	let usng = mcode usng in
+	let atnm = ident atnm in
+	let dotdot = mcode dotdot in
+	let args = dots expression args in
+	let rb1 = mcode rb1 in
+	let rb2 = mcode rb2 in
+	Ast.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2))
 
 and attr_arg a =
   rewrap a no_isos
@@ -1260,6 +1388,7 @@ and statement_dots l = dots statement l
 and anything = function
     Ast0.DotsExprTag(d) -> Ast.ExprDotsTag(expression_dots d)
   | Ast0.DotsParamTag(d) -> Ast.ParamDotsTag(parameter_list d)
+  | Ast0.DotsTemplateParamTag(d) -> failwith "TODO: FIXME: "
   | Ast0.DotsInitTag(d) -> failwith "not possible"
   | Ast0.DotsStmtTag(d) -> Ast.StmtDotsTag(statement_dots d)
   | Ast0.DotsDeclTag(d) -> Ast.AnnDeclDotsTag(declaration_dots d)
@@ -1275,6 +1404,7 @@ and anything = function
      failwith "only in isos, not converted to ast"
   | Ast0.TypeCTag(d) -> Ast.FullTypeTag(typeC false d)
   | Ast0.ParamTag(d) -> Ast.ParamTag(parameterTypeDef d)
+  | Ast0.TemplateParamTag(d) -> Ast.TemplateParamTag(templateParameterTypeDef d)
   | Ast0.InitTag(d) -> Ast.InitTag(initialiser d)
   | Ast0.DeclTag(d) -> Ast.DeclarationTag(declaration d)
   | Ast0.FieldTag(d) -> Ast.FieldTag(field d)
