@@ -46,7 +46,8 @@ and 'a replacement = REPLACEMENT of 'a list list * count | NOREPLACEMENT
 and 'a mcode = 'a * info * mcodekind * meta_pos list (* pos variables *)
  (* pos is an offset indicating where in the C code the mcodekind has an
  effect *)
-and adjacency = ALLMINUS | ADJ of int
+and adj = { counter : int; mutable ender : bool; }
+and adjacency = ALLMINUS | ADJ of adj
  and mcodekind =
     MINUS       of pos * int list * adjacency * anything replacement
   | CONTEXT     of pos * anything befaft
@@ -89,6 +90,7 @@ and metavar =
   | MetaParamListDecl of arity * meta_name (*name*) * list_len (*len*)
   | MetaBinaryOperatorDecl of arity * meta_name
   | MetaAssignmentOperatorDecl of arity * meta_name
+  | MetaPragmaInfoDecl of arity * meta_name
   | MetaConstDecl of
       arity * meta_name (* name *) * fullType list option
   | MetaErrDecl of arity * meta_name (* name *)
@@ -175,16 +177,25 @@ and base_expression =
   | Unary          of expression * unaryOp mcode
   | Binary         of expression * binaryOp * expression
   | Nested         of expression * binaryOp * expression
-  | ArrayAccess    of expression * string mcode (* [ *) * expression *
+  | ArrayAccess    of expression * string mcode (* [ *) * expression dots *
 	              string mcode (* ] *)
   | RecordAccess   of expression * string mcode (* . *) * ident
   | RecordPtAccess of expression * string mcode (* -> *) * ident
-  | Cast           of string mcode (* ( *) * fullType * attr list *
+  | QualifiedAccess of fullType option * string mcode (* :: *) * ident
+  | Cast           of string mcode (* ( *) * fullType *
                       string mcode (* ) *) * expression
 
   | SizeOfExpr     of string mcode (* sizeof *) * expression
   | SizeOfType     of string mcode (* sizeof *) * string mcode (* ( *) *
                       fullType * string mcode (* ) *)
+  | Delete         of string mcode (* delete *) * expression
+  | DeleteArr      of string mcode (* delete *) * string mcode (* [ *) * string mcode (* ] *) *
+	              expression
+  | New            of string mcode (* new *) * arguments option *
+	           string mcode (* ( *) option * fullType *
+                   string mcode (* ) *) option * arguments option
+  | TemplateInst   of expression (* name *) * string mcode (* < *) *
+                      expression dots * string mcode (* > *)
   | TypeExp        of fullType
 
   | Paren          of string mcode (* ( *) * expression *
@@ -221,6 +232,7 @@ and base_expression =
 and 'expression generic_constraints =
     CstrFalse
   | CstrTrue
+  | CstrExit
   | CstrAnd of 'expression generic_constraints list
   | CstrOr of 'expression generic_constraints list
   | CstrNot of 'expression generic_constraints
@@ -259,6 +271,8 @@ and form = ANY | ID | LocalID| GlobalID | CONST (* form for MetaExp *)
 
 and expression = base_expression wrap
 
+and arguments = string mcode (* ( *) * expression dots * string mcode (* ) *)
+
 and listlen =
     MetaListLen of meta_name mcode * constraints * keep_binding * inherited
   | CstListLen of int
@@ -280,7 +294,8 @@ and base_string_format =
 
 and string_format = base_string_format wrap
 
-and  unaryOp = GetRef | GetRefLabel | DeRef | UnPlus |  UnMinus | Tilde | Not
+and  unaryOp =
+      GetRef | GetRefLabel | DeRef | UnPlus |  UnMinus | Tilde of string | Not of string
 and  base_assignOp =
     SimpleAssign of simpleAssignOp mcode
   | OpAssign of arithOp mcode
@@ -297,8 +312,11 @@ and  base_binaryOp =
       meta_name mcode * constraints * keep_binding * inherited
 and binaryOp = base_binaryOp wrap
 and  arithOp =
-    Plus | Minus | Mul | Div | Mod | DecLeft | DecRight | And | Or | Xor | Min | Max
-and  logicalOp = Inf | Sup | InfEq | SupEq | Eq | NotEq | AndLog | OrLog
+    Plus | Minus | Mul | Div | Mod | DecLeft | DecRight
+  | And of string | Or of string | Xor of string
+  | Min | Max
+and  logicalOp = Inf | Sup | InfEq | SupEq | Eq | NotEq of string
+  | AndLog of string | OrLog of string
 
 and constant =
     String of string * isWchar
@@ -312,8 +330,7 @@ and isWchar = IsWchar | IsUchar | Isuchar | Isu8char | IsChar
 (* Types *)
 
 and base_fullType =
-    Type            of bool (* true if all minus *) *
-                       const_vol mcode list * typeC
+    Type            of bool (* true if all minus *) * cvattr list * typeC * cvattr list
   | AsType          of fullType * fullType (* as type, always metavar *)
   | DisjType        of fullType list
   | ConjType        of fullType list
@@ -332,8 +349,8 @@ and base_typeC =
 	               expression *
 	               string mcode option (* , *) * expression option *
 	               string mcode (* ) *) (* IBM C only *)
-  | EnumName        of string mcode (*enum*) * ident option (* name *)
-  | EnumDef  of fullType (* either EnumName or metavar *) *
+  | EnumName        of string mcode (*enum*) * structUnion mcode option * ident option (* name *)
+  | EnumDef  of fullType (* either EnumName or metavar *) * enum_base option *
 	string mcode (* { *) * enum_decl dots * string mcode (* } *)
   | StructUnionName of structUnion mcode * ident option (* name *)
   | StructUnionDef  of fullType (* either StructUnionName or metavar *) *
@@ -343,7 +360,10 @@ and base_typeC =
   | TypeOfType      of string mcode (* typeof *) * string mcode (* ( *) *
                        fullType * string mcode (* ) *)
   | TypeName        of string mcode
+  | QualifiedType   of fullType option * string mcode (* :: *) * ident
   | AutoType        of string mcode (* auto *) (* c++ >= 11 *)
+  | TemplateType    of fullType (* name *) * string mcode (* < *) *
+                       expression dots * string mcode (* > *)
 
   | MetaType        of meta_name mcode * constraints * keep_binding *
 	inherited
@@ -358,7 +378,7 @@ and baseType = VoidType | CharType | ShortType | ShortIntType | IntType
 | SizeType | SSizeType | PtrDiffType
 | BoolType | Unknown
 
-and structUnion = Struct | Union
+and structUnion = Struct | Union | Class
 
 and sign = Signed | Unsigned
 
@@ -370,23 +390,23 @@ and const_vol = Const | Volatile
    split out into multiple declarations of a single variable each. *)
 
 and base_declaration =
-  | Init of storage mcode option * fullType * attr list * ident *
+  | Init of alignas option * storage mcode option * fullType * ident *
         attr list * string mcode (*=*) * initialiser * string mcode (*;*)
-  | UnInit of storage mcode option * fullType * attr list * ident *
+  | UnInit of alignas option * storage mcode option * fullType * ident *
 	attr list * string mcode (* ; *)
   | FunProto of
-	fninfo list * attr list * ident (* name *) *
+	fninfo list * ident (* name *) *
 	string mcode (* ( *) * parameter_list *
 	(string mcode (* , *) * string mcode (* ...... *) ) option *
 	string mcode (* ) *) * string mcode (* ; *)
-  | TyDecl of fullType * attr list * string mcode (* ; *)
-  | MacroDecl of storage mcode option *
+  | TyDecl of fullType * string mcode (* ; *)
+  | MacroDecl of storage mcode option * attr list *
 	ident (* name *) * string mcode (* ( *) *
         expression dots * string mcode (* ) *) *
         attr list * string mcode (* ; *)
-  | MacroDeclInit of storage mcode option *
+  | MacroDeclInit of storage mcode option * attr list *
 	ident (* name *) * string mcode (* ( *) *
-        expression dots * string mcode (* ) *) * string mcode (*=*) *
+        expression dots * string mcode (* ) *) * attr list * string mcode (*=*) *
         initialiser * string mcode (* ; *)
   | Typedef of string mcode (*typedef*) * fullType * typeC * string mcode (*;*)
   | DisjDecl   of declaration list
@@ -398,6 +418,9 @@ and base_declaration =
 
 and declaration = base_declaration wrap
 
+and alignas = Align of string mcode (* alignas *) * string mcode (* ( *)
+					* expression * string mcode (* ) *)
+
 and base_annotated_decl =
     DElem of mcodekind (* before the decl *) * bool (* true if all minus *) *
       declaration
@@ -408,7 +431,12 @@ and annotated_decl = base_annotated_decl wrap
 (* Field declaration *)
 
 and base_field =
-    Field of fullType * ident option * bitfield option * string mcode (* ; *)
+    Field of fullType * ident option * bitfield option * attr list * string mcode (* ; *)
+  | MacroDeclField of
+	ident (* name *) * string mcode (* ( *) *
+        expression dots * string mcode (* ) *) *
+        attr list * string mcode (* ; *)
+  | CppField  of directive
   | MetaField of meta_name mcode * constraints * keep_binding * inherited
   | MetaFieldList of meta_name mcode * listlen * constraints * keep_binding *
 	inherited
@@ -433,7 +461,7 @@ and base_enum_decl =
   | EnumDots of string mcode (* ... *) * enum_decl option (* whencode *)
 
 and enum_decl = base_enum_decl wrap
-
+and enum_base = string mcode (* : *) * fullType
 
 (* --------------------------------------------------------------------- *)
 (* Initializers *)
@@ -470,8 +498,7 @@ and initialiser = base_initialiser wrap
 (* Parameter *)
 
 and base_parameterTypeDef =
-    VoidParam     of fullType * attr list
-  | Param         of fullType * attr list * ident option * attr list
+    Param         of fullType * ident option * attr list
 
   | MetaParam     of meta_name mcode * constraints * keep_binding * inherited
   | MetaParamList of meta_name mcode * listlen * constraints * keep_binding *
@@ -494,6 +521,7 @@ and parameter_list = parameterTypeDef dots
 
 and base_define_param =
     DParam        of ident
+  | DParamEll     of ident * string mcode (* ... *)
   | MetaDParamList of meta_name mcode * listlen * constraints * keep_binding *
 	inherited
   | DPComma       of string mcode
@@ -550,9 +578,7 @@ and base_rule_elem =
 	             expression * string mcode (* ) *) *
                      string mcode (* ; *)
   | ForHeader     of string mcode (* for *) * string mcode (* ( *) *
-                     forinfo *
-	             expression option * string mcode (*;*) *
-                     expression option * string mcode (* ) *)
+                     forinfo * string mcode (* ) *)
   | IteratorHeader of ident (* name *) * string mcode (* ( *) *
 	             expression dots * string mcode (* ) *)
   | SwitchHeader  of string mcode (* switch *) * string mcode (* ( *) *
@@ -578,38 +604,62 @@ and base_rule_elem =
   | Ty            of fullType (* only at top level *)
   | TopId         of ident (* only at top level *)
   | TopInit       of initialiser (* only at top level *)
-  | Include       of string mcode (*#include*) * inc_file mcode (*file *)
-  | MetaInclude   of string mcode (* #include *) * expression (* file *)
+  | CppTop        of directive
   | Undef         of string mcode (* #define *) * ident (* name *)
   | DefineHeader  of string mcode (* #define *) * ident (* name *) *
 	             define_parameters (*params*)
-  | Pragma        of string mcode (* #pragma *) * ident * pragmainfo
+  | TemplateDefinitionHeader
+                  of string mcode (* template *) * string mcode (* < *) *
+                      template_parameter_list * string mcode (* > *)
   | Case          of string mcode (* case *) * expression * string mcode (*:*)
   | Default       of string mcode (* default *) * string mcode (*:*)
   | AsRe          of rule_elem * rule_elem (* always { and MetaStmtList *)
   | DisjRuleElem  of rule_elem list
 
+and base_templateParameterTypeDef =
+    TypenameOrClassParam of string mcode (* typename|class *) * ident (* name *) *
+	(string mcode (* = *) * fullType) option
+  | VarNameParam of fullType * ident (* name *) * (string mcode (* = *) * initialiser) option
+  | TPComma of string mcode
+  | TPDots of string mcode (* ... *)
+  (* Note: TemplateParam not supported yet. *)
+and templateParameterTypeDef = base_templateParameterTypeDef wrap
+and template_parameter_list = templateParameterTypeDef dots
+
 and base_pragmainfo =
     PragmaString of string mcode
   | PragmaDots of string mcode
+  | MetaPragmaInfo of meta_name mcode * constraints * keep_binding * inherited
 
 and pragmainfo = base_pragmainfo wrap
 
 and forinfo =
-    ForExp of expression option * string mcode (*;*)
-  | ForDecl of annotated_decl
+    ForExp of expression option * string mcode (*;*) *
+	expression option * string mcode (*;*) *
+        expression option
+  | ForDecl of annotated_decl * expression option * string mcode (*;*) *
+        expression option
+  | ForRange of annotated_decl * initialiser
 
 and fninfo =
     FStorage of storage mcode
   | FType of fullType
   | FInline of string mcode
-  | FAttr of attr
+
+and cvattr =
+    CV of const_vol mcode
+  | Attr of attr
 
 and base_attr =
     Attribute of attr_arg
   | GccAttribute of string mcode (* __attribute__ *) *
                     string mcode (* ( *) * string mcode (* ( *) *
-                    attr_arg * string mcode (* ) *) * string mcode (* ) *)
+                    expression dots * string mcode (* ) *) * string mcode (* ) *)
+  | CxxAttribute of string mcode (* [[ *) *
+                    expression dots * string mcode (* ] *) * string mcode (* ] *)
+  | CxxAttributeUsing of string mcode (* [[ *) * string mcode (* using *) *
+                    ident * string mcode (* : *) *
+                    expression dots * string mcode (* ] *) * string mcode (* ] *)
 
 and attr = base_attr wrap
 
@@ -647,6 +697,7 @@ and base_statement =
 	             dots_whencode list * dots_whencode list
   | FunDecl       of rule_elem (* header *) * rule_elem (* { *) *
      	             statement dots * rule_elem (* } *) * end_info (*exit*)
+  | TemplateDefinition of rule_elem * statement
   | Define        of rule_elem (* header *) * statement dots
   | AsStmt        of statement * statement (* as statement, always metavar *)
   | Dots          of string mcode (* ... *) *
@@ -654,6 +705,20 @@ and base_statement =
 	             dots_whencode list * dots_whencode list
   | OptStm        of statement
 
+and base_directive =
+    Include       of string mcode (*#include*) * inc_file mcode (*file*)
+  | MetaInclude   of string mcode (* #include *) * expression (* file *)
+  | Pragma        of string mcode (* #pragma *) * ident * pragmainfo
+  | UsingNamespace of string mcode (*using*) * string mcode (*namespace*) *
+      ident (*name*) * string mcode (*;*)
+  | UsingTypename of string mcode (*using*) * ident (*name*) *
+      string mcode (*=*) * string mcode option (*typename*) *
+      fullType (*full_type*) * string mcode (*;*)
+  | UsingMember of string mcode (*using*) * ident (*name*) *
+      string mcode (*;*)
+
+and directive = base_directive wrap
+
 and ('a,'b) whencode =
     WhenNot of 'a
   | WhenAlways of 'b
@@ -793,6 +858,7 @@ and anything =
   | Code                of top_level
   | ExprDotsTag         of expression dots
   | ParamDotsTag        of parameterTypeDef dots
+  | TemplateParamDotsTag of templateParameterTypeDef dots
   | StmtDotsTag         of statement dots
   | AnnDeclDotsTag      of annotated_decl dots
   | AnnFieldDotsTag     of annotated_field dots
@@ -800,6 +866,7 @@ and anything =
   | DefParDotsTag       of define_param dots
   | TypeCTag            of typeC
   | ParamTag            of parameterTypeDef
+  | TemplateParamTag    of templateParameterTypeDef
   | SgrepStartTag       of string
   | SgrepEndTag         of string
 
@@ -878,7 +945,7 @@ val string_of_binaryOp : binaryOp -> string
 
 val string_of_sign : sign -> string
 val string_of_baseType : baseType -> string
-val string_of_const_vol : const_vol list -> string
+val string_of_const_vol : const_vol -> string
 val string_of_structUnion : structUnion -> string
 val string_of_typeC : typeC -> string
 val string_of_fullType : fullType -> string
@@ -900,7 +967,7 @@ type 'a transformer = {
     baseType: (baseType -> string mcode list -> 'a) option;
     decimal: (string mcode -> string mcode -> expression ->
       string mcode option -> expression option -> string mcode -> 'a) option;
-    enumName: (string mcode -> ident option -> 'a) option;
+    enumName: (string mcode -> structUnion mcode option -> ident option -> 'a) option;
     structUnionName: (structUnion mcode -> ident option -> 'a) option;
     typeName: (string mcode -> 'a) option;
     metaType: (meta_name mcode -> constraints -> keep_binding ->
