@@ -17,7 +17,12 @@ module VT0 = Visitor_ast0_types
 (* all fresh identifiers *)
 let fresh_table = (Hashtbl.create(51) : (Ast.meta_name, unit) Hashtbl.t)
 
-let warning s = Printf.fprintf stderr "warning: %s\n" s
+let warning line s =
+  let line =
+    if line > 0
+    then Printf.sprintf "line %d: " line
+    else "" in
+  Printf.eprintf "warning: %s%s\n" line s
 
 let promote name = (name,(),Ast0.default_info(),(),None,-1)
 
@@ -45,6 +50,21 @@ let check_table table minus (name,_,info,_,_,_) =
 	with Not_found -> ()))
   else (try (find_loop table name) := true with Not_found -> ())
 
+let ast_check_table table minus name =
+  if minus
+  then
+    (try (find_loop table name) := true
+    with
+      Not_found ->
+	(try
+	  Hashtbl.find fresh_table name;
+	  let (_,name) = name in
+	  failwith
+	    (Printf.sprintf
+	       "unexpected use of a fresh identifier %s" name)
+	with Not_found -> ()))
+  else (try (find_loop table name) := true with Not_found -> ())
+
 let get_opt fn = Common.do_option fn
 
 (* --------------------------------------------------------------------- *)
@@ -57,40 +77,47 @@ let dots fn d = List.iter fn (Ast0.unwrap d)
 
 type context = ID | FIELD | FN | GLOBAL
 
+let warned = ref ([] : string list)
+
 (* heuristic for distinguishing ifdef variables from undeclared metavariables*)
 let is_ifdef name =
   String.length name > 2 && Stdcompat.String.uppercase_ascii name = name
 
 let rec ident context old_metas table minus i =
   match Ast0.unwrap i with
-    Ast0.Id((name,_,info,_,_,_) : string Ast0.mcode) ->
+    Ast0.Id((name,_,info,_,_,_) : string Ast0.mcode) when not (List.mem name !warned) ->
       let rl = info.Ast0.pos_info.Ast0.line_start in
       let is_plus i =
  	match Ast0.get_mcodekind i with Ast0.PLUS _ -> true | _ -> false in
-      let err =
-	if List.exists (function x -> x = name) old_metas
-	    && (minus || is_plus i)
-	then
-	  begin
-	    warning
-	      (Printf.sprintf
-		 "line %d: %s, previously declared as a metavariable, is used as an identifier" rl name);
-	    true
-	  end
-	else false in
-      (match context with
-	ID ->
-	  if not (is_ifdef name) && minus &&
-	    not err(* warn only once per id *) && not info.Ast0.isSymbolIdent
-	  then
-	    warning
-	      (Printf.sprintf "line %d: should %s be a metavariable?" rl name)
-      | _ -> ())
-  | Ast0.MetaId(name,_,seedval,_) ->
+      if List.exists (function x -> x = name) old_metas
+	  && (minus || is_plus i)
+      then
+	begin
+	  warned := name :: !warned;
+	  warning rl
+	    (Printf.sprintf
+	       "%s, previously declared as a metavariable, is used as an identifier" name)
+	end
+      else
+	(match context with
+	  ID ->
+	    warned := name :: !warned;
+	    if not (is_ifdef name) && minus && not info.Ast0.isSymbolIdent
+	    then
+	      warning rl
+		(Printf.sprintf "should %s be a metavariable?" name)
+	| _ -> ())
+  | Ast0.Id _ -> ()
+  | Ast0.MetaId(name,cstr,seedval,_) ->
       check_table table minus name;
+      constraints table minus cstr;
       seed table minus seedval
-  | Ast0.MetaFunc(name,_,_) -> check_table table minus name
-  | Ast0.MetaLocalFunc(name,_,_) -> check_table table minus name
+  | Ast0.MetaFunc(name,cstr,_) ->
+      check_table table minus name;
+      constraints table minus cstr
+  | Ast0.MetaLocalFunc(name,cstr,_) ->
+      check_table table minus name;
+      constraints table minus cstr
   | Ast0.AsIdent(id,asid) -> failwith "not generated yet"
   | Ast0.DisjId(_,id_list,_,_) | Ast0.ConjId(_,id_list,_,_) ->
       List.iter (ident context old_metas table minus) id_list
@@ -114,20 +141,24 @@ and seed table minus = function
 (* --------------------------------------------------------------------- *)
 (* Operators *)
 
-let assignOp context old_metas table minus op = match Ast0.unwrap op with
+and assignOp context old_metas table minus op = match Ast0.unwrap op with
   | Ast0.SimpleAssign _ -> ()
   | Ast0.OpAssign _ -> ()
-  | Ast0.MetaAssign (name,_,_) -> check_table table minus name
+  | Ast0.MetaAssign (name,cstr,_) ->
+      check_table table minus name;
+      constraints table minus cstr
 
-let binaryOp context old_metas table minus op = match Ast0.unwrap op with
+and binaryOp context old_metas table minus op = match Ast0.unwrap op with
   | Ast0.Arith _ -> ()
   | Ast0.Logical _ -> ()
-  | Ast0.MetaBinary (name,_,_) -> check_table table minus name
+  | Ast0.MetaBinary (name,cstr,_) ->
+      check_table table minus name;
+      constraints table minus cstr
 
 (* --------------------------------------------------------------------- *)
 (* Expression *)
 
-let rec expression context old_metas table minus e =
+and expression context old_metas table minus e =
   match Ast0.unwrap e with
     Ast0.Ident(id) ->
       ident context old_metas table minus id
@@ -163,25 +194,36 @@ let rec expression context old_metas table minus e =
       expression ID old_metas table minus right
   | Ast0.Paren(lp,exp,rp) ->
       expression ID old_metas table minus exp
-  | Ast0.ArrayAccess(exp1,lb,exp2,rb) ->
-      expression ID old_metas table minus exp1;
-      expression ID old_metas table minus exp2
+  | Ast0.ArrayAccess(fn,lb,args,rb) ->
+      expression FN old_metas table minus fn;
+      dots (expression ID old_metas table minus) args
   | Ast0.RecordAccess(exp,pt,field) ->
       expression ID old_metas table minus exp;
       ident FIELD old_metas table minus field
   | Ast0.RecordPtAccess(exp,ar,field) ->
       expression ID old_metas table minus exp;
       ident FIELD old_metas table minus field
-  | Ast0.Cast(lp,ty,attr,rp,exp) ->
+  | Ast0.QualifiedAccess(ty,coloncolon,field) ->
+      get_opt (typeC old_metas table minus) ty;
+      ident FIELD old_metas table minus field
+  | Ast0.Cast(lp,ty,rp,exp) ->
       typeC old_metas table minus ty;
-      List.iter (attribute old_metas table minus) attr;
       expression ID old_metas table minus exp
   | Ast0.SizeOfExpr(szf,exp) -> expression ID old_metas table minus exp
   | Ast0.SizeOfType(szf,lp,ty,rp) -> typeC old_metas table minus ty
+  | Ast0.Delete(dlt, exp) -> expression ID old_metas table minus exp
+  | Ast0.DeleteArr(dlt,lb,rb,exp) -> expression ID old_metas table minus exp
+  | Ast0.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) ->
+      typeC old_metas table minus ty;
+      get_opt (check_args (expression ID old_metas table minus)) pp_opt;
+      get_opt (check_args (expression ID old_metas table minus)) args_opt;
+  | Ast0.TemplateInst(name,lab,args,rab) ->
+      expression ID old_metas table minus name;
+      dots (expression ID old_metas table minus) args
   | Ast0.TypeExp(ty) -> typeC old_metas table minus ty
   | Ast0.Constructor(lp,ty,rp,init) ->
       typeC old_metas table minus ty; initialiser old_metas table minus init
-  | Ast0.MetaExpr(name,_,Some tys,_,_,bitfield) ->
+  | Ast0.MetaExpr(name,cstr,Some tys,_,_,bitfield) ->
       List.iter
 	(function x ->
 	  List.iter
@@ -189,12 +231,15 @@ let rec expression context old_metas table minus e =
 	    (Ast0.meta_names_of_typeC x))
 	tys;
       check_table table minus name;
+      constraints table minus cstr;
       Common.do_option (check_len table minus) bitfield
-  | Ast0.MetaExpr(name,_,_,_,_,_) | Ast0.MetaErr(name,_,_) ->
-      check_table table minus name
-  | Ast0.MetaExprList(name,len,_,_) ->
+  | Ast0.MetaExpr(name,cstr,_,_,_,_) | Ast0.MetaErr(name,cstr,_) ->
       check_table table minus name;
-      check_len table minus len
+      constraints table minus cstr
+  | Ast0.MetaExprList(name,len,cstr,_) ->
+      check_table table minus name;
+      check_len table minus len;
+      constraints table minus cstr
   | Ast0.AsExpr(exp,asexp) -> failwith "not generated yet"
   | Ast0.AsSExpr(exp,asstm) -> failwith "not generated yet"
   | Ast0.DisjExpr(_,exps,_,_)
@@ -211,15 +256,25 @@ let rec expression context old_metas table minus e =
 
 and check_len table minus len =
   match len with
-    Ast0.MetaListLen (lenname, _) -> check_table table minus lenname
+    Ast0.MetaListLen (lenname,cstr) ->
+      check_table table minus lenname;
+      constraints table minus cstr
   | _ -> ()
 
+and check_args fn (_,args,_) =
+    dots fn args
 (* --------------------------------------------------------------------- *)
 (* Types *)
 
 and typeC old_metas table minus t =
   match Ast0.unwrap t with
-    Ast0.ConstVol(cv,ty) -> typeC old_metas table minus ty
+    Ast0.ConstVol(cvbefore,ty,cvafter) ->
+      let do_cva = function
+	  Ast0.CV cv -> ()
+	| Ast0.Attr attr -> attribute old_metas table minus attr in
+      List.iter do_cva cvbefore;
+      typeC old_metas table minus ty;
+      List.iter do_cva cvafter
   | Ast0.Signed(sgn,ty) ->
       get_opt (typeC old_metas table minus) ty
   | Ast0.Pointer(ty,star) -> typeC old_metas table minus ty
@@ -236,16 +291,21 @@ and typeC old_metas table minus t =
       get_opt (expression ID old_metas table minus) precision_opt
   | Ast0.TypeOfExpr(_, _, exp, _) -> expression ID old_metas table minus exp
   | Ast0.TypeOfType(_, _, ty, _) -> typeC old_metas table minus ty
-  | Ast0.MetaType(name,_,_) ->
-      check_table table minus name
+  | Ast0.TemplateType(name,lab,args,rab) ->
+      typeC old_metas table minus name;
+      dots (expression ID old_metas table minus) args
+  | Ast0.MetaType(name,cstr,_) ->
+      check_table table minus name;
+      constraints table minus cstr
   | Ast0.AsType(ty,asty) -> failwith "not generated yet"
   | Ast0.DisjType(_,types,_,_)
   | Ast0.ConjType(_,types,_,_) ->
       List.iter (typeC old_metas table minus) types
-  | Ast0.EnumName(en,Some id) -> ident GLOBAL old_metas table minus id
-  | Ast0.EnumDef(ty,lb,ids,rb) ->
+  | Ast0.EnumName(en, key, Some id) -> ident GLOBAL old_metas table minus id
+  | Ast0.EnumDef(ty,base,lb,ids,rb) ->
       typeC old_metas table minus ty;
-      dots (enum_decl GLOBAL old_metas table minus) ids
+      Common.do_option (function (td,ty) -> typeC old_metas table minus ty) base;
+      (dots (enum_decl GLOBAL old_metas table minus)) ids
   | Ast0.StructUnionName(su,Some id) -> ident GLOBAL old_metas table minus id
   | Ast0.StructUnionDef(ty,lb,decls,rb) ->
       typeC old_metas table minus ty;
@@ -261,13 +321,15 @@ and typeC old_metas table minus t =
 
 and declaration context old_metas table minus d =
   match Ast0.unwrap d with
-    Ast0.MetaDecl(name,_,_) ->
-      check_table table minus name
+    Ast0.MetaDecl(name,cstr,_) ->
+      check_table table minus name;
+      constraints table minus cstr
   | Ast0.AsDecl(decl,asdecl) -> failwith "not generated yet"
-  | Ast0.Init(stg,ty,id,attr,eq,ini,sem) ->
+  | Ast0.Init(algn,stg,ty,id,endattr,eq,ini,sem) ->
+      alignas ID old_metas table minus algn;
       typeC old_metas table minus ty;
       ident context old_metas table minus id;
-      List.iter (attribute old_metas table minus) attr;
+      List.iter (attribute old_metas table minus) endattr;
       (match Ast0.unwrap ini with
 	Ast0.InitExpr exp ->
 	  expression ID old_metas table minus exp
@@ -278,26 +340,30 @@ and declaration context old_metas table minus d =
 	    failwith "complex initializer specification not allowed in - code"
 	  else*)
 	    initialiser old_metas table minus ini)
-  | Ast0.UnInit(stg,ty,id,attr,sem) ->
-      typeC old_metas table minus ty; ident context old_metas table minus id;
-      List.iter (attribute old_metas table minus) attr
+  | Ast0.UnInit(algn,stg,ty,id,endattr,sem) ->
+      alignas ID old_metas table minus algn;
+      typeC old_metas table minus ty;
+      ident context old_metas table minus id;
+      List.iter (attribute old_metas table minus) endattr
   | Ast0.FunProto(fi,name,lp1,params,va,rp1,sem) ->
       ident FN old_metas table minus name;
       List.iter (fninfo old_metas table minus) fi;
       parameter_list old_metas table minus params
-  | Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
-      ident GLOBAL old_metas table minus name;
-      List.iter (attribute old_metas table minus) attr;
-      dots (expression ID old_metas table minus) args
-  | Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
+  | Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
+      List.iter (attribute old_metas table minus) preattr;
       ident GLOBAL old_metas table minus name;
       dots (expression ID old_metas table minus) args;
+      List.iter (attribute old_metas table minus) attr
+  | Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
+      List.iter (attribute old_metas table minus) preattr;
+      ident GLOBAL old_metas table minus name;
+      dots (expression ID old_metas table minus) args;
+      List.iter (attribute old_metas table minus) attr;
       (match Ast0.unwrap ini with
 	Ast0.InitExpr exp -> expression ID old_metas table minus exp
       |	_ -> initialiser old_metas table minus ini)
-  | Ast0.TyDecl(ty,attr,sem) ->
-      typeC old_metas table minus ty;
-      List.iter (attribute old_metas table minus) attr
+  | Ast0.TyDecl(ty,sem) ->
+      typeC old_metas table minus ty
   | Ast0.Typedef(stg,ty,id,sem) ->
       typeC old_metas table minus ty;
       typeC old_metas table minus id
@@ -307,21 +373,34 @@ and declaration context old_metas table minus d =
   | Ast0.OptDecl(_) ->
       failwith "unexpected code"
 
+and alignas context old_metas table minus = function
+    None -> ()
+  | Some(Ast0.Align(a,lp,e,rp)) -> expression context old_metas table minus e
+
 (* --------------------------------------------------------------------- *)
 (* Field declaration *)
 
 and field context old_metas table minus d =
   match Ast0.unwrap d with
-    Ast0.MetaField(name,_,_) ->
-      check_table table minus name
-  | Ast0.MetaFieldList(name,len,_,_) ->
+    Ast0.MetaField(name,cstr,_) ->
       check_table table minus name;
-      check_len table minus len
-  | Ast0.Field(ty,id,bf,sem) ->
+      constraints table minus cstr
+  | Ast0.MetaFieldList(name,len,cstr,_) ->
+      check_table table minus name;
+      check_len table minus len;
+      constraints table minus cstr
+  | Ast0.Field(ty,id,bf,endattr,sem) ->
       typeC old_metas table minus ty;
       Common.do_option (ident context old_metas table minus) id;
       let bitfield (c, e) = expression context old_metas table minus e in
-      Common.do_option bitfield bf
+      Common.do_option bitfield bf;
+      List.iter (attribute old_metas table minus) endattr
+  | Ast0.MacroDeclField(name,lp,args,rp,attr,sem) ->
+      ident GLOBAL old_metas table minus name;
+      dots (expression ID old_metas table minus) args;
+      List.iter (attribute old_metas table minus) attr
+  | Ast0.CppField(di) ->
+      directive old_metas table minus di
   | Ast0.DisjField(_,decls,_,_)
   | Ast0.ConjField(_,decls,_,_) ->
       List.iter (field ID old_metas table minus) decls
@@ -344,11 +423,13 @@ and enum_decl context old_metas table minus d =
 
 and initialiser old_metas table minus ini =
   match Ast0.unwrap ini with
-    Ast0.MetaInit(name,_,_) ->
-      check_table table minus name
-  | Ast0.MetaInitList(name,len,_,_) ->
+    Ast0.MetaInit(name,cstr,_) ->
       check_table table minus name;
-      check_len table minus len
+      constraints table minus cstr
+  | Ast0.MetaInitList(name,len,cstr,_) ->
+      check_table table minus name;
+      check_len table minus len;
+      constraints table minus cstr
   | Ast0.AsInit(ini,asini) -> failwith "not generated yet"
   | Ast0.InitExpr(exp) -> expression ID old_metas table minus exp
   | Ast0.InitList(lb,initlist,rb,ordered) ->
@@ -385,11 +466,13 @@ and parameterTypeDef old_metas table minus param =
       get_opt (ident ID old_metas table minus) id;
       typeC old_metas table minus ty;
       List.iter (attribute old_metas table minus) attr
-  | Ast0.MetaParam(name,_,_) ->
-      check_table table minus name
-  | Ast0.MetaParamList(name,len,_,_) ->
+  | Ast0.MetaParam(name,cstr,_) ->
       check_table table minus name;
-      check_len table minus len
+      constraints table minus cstr
+  | Ast0.MetaParamList(name,len,cstr,_) ->
+      check_table table minus name;
+      check_len table minus len;
+      constraints table minus cstr
   | _ -> () (* no metavariable subterms *)
 
 and parameter_list old_metas table minus =
@@ -404,15 +487,17 @@ and string_fragment old_metas table minus e =
   | Ast0.FormatFragment(pct,fmt) ->
       string_format old_metas table minus fmt
   | Ast0.Strdots dots -> ()
-  | Ast0.MetaFormatList(pct,name,_,len) ->
+  | Ast0.MetaFormatList(pct,name,cstr,len) ->
       check_table table minus name;
+      constraints table minus cstr;
       check_len table minus len
 
 and string_format old_metas table minus e =
   match Ast0.unwrap e with
     Ast0.ConstantFormat(str) -> ()
-  | Ast0.MetaFormat(name,_) ->
-      check_table table minus name
+  | Ast0.MetaFormat(name,cstr) ->
+      check_table table minus name;
+      constraints table minus cstr
 
 (* --------------------------------------------------------------------- *)
 (* Top-level code *)
@@ -436,14 +521,19 @@ and statement old_metas table minus s =
   | Ast0.Do(d,body,wh,lp,exp,rp,sem) ->
       statement old_metas table minus body;
       expression ID old_metas table minus exp
-  | Ast0.For(fr,lp,first,exp2,sem2,exp3,rp,body,_) ->
+  | Ast0.For(fr,lp,first,rp,body,_) ->
       (match Ast0.unwrap first with
-	Ast0.ForExp(exp1,sem1) ->
-	  get_opt (expression ID old_metas table minus) exp1
-      |	Ast0.ForDecl (_,decl) ->
-	  declaration ID old_metas table minus decl);
-      get_opt (expression ID old_metas table minus) exp2;
-      get_opt (expression ID old_metas table minus) exp3;
+	Ast0.ForExp(exp1,sem1,exp2,sem2,exp3) ->
+	  get_opt (expression ID old_metas table minus) exp1;
+	  get_opt (expression ID old_metas table minus) exp2;
+	  get_opt (expression ID old_metas table minus) exp3
+      |	Ast0.ForDecl (_,decl,exp2,sem2,exp3) ->
+	  declaration ID old_metas table minus decl;
+	  get_opt (expression ID old_metas table minus) exp2;
+	  get_opt (expression ID old_metas table minus) exp3
+      |	Ast0.ForRange (_,decl,ini) ->
+	  declaration ID old_metas table minus decl;
+	  initialiser old_metas table minus ini);
       statement old_metas table minus body
   | Ast0.Iterator(nm,lp,args,rp,body,_) ->
       ident GLOBAL old_metas table minus nm;
@@ -456,10 +546,13 @@ and statement old_metas table minus s =
   | Ast0.ReturnExpr(ret,exp,sem) -> expression ID old_metas table minus exp
   | Ast0.Exec(exec,lang,code,sem) ->
       dots (exec_code ID old_metas table minus) code
-  | Ast0.MetaStmt(name,_,_) ->     check_table table minus name
-  | Ast0.MetaStmtList(name,len,_,_) ->
+  | Ast0.MetaStmt(name,cstr,_) ->
       check_table table minus name;
-      check_len table minus len
+      constraints table minus cstr
+  | Ast0.MetaStmtList(name,len,cstr,_) ->
+      check_table table minus name;
+      check_len table minus len;
+      constraints table minus cstr;
   | Ast0.AsStmt(stm,asstm) -> failwith "not generated yet"
   | Ast0.Exp(exp) -> expression ID old_metas table minus exp
   | Ast0.TopExp(exp) -> expression ID old_metas table minus exp
@@ -480,37 +573,53 @@ and statement old_metas table minus s =
 	(whencode (dots (statement old_metas table minus))
 	   (statement old_metas table minus)
 	   (expression ID old_metas table minus)) x
-  | Ast0.FunDecl(_,fi,name,lp,params,va,rp,lbrace,body,rbrace,_) ->
+  | Ast0.FunDecl(_,fi,name,lp,params,va,rp,attrs,lbrace,body,rbrace,_) ->
       ident FN old_metas table minus name;
       List.iter (fninfo old_metas table minus) fi;
       parameter_list old_metas table minus params;
+      List.iter (attribute old_metas table minus) attrs;
       dots (statement old_metas table minus) body
-  | Ast0.Include(inc,s) -> () (* no metavariables possible *)
-  | Ast0.MetaInclude(inc,s) -> expression ID old_metas table minus s
   | Ast0.Undef(def,id) ->
       ident GLOBAL old_metas table minus id
   | Ast0.Define(def,id,params,body) ->
       ident GLOBAL old_metas table minus id;
       define_parameters old_metas table minus params;
       dots (statement old_metas table minus) body
-  | Ast0.Pragma(prg,id,body) ->
-      ident GLOBAL old_metas table minus id;
-      pragmainfo old_metas table minus body
   | Ast0.Label(i,_) -> ident ID old_metas table minus i
   | Ast0.Goto(_,i,_) -> ident ID old_metas table minus i
   | _ -> () (* no metavariable subterms *)
 
 and pragmainfo old_metas table minus pi =
   match Ast0.unwrap pi with
-      Ast0.PragmaString(s) -> ()
-    | Ast0.PragmaDots (dots) -> ()
+    Ast0.PragmaString(s) -> ()
+  | Ast0.PragmaDots (dots) -> ()
+  | Ast0.MetaPragmaInfo (name,cstr,_) ->
+      check_table table minus name;
+      constraints table minus cstr
+
+and directive old_metas table minus di =
+  match Ast0.unwrap di with
+    Ast0.Include(inc,s) -> () (* no metavariables possible *)
+  | Ast0.MetaInclude(inc,s) -> expression ID old_metas table minus s
+  | Ast0.Pragma(prg,id,body) ->
+      ident GLOBAL old_metas table minus id;
+      pragmainfo old_metas table minus body
+  | Ast0.UsingNamespace(using,ns,nm,semi) ->
+      ident GLOBAL old_metas table minus nm
+  | Ast0.UsingTypename(using,nm,eq,tynm,ty,semi) ->
+      ident GLOBAL old_metas table minus nm;
+      typeC old_metas table minus ty
+  | Ast0.UsingMember(using,nm,semi) ->
+      ident GLOBAL old_metas table minus nm
 
 and define_param old_metas table minus p =
   match Ast0.unwrap p with
     Ast0.DParam(id) -> ident GLOBAL old_metas table minus id
-  | Ast0.MetaDParamList(name,len,_,_) ->
+  | Ast0.DParamEll(id,dots) -> ident GLOBAL old_metas table minus id
+  | Ast0.MetaDParamList(name,len,cstr,_) ->
       check_table table minus name;
-      check_len table minus len
+      check_len table minus len;
+      constraints table minus cstr
   | Ast0.DPComma(_) | Ast0.DPdots(_) ->
       () (* no metavariable subterms *)
   | Ast0.OptDParam(dp)    -> define_param old_metas table minus dp
@@ -524,12 +633,23 @@ and fninfo old_metas table minus = function
     Ast0.FStorage(stg) -> ()
   | Ast0.FType(ty) -> typeC old_metas table minus ty
   | Ast0.FInline(inline) -> ()
-  | Ast0.FAttr(attr) -> attribute old_metas table minus attr
 
 and attribute old_metas table minus x =
   match Ast0.unwrap x with
-    Ast0.MetaAttribute(name,_,_) ->
-      check_table table minus name
+    Ast0.Attribute(arg) ->
+      attr_arg old_metas table minus arg
+  | Ast0.GccAttribute _ -> ()
+  | Ast0.CxxAttribute _ -> ()
+  | Ast0.CxxAttributeUsing (lb1,usng,atnm,dotdot,args,rb1,rb2) ->
+      ident GLOBAL old_metas table minus atnm
+
+and attr_arg old_metas table minus x =
+  match Ast0.unwrap x with
+    Ast0.MetaAttr(name,cstr,_) ->
+      check_table table minus name;
+      constraints table minus cstr
+  | Ast0.MacroAttrArgs(attr,lp,args,rp) ->
+      dots (expression ID old_metas table minus) args
   | _ -> ()
 
 and whencode notfn alwaysfn expression = function
@@ -556,6 +676,24 @@ and exec_code context old_metas table minus e =
   | Ast0.ExecToken(tok) -> ()
   | Ast0.ExecDots(dots) -> ()
 
+and constraints table minus = function
+    Ast.CstrFalse -> ()
+  | Ast.CstrTrue -> ()
+  | Ast.CstrExit -> ()
+  | Ast.CstrAnd l -> List.iter (constraints table minus) l
+  | Ast.CstrOr l -> List.iter (constraints table minus) l
+  | Ast.CstrNot c -> constraints table minus c
+  | Ast.CstrConstant cst -> ()
+  | Ast.CstrOperator op -> ()
+  | Ast.CstrMeta_name name -> ()
+  | Ast.CstrRegexp (s,re) -> ()
+  | Ast.CstrScript(local,script) -> ()
+  | Ast.CstrExpr s -> ()
+  | Ast.CstrSub names -> ()
+  | Ast.CstrType t ->
+      List.iter (ast_check_table table minus)
+	(Free_vars.collect_all_refs.Visitor_ast.combiner_fullType t)
+
 (* --------------------------------------------------------------------- *)
 (* Rules *)
 
@@ -602,14 +740,7 @@ let positions rname table rules =
   let bind x y = () in
   let donothing r k e = k e in
   let fn =
-    V0.flat_combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode mcode mcode
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing in
-
+    V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=donothing} () in
   List.iter fn.VT0.combiner_rec_top_level rules
 
 let dup_positions rules =
@@ -672,13 +803,8 @@ let dup_positions rules =
 
   let donothing r k e = k e in
   let fn =
-    V0.flat_combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode mcode mcode
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing expression donothing donothing typeC
-      donothing donothing declaration field donothing statement
-      donothing donothing donothing donothing donothing in
+    V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=donothing}
+      ~expr:expression ~ty:typeC ~decl:declaration ~field:field ~stmt:statement () in
 
   let res =
     List.sort compare
@@ -715,7 +841,7 @@ let check_all_marked rname err table after_err =
 	if not (!cell)
 	then
 	  let (_,name) = name in
-	  warning
+	  warning (-1)
 	    (Printf.sprintf "%s: %s %s not used %s" rname err name after_err))
     table
 
@@ -733,6 +859,7 @@ let check_both_marked rname required optional =
     optional
 
 let check_meta rname old_metas inherited_metavars metavars minus plus =
+  warned := [];
   let old_metas =
     List.map (function (_,x) -> x) (List.map Ast.get_meta_name old_metas) in
   let (fresh,other) =
