@@ -409,12 +409,12 @@ let rec expression e =
       let rp = normal_mcode rp in
       mkres e (Ast0.Paren(lp,expression exp,rp))
 	(promote_mcode lp) (promote_mcode rp)
-  | Ast0.ArrayAccess(exp1,lb,exp2,rb) ->
-      let exp1 = expression exp1 in
+  | Ast0.ArrayAccess(fn,lb,args,rb) ->
+      let fn = expression fn in
       let lb = normal_mcode lb in
-      let exp2 = expression exp2 in
+      let args = dots is_exp_dots (Some(promote_mcode lb)) expression args in
       let rb = normal_mcode rb in
-      mkres e (Ast0.ArrayAccess(exp1,lb,exp2,rb)) exp1 (promote_mcode rb)
+      mkres e (Ast0.ArrayAccess(fn,lb,args,rb)) fn (promote_mcode rb)
   | Ast0.RecordAccess(exp,pt,field) ->
       let exp = expression exp in
       let pt = normal_mcode pt in
@@ -425,12 +425,22 @@ let rec expression e =
       let ar = normal_mcode ar in
       let field = ident field in
       mkres e (Ast0.RecordPtAccess(exp,ar,field)) exp field
-  | Ast0.Cast(lp,ty,attr,rp,exp) ->
+  | Ast0.QualifiedAccess(None,coloncolon,field) ->
+      let coloncolon = normal_mcode coloncolon in
+      let field = ident field in
+      mkres e (Ast0.QualifiedAccess(None,coloncolon,field))
+        (promote_mcode coloncolon) field
+  | Ast0.QualifiedAccess(Some ty,coloncolon,field) ->
+      let ty = typeC ty in
+      let coloncolon = normal_mcode coloncolon in
+      let field = ident field in
+      mkres e (Ast0.QualifiedAccess(Some ty,coloncolon,field))
+          ty field
+  | Ast0.Cast(lp,ty,rp,exp) ->
       let lp = normal_mcode lp in
       let exp = expression exp in
-      let attr = List.map attribute attr in
       let rp = normal_mcode rp in
-      mkres e (Ast0.Cast(lp,typeC ty,attr,rp,exp)) (promote_mcode lp) exp
+      mkres e (Ast0.Cast(lp,typeC ty,rp,exp)) (promote_mcode lp) exp
   | Ast0.SizeOfExpr(szf,exp) ->
       let szf = normal_mcode szf in
       let exp = expression exp in
@@ -441,6 +451,34 @@ let rec expression e =
       let rp = normal_mcode rp in
       mkres e (Ast0.SizeOfType(szf,lp,typeC ty,rp))
         (promote_mcode szf)  (promote_mcode rp)
+  | Ast0.Delete(dlt,exp) ->
+      let dlt = normal_mcode dlt in
+      let exp = expression exp in
+      mkres e (Ast0.Delete(dlt,exp)) (promote_mcode dlt) exp
+  | Ast0.DeleteArr(dlt,lb,rb,exp) ->
+      let dlt = normal_mcode dlt in
+      let lb = normal_mcode lb in
+      let rb = normal_mcode rb in
+      let exp = expression exp in
+      mkres e (Ast0.DeleteArr(dlt,lb,rb, exp)) (promote_mcode dlt) exp
+  | Ast0.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) -> (* HIGH doubts in this one *)
+      let nw = normal_mcode nw in
+      let pp_opt = get_option normal_args pp_opt in
+      let lp_opt = get_option normal_mcode lp_opt in
+      let ty = typeC ty in
+      let rp_opt = get_option normal_mcode rp_opt in
+      let args_opt = get_option normal_args args_opt in
+      let tmp = Ast0.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) in
+      (match (args_opt,rp_opt) with
+      | (None,None) -> mkres e tmp (promote_mcode nw) ty
+      | (None,Some x) -> mkres e tmp (promote_mcode nw) (promote_mcode x)
+      | (Some (lp,args,rp),_) -> mkres e tmp (promote_mcode nw) (promote_mcode rp))
+  | Ast0.TemplateInst(tn,lp,args,rp) ->
+      let tn = expression tn in
+      let lp = normal_mcode lp in
+      let args = dots is_exp_dots (Some(promote_mcode lp)) expression args in
+      let rp = normal_mcode rp in
+      mkres e (Ast0.TemplateInst(tn,lp,args,rp)) tn (promote_mcode rp)
   | Ast0.TypeExp(ty) ->
       let ty = typeC ty in mkres e (Ast0.TypeExp(ty)) ty ty
   | Ast0.Constructor(lp,ty,rp,init) ->
@@ -500,6 +538,12 @@ let rec expression e =
       let exp = expression exp in
       mkres e (Ast0.AsSExpr(exp,asexp)) exp exp
 
+and normal_args (lp,args,rp) =
+  let lp = normal_mcode lp in
+  let args = dots is_exp_dots (Some(promote_mcode lp)) expression args in
+  let rp = normal_mcode rp in
+  (lp,args,rp)
+
 and expression_dots x = dots is_exp_dots None expression x
 
 and string_fragment e =
@@ -544,10 +588,35 @@ and string_format e =
 
 and typeC t =
   match Ast0.unwrap t with
-    Ast0.ConstVol(cv,ty) ->
-      let cv = List.map normal_mcode cv in
+    Ast0.ConstVol(cvbefore,ty,cvafter) ->
+      let do_cvattr = function
+	  Ast0.CV cv -> Ast0.CV (normal_mcode cv)
+	| Ast0.Attr attr -> Ast0.Attr (attribute attr) in
+      let cvbefore = List.map do_cvattr cvbefore in
+      let cvafter = List.map do_cvattr cvafter in
       let ty = typeC ty in
-      mkres t (Ast0.ConstVol(cv,ty)) (promote_mcode (List.hd cv)) ty
+      (* long due to polymorphism issue *)
+      (match (cvbefore,List.rev cvafter) with
+	([],[]) -> failwith "useless ConstVol"
+      | (Ast0.CV cv::_,[]) ->
+	  mkres t (Ast0.ConstVol(cvbefore,ty,cvafter)) (promote_mcode cv) ty
+      | (Ast0.Attr attr::_,[]) ->
+	  mkres t (Ast0.ConstVol(cvbefore,ty,cvafter)) attr ty
+      | ([],Ast0.CV cv::_) ->
+	  mkres t (Ast0.ConstVol(cvbefore,ty,cvafter)) ty (promote_mcode cv)
+      | ([],Ast0.Attr attr::_) ->
+	  mkres t (Ast0.ConstVol(cvbefore,ty,cvafter)) ty attr
+      | (Ast0.CV cvb::_,Ast0.CV cva::_) ->
+	  mkres t (Ast0.ConstVol(cvbefore,ty,cvafter))
+	    (promote_mcode cvb) (promote_mcode cva)
+      | (Ast0.CV cvb::_,Ast0.Attr attra::_) ->
+	  mkres t (Ast0.ConstVol(cvbefore,ty,cvafter))
+	    (promote_mcode cvb) attra
+      | (Ast0.Attr attrb::_,Ast0.CV cva::_) ->
+	  mkres t (Ast0.ConstVol(cvbefore,ty,cvafter))
+	    attrb (promote_mcode cva)
+      | (Ast0.Attr attrb::_,Ast0.Attr attra::_) ->
+	  mkres t (Ast0.ConstVol(cvbefore,ty,cvafter)) attrb attra)
   | Ast0.BaseType(ty,strings) ->
       let strings = List.map normal_mcode strings in
       let first = List.hd strings in
@@ -591,21 +660,25 @@ and typeC t =
       let rp = normal_mcode rp in
       mkres t (Ast0.Decimal(dec,lp,length,comma,precision_opt,rp))
 	(promote_mcode dec) (promote_mcode rp)
-  | Ast0.EnumName(kind,Some name) ->
+  | Ast0.EnumName(kind, key, Some name) ->
       let kind = normal_mcode kind in
+      let key = get_option normal_mcode key in
       let name = ident name in
-      mkres t (Ast0.EnumName(kind,Some name)) (promote_mcode kind) name
-  | Ast0.EnumName(kind,None) ->
+      mkres t (Ast0.EnumName(kind,key,Some name)) (promote_mcode kind) name
+  | Ast0.EnumName(kind,key,None) ->(* Never needed, but removing will cause warnings *)
       let kind = normal_mcode kind in
       let mc = promote_mcode kind in
-      mkres t (Ast0.EnumName(kind,None)) mc mc
-  | Ast0.EnumDef(ty,lb,ids,rb) ->
+      mkres t (Ast0.EnumName(kind,key,None)) mc mc
+  | Ast0.EnumDef(ty,base,lb,ids,rb) ->
       let ty = typeC ty in
+      let base = get_option (function (td, ty) -> (normal_mcode td, typeC ty)) base in
       let lb = normal_mcode lb in
       let ids =
-        dots is_enum_decl_dots (Some(promote_mcode lb)) enum_decl ids in
+        (dots is_enum_decl_dots
+	  (Some (Common.default (promote_mcode lb) (function (td,ty) -> promote_mcode td) base))
+	  enum_decl) ids in
       let rb = normal_mcode rb in
-      mkres t (Ast0.EnumDef(ty,lb,ids,rb)) ty (promote_mcode rb)
+      mkres t (Ast0.EnumDef(ty,base,lb,ids,rb)) ty (promote_mcode rb)
   | Ast0.StructUnionName(kind,Some name) ->
       let kind = normal_mcode kind in
       let name = ident name in
@@ -634,10 +707,25 @@ and typeC t =
       let rp = normal_mcode rp in
       mkres t (Ast0.TypeOfType(tf,lp,typeC ty,rp))
         (promote_mcode tf) (promote_mcode rp)
+  | Ast0.QualifiedType(Some ty,coloncolon,name) ->
+      let ty = typeC ty in
+      let coloncolon = normal_mcode coloncolon in
+      let name = ident name in
+      mkres t (Ast0.QualifiedType(Some ty,coloncolon,name)) ty name
+  | Ast0.QualifiedType(None,coloncolon,name) ->
+      let coloncolon = normal_mcode coloncolon in
+      let name = ident name in
+      mkres t (Ast0.QualifiedType(None,coloncolon,name)) (promote_mcode coloncolon) name
   | Ast0.TypeName(name) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
       mkres t (Ast0.TypeName(name)) ln ln
+  | Ast0.TemplateType(tn,lp,args,rp) ->
+      let tn = typeC tn in
+      let lp = normal_mcode lp in
+      let args = dots is_exp_dots (Some(promote_mcode lp)) expression args in
+      let rp = normal_mcode rp in
+      mkres t (Ast0.TemplateType(tn,lp,args,rp)) tn (promote_mcode rp)
   | Ast0.AutoType(auto) ->
       let auto = normal_mcode auto in
       let la = promote_mcode auto in
@@ -663,46 +751,59 @@ and typeC t =
 (* Even if the Cocci program specifies a list of declarations, they are
    split out into multiple declarations of a single variable each. *)
 
+and alignas (Ast0.Align(al,lp,expr,rp)) =
+      let al = normal_mcode al in
+      let lp = normal_mcode lp in
+      let expr = expression expr in
+      let rp = normal_mcode rp in
+      (Ast0.Align(al,lp,expr,rp), al)
+
 and declaration d =
   match Ast0.unwrap d with
     Ast0.MetaDecl(name,a,b) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
       mkres d (Ast0.MetaDecl(name,a,b)) ln ln
-  | Ast0.Init(stg,ty,midattr,id,endattr,eq,exp,sem) ->
+  | Ast0.Init(al,stg,ty,id,endattr,eq,exp,sem) ->
       let ty = typeC ty in
-      let midattr = List.map attribute midattr in
       let id = ident id in
       let endattr = List.map attribute endattr in
       let eq = normal_mcode eq in
       let exp = initialiser exp in
       let sem = normal_mcode sem in
-      (match stg with
-	None ->
-	  mkres d (Ast0.Init(stg,ty,midattr,id,endattr,eq,exp,sem)) ty (promote_mcode sem)
-      | Some x ->
+      (match (al,stg) with
+	(None,None) ->
+	  mkres d (Ast0.Init(al,stg,ty,id,endattr,eq,exp,sem)) ty (promote_mcode sem)
+      | (None,Some x)  ->
 	  let stg = Some (normal_mcode x) in
-	  mkres d (Ast0.Init(stg,ty,midattr,id,endattr,eq,exp,sem))
+	  mkres d (Ast0.Init(al,stg,ty,id,endattr,eq,exp,sem))
+	    (promote_mcode x) (promote_mcode sem)
+      | (Some al, _) ->
+          let (al,x) = alignas al in
+          mkres d (Ast0.Init(Some al,stg,ty,id,endattr,eq,exp,sem))
 	    (promote_mcode x) (promote_mcode sem))
-  | Ast0.UnInit(stg,ty,midattr,id,endattr,sem) ->
+  | Ast0.UnInit(al,stg,ty,id,endattr,sem) ->
       let ty = typeC ty in
-      let midattr = List.map attribute midattr in
       let id = ident id in
       let endattr = List.map attribute endattr in
       let sem = normal_mcode sem in
-      (match stg with
-	None ->
-	  mkres d (Ast0.UnInit(stg,ty,midattr,id,endattr,sem)) ty (promote_mcode sem)
-      | Some x ->
+      (match (al,stg) with
+	(None,None) ->
+	  mkres d (Ast0.UnInit(al,stg,ty,id,endattr,sem)) ty (promote_mcode sem)
+      | (None,Some x)  ->
 	  let stg = Some (normal_mcode x) in
-	  mkres d (Ast0.UnInit(stg,ty,midattr,id,endattr,sem))
+	  mkres d (Ast0.UnInit(al,stg,ty,id,endattr,sem))
+	    (promote_mcode x) (promote_mcode sem)
+      | (Some al, _) ->
+          let (al,x) = alignas al in
+          mkres d (Ast0.UnInit(Some al,stg,ty,id,endattr,sem))
 	    (promote_mcode x) (promote_mcode sem))
-  | Ast0.FunProto(fninfo,attr,name,lp1,params,va1,rp1,sem) ->
+
+  | Ast0.FunProto(fninfo,name,lp1,params,va1,rp1,sem) ->
       let fninfo =
 	List.map
 	  (function Ast0.FType(ty) -> Ast0.FType(typeC ty) | x -> x)
 	  fninfo in
-      let attr = List.map attribute attr in
       let name = ident name in
       let lp1 = normal_mcode lp1 in
       let params = parameter_list (Some(promote_mcode lp1)) params in
@@ -711,15 +812,15 @@ and declaration d =
         | Some (c1,e1) -> Some (normal_mcode c1, normal_mcode e1) in
       let rp1 = normal_mcode rp1 in
       let sem = normal_mcode sem in
-      let res = Ast0.FunProto(fninfo,attr,name,lp1,params,va1,rp1,sem) in
+      let res = Ast0.FunProto(fninfo,name,lp1,params,va1,rp1,sem) in
       let right = promote_mcode sem in
       (match fninfo with
 	  [] -> mkres d res name right
 	| Ast0.FStorage(stg)::_ -> mkres d res (promote_mcode stg) right
 	| Ast0.FType(ty)::_ -> mkres d res ty right
-	| Ast0.FInline(inline)::_ -> mkres d res (promote_mcode inline) right
-	| Ast0.FAttr(attr)::_ -> mkres d res attr right)
-  | Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
+	| Ast0.FInline(inline)::_ -> mkres d res (promote_mcode inline) right)
+  | Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
+      let preattr = List.map attribute preattr in
       let name = ident name in
       let lp = normal_mcode lp in
       let args = dots is_exp_dots (Some(promote_mcode lp)) expression args in
@@ -728,33 +829,44 @@ and declaration d =
       let sem = normal_mcode sem in
       (match stg with
 	None ->
-	  mkres d (Ast0.MacroDecl(None,name,lp,args,rp,attr,sem))
-	    name (promote_mcode sem)
+	  (match preattr with
+	    [] ->
+	      mkres d (Ast0.MacroDecl(None,preattr,name,lp,args,rp,attr,sem))
+		name (promote_mcode sem)
+	  | lattr::_ ->
+	      mkres d (Ast0.MacroDecl(None,preattr,name,lp,args,rp,attr,sem))
+		lattr (promote_mcode sem))
       | Some x ->
 	  let stg = Some (normal_mcode x) in
-	  mkres d (Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem))
+	  mkres d (Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem))
 	    (promote_mcode x) (promote_mcode sem))
-  | Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
+  | Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
+      let preattr = List.map attribute preattr in
       let name = ident name in
       let lp = normal_mcode lp in
       let args = dots is_exp_dots (Some(promote_mcode lp)) expression args in
       let rp = normal_mcode rp in
+      let attr = List.map attribute attr in
       let eq = normal_mcode eq in
       let ini = initialiser ini in
       let sem = normal_mcode sem in
       (match stg with
 	None ->
-	  mkres d (Ast0.MacroDeclInit(None,name,lp,args,rp,eq,ini,sem))
-	    name (promote_mcode sem)
+	  (match preattr with
+	    [] ->
+	      mkres d (Ast0.MacroDeclInit(None,[],name,lp,args,rp,attr,eq,ini,sem))
+		name (promote_mcode sem)
+	  | lattr::_ ->
+	      mkres d (Ast0.MacroDeclInit(None,preattr,name,lp,args,rp,attr,eq,ini,sem))
+		lattr (promote_mcode sem))
       | Some x ->
 	  let stg = Some (normal_mcode x) in
-	  mkres d (Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem))
+	  mkres d (Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem))
 	    (promote_mcode x) (promote_mcode sem))
-  | Ast0.TyDecl(ty,attr,sem) ->
+  | Ast0.TyDecl(ty,sem) ->
       let ty = typeC ty in
-      let attr = List.map attribute attr in
       let sem = normal_mcode sem in
-      mkres d (Ast0.TyDecl(ty,attr,sem)) ty (promote_mcode sem)
+      mkres d (Ast0.TyDecl(ty,sem)) ty (promote_mcode sem)
   | Ast0.Typedef(stg,ty,id,sem) ->
       let stg = normal_mcode stg in
       let ty = typeC ty in
@@ -793,13 +905,26 @@ and field d =
       let name = normal_mcode name in
       let ln = promote_mcode name in
       mkres d (Ast0.MetaFieldList(name,a,b,c)) ln ln
-  | Ast0.Field(ty,id,bf,sem) ->
+  | Ast0.Field(ty,id,bf,endattr,sem) ->
       let ty = typeC ty in
       let id = Common.map_option ident id in
       let bitfield (c, e) = (normal_mcode c, expression e) in
       let bf = Common.map_option bitfield bf in
+      let endattr = List.map attribute endattr in
       let sem = normal_mcode sem in
-      mkres d (Ast0.Field(ty,id,bf,sem)) ty (promote_mcode sem)
+      mkres d (Ast0.Field(ty,id,bf,endattr,sem)) ty (promote_mcode sem)
+  | Ast0.MacroDeclField(name,lp,args,rp,attr,sem) ->
+      let name = ident name in
+      let lp = normal_mcode lp in
+      let args = dots is_exp_dots (Some(promote_mcode lp)) expression args in
+      let rp = normal_mcode rp in
+      let attr = List.map attribute attr in
+      let sem = normal_mcode sem in
+      mkres d (Ast0.MacroDeclField(name,lp,args,rp,attr,sem))
+	name (promote_mcode sem)
+  | Ast0.CppField(di) ->
+      let di = directive di in
+      mkres d (Ast0.CppField(di)) di di
   | Ast0.DisjField(starter,decls,mids,ender) ->
       do_disj d starter decls mids ender field
 	(fun starter decls mids ender ->
@@ -837,6 +962,58 @@ and enum_decl d =
       let ln = promote_mcode dots in
       mkres d (Ast0.EnumDots(dots,whencode)) ln ln
 
+and pragmainfo pi =
+  match Ast0.unwrap pi with
+    Ast0.PragmaString(s) ->
+      let s = normal_mcode s in
+      let ln = promote_mcode s in
+      mkres pi (Ast0.PragmaString(s)) ln ln
+  | Ast0.PragmaDots(dots) ->
+      let dots = bad_mcode dots in
+      let ln = promote_mcode dots in
+      mkres pi (Ast0.PragmaDots(dots)) ln ln
+  | Ast0.MetaPragmaInfo(mv0, c, pure) ->
+      let mv0 = normal_mcode mv0 in
+      let ln = promote_mcode mv0 in
+      mkres pi (Ast0.MetaPragmaInfo(mv0, c, pure)) ln ln
+
+and directive d =
+  match Ast0.unwrap d with
+    Ast0.Include(inc,stm) ->
+      let inc = normal_mcode inc in
+      let stm = normal_mcode stm in
+      mkres d (Ast0.Include(inc,stm)) (promote_mcode inc) (promote_mcode stm)
+  | Ast0.MetaInclude(inc,s) ->
+      let inc = normal_mcode inc in
+      let s = expression s in
+      mkres d (Ast0.MetaInclude(inc,s)) (promote_mcode inc) s
+  | Ast0.Pragma(prg,id,body) ->
+      let prg = normal_mcode prg in
+      let id = ident id in
+      let body = pragmainfo body in
+      mkres d (Ast0.Pragma(prg,id,body)) (promote_mcode prg) body
+  | Ast0.UsingNamespace(usng,nmspc,name,sem) ->
+      let usng = normal_mcode usng in
+      let nmspc = normal_mcode nmspc in
+      let name = ident name in
+      let sem = normal_mcode sem in
+      mkres d (Ast0.UsingNamespace(usng,nmspc,name,sem))
+        (promote_mcode usng) (promote_mcode sem)
+  | Ast0.UsingTypename(usng,name,eq,tn,ty,sem) ->
+      let usng = normal_mcode usng in
+      let name = ident name in
+      let eq = normal_mcode eq in
+      let tn = get_option normal_mcode tn in
+      let ty = typeC ty in
+      let sem = normal_mcode sem in
+      mkres d (Ast0.UsingTypename(usng,name,eq,tn,ty,sem))
+        (promote_mcode usng) (promote_mcode sem)
+  | Ast0.UsingMember(usng,name,sem) ->
+      let usng = normal_mcode usng in
+      let name = ident name in
+      let sem = normal_mcode sem in
+      mkres d (Ast0.UsingMember(usng,name,sem))
+        (promote_mcode usng) (promote_mcode sem)
 
 (* --------------------------------------------------------------------- *)
 (* Initializer *)
@@ -915,16 +1092,35 @@ and attribute attr =
     Ast0.Attribute(a) ->
       let a = attr_arg a in
       mkres attr (Ast0.Attribute(a)) a a
-  | Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) ->
+  | Ast0.GccAttribute(attr_,lp1,lp2,args,rp1,rp2) ->
       let attr_ = normal_mcode attr_ in
       let lp1 = normal_mcode lp1 in
       let lp2 = normal_mcode lp2 in
-      let arg = attr_arg arg in
+      let args = dots is_exp_dots (Some(promote_mcode lp2)) expression args in
       let rp1 = normal_mcode rp1 in
       let rp2 = normal_mcode rp2 in
       let ln1 = promote_mcode attr_ in
       let ln2 = promote_mcode rp2 in
-      mkres attr (Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2)) ln1 ln2
+      mkres attr (Ast0.GccAttribute(attr_,lp1,lp2,args,rp1,rp2)) ln1 ln2
+  | Ast0.CxxAttribute(lb1,args,rb1,rb2) ->
+      let lb1 = normal_mcode lb1 in
+      let args = dots is_exp_dots (Some(promote_mcode lb1)) expression args in
+      let rb1 = normal_mcode rb1 in
+      let rb2 = normal_mcode rb2 in
+      let ln1 = promote_mcode lb1 in
+      let ln2 = promote_mcode rb2 in
+      mkres attr (Ast0.CxxAttribute(lb1,args,rb1,rb2)) ln1 ln2
+  | Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2) ->
+      let lb1 = normal_mcode lb1 in
+      let usng = normal_mcode usng in
+      let atnm = ident atnm in
+      let dotdot = normal_mcode dotdot in
+      let args = dots is_exp_dots (Some(promote_mcode dotdot)) expression args in
+      let rb1 = normal_mcode rb1 in
+      let rb2 = normal_mcode rb2 in
+      let ln1 = promote_mcode lb1 in
+      let ln2 = promote_mcode rb2 in
+      mkres attr (Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2)) ln1 ln2
 
 and attr_arg arg =
   match Ast0.unwrap arg with
@@ -950,35 +1146,30 @@ and is_param_dots p =
     Ast0.Pdots(_) -> true
   | _ -> false
 
+and is_template_param_dots p =
+  match Ast0.unwrap p with
+    Ast0.TPDots(_) -> true
+  | _ -> false
+
 and parameterTypeDef p =
   match Ast0.unwrap p with
-    Ast0.VoidParam(ty,attr) ->
-      let attr = List.map attribute attr in
-      let ty = typeC ty in
-      (match attr with
-        [] -> mkres p (Ast0.VoidParam(ty,attr)) ty ty
-      | l ->
-          let lattr = List.hd (List.rev l) in
-          mkres p (Ast0.VoidParam(ty,attr)) ty lattr)
-  | Ast0.Param(ty,midattr,Some id,attr) ->
+    Ast0.Param(ty,Some id,attr) ->
       let id = ident id in
-      let midattr = List.map attribute midattr in
       let ty = typeC ty in
       let attr = List.map attribute attr in
       (match attr with
-        [] -> mkres p (Ast0.Param(ty,midattr,Some id,attr)) ty id
+        [] -> mkres p (Ast0.Param(ty,Some id,attr)) ty id
       | l ->
           let lattr = List.hd (List.rev l) in
-          mkres p (Ast0.Param(ty,midattr,Some id,attr)) ty lattr)
-  | Ast0.Param(ty,midattr,None,attr) ->
-      assert (midattr = []);
+          mkres p (Ast0.Param(ty,Some id,attr)) ty lattr)
+  | Ast0.Param(ty,None,attr) ->
       let attr = List.map attribute attr in
       let ty = typeC ty in
       (match attr with
-        [] -> mkres p (Ast0.Param(ty,[],None,attr)) ty ty
+        [] -> mkres p (Ast0.Param(ty,None,attr)) ty ty
       | l ->
           let lattr = List.hd (List.rev l) in
-          mkres p (Ast0.Param(ty,[],None,attr)) ty lattr)
+          mkres p (Ast0.Param(ty,None,attr)) ty lattr)
   | Ast0.MetaParam(name,a,b) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
@@ -1000,7 +1191,35 @@ and parameterTypeDef p =
       mkres p (Ast0.OptParam(res)) res res
   | Ast0.AsParam _ -> failwith "not possible"
 
+and templateParameterTypeDef p =
+    (match Ast0.unwrap p with
+      Ast0.TypenameOrClassParam(tyorcl,id,None) ->
+	let ln = promote_mcode tyorcl in
+	mkres p (Ast0.TypenameOrClassParam(normal_mcode tyorcl,ident id,None)) ln ln
+    | Ast0.TypenameOrClassParam(tyorcl,id,Some (eq,ty)) ->
+	let ln = promote_mcode tyorcl in
+	mkres p
+	  (Ast0.TypenameOrClassParam
+	     (normal_mcode tyorcl,ident id,Some (normal_mcode eq,typeC ty))) ln ln
+    | Ast0.VarNameParam(ty,id,None) ->
+	let ty = typeC ty in
+	mkres p (Ast0.VarNameParam(ty,ident id,None)) ty ty
+    | Ast0.VarNameParam(ty,id,Some (eq,ini)) ->
+	let ty = typeC ty in
+	mkres p
+	  (Ast0.VarNameParam
+	     (typeC ty,ident id,Some (normal_mcode eq,initialiser ini))) ty ty
+    | Ast0.TPComma(cm) ->
+	let ln = promote_mcode cm in
+	mkres p (Ast0.TPComma(normal_mcode cm)) ln ln
+    | Ast0.TPDots(dots) ->
+	let ln = promote_mcode dots in
+	mkres p (Ast0.TPDots(normal_mcode dots )) ln ln
+    (* Note: TemplateParam not supported yet. *)
+    )
+
 and parameter_list prev = dots is_param_dots prev parameterTypeDef
+and template_parameter_list prev = dots is_template_param_dots prev templateParameterTypeDef
 
 (* for export *)
 let parameter_dots x = dots is_param_dots None parameterTypeDef x
@@ -1016,6 +1235,11 @@ let rec define_param p =
   match Ast0.unwrap p with
     Ast0.DParam(id) ->
       let id = ident id in mkres p (Ast0.DParam(id)) id id
+  | Ast0.DParamEll(id,dots) ->
+      let dots = normal_mcode dots in
+      let ln = promote_mcode dots in
+      let id = ident id in
+      mkres p (Ast0.DParamEll(id,dots)) id ln
   | Ast0.MetaDParamList(name,a,b,c) ->
       let name = normal_mcode name in
       let ln = promote_mcode name in
@@ -1123,31 +1347,43 @@ let rec statement s =
 	let rp = normal_mcode rp in
 	mkres s (Ast0.Do(d,body,wh,lp,exp,rp,sem))
 	  (promote_mcode d) (promote_mcode sem)
-    | Ast0.For(fr,lp,first,exp2,sem2,exp3,rp,body,(_,aft,adj)) ->
+    | Ast0.For(fr,lp,first,rp,body,(_,aft,adj)) ->
 	let fr = normal_mcode fr in
 	let lp = normal_mcode lp in
 	let first =
 	  match Ast0.unwrap first with
-	    Ast0.ForExp(None,sem1) ->
+	    Ast0.ForExp(exp1,sem1,exp2,sem2,exp3) ->
+	      let exp1 = get_option expression exp1 in
 	      let sem1 = normal_mcode sem1 in
-	      mkres first (Ast0.ForExp(None,sem1))
-		(promote_mcode sem1) (promote_mcode sem1)
-	  | Ast0.ForExp(Some exp1,sem1) ->
-	      let exp1 = expression exp1 in
-	      let sem1 = normal_mcode sem1 in
-	      mkres first (Ast0.ForExp(Some exp1,sem1))
-		exp1 (promote_mcode sem1)
-	  | Ast0.ForDecl((_,bef),decl) ->
+	      let exp2 = get_option expression exp2 in
+	      let sem2 = normal_mcode sem2 in
+	      let exp3 = get_option expression exp3 in
+	      let res = Ast0.ForExp(exp1,sem1,exp2,sem2,exp3) in
+	      (match exp1,exp3 with
+		(Some exp1,Some exp3) -> mkres first res exp1 exp3
+	      | (Some exp1,None) -> mkres first res exp1 (promote_mcode sem2)
+	      | (None,Some exp3) -> mkres first res (promote_mcode sem1) exp3
+	      | (None,None) ->
+		  mkres first res (promote_mcode sem1) (promote_mcode sem2))
+	  | Ast0.ForDecl((_,bef),decl,exp2,sem2,exp3) ->
 	      let decl = declaration decl in
 	      let (leftinfo,decl) = promote_to_statement_start decl bef in
-	      mkres first (Ast0.ForDecl ((leftinfo,bef),decl)) decl decl in
-	let exp2 = get_option expression exp2 in
-	let sem2 = normal_mcode sem2 in
-	let exp3 = get_option expression exp3 in
+	      let exp2 = get_option expression exp2 in
+	      let sem2 = normal_mcode sem2 in
+	      let exp3 = get_option expression exp3 in
+	      let res = Ast0.ForDecl ((leftinfo,bef),decl,exp2,sem2,exp3) in
+	      (match exp3 with
+		Some exp3 -> mkres first res decl  exp3
+	      | None -> mkres first res decl (promote_mcode sem2))
+	  | Ast0.ForRange((_,bef),decl,ini) ->
+	      let decl = declaration decl in
+	      let (leftinfo,decl) = promote_to_statement_start decl bef in
+	      let ini = initialiser ini in
+	      mkres first (Ast0.ForRange ((leftinfo,bef),decl,ini)) decl ini in
 	let rp = normal_mcode rp in
 	let body = statement body in
 	let (rightinfo,right,body) = promote_to_statement_end body aft in
-	mkres s (Ast0.For(fr,lp,first,exp2,sem2,exp3,rp,body,
+	mkres s (Ast0.For(fr,lp,first,rp,body,
 			  (rightinfo,aft,adj)))
 	  (promote_mcode fr) right
     | Ast0.Iterator(nm,lp,args,rp,body,(_,aft,adj)) ->
@@ -1334,17 +1570,18 @@ let rec statement s =
 	  [] -> mkres s res name right
 	| Ast0.FStorage(stg)::_ -> mkres s res (promote_mcode stg) right
 	| Ast0.FType(ty)::_ -> mkres s res ty right
-	| Ast0.FInline(inline)::_ -> mkres s res (promote_mcode inline) right
-	| Ast0.FAttr(attr)::_ -> mkres s res attr right)
+	| Ast0.FInline(inline)::_ -> mkres s res (promote_mcode inline) right)
 
-    | Ast0.Include(inc,stm) ->
-	let inc = normal_mcode inc in
-	let stm = normal_mcode stm in
-	mkres s (Ast0.Include(inc,stm)) (promote_mcode inc) (promote_mcode stm)
-    | Ast0.MetaInclude(inc,s) ->
-	let inc = normal_mcode inc in
-	let s = expression s in
-	mkres s (Ast0.MetaInclude(inc,s)) (promote_mcode inc) s
+    | Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt) ->
+	let tmpkw = normal_mcode tmpkw in
+	let lab = normal_mcode lab in
+	let params = template_parameter_list (Some(promote_mcode lab)) params in
+	let rab = normal_mcode rab in
+	mkres s (Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt))
+          (promote_mcode tmpkw) (statement stmt)
+    | Ast0.CppTop(di) ->
+	let di = directive di in
+	mkres s (Ast0.CppTop(di)) di di
 
     | Ast0.Undef(def,id) ->
 	let def = normal_mcode def in
@@ -1359,11 +1596,6 @@ let rec statement s =
 	    let (params,prev) = define_parameters params right in
 	    let body = dots is_stm_dots (Some prev) statement body in
 	    mkres s (Ast0.Define(def,id,params,body)) (promote_mcode def) body)
-    | Ast0.Pragma(prg,id,body) ->
-	let prg = normal_mcode prg in
-	let id = ident id in
-	let body = pragmainfo body in
-	mkres s (Ast0.Pragma(prg,id,body)) (promote_mcode prg) body
     | Ast0.OptStm(stm) ->
 	let stm = statement stm in mkres s (Ast0.OptStm(stm)) stm stm
     | Ast0.AsStmt(stm,asstm) ->
@@ -1396,22 +1628,6 @@ and leftfninfo fninfo name bef = (* cases on what is leftmost *)
       (leftinfo,
        Ast0.FInline(set_mcode_info inline (Ast0.get_info inlinfo))::rest,
        name)
-  | Ast0.FAttr(attr)::rest ->
-      let attr = attribute attr in
-      let (leftinfo,attr) =
-	promote_to_statement_start attr bef in
-      (leftinfo,Ast0.FAttr(attr)::rest,name)
-
-and pragmainfo pi =
-  match Ast0.unwrap pi with
-    Ast0.PragmaString(s) ->
-      let s = bad_mcode s in
-      let ln = promote_mcode s in
-      mkres pi (Ast0.PragmaString(s)) ln ln
-  | Ast0.PragmaDots(dots) ->
-      let dots = bad_mcode dots in
-      let ln = promote_mcode dots in
-      mkres pi (Ast0.PragmaDots(dots)) ln ln
 
 and case_line c =
   match Ast0.unwrap c with
