@@ -24,6 +24,7 @@ let set_mcodekind x mcodekind =
     Ast0.DotsExprTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DotsInitTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DotsParamTag(d) -> Ast0.set_mcodekind d mcodekind
+  | Ast0.DotsTemplateParamTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DotsStmtTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DotsDeclTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DotsFieldTag(d) -> Ast0.set_mcodekind d mcodekind
@@ -38,6 +39,7 @@ let set_mcodekind x mcodekind =
       failwith "not possible - iso only"
   | Ast0.TypeCTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.ParamTag(d) -> Ast0.set_mcodekind d mcodekind
+  | Ast0.TemplateParamTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.DeclTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.FieldTag(d) -> Ast0.set_mcodekind d mcodekind
   | Ast0.EnumDeclTag(d) -> Ast0.set_mcodekind d mcodekind
@@ -61,6 +63,7 @@ let set_index x index =
     Ast0.DotsExprTag(d) -> Ast0.set_index d index
   | Ast0.DotsInitTag(d) -> Ast0.set_index d index
   | Ast0.DotsParamTag(d) -> Ast0.set_index d index
+  | Ast0.DotsTemplateParamTag(d) -> Ast0.set_index d index
   | Ast0.DotsStmtTag(d) -> Ast0.set_index d index
   | Ast0.DotsDeclTag(d) -> Ast0.set_index d index
   | Ast0.DotsFieldTag(d) -> Ast0.set_index d index
@@ -75,6 +78,7 @@ let set_index x index =
       failwith "not possible - iso only"
   | Ast0.TypeCTag(d) -> Ast0.set_index d index
   | Ast0.ParamTag(d) -> Ast0.set_index d index
+  | Ast0.TemplateParamTag(d) -> Ast0.set_index d index
   | Ast0.InitTag(d) -> Ast0.set_index d index
   | Ast0.DeclTag(d) -> Ast0.set_index d index
   | Ast0.FieldTag(d) -> Ast0.set_index d index
@@ -97,6 +101,7 @@ let get_index = function
     Ast0.DotsExprTag(d) -> Index.expression_dots d
   | Ast0.DotsInitTag(d) -> Index.initialiser_dots d
   | Ast0.DotsParamTag(d) -> Index.parameter_dots d
+  | Ast0.DotsTemplateParamTag(d) -> Index.template_parameter_dots d
   | Ast0.DotsStmtTag(d) -> Index.statement_dots d
   | Ast0.DotsDeclTag(d) -> Index.declaration_dots d
   | Ast0.DotsFieldTag(d) -> Index.field_dots d
@@ -111,6 +116,7 @@ let get_index = function
       failwith "not possible - iso only"
   | Ast0.TypeCTag(d) -> Index.typeC d
   | Ast0.ParamTag(d) -> Index.parameterTypeDef d
+  | Ast0.TemplateParamTag(d) -> Index.templateParameterTypeDef d
   | Ast0.InitTag(d) -> Index.initialiser d
   | Ast0.DeclTag(d) -> Index.declaration d
   | Ast0.FieldTag(d) -> Index.field d
@@ -178,18 +184,12 @@ let collect_plus_lines top =
       | Ast0.IfThenElse(_,_,_,_,_,_,_,(info,aft,adj))
       | Ast0.Iterator(_,_,_,_,_,(info,aft,adj))
       | Ast0.While(_,_,_,_,_,(info,aft,adj))
-      | Ast0.For(_,_,_,_,_,_,_,_,(info,aft,adj)) ->
+      | Ast0.For(_,_,_,_,_,(info,aft,adj)) ->
 	  bind (k s) (mcode info aft)
       |	_ -> k s in
   let fn =
-    V0.flat_combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing statement donothing donothing donothing
-      donothing donothing donothing in
+    V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=donothing}
+      ~stmt:statement () in
   fn.VT0.combiner_rec_top_level top
 
 (* --------------------------------------------------------------------- *)
@@ -406,6 +406,14 @@ let classify is_minus all_marked table code =
 	  disj_cases e starter expr_list r.VT0.combiner_rec_expression ender
       |	_ -> k e) in
 
+  (* (Ast0.FunCall(_,lp1,_,rp1),Ast0.FunCall(_,lp2,_,rp2)) ->
+      equal_mcode lp1 lp2 && equal_mcode rp1 rp2 *)
+
+  let alignas r (Ast0.Align(al,lp,exp,rp)) =
+	bind (mcode al)
+	  (bind (mcode lp)
+	    (bind (r.VT0.combiner_rec_expression exp)
+	      (mcode rp))) in
 
   (* not clear why we have the next cases, since DisjDecl and
   as it only comes from isos *)
@@ -425,32 +433,26 @@ let classify is_minus all_marked table code =
 	   term structure.  In (all?) other such cases, we visit the terms
 	   using rebuilder, which just visits the subterms, rather than
 	   reordering their components. *)
-      |	Ast0.Init(stg,ty,midattr,id,endattr,eq,ini,sem) ->
-	  bind (match stg with Some stg -> mcode stg | _ -> option_default)
-	    (bind (r.VT0.combiner_rec_typeC ty)
-              (bind
-                 (List.fold_right bind
-                   (List.map r.VT0.combiner_rec_attribute midattr)
-                    option_default)
+      |	Ast0.Init(al,stg,ty,id,endattr,eq,ini,sem) ->
+	  bind (match al with Some al -> alignas r al | _ -> option_default)
+	    (bind (match stg with Some stg -> mcode stg | _ -> option_default)
+	      (bind (r.VT0.combiner_rec_typeC ty)
                  (bind (r.VT0.combiner_rec_ident id)
                     (bind
                        (List.fold_right bind
-                         (List.map r.VT0.combiner_rec_attribute endattr)
+                          (List.map r.VT0.combiner_rec_attribute endattr)
                           option_default)
                        (bind (mcode eq)
                           (bind (r.VT0.combiner_rec_initialiser ini)
                              (mcode sem)))))))
-      | Ast0.UnInit(stg,ty,midattr,id,endattr,sem) ->
-	  bind (match stg with Some stg -> mcode stg | _ -> option_default)
-	    (bind (r.VT0.combiner_rec_typeC ty)
-              (bind
-                 (List.fold_right bind
-                   (List.map r.VT0.combiner_rec_attribute midattr)
-                    option_default)
+      | Ast0.UnInit(al,stg,ty,id,endattr,sem) ->
+          bind (match al with Some al -> alignas r al | _ -> option_default)
+	    (bind (match stg with Some stg -> mcode stg | _ -> option_default)
+	      (bind (r.VT0.combiner_rec_typeC ty)
                  (bind (r.VT0.combiner_rec_ident id)
                     (bind
                        (List.fold_right bind
-                         (List.map r.VT0.combiner_rec_attribute endattr)
+                          (List.map r.VT0.combiner_rec_attribute endattr)
                           option_default)
                        (mcode sem)))))
       |	_ -> k e) in
@@ -463,11 +465,16 @@ let classify is_minus all_marked table code =
 	  disj_cases e starter decls r.VT0.combiner_rec_field ender
       | Ast0.Fdots(dots,whencode) ->
 	  k (Ast0.rewrap e (Ast0.Fdots(dots,None)))
-      | Ast0.Field(ty,id,bf,sem) ->
+      | Ast0.Field(ty,id,bf,endattr,sem) ->
 	  let bitfield (_c, e) = r.VT0.combiner_rec_expression e in
 	  bind (r.VT0.combiner_rec_typeC ty)
 	    (bind (Common.default option_default r.VT0.combiner_rec_ident id)
-	       (bind (Common.default option_default bitfield bf) (mcode sem)))
+	       (bind (Common.default option_default bitfield bf)
+                  (bind
+                     (List.fold_right bind
+                        (List.map r.VT0.combiner_rec_attribute endattr)
+                        option_default)
+		     (mcode sem))))
       |	_ -> k e) in
 
   let enum_decl r k e =
@@ -483,16 +490,26 @@ let classify is_minus all_marked table code =
   let param r k e =
     compute_result Ast0.param e
       (match Ast0.unwrap e with
-	Ast0.Param(ty,midattr,Some id,attr) ->
+	Ast0.Param(ty,Some id,attr) ->
 	  (* needed for the same reason as in the Init and UnInit cases *)
 	  bind (r.VT0.combiner_rec_typeC ty)
-            (bind (List.fold_right bind
-              (List.map r.VT0.combiner_rec_attribute midattr)
-              option_default)
-              (bind (r.VT0.combiner_rec_ident id)
-                (List.fold_right bind
-                   (List.map r.VT0.combiner_rec_attribute attr)
-                   option_default)))
+            (bind (r.VT0.combiner_rec_ident id)
+               (List.fold_right bind
+                  (List.map r.VT0.combiner_rec_attribute attr)
+                  option_default))
+      |	_ -> k e) in
+
+  let template_param r k e =
+    compute_result Ast0.template_param e
+      (match Ast0.unwrap e with
+	Ast0.VarNameParam(ty,id,Some(eq,ini)) ->
+	  (* needed for the same reason as in the Init and UnInit cases *)
+	  bind (r.VT0.combiner_rec_typeC ty)
+            (bind (r.VT0.combiner_rec_ident id)
+	       (bind (mcode eq) (r.VT0.combiner_rec_initialiser ini)))
+      |	Ast0.VarNameParam(ty,id,None) ->
+	  (* needed for the same reason as in the Init and UnInit cases *)
+	  bind (r.VT0.combiner_rec_typeC ty) (r.VT0.combiner_rec_ident id)
       |	_ -> k e) in
 
   let typeC r k e =
@@ -550,7 +567,7 @@ let classify is_minus all_marked table code =
       | Ast0.IfThenElse(start,_,_,_,_,_,_,(info,aft,adj))
       | Ast0.Iterator(_,start,_,_,_,(info,aft,adj))
       | Ast0.While(start,_,_,_,_,(info,aft,adj))
-      | Ast0.For(start,_,_,_,_,_,_,_,(info,aft,adj)) ->
+      | Ast0.For(start,_,_,_,_,(info,aft,adj)) ->
 	  let mcode_info (_,_,info,_,_,_) = info in
 	  bind (k s) (nc_mcode ((),(),mcode_info start,aft,(),adj))
       |	_ -> k s) in
@@ -561,18 +578,20 @@ let classify is_minus all_marked table code =
   let do_top builder r k e = compute_result builder e (k e) in
 
   let combiner =
-    V0.flat_combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode mcode mcode mcode
-      (do_nothing Ast0.dotsExpr) (do_nothing Ast0.dotsInit)
-      (do_nothing Ast0.dotsParam) (do_nothing Ast0.dotsStmt)
-      (do_nothing Ast0.dotsDecl) (do_nothing Ast0.dotsField)
-      (do_nothing Ast0.dotsEnumDecl) (do_nothing Ast0.dotsCase)
-      (do_nothing Ast0.dotsDefParam)
-      ident expression (do_nothing Ast0.assignOp) (do_nothing Ast0.binaryOp)
-      typeC initialiser param declaration field enum_decl
-      statement (do_nothing Ast0.forinfo) case_line string_fragment
-      (do_nothing Ast0.attr) (do_nothing Ast0.attr_arg) (do_top Ast0.top) in
+    V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=(fun r k e -> k e)}
+      ~dotsexpr:(do_nothing Ast0.dotsExpr) ~dotsinit:(do_nothing Ast0.dotsInit)
+      ~dotsparam:(do_nothing Ast0.dotsParam) ~dotstemplateparam:(do_nothing Ast0.dotsTemplateParam)
+      ~dotsstmt:(do_nothing Ast0.dotsStmt)
+      ~dotsdecl:(do_nothing Ast0.dotsDecl) ~dotsfield:(do_nothing Ast0.dotsField)
+      ~dotsenumdecl:(do_nothing Ast0.dotsEnumDecl) ~dotscase:(do_nothing Ast0.dotsCase)
+      ~dotsdefpar:(do_nothing Ast0.dotsDefParam)
+      ~ident:ident ~expr:expression ~assignOp:(do_nothing Ast0.assignOp)
+      ~binaryOp:(do_nothing Ast0.binaryOp)
+      ~ty:typeC ~init:initialiser ~param:param ~template_param:template_param
+      ~decl:declaration ~field:field ~enumdecl:enum_decl
+      ~stmt:statement ~forinfo:(do_nothing Ast0.forinfo) ~case:case_line
+      ~string_fragment:string_fragment ~attribute:(do_nothing Ast0.attr)
+      ~attr_arg:(do_nothing Ast0.attr_arg) ~top:(do_top Ast0.top) () in
   combiner.VT0.combiner_rec_top_level code
 
 (* --------------------------------------------------------------------- *)
@@ -604,6 +623,15 @@ let equal_option e1 e2 =
   | (None, None) -> true
   | _ -> false
 
+let equal_args args1 args2=
+  match (args1, args2) with
+  | (Some args1, Some args2) ->
+    let (lp1,_,rp1) = args1 in
+    let (lp2,_,rp2) = args2 in
+    equal_mcode lp1 lp2 && equal_mcode rp1 rp2
+  | (None, None) -> true
+  | _ -> false
+
 let dots fn d1 d2 =
   List.length (Ast0.unwrap d1) = List.length (Ast0.unwrap d2)
 
@@ -623,9 +651,17 @@ let equal_attribute a1 a2 =
       equal_attr_arg attr1 attr2
   | (Ast0.GccAttribute(attr_1,lp11,lp21,arg1,rp11,rp21),
      Ast0.GccAttribute(attr_2,lp12,lp22,arg2,rp12,rp22)) ->
-      equal_mcode attr_1 attr_2 && equal_mcode lp11 lp12 &&
-      equal_mcode lp21 lp22 && equal_attr_arg arg1 arg2 &&
+      equal_mcode attr_1 attr_2 &&
+      equal_mcode lp11 lp12 && equal_mcode lp21 lp22 &&
       equal_mcode rp11 rp12 && equal_mcode rp21 rp22
+  | (Ast0.CxxAttribute(lb11,arg1,rb11,rb21),
+     Ast0.CxxAttribute(lb12,arg2,rb12,rb22)) ->
+      equal_mcode lb11 lb12 && equal_mcode rb11 rb12 && equal_mcode rb21 rb22
+  | (Ast0.CxxAttributeUsing(lb11,usng1,atnm1,dotdot1,arg1,rb11,rb21),
+     Ast0.CxxAttributeUsing(lb12,usng2,atnm2,dotdot2,arg2,rb12,rb22)) ->
+      equal_mcode lb11 lb12 && equal_mcode usng1 usng2 &&
+      equal_mcode dotdot1 dotdot2 && equal_mcode rb11 rb12 &&
+      equal_mcode rb21 rb22
   | _ -> false
 
 let equal_ident i1 i2 =
@@ -673,15 +709,21 @@ let rec equal_expression e1 e2 =
       equal_mcode pt1 pt2
   | (Ast0.RecordPtAccess(_,ar1,_),Ast0.RecordPtAccess(_,ar2,_)) ->
       equal_mcode ar1 ar2
-  | (Ast0.Cast(lp1,_,ar1,rp1,_),Ast0.Cast(lp2,_,ar2,rp2,_)) ->
-      equal_mcode lp1 lp2 &&
-      (List.length ar1) = (List.length ar2) &&
-      List.for_all2 equal_attribute ar1 ar2 &&
-      equal_mcode rp1 rp2
+  | (Ast0.QualifiedAccess(_,coloncolon1,_),Ast0.QualifiedAccess(_,coloncolon2,_)) ->
+      equal_mcode coloncolon1 coloncolon2
+  | (Ast0.Cast(lp1,_,rp1,_),Ast0.Cast(lp2,_,rp2,_)) ->
+      equal_mcode lp1 lp2 && equal_mcode rp1 rp2
   | (Ast0.SizeOfExpr(szf1,_),Ast0.SizeOfExpr(szf2,_)) ->
       equal_mcode szf1 szf2
   | (Ast0.SizeOfType(szf1,lp1,_,rp1),Ast0.SizeOfType(szf2,lp2,_,rp2)) ->
       equal_mcode szf1 szf2 && equal_mcode lp1 lp2 && equal_mcode rp1 rp2
+  | (Ast0.Delete(dlt1,_),Ast0.Delete(dlt2,_)) ->
+      equal_mcode dlt1 dlt2
+  | (Ast0.DeleteArr(dlt1,lb1,rb1,_),Ast0.DeleteArr(dlt2,lb2,rb2,_)) ->
+      equal_mcode dlt1 dlt2 && equal_mcode lb1 lb2 && equal_mcode rb1 rb2
+  | (Ast0.New(new1,pp1_opt,lp1,_,rp1,args1_opt), Ast0.New(new2,pp2_opt,lp2,_,rp2,args2_opt)) ->
+      equal_mcode new1 new2 && equal_args pp1_opt pp2_opt &&
+      equal_option lp1 lp2 && equal_option rp1 rp2 && equal_args args1_opt args2_opt
   | (Ast0.TypeExp(_),Ast0.TypeExp(_)) -> true
   | (Ast0.Constructor(lp1,_,rp1,_),Ast0.Constructor(lp2,_,rp2,_)) ->
       equal_mcode lp1 lp2 && equal_mcode rp1 rp2
@@ -721,7 +763,27 @@ and binaryOp_equal op1 op2 =
 
 let equal_typeC t1 t2 =
   match (Ast0.unwrap t1,Ast0.unwrap t2) with
-    (Ast0.ConstVol(cv1,_),Ast0.ConstVol(cv2,_)) -> List.for_all2 equal_mcode cv1 cv2
+    (Ast0.ConstVol(cvbeforea,_,cvaftera),Ast0.ConstVol(cvbeforeb,_,cvafterb)) ->
+      let get_cvs =
+	List.fold_left
+	  (fun prev ->
+	    function
+		Ast0.CV cv -> cv::prev
+	      | Ast0.Attr attr -> prev)
+	  [] in
+      let get_attrs =
+	List.fold_left
+	  (fun prev ->
+	    function
+		Ast0.CV cv -> prev
+	      | Ast0.Attr attr -> attr::prev)
+	  [] in
+      let cva = List.sort compare(get_cvs (cvbeforea@cvaftera)) in
+      let cvb = List.sort compare(get_cvs (cvbeforeb@cvafterb)) in
+      let attra = List.sort compare(get_attrs (cvbeforea@cvaftera)) in
+      let attrb = List.sort compare(get_attrs (cvbeforeb@cvafterb)) in
+      List.for_all2 equal_mcode cva cvb &&
+      List.for_all2 equal_attribute attra attrb
   | (Ast0.BaseType(ty1,stringsa),Ast0.BaseType(ty2,stringsb)) ->
       List.for_all2 equal_mcode stringsa stringsb
   | (Ast0.Signed(sign1,_),Ast0.Signed(sign2,_)) ->
@@ -738,12 +800,16 @@ let equal_typeC t1 t2 =
      Ast0.Decimal(dec2,lp2,_,comma2,_,rp2)) ->
        equal_mcode dec1 dec2 && equal_mcode lp1 lp2 &&
        equal_option comma1 comma2 && equal_mcode rp1 rp2
-  | (Ast0.EnumName(kind1,_),Ast0.EnumName(kind2,_)) ->
-      equal_mcode kind1 kind2
-  | (Ast0.EnumDef(_,lb1,_,rb1),Ast0.EnumDef(_,lb2,_,rb2)) ->
+  | (Ast0.EnumName(kind1,key1,_),Ast0.EnumName(kind2,key2,_)) ->
+      equal_mcode kind1 kind2 && equal_option key1 key2
+  | (Ast0.EnumDef(_,base1,lb1,_,rb1),Ast0.EnumDef(_,base2,lb2,_,rb2)) ->
        let tru1 = equal_mcode lb1 lb2 in
-       let tru2 = equal_mcode rb1 rb2 in
-       tru1 && tru2
+       let tru2 = (match (base1,base2) with
+           (None, None) -> true
+         | (Some (td1, _), Some (td2, _)) -> equal_mcode td1 td2
+         | _ -> false) in
+       let tru3 = equal_mcode rb1 rb2 in
+       tru1 && tru2 && tru3
   | (Ast0.StructUnionName(kind1,_),Ast0.StructUnionName(kind2,_)) ->
       equal_mcode kind1 kind2
   | (Ast0.StructUnionDef(_,lb1,_,rb1),
@@ -753,6 +819,8 @@ let equal_typeC t1 t2 =
       equal_mcode tf1 tf2 && equal_mcode lp1 lp2 && equal_mcode rp1 rp2
   | (Ast0.TypeOfType(tf1,lp1,_,rp1),Ast0.TypeOfType(tf2,lp2,_,rp2)) ->
       equal_mcode tf1 tf2 && equal_mcode lp1 lp2 && equal_mcode rp1 rp2
+  | (Ast0.QualifiedType(_,coloncolon1,_),Ast0.QualifiedType(_,coloncolon2,_)) ->
+      equal_mcode coloncolon1 coloncolon2
   | (Ast0.TypeName(name1),Ast0.TypeName(name2)) -> equal_mcode name1 name2
   | (Ast0.AutoType(auto1),Ast0.AutoType(auto2)) -> equal_mcode auto1 auto2
   | (Ast0.MetaType(name1,_,_),Ast0.MetaType(name2,_,_)) ->
@@ -775,30 +843,25 @@ let equal_fninfo x y =
     (Ast0.FStorage(s1),Ast0.FStorage(s2)) -> equal_mcode s1 s2
   | (Ast0.FType(_),Ast0.FType(_)) -> true
   | (Ast0.FInline(i1),Ast0.FInline(i2)) -> equal_mcode i1 i2
-  | (Ast0.FAttr(i1),Ast0.FAttr(i2)) -> equal_attribute i1 i2
   | _ -> false
 
 let equal_declaration d1 d2 =
   match (Ast0.unwrap d1,Ast0.unwrap d2) with
     (Ast0.MetaDecl(name1,_,_),Ast0.MetaDecl(name2,_,_)) ->
       equal_mcode name1 name2
-  | (Ast0.Init(stg1,_,midattr1,_,endattr1,eq1,_,sem1),
-     Ast0.Init(stg2,_,midattr2,_,endattr2,eq2,_,sem2)) ->
+  | (Ast0.Init(_,stg1,_,_,endattr1,eq1,_,sem1),
+     Ast0.Init(_,stg2,_,_,endattr2,eq2,_,sem2)) ->
       equal_option stg1 stg2 &&
-      (List.length midattr1) = (List.length midattr2) &&
-      List.for_all2 equal_attribute midattr1 midattr2 &&
       (List.length endattr1) = (List.length endattr2) &&
       List.for_all2 equal_attribute endattr1 endattr2 &&
       equal_mcode eq1 eq2 && equal_mcode sem1 sem2
-  | (Ast0.UnInit(stg1,_,midattr1,_,endattr1,sem1),Ast0.UnInit(stg2,_,midattr2,_,endattr2,sem2)) ->
+  | (Ast0.UnInit(_,stg1,_,_,endattr1,sem1),Ast0.UnInit(_,stg2,_,_,endattr2,sem2)) ->
       equal_option stg1 stg2 &&
-      (List.length midattr1) = (List.length midattr2) &&
-      List.for_all2 equal_attribute midattr1 midattr2 &&
       (List.length endattr1) = (List.length endattr2) &&
       List.for_all2 equal_attribute endattr1 endattr2 &&
       equal_mcode sem1 sem2
-  | (Ast0.FunProto(fninfo1,attr1,name1,lp1,p1,va1,rp1,sem1),
-     Ast0.FunProto(fninfo2,attr2,name2,lp2,p2,va2,rp2,sem2)) ->
+  | (Ast0.FunProto(fninfo1,name1,lp1,p1,va1,rp1,sem1),
+     Ast0.FunProto(fninfo2,name2,lp2,p2,va2,rp2,sem2)) ->
        let equal_varargs va1 va2 = match (va1,va2) with
          | None, None -> true
          | Some (c1, e1), Some (c2, e2) ->
@@ -806,24 +869,26 @@ let equal_declaration d1 d2 =
          | _ -> false in
        (List.length fninfo1) = (List.length fninfo2) &&
        List.for_all2 equal_fninfo fninfo1 fninfo2 &&
-      (List.length attr1) = (List.length attr2) &&
-      List.for_all2 equal_attribute attr1 attr2 &&
        equal_mcode lp1 lp2 && equal_varargs va1 va2 &&
        equal_mcode rp1 rp2 && equal_mcode sem1 sem2
-  | (Ast0.MacroDecl(stg1,nm1,lp1,_,rp1,attr1,sem1),
-     Ast0.MacroDecl(stg2,nm2,lp2,_,rp2,attr2,sem2)) ->
-      equal_option stg1 stg2 &&
-      (List.length attr1) = (List.length attr2) &&
-      List.for_all2 equal_attribute attr1 attr2 &&
-      equal_mcode lp1 lp2 && equal_mcode rp1 rp2 && equal_mcode sem1 sem2
-  | (Ast0.MacroDeclInit(stg1,nm1,lp1,_,rp1,eq1,_,sem1),
-     Ast0.MacroDeclInit(stg2,nm2,lp2,_,rp2,eq2,_,sem2)) ->
+  | (Ast0.MacroDecl(stg1,preattr1,nm1,lp1,_,rp1,attr1,sem1),
+     Ast0.MacroDecl(stg2,preattr2,nm2,lp2,_,rp2,attr2,sem2)) ->
        equal_option stg1 stg2 &&
+       (List.length preattr1) = (List.length preattr2) &&
+       List.for_all2 equal_attribute preattr1 preattr2 &&
+       (List.length attr1) = (List.length attr2) &&
+       List.for_all2 equal_attribute attr1 attr2 &&
+       equal_mcode lp1 lp2 && equal_mcode rp1 rp2 && equal_mcode sem1 sem2
+  | (Ast0.MacroDeclInit(stg1,preattr1,nm1,lp1,_,rp1,attr1,eq1,_,sem1),
+     Ast0.MacroDeclInit(stg2,preattr2,nm2,lp2,_,rp2,attr2,eq2,_,sem2)) ->
+       equal_option stg1 stg2 &&
+       (List.length preattr1) = (List.length preattr2) &&
+       List.for_all2 equal_attribute preattr1 preattr2 &&
+       (List.length attr1) = (List.length attr2) &&
+       List.for_all2 equal_attribute attr1 attr2 &&
        equal_mcode lp1 lp2 && equal_mcode rp1 rp2 && equal_mcode eq1 eq2
 	 && equal_mcode sem1 sem2
-  | (Ast0.TyDecl(_,attr1,sem1),Ast0.TyDecl(_,attr2,sem2)) ->
-       (List.length attr1) = (List.length attr2) &&
-       List.for_all2 equal_attribute attr1 attr2 && equal_mcode sem1 sem2
+  | (Ast0.TyDecl(_,sem1),Ast0.TyDecl(_,sem2)) -> equal_mcode sem1 sem2
   | (Ast0.OptDecl(_),Ast0.OptDecl(_)) -> true
   | (Ast0.DisjDecl(starter1,_,mids1,ender1),
      Ast0.DisjDecl(starter2,_,mids2,ender2))
@@ -834,13 +899,44 @@ let equal_declaration d1 d2 =
        equal_mcode ender1 ender2
   | _ -> false
 
+let equal_directive di1 di2 =
+  match (Ast0.unwrap di1,Ast0.unwrap di2) with
+    (Ast0.Include(inc1,name1),Ast0.Include(inc2,name2)) ->
+      equal_mcode inc1 inc2 && equal_mcode name1 name2
+  | (Ast0.MetaInclude(inc1,name1),Ast0.MetaInclude(inc2,name2)) ->
+      equal_mcode inc1 inc2
+  | (Ast0.Pragma(prg1,id1,body1),Ast0.Pragma(prg2,id2,body2)) ->
+      equal_mcode prg1 prg2
+  | (Ast0.UsingNamespace(usng1,nmspc1,s1,sem1),
+     Ast0.UsingNamespace(usng2,nmspc2,s2,sem2)) ->
+       equal_mcode usng1 usng2 && equal_mcode nmspc1 nmspc2 &&
+       equal_mcode sem1 sem2
+  | (Ast0.UsingTypename(usng1,name1,eq1,tn1,ty1,sem1),
+     Ast0.UsingTypename(usng2,name2,eq2,tn2,ty2,sem2)) ->
+       equal_mcode usng1 usng2 && equal_mcode eq1 eq2 &&
+       equal_mcode sem1 sem2
+  | (Ast0.UsingMember(usng1,_,sem1),
+     Ast0.UsingMember(usng2,_,sem2)) ->
+       equal_mcode usng1 usng2 && equal_mcode sem1 sem2
+  | _ -> false
+
 let equal_field d1 d2 =
   match (Ast0.unwrap d1,Ast0.unwrap d2) with
-     (Ast0.MetaField(name1,_,_),Ast0.MetaField(name2,_,_))
+    (Ast0.MetaField(name1,_,_),Ast0.MetaField(name2,_,_))
   | (Ast0.MetaFieldList(name1,_,_,_),Ast0.MetaFieldList(name2,_,_,_)) ->
       equal_mcode name1 name2
-  | (Ast0.Field(_,_,_bf1,sem1),Ast0.Field(_,_,_bf2,sem2)) ->
+  | (Ast0.Field(_,_,_bf1,endattr1,sem1),Ast0.Field(_,_,_bf2,endattr2,sem2)) ->
+      (List.length endattr1) = (List.length endattr2) &&
+      List.for_all2 equal_attribute endattr1 endattr2 &&
       equal_mcode sem1 sem2
+  | (Ast0.MacroDeclField(_,lp1,_,rp1,endattr1,sem1),
+     Ast0.MacroDeclField(_,lp2,_,rp2,endattr2,sem2)) ->
+      (List.length endattr1) = (List.length endattr2) &&
+      List.for_all2 equal_attribute endattr1 endattr2 &&
+      equal_mcode lp1 lp2 && equal_mcode lp1 lp2 &&
+      equal_mcode sem1 sem2
+  | (Ast0.CppField di1,Ast0.CppField di2) -> equal_directive di1 di2
+       
   | (Ast0.Fdots(dots1,_),Ast0.Fdots(dots2,_)) -> equal_mcode dots1 dots2
   | (Ast0.OptField(_),Ast0.OptField(_)) -> true
   | (Ast0.DisjField(starter1,_,mids1,ender1),
@@ -901,14 +997,9 @@ let equal_initialiser i1 i2 =
 
 let equal_parameterTypeDef p1 p2 =
   match (Ast0.unwrap p1,Ast0.unwrap p2) with
-    (Ast0.VoidParam(_,ar1),Ast0.VoidParam(_,ar2)) ->
+    (Ast0.Param(_,_,ar1),Ast0.Param(_,_,ar2)) ->
       (List.length ar1) = (List.length ar2) &&
       List.for_all2 equal_attribute ar1 ar2
-  | (Ast0.Param(_,mar1,_,ar1),Ast0.Param(_,mar2,_,ar2)) ->
-      (List.length ar1) = (List.length ar2) &&
-      List.for_all2 equal_attribute ar1 ar2 &&
-      (List.length mar1) = (List.length mar2) &&
-      List.for_all2 equal_attribute mar1 mar2
   | (Ast0.MetaParam(name1,_,_),Ast0.MetaParam(name2,_,_))
   | (Ast0.MetaParamList(name1,_,_,_),Ast0.MetaParamList(name2,_,_,_)) ->
       equal_mcode name1 name2
@@ -917,6 +1008,18 @@ let equal_parameterTypeDef p1 p2 =
   | (Ast0.OptParam(_),Ast0.OptParam(_)) -> true
   | _ -> false
 
+let equal_templateParameterTypeDef p1 p2 =
+  match (Ast0.unwrap p1,Ast0.unwrap p2) with
+    (Ast0.TypenameOrClassParam(tn1,_,Some(eq1,_)),
+     Ast0.TypenameOrClassParam(tn2,_,Some(eq2,_))) ->
+       equal_mcode tn1 tn2 && equal_mcode eq1 eq2
+  | (Ast0.TypenameOrClassParam(tn1,_,None),
+     Ast0.TypenameOrClassParam(tn2,_,None)) -> equal_mcode tn1 tn2
+  | (Ast0.VarNameParam(_,_,Some(eq1,_)),
+     Ast0.VarNameParam(_,_,Some(eq2,_))) -> equal_mcode eq1 eq2
+  | (Ast0.VarNameParam(_,_,None),Ast0.VarNameParam(_,_,None)) -> true
+  | _-> false
+
 let equal_statement s1 s2 =
   match (Ast0.unwrap s1,Ast0.unwrap s2) with
     (Ast0.FunDecl(_,fninfo1,_,lp1,_,_,rp1,_,lbrace1,_,rbrace1,_),
@@ -941,17 +1044,18 @@ let equal_statement s1 s2 =
   | (Ast0.Do(d1,_,whl1,lp1,_,rp1,sem1),Ast0.Do(d2,_,whl2,lp2,_,rp2,sem2)) ->
       equal_mcode whl1 whl2 && equal_mcode d1 d2 &&
       equal_mcode lp1 lp2 && equal_mcode rp1 rp2 && equal_mcode sem1 sem2
-  | (Ast0.For(fr1,lp1,first1,_,sem21,_,rp1,_,_),
-     Ast0.For(fr2,lp2,first2,_,sem22,_,rp2,_,_)) ->
+  | (Ast0.For(fr1,lp1,first1,rp1,_,_),
+     Ast0.For(fr2,lp2,first2,rp2,_,_)) ->
        let first =
 	 match (Ast0.unwrap first1,Ast0.unwrap first2) with
-	   (Ast0.ForExp(_,sem1),Ast0.ForExp(_,sem2)) ->
-	     equal_mcode sem1 sem2
-	 | (Ast0.ForDecl _,Ast0.ForDecl _) -> true
+	   (Ast0.ForExp(_,sem1,_,sem21,_),Ast0.ForExp(_,sem2,_,sem22,_)) ->
+	     equal_mcode sem1 sem2 && equal_mcode sem21 sem22
+	 | (Ast0.ForDecl(_,_,_,sem21,_),Ast0.ForDecl(_,_,_,sem22,_)) ->
+	     equal_mcode sem21 sem22
+	 | (Ast0.ForRange(_,_,_),Ast0.ForRange(_,_,_)) -> true
 	 | _ -> false in
        equal_mcode fr1 fr2 && equal_mcode lp1 lp2 &&
-       first && equal_mcode sem21 sem22 &&
-       equal_mcode rp1 rp2
+       first && equal_mcode rp1 rp2
   | (Ast0.Iterator(nm1,lp1,_,rp1,_,_),Ast0.Iterator(nm2,lp2,_,rp2,_,_)) ->
       equal_mcode lp1 lp2 && equal_mcode rp1 rp2
   | (Ast0.Switch(switch1,lp1,_,rp1,lb1,_,_,rb1),
@@ -990,16 +1094,11 @@ let equal_statement s1 s2 =
   | (Ast0.TopId(_),Ast0.TopId(_)) -> true
   | (Ast0.TopInit(_),Ast0.TopInit(_)) -> true
   | (Ast0.Dots(d1,_),Ast0.Dots(d2,_)) -> equal_mcode d1 d2
-  | (Ast0.Include(inc1,name1),Ast0.Include(inc2,name2)) ->
-      equal_mcode inc1 inc2 && equal_mcode name1 name2
-  | (Ast0.MetaInclude(inc1,name1),Ast0.MetaInclude(inc2,name2)) ->
-      equal_mcode inc1 inc2
   | (Ast0.Undef(def1,_),Ast0.Undef(def2,_)) ->
       equal_mcode def1 def2
   | (Ast0.Define(def1,_,_,_),Ast0.Define(def2,_,_,_)) ->
       equal_mcode def1 def2
-  | (Ast0.Pragma(prg1,_,_),Ast0.Pragma(prg2,_,_)) ->
-      equal_mcode prg1 prg2
+  | (Ast0.CppTop di1,Ast0.CppTop di2) -> equal_directive di1 di2
   | (Ast0.OptStm(_),Ast0.OptStm(_)) -> true
   | _ -> false
 
@@ -1020,6 +1119,7 @@ let equal_case_line c1 c2 =
 let equal_define_param d1 d2 =
   match (Ast0.unwrap d1,Ast0.unwrap d2) with
     (Ast0.DParam _,Ast0.DParam _) -> true
+  | (Ast0.DParamEll _,Ast0.DParamEll _) -> true
   | (Ast0.MetaDParamList(name1,_,_,_),Ast0.MetaDParamList(name2,_,_,_)) ->
       equal_mcode name1 name2
   | (Ast0.DPComma cm1,Ast0.DPComma cm2) -> equal_mcode cm1 cm2
@@ -1041,6 +1141,8 @@ let root_equal e1 e2 =
     (Ast0.DotsExprTag(d1),Ast0.DotsExprTag(d2)) -> dots equal_expression d1 d2
   | (Ast0.DotsParamTag(d1),Ast0.DotsParamTag(d2)) ->
       dots equal_parameterTypeDef d1 d2
+  | (Ast0.DotsTemplateParamTag(d1),Ast0.DotsTemplateParamTag(d2)) ->
+      dots equal_templateParameterTypeDef d1 d2
   | (Ast0.DotsStmtTag(d1),Ast0.DotsStmtTag(d2)) -> dots equal_statement d1 d2
   | (Ast0.DotsDeclTag(d1),Ast0.DotsDeclTag(d2)) -> dots equal_declaration d1 d2
   | (Ast0.DotsFieldTag(d1),Ast0.DotsFieldTag(d2)) -> dots equal_field d1 d2
@@ -1056,6 +1158,8 @@ let root_equal e1 e2 =
   | (Ast0.ArgExprTag(d),_) -> failwith "not possible - iso only"
   | (Ast0.TypeCTag(t1),Ast0.TypeCTag(t2)) -> equal_typeC t1 t2
   | (Ast0.ParamTag(p1),Ast0.ParamTag(p2)) -> equal_parameterTypeDef p1 p2
+  | (Ast0.TemplateParamTag(p1),Ast0.TemplateParamTag(p2)) ->
+      equal_templateParameterTypeDef p1 p2
   | (Ast0.InitTag(d1),Ast0.InitTag(d2)) -> equal_initialiser d1 d2
   | (Ast0.DeclTag(d1),Ast0.DeclTag(d2)) -> equal_declaration d1 d2
   | (Ast0.FieldTag(d1),Ast0.FieldTag(d2)) -> equal_field d1 d2
@@ -1097,14 +1201,7 @@ let contextify_all =
   let option_default = () in
   let mcode x = () in
   let donothing r k e = Ast0.set_mcodekind e (default_context()); k e in
-
-  V0.flat_combiner bind option_default
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing
+  V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=donothing} ()
 
 let contextify_whencode =
   let bind x y = () in
@@ -1141,11 +1238,8 @@ let contextify_whencode =
     | _ -> () in
 
   let combiner =
-    V0.combiner bind option_default
-      {V0.combiner_functions with
-	VT0.combiner_exprfn = expression;
-	VT0.combiner_initfn = initialiser;
-	VT0.combiner_stmtfn = statement} in
+    V0.combiner_default bind option_default
+      ~expr:expression ~init:initialiser ~stmt:statement () in
   combiner.VT0.combiner_rec_top_level
 
 (* --------------------------------------------------------------------- *)
@@ -1273,9 +1367,8 @@ let rec is_toplevel s =
       (match Ast0.unwrap fc with
 	Ast0.FunCall(_,_,_,_) -> true
       |	_ -> false)
-  | Ast0.Include(_,_) -> true
   | Ast0.Undef(_,_) -> true
-  | Ast0.Pragma(_,_,_) -> true
+  | Ast0.CppTop(_) -> true
   | Ast0.Define(_,_,_,_) -> true
   | _ -> false
 
