@@ -21,6 +21,20 @@ let disjmult3 e1 e2 e3 k =
 	      e2))
        e1)
 
+let disjmult4 e1 e2 e3 e4 k =
+  List.concat
+    (List.map
+       (function e1 ->
+	 List.concat
+	   (List.map
+	      (function e2 ->
+		List.concat
+		  (List.map
+		     (function e3 -> List.map (function e4 -> k e1 e2 e3 e4) e4)
+		     e3))
+	      e2))
+       e1)
+
 let rec disjmult f = function
     [] -> [[]]
   | x::xs ->
@@ -47,9 +61,11 @@ let disjdots f d =
 
 let rec disjty ft =
   match Ast.unwrap ft with
-    Ast.Type(allminus,cv,ty) ->
+    Ast.Type(allminus,cvbefore,ty,cvafter) ->
       let ty = disjtypeC ty in
-      List.map (function ty -> Ast.rewrap ft (Ast.Type(allminus,cv,ty))) ty
+      List.map
+	(function ty -> Ast.rewrap ft (Ast.Type(allminus,cvbefore,ty,cvafter)))
+	ty
   | Ast.AsType(ty,asty) -> (* as ty doesn't contain disj *)
       let ty = disjty ty in
       List.map (function ty -> Ast.rewrap ft (Ast.AsType(ty,asty))) ty
@@ -85,18 +101,24 @@ and disjtypeC bty =
       disjmult2 (disjexp length) (disjoption disjexp precision_opt)
 	(function length -> function precision_opt ->
 	  Ast.rewrap bty (Ast.Decimal(dec,lp,length,comma,precision_opt,rp)))
-  | Ast.EnumName(enum,name) ->
+  | Ast.EnumName(enum,key,name) ->
       let name = disjoption disjident name in
-      List.map (function name -> Ast.rewrap bty (Ast.EnumName(enum,name))) name
+      List.map (function name -> Ast.rewrap bty (Ast.EnumName(enum,key,name))) name
   | Ast.StructUnionName(su,name) ->
       let name = disjoption disjident name in
       List.map
 	(function name -> Ast.rewrap bty (Ast.StructUnionName(su,name)))
 	name
-  | Ast.EnumDef(ty,lb,ids,rb) ->
-      disjmult2 (disjty ty) (disjdots disjenumdecl ids)
-	(function ty -> function ids ->
-	  Ast.rewrap bty (Ast.EnumDef(ty,lb,ids,rb)))
+  | Ast.EnumDef(ty,base,lb,ids,rb) ->
+      disjmult3 (disjty ty)
+	(disjoption
+	   (function (td, ty1) ->
+	      let ty1 = disjty ty1 in
+	      List.map (function ty1 -> (td, ty1)) ty1)
+	   base)
+	((disjdots disjenumdecl) ids)
+	(fun ty base ids ->
+	  Ast.rewrap bty (Ast.EnumDef(ty,base,lb,ids,rb)))
   | Ast.StructUnionDef(ty,lb,decls,rb) ->
       disjmult2 (disjty ty) (disjdots anndisjfield decls)
 	(function ty -> function decls ->
@@ -109,6 +131,13 @@ and disjtypeC bty =
       let ty = disjty ty in
       List.map
 	(function ty -> Ast.rewrap bty (Ast.TypeOfType(tf,lp,ty,rp))) ty
+  | Ast.TemplateType(tn,lp,args,rp) ->
+      disjmult2 (disjty tn) (disjdots disjexp args)
+	(fun tn args ->
+	  Ast.rewrap bty (Ast.TemplateType(tn,lp,args,rp)))
+  | Ast.QualifiedType(ty,coloncolon,name) ->
+      disjmult2 (disjoption disjty ty) (disjident name)
+      (function ty -> function name -> Ast.rewrap bty (Ast.QualifiedType(ty,coloncolon,name)))
   | Ast.TypeName(_) | Ast.AutoType(_) | Ast.MetaType(_,_,_,_) -> [bty]
 
 and anndisjdecl d =
@@ -209,20 +238,26 @@ and disjexp e =
   | Ast.Paren(lp,exp,rp) ->
       let exp = disjexp exp in
       List.map (function exp -> Ast.rewrap e (Ast.Paren(lp,exp,rp))) exp
-  | Ast.ArrayAccess(exp1,lb,exp2,rb) ->
-      disjmult2 (disjexp exp1) (disjexp exp2)
-	(function exp1 -> function exp2 ->
-	  Ast.rewrap e (Ast.ArrayAccess(exp1,lb,exp2,rb)))
+  | Ast.ArrayAccess(fn,lb,args,rb) ->
+      disjmult2 (disjexp fn) (disjdots disjexp args)
+	(function fn -> function args ->
+	  Ast.rewrap e (Ast.ArrayAccess(fn,lb,args,rb)))
   | Ast.RecordAccess(exp,pt,field) ->
       disjmult2 (disjexp exp) (disjident field)
 	(fun exp field -> Ast.rewrap e (Ast.RecordAccess(exp,pt,field)))
   | Ast.RecordPtAccess(exp,ar,field) ->
       disjmult2 (disjexp exp) (disjident field)
 	(fun exp field -> Ast.rewrap e (Ast.RecordPtAccess(exp,ar,field)))
-  | Ast.Cast(lp,ty,attr,rp,exp) ->
+  | Ast.QualifiedAccess(Some ty,coloncolon,field) ->
+      disjmult2 (disjty ty) (disjident field)
+    (fun ty field -> Ast.rewrap e (Ast.QualifiedAccess(Some ty,coloncolon,field)))
+  | Ast.QualifiedAccess(None,coloncolon,field) ->
+      let field = disjident field in 
+      List.map (function field -> Ast.rewrap e (Ast.QualifiedAccess(None,coloncolon,field))) field
+  | Ast.Cast(lp,ty,rp,exp) ->
       disjmult2 (disjty ty) (disjexp exp)
 	(function ty -> function exp ->
-          Ast.rewrap e (Ast.Cast(lp,ty,attr,rp,exp)))
+          Ast.rewrap e (Ast.Cast(lp,ty,rp,exp)))
   | Ast.SizeOfExpr(szf,exp) ->
       let exp = disjexp exp in
       List.map (function exp -> Ast.rewrap e (Ast.SizeOfExpr(szf,exp))) exp
@@ -230,6 +265,20 @@ and disjexp e =
       let ty = disjty ty in
       List.map
 	(function ty -> Ast.rewrap e (Ast.SizeOfType(szf,lp,ty,rp))) ty
+  | Ast.Delete(dlt, exp) ->
+      let exp = disjexp exp in
+      List.map (function exp -> Ast.rewrap e (Ast.Delete(dlt, exp))) exp
+  | Ast.DeleteArr(dlt,lb,rb,exp) ->
+      let exp = disjexp exp in
+      List.map (function exp -> Ast.rewrap e (Ast.DeleteArr(dlt,lb,rb,exp))) exp
+  | Ast.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) ->
+      disjmult3 (disjoption disjargs pp_opt) (disjty ty) (disjoption disjargs args_opt)
+	(fun pp_opt ty args_opt ->
+	  Ast.rewrap e (Ast.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt)))
+  | Ast.TemplateInst(tn,lp,args,rp) ->
+      disjmult2 (disjexp tn) (disjdots disjexp args)
+	(fun tn args ->
+	  Ast.rewrap e (Ast.TemplateInst(tn,lp,args,rp)))
   | Ast.TypeExp(ty) ->
       let ty = disjty ty in
       List.map (function ty -> Ast.rewrap e (Ast.TypeExp(ty))) ty
@@ -258,12 +307,15 @@ and disjexp e =
       let exp = disjexp exp in
       List.map (function exp -> Ast.rewrap e (Ast.OptExp(exp))) exp
 
+and disjargs (lp,args,rp) =
+  let args = disjdots disjexp args in
+  List.map (function args -> (lp,args,rp)) args
+
 and disjparam p =
   match Ast.unwrap p with
-    Ast.VoidParam(ty,attr) -> [p] (* void is the only possible value *)
-  | Ast.Param(ty,midattr,id,attr) ->
+    Ast.Param(ty,id,attr) ->
       disjmult2 (disjty ty) (disjoption disjident id)
-	(fun ty id -> Ast.rewrap p (Ast.Param(ty,midattr,id,attr)))
+	(fun ty id -> Ast.rewrap p (Ast.Param(ty,id,attr)))
   | Ast.AsParam(pm,asexp) -> (* as exp doesn't contain disj *)
       let pm = disjparam pm in
       List.map (function pm -> Ast.rewrap p (Ast.AsParam(pm,asexp))) pm
@@ -273,6 +325,26 @@ and disjparam p =
       let param = disjparam param in
       List.map (function param -> Ast.rewrap p (Ast.OptParam(param))) param
 
+and disjtemplateparam p =
+  match Ast.unwrap p with
+    Ast.TypenameOrClassParam(tyorcl,id,Some(eq,ty)) ->
+      disjmult2 (disjident id) (disjty ty)
+	(fun id ty ->
+	  Ast.rewrap p (Ast.TypenameOrClassParam(tyorcl,id,Some(eq,ty))))
+  | Ast.TypenameOrClassParam(tyorcl,id,None) ->
+      let id = disjident id in
+      List.map
+	(fun id -> Ast.rewrap p (Ast.TypenameOrClassParam(tyorcl,id,None)))
+	id
+  | Ast.VarNameParam(ty,id,Some (eq,ini)) ->
+      disjmult3 (disjty ty) (disjident id) (disjini ini)
+	(fun ty id exp -> Ast.rewrap p (Ast.VarNameParam(ty,id,Some(eq,ini))))
+  | Ast.VarNameParam(ty,id,None) ->
+      disjmult2 (disjty ty) (disjident id)
+	(fun ty id -> Ast.rewrap p (Ast.VarNameParam(ty,id,None)))
+  | Ast.TPComma(comma) -> [p]
+  | Ast.TPDots(dots) -> [p]
+
 and disjini i =
   match Ast.unwrap i with
     Ast.MetaInit(_,_,_,_) | Ast.MetaInitList(_,_,_,_,_) -> [i]
@@ -325,35 +397,42 @@ and disjfninfo = function
   | Ast.FType(ty) -> List.map (function ty -> Ast.FType(ty)) (disjty ty)
   | fi -> [fi]
 
+and disjalign = function
+    None -> [None]
+  | Some(Ast.Align(a,lp,e,rp)) ->
+      let e = disjexp e in
+      List.map (function e -> Some(Ast.Align(a,lp,e,rp))) e
+
 and disjdecl d =
   match Ast.unwrap d with
     Ast.MetaDecl(_,_,_,_) -> [d]
   | Ast.AsDecl(decl,asdecl) ->
       let decl = disjdecl decl in
       List.map (function decl -> Ast.rewrap d (Ast.AsDecl(decl,asdecl))) decl
-  | Ast.Init(stg,ty,midattr,id,endattr,eq,ini,sem) ->
-      disjmult3 (disjty ty) (disjident id) (disjini ini)
-	(fun ty id ini ->
-	  Ast.rewrap d (Ast.Init(stg,ty,midattr,id,endattr,eq,ini,sem)))
-  | Ast.UnInit(stg,ty,midattr,id,endattr,sem) ->
-      disjmult2 (disjty ty) (disjident id)
-	(fun ty id ->
-	  Ast.rewrap d (Ast.UnInit(stg,ty,midattr,id,endattr,sem)))
-  | Ast.FunProto(fninfo,attr,name,lp1,params,va,rp1,sem) ->
+  | Ast.Init(align,stg,ty,id,endattr,eq,ini,sem) ->
+      disjmult4 (disjalign align) (disjty ty) (disjident id) (disjini ini)
+	(fun al ty id ini ->
+	  Ast.rewrap d (Ast.Init(al,stg,ty,id,endattr,eq,ini,sem)))
+  | Ast.UnInit(align,stg,ty,id,endattr,sem) ->
+      disjmult3 (disjalign align) (disjty ty) (disjident id)
+	(fun al ty id ->
+	  Ast.rewrap d (Ast.UnInit(al,stg,ty,id,endattr,sem)))
+  | Ast.FunProto(fninfo,name,lp1,params,va,rp1,sem) ->
       disjmult2 (disjmult disjfninfo fninfo) (disjident name)
 	(fun fninfo name ->
-	  Ast.rewrap d (Ast.FunProto(fninfo,attr,name,lp1,params,va,rp1,sem)))
-  | Ast.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
+	  Ast.rewrap d (Ast.FunProto(fninfo,name,lp1,params,va,rp1,sem)))
+  | Ast.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
       disjmult2 (disjident name) (disjdots disjexp args)
 	(fun name args ->
-	  Ast.rewrap d (Ast.MacroDecl(stg,name,lp,args,rp,attr,sem)))
-  | Ast.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
+	  Ast.rewrap d (Ast.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem)))
+  | Ast.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
       disjmult3 (disjident name) (disjdots disjexp args) (disjini ini)
 	(fun name args ini ->
-	  Ast.rewrap d (Ast.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem)))
-  | Ast.TyDecl(ty,attr,sem) ->
+	  Ast.rewrap d
+	    (Ast.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem)))
+  | Ast.TyDecl(ty,sem) ->
       let ty = disjty ty in
-      List.map (function ty -> Ast.rewrap d (Ast.TyDecl(ty,attr,sem))) ty
+      List.map (function ty -> Ast.rewrap d (Ast.TyDecl(ty,sem))) ty
   | Ast.Typedef(stg,ty,id,sem) ->
       let ty = disjty ty in (* disj not allowed in id *)
       List.map (function ty -> Ast.rewrap d (Ast.Typedef(stg,ty,id,sem))) ty
@@ -370,11 +449,35 @@ and disjfield d =
   match Ast.unwrap d with
     Ast.MetaField(_,_,_,_)
   | Ast.MetaFieldList(_,_,_,_,_) -> [d]
-  | Ast.Field(ty,id,bf,sem) ->
+  | Ast.Field(ty,id,bf,endattr,sem) ->
       let disjbf (c, e) = List.map (fun e -> (c, e)) (disjexp e) in
       disjmult3 (disjty ty) (disjoption disjident id) (disjoption disjbf bf)
 	(fun ty id bf ->
-	  Ast.rewrap d (Ast.Field(ty,id,bf,sem)))
+	  Ast.rewrap d (Ast.Field(ty,id,bf,endattr,sem)))
+  | Ast.MacroDeclField(name,lp,args,rp,attr,sem) ->
+      disjmult2 (disjident name) (disjdots disjexp args)
+	(fun name args ->
+	  Ast.rewrap d (Ast.MacroDeclField(name,lp,args,rp,attr,sem)))
+  | Ast.CppField di ->
+      let di = disjdirective di in
+      List.map (fun di -> Ast.rewrap d (Ast.CppField di)) di
+
+and disjdirective di =
+  match Ast.unwrap di with
+    Ast.Include _ | Ast.MetaInclude _ | Ast.Pragma _ -> [di]
+  | Ast.UsingNamespace(usng,nmspc,name,sem) ->
+      let name = disjident name in
+      List.map
+	(function name ->
+	  Ast.rewrap di (Ast.UsingNamespace(usng,nmspc,name,sem)))
+	name
+  | Ast.UsingTypename(usng,name,eq,tn,ty,sem) ->
+      disjmult2 (disjident name) (disjty ty)
+	(fun name ty ->
+	  Ast.rewrap di (Ast.UsingTypename(usng,name,eq,tn,ty,sem)))
+  | Ast.UsingMember(usng,name,sem) ->
+      let name = disjident name in
+      List.map (function name -> Ast.rewrap di (Ast.UsingMember(usng,name,sem))) name
 
 let generic_orify_rule_elem f re exp rebuild =
   match f exp with
@@ -400,6 +503,10 @@ let rec disj_rule_elem r k re =
 	(fun (name,params) ->
 	  Ast.rewrap re
 	    (Ast.FunHeader(bef,allminus,fninfo,name,lp,params,va,rp,attrs)))
+  | Ast.TemplateDefinitionHeader(tmpkw,lab,params,rab) ->
+      generic_orify_rule_elem (disjdots disjtemplateparam) re params
+	(function params ->
+	  Ast.rewrap re (Ast.TemplateDefinitionHeader(tmpkw,lab,params,rab)))
   | Ast.Decl decl ->
       orify_rule_elem_anndecl re decl
 	(function decl -> Ast.rewrap re (Ast.Decl decl))
@@ -420,19 +527,43 @@ let rec disj_rule_elem r k re =
   | Ast.WhileTail(whl,lp,exp,rp,sem) ->
       orify_rule_elem re exp
 	(function exp -> Ast.rewrap re (Ast.WhileTail(whl,lp,exp,rp,sem)))
-  | Ast.ForHeader(fr,lp,first,e2,sem2,e3,rp) ->
-      let disjfirst = function
-	  Ast.ForExp(e1,sem1) ->
-	    List.map (function e1 -> Ast.ForExp(e1,sem1))
-	      (disjoption disjexp e1)
-	| Ast.ForDecl decl ->
-	    List.map (function decl -> Ast.ForDecl decl) (anndisjdecl decl) in
-      generic_orify_rule_elem
-	(disjmult_two disjfirst (disjoption disjexp)) re (first,[e2;e3])
-        (function
-            (first,[exp2;exp3]) ->
-              Ast.rewrap re (Ast.ForHeader(fr,lp,first,exp2,sem2,exp3,rp))
-          |  _ -> failwith "not possible")
+  | Ast.ForHeader(fr,lp,first,rp) ->
+      (match first with
+	Ast.ForExp(e1,sem1,e2,sem2,e3) ->
+	  generic_orify_rule_elem (disjoption disjexp) re e1
+	    (function e1 ->
+	      let first = Ast.ForExp(e1,sem1,e2,sem2,e3) in
+	      let re = Ast.rewrap re (Ast.ForHeader(fr,lp,first,rp)) in
+	      generic_orify_rule_elem (disjoption disjexp) re e2
+		(function e2 ->
+		  let first = Ast.ForExp(e1,sem1,e2,sem2,e3) in
+		  let re = Ast.rewrap re (Ast.ForHeader(fr,lp,first,rp)) in
+		  generic_orify_rule_elem (disjoption disjexp) re e3
+		    (function e3 ->
+		      let first = Ast.ForExp(e1,sem1,e2,sem2,e3) in
+		      Ast.rewrap re (Ast.ForHeader(fr,lp,first,rp)))))
+      | Ast.ForDecl(decl,e2,sem2,e3) ->
+	  generic_orify_rule_elem anndisjdecl re decl
+	    (function decl ->
+	      let first = Ast.ForDecl(decl,e2,sem2,e3) in
+	      let re = Ast.rewrap re (Ast.ForHeader(fr,lp,first,rp)) in
+	      generic_orify_rule_elem (disjoption disjexp) re e2
+		(function e2 ->
+		  let first = Ast.ForDecl(decl,e2,sem2,e3) in
+		  let re = Ast.rewrap re (Ast.ForHeader(fr,lp,first,rp)) in
+		  generic_orify_rule_elem (disjoption disjexp) re e3
+		    (function e3 ->
+		      let first = Ast.ForDecl(decl,e2,sem2,e3) in
+		      Ast.rewrap re (Ast.ForHeader(fr,lp,first,rp)))))
+      | Ast.ForRange(decl,i2) ->
+	  generic_orify_rule_elem anndisjdecl re decl
+	    (function decl ->
+	      let first = Ast.ForRange(decl,i2) in
+	      let re = Ast.rewrap re (Ast.ForHeader(fr,lp,first,rp)) in
+	      generic_orify_rule_elem disjini re i2
+		(function i2 ->
+		  let first = Ast.ForRange(decl,i2) in
+		  Ast.rewrap re (Ast.ForHeader(fr,lp,first,rp)))))
   | Ast.IteratorHeader(whl,lp,args,rp) ->
       generic_orify_rule_elem (disjdots disjexp) re args
 	(function args -> Ast.rewrap re (Ast.IteratorHeader(whl,lp,args,rp)))
@@ -458,10 +589,11 @@ let rec disj_rule_elem r k re =
   | Ast.TopInit(init) ->
       orify_rule_elem_ini re init
 	(function init -> Ast.rewrap init (Ast.TopInit(init)))
-  | Ast.Include(inc,_) | Ast.MetaInclude(inc,_) -> re
+  | Ast.CppTop di ->
+      generic_orify_rule_elem disjdirective re di
+	(fun di -> Ast.rewrap re (Ast.CppTop di))
   | Ast.Undef(def,id) -> re
   | Ast.DefineHeader(def,id,params) -> re
-  | Ast.Pragma(prg,id,body) -> re
   | Ast.Default(def,colon) -> re
   | Ast.Case(case,exp,colon) ->
       orify_rule_elem re exp
@@ -476,14 +608,8 @@ let rec disj_rule_elem r k re =
 let disj_all =
   let mcode x = x in
   let donothing r k e = k e in
-  V.rebuilder
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode mcode mcode mcode
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing disj_rule_elem donothing donothing donothing donothing
-    donothing donothing
+  V.rebuilder{V.rmcode=mcode} {V.rdonothing=donothing} ~rule:disj_rule_elem donothing
+   
 
 (* ----------------------------------------------------------------------- *)
 (* collect iso information at the rule_elem level *)
@@ -494,14 +620,8 @@ let collect_all_isos =
   let mcode r x = [] in
   let donothing r k e = Common.union_set (Ast.get_isos e) (k e) in
   let doanything r k e = k e in
-  V.combiner bind option_default
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode mcode mcode mcode
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing
-    doanything donothing doanything donothing donothing donothing donothing
-    donothing donothing donothing doanything
+  V.combiner bind option_default {V.cmcode=mcode} {V.cdonothing=donothing}
+    ~annotated_field:doanything doanything
 
 let collect_iso_info =
   let mcode x = x in
@@ -512,14 +632,7 @@ let collect_iso_info =
     | _ ->
 	let isos = collect_all_isos.V.combiner_rule_elem e in
 	Ast.set_isos e isos in
-  V.rebuilder
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode mcode mcode mcode
-    donothing donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing
-    donothing donothing donothing donothing donothing rule_elem donothing donothing
-    donothing donothing donothing donothing
+  V.rebuilder {V.rmcode=mcode} {V.rdonothing=donothing} ~rule:rule_elem donothing
 
 (* ----------------------------------------------------------------------- *)
 
