@@ -5,8 +5,8 @@
  *)
 
 (* create an index for each constructor *)
-(* current max is 193, also unused: 8-9, 15, 42, 46, 57, 65, 85-86,
- 113-115 *)
+(* current max is 202, also unused: {9,15,161}
+*)
 
 (* doesn't really work - requires that identical terms with no token
 subterms (eg dots) not appear on the same line *)
@@ -25,6 +25,7 @@ let dots d =
 let expression_dots x  = 1 :: dots x
 let initialiser_dots x = 2 :: dots x
 let parameter_dots x =   3 :: dots x
+let template_parameter_dots x =   196 :: dots x
 let statement_dots x =   4 :: dots x
 let declaration_dots x = 5 :: dots x
 let field_dots x = 8 :: dots x
@@ -61,9 +62,14 @@ let expression e =
   | Ast0.ArrayAccess(exp1,lb,exp2,rb) -> [27]
   | Ast0.RecordAccess(exp,pt,field) -> [28]
   | Ast0.RecordPtAccess(exp,ar,field) -> [29]
-  | Ast0.Cast(lp,ty,attr,rp,exp) -> [30]
+  | Ast0.QualifiedAccess(ty,coloncolon,field) -> [111]
+  | Ast0.Cast(lp,ty,rp,exp) -> [30]
   | Ast0.SizeOfExpr(szf,exp) -> [98] (* added after *)
   | Ast0.SizeOfType(szf,lp,ty,rp) -> [99] (* added after *)
+  | Ast0.Delete(dlt,exp) -> [85]
+  | Ast0.DeleteArr(dlt,lb,rb,exp) -> [86]
+  | Ast0.New(nw,pp_opt,lp_opt,ty,rp_opt,args) -> [57]
+  | Ast0.TemplateInst(name,lab,args,rab) -> [202]
   | Ast0.TypeExp(ty) -> [123] (* added after *)
   | Ast0.Constructor(lp,ty,rp,init) -> [155]
   | Ast0.MetaErr(name,_,_) -> [32]
@@ -89,7 +95,7 @@ let binaryOp op = match Ast0.unwrap op with
 
 let typeC t =
   match Ast0.unwrap t with
-    Ast0.ConstVol(cv,ty) -> [44]
+    Ast0.ConstVol(cvbefore,ty,cvafter) -> [44]
   | Ast0.BaseType(ty,strings) -> [48]
   | Ast0.Signed(sign,ty) -> [129]
   | Ast0.Pointer(ty,star) -> [49]
@@ -97,13 +103,15 @@ let typeC t =
   | Ast0.FunctionType(ty,lp,params,rp) -> [139]
   | Ast0.Array(ty,lb,size,rb) -> [50]
   | Ast0.Decimal(dec,lp,length,comma,precision_opt,rp) -> [160]
-  | Ast0.EnumName(kind,name) -> [146]
-  | Ast0.EnumDef(ty,lb,decls,rb) -> [150]
+  | Ast0.EnumName(kind,key,name) -> [146]
+  | Ast0.EnumDef(ty,base,lb,decls,rb) -> [150]
   | Ast0.StructUnionName(kind,name) -> [51]
   | Ast0.StructUnionDef(ty,lb,decls,rb) -> [117]
   | Ast0.TypeOfExpr(tf,lp,exp,rp) -> [135]
   | Ast0.TypeOfType(tf,lp,ty,rp) -> [136]
   | Ast0.TypeName(name) -> [52]
+  | Ast0.QualifiedType(ty,coloncolon,name) -> [118]
+  | Ast0.TemplateType(_,_,_,_) -> [195]
   | Ast0.AutoType _ -> [192]
   | Ast0.MetaType(name,_,_) -> [53]
   | Ast0.DisjType(_,type_list,_,_) -> [130]
@@ -114,12 +122,12 @@ let typeC t =
 let declaration d =
   match Ast0.unwrap d with
     Ast0.MetaDecl(name,_,_) -> [148]
-  | Ast0.Init(stg,ty,midattr,id,endattr,eq,exp,sem) -> [54]
-  | Ast0.UnInit(stg,ty,midattr,id,endattr,sem) -> [55]
-  | Ast0.FunProto(fi,attr,name,lp1,params,va,rp1,sem) -> [132]
-  | Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem) -> [137]
-  | Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) -> [157]
-  | Ast0.TyDecl(ty,attr,sem) -> [116]
+  | Ast0.Init(al,stg,ty,id,endattr,eq,exp,sem) -> [54]
+  | Ast0.UnInit(al,stg,ty,id,endattr,sem) -> [55]
+  | Ast0.FunProto(fi,name,lp1,params,va,rp1,sem) -> [132]
+  | Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) -> [137]
+  | Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) -> [157]
+  | Ast0.TyDecl(ty,sem) -> [116]
   | Ast0.Typedef(stg,ty,id,sem) -> [143]
   | Ast0.DisjDecl(_,decls,_,_) -> [97] (* added after *)
   | Ast0.ConjDecl(_,decls,_,_) -> [88] (* added after *)
@@ -130,7 +138,9 @@ let field d =
   match Ast0.unwrap d with
   | Ast0.MetaField(name,_,_) -> [149]
   | Ast0.MetaFieldList(name,_,_,_) -> [152]
-  | Ast0.Field(ty,id,_bf,sem) -> [55]
+  | Ast0.Field(ty,id,_bf,_endattr,sem) -> [55]
+  | Ast0.MacroDeclField(name,lp,args,rp,attr,sem) -> [59]
+  | Ast0.CppField(_) -> [202]
   | Ast0.DisjField(_,decls,_,_) -> [189] (* added after *)
   | Ast0.ConjField(_,decls,_,_) -> [190] (* added after *)
   | Ast0.Fdots(dots,whencode) -> [133]
@@ -157,8 +167,7 @@ let initialiser i =
 
 let parameterTypeDef p =
   match Ast0.unwrap p with
-    Ast0.VoidParam(ty,attr) -> [59]
-  | Ast0.Param(ty,midattr,id,attr) -> [60]
+    Ast0.Param(ty,id,attr) -> [60]
   | Ast0.MetaParam(name,_,_) -> [61]
   | Ast0.MetaParamList(name,_,_,_) -> [62]
   | Ast0.PComma(cm) -> [63]
@@ -166,9 +175,17 @@ let parameterTypeDef p =
   | Ast0.OptParam(param) -> [66]
   | Ast0.AsParam _ -> failwith "not possible"
 
+let templateParameterTypeDef p =
+  match Ast0.unwrap p with
+    Ast0.TypenameOrClassParam(tyorcl,id,eqtyopt) -> [198]
+  | Ast0.VarNameParam(ty,id,eqexpopt) -> [199]
+  | Ast0.TPComma(_) -> [200]
+  | Ast0.TPDots(_) -> [201]
+
 let statement s =
   match Ast0.unwrap s with
     Ast0.FunDecl(bef,fninfo,name,lp,params,va,rp,attrs,lbrace,body,rbrace,aft) -> [68]
+  | Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt) -> [197]
   | Ast0.Decl(bef,decl) -> [69]
   | Ast0.Seq(lbrace,body,rbrace) -> [70]
   | Ast0.ExprStatement(exp,sem) -> [71]
@@ -176,7 +193,7 @@ let statement s =
   | Ast0.IfThenElse(iff,lp,exp,rp,branch1,els,branch2,aft) -> [73]
   | Ast0.While(whl,lp,exp,rp,body,_) -> [74]
   | Ast0.Do(d,body,whl,lp,exp,rp,sem) -> [75]
-  | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,_) -> [76]
+  | Ast0.For(fr,lp,first,rp,body,_) -> [76]
   | Ast0.Iterator(nm,lp,args,rp,body,_) -> [142]
   | Ast0.Switch(switch,lp,exp,rp,lb,decls,cases,rb) -> [125]
   | Ast0.Break(br,sem) -> [100]
@@ -196,24 +213,24 @@ let statement s =
   | Ast0.Ty(ty) -> [124]
   | Ast0.TopId(ty) -> [186]
   | Ast0.TopInit(init) -> [146]
+  | Ast0.CppTop(di) -> [194]
   | Ast0.Dots(d,whencode) -> [84]
-  | Ast0.Include(inc,name) -> [118]
-  | Ast0.MetaInclude(inc,name) -> [111]
   | Ast0.Undef(def,id) -> [151]
   | Ast0.Define(def,id,params,body) -> [119]
-  | Ast0.Pragma(prg,id,body) -> [161]
   | Ast0.OptStm(re) -> [87]
   | Ast0.AsStmt _ -> failwith "not possible"
 
 let forinfo fi =
   match Ast0.unwrap fi with
-    Ast0.ForExp(exp,sem) -> [158]
-  | Ast0.ForDecl (bef,decl) -> [159]
+    Ast0.ForExp(exp,sem,exp2,sem2,exp3) -> [158]
+  | Ast0.ForDecl (bef,decl,exp2,sem2,exp3) -> [159]
+  | Ast0.ForRange (bef,decl,ini) -> [65]
 
 and pragmainfo pi =
   match Ast0.unwrap pi with
     Ast0.PragmaString(s) -> [163]
   | Ast0.PragmaDots (dots) -> [164]
+  | Ast0.MetaPragmaInfo(_,_,_) -> [8]
 
 let case_line c =
   match Ast0.unwrap c with
@@ -233,6 +250,8 @@ let attribute a =
   match Ast0.unwrap a with
     Ast0.Attribute(attr) -> [39]
   | Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) -> [162]
+  | Ast0.CxxAttribute(lb1,arg,rb1,rb2) -> [46]
+  | Ast0.CxxAttributeUsing(lb1,usng,dotdot,atnm,arg,rb1,rb2) -> [42]
 
 let attr_arg a =
   match Ast0.unwrap a with
