@@ -32,8 +32,6 @@ let collect_context e =
   let bind x y = x @ y in
   let option_default = [] in
 
-  let mcode _ = [] in
-
   let donothing builder r k e =
     match Ast0.get_mcodekind e with
       Ast0.CONTEXT(_) -> (builder e) :: (k e)
@@ -69,21 +67,22 @@ it *)
   let topfn r k e = Ast0.TopTag(e) :: (k e) in
 
   let res =
-    V0.flat_combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode
-      (donothing Ast0.dotsExpr) (donothing Ast0.dotsInit)
-      (donothing Ast0.dotsParam) (donothing Ast0.dotsStmt)
-      (donothing Ast0.dotsDecl) (donothing Ast0.dotsField)
-      (donothing Ast0.dotsCase)
-      (donothing Ast0.dotsDefParam)
-      (donothing Ast0.ident) expression  (donothing Ast0.assignOp)
-      (donothing Ast0.binaryOp)
-      (donothing Ast0.typeC) initialiser
-      (donothing Ast0.param) (donothing Ast0.decl)
-      (donothing Ast0.field) statement
-      (donothing Ast0.forinfo) (donothing Ast0.case_line)
-      (donothing Ast0.string_fragment) topfn in
+    V0.combiner_default bind option_default
+      ~dotsexpr:(donothing Ast0.dotsExpr) ~dotsinit:(donothing Ast0.dotsInit)
+      ~dotsparam:(donothing Ast0.dotsParam) ~dotstemplateparam:(donothing Ast0.dotsTemplateParam)
+      ~dotsstmt:(donothing Ast0.dotsStmt)
+      ~dotsdecl:(donothing Ast0.dotsDecl) ~dotsfield:(donothing Ast0.dotsField)
+      ~dotsenumdecl:(donothing Ast0.dotsEnumDecl) ~dotscase:(donothing Ast0.dotsCase)
+      ~dotsdefpar:(donothing Ast0.dotsDefParam)
+      ~ident:(donothing Ast0.ident) ~expr:expression ~assignOp:(donothing Ast0.assignOp)
+      ~binaryOp:(donothing Ast0.binaryOp)
+      ~ty:(donothing Ast0.typeC) ~init:initialiser
+      ~param:(donothing Ast0.param) ~template_param:(donothing Ast0.template_param)
+      ~decl:(donothing Ast0.decl) ~field:(donothing Ast0.field)
+      ~enumdecl:(donothing Ast0.enum_decl) ~stmt:statement
+      ~forinfo:(donothing Ast0.forinfo) ~case:(donothing Ast0.case_line)
+      ~string_fragment:(donothing Ast0.string_fragment) ~attribute:(donothing Ast0.attr)
+      ~attr_arg:(donothing Ast0.attr_arg) ~top:topfn () in
   res.VT0.combiner_rec_top_level e
 
 (* --------------------------------------------------------------------- *)
@@ -111,9 +110,11 @@ let create_root_token_table minus =
 	    Ast0.DotsExprTag(d) -> Ast0.get_index d
 	  | Ast0.DotsInitTag(d) -> Ast0.get_index d
 	  | Ast0.DotsParamTag(d) -> Ast0.get_index d
+	  | Ast0.DotsTemplateParamTag(d) -> Ast0.get_index d
 	  | Ast0.DotsStmtTag(d) -> Ast0.get_index d
 	  | Ast0.DotsDeclTag(d) -> Ast0.get_index d
 	  | Ast0.DotsFieldTag(d) -> Ast0.get_index d
+	  | Ast0.DotsEnumDeclTag(d) -> Ast0.get_index d
 	  | Ast0.DotsCaseTag(d) -> Ast0.get_index d
 	  | Ast0.DotsDefParamTag(d) -> Ast0.get_index d
 	  | Ast0.IdentTag(d) -> Ast0.get_index d
@@ -124,13 +125,17 @@ let create_root_token_table minus =
 	      failwith "not possible - iso only"
 	  | Ast0.TypeCTag(d) -> Ast0.get_index d
 	  | Ast0.ParamTag(d) -> Ast0.get_index d
+	  | Ast0.TemplateParamTag(d) -> Ast0.get_index d
 	  | Ast0.InitTag(d) -> Ast0.get_index d
 	  | Ast0.DeclTag(d) -> Ast0.get_index d
 	  | Ast0.FieldTag(d) -> Ast0.get_index d
+	  | Ast0.EnumDeclTag(d) -> Ast0.get_index d
 	  | Ast0.StmtTag(d) -> Ast0.get_index d
 	  | Ast0.ForInfoTag(d) -> Ast0.get_index d
 	  | Ast0.CaseLineTag(d) -> Ast0.get_index d
 	  | Ast0.StringFragmentTag(d) -> Ast0.get_index d
+	  | Ast0.AttributeTag(d) -> Ast0.get_index d
+	  | Ast0.AttrArgTag(d) -> Ast0.get_index d
 	  | Ast0.TopTag(d) -> Ast0.get_index d
 	  | Ast0.IsoWhenTag(_) -> failwith "only within iso phase"
 	  | Ast0.IsoWhenTTag(_) -> failwith "only within iso phase"
@@ -186,14 +191,42 @@ bind to that; not good for isomorphisms *)
     let l = Ast0.unwrap d in
     multibind (List.map f l) in
 
-  (* just unfavor all commas *)
   let edots r k d =
+    let l =
+      match Ast0.unwrap d with
+      | [] -> []
+      | el ->
+          (* transform an expression into a tuple (bool, expression)
+           * with the boolean indicating if the previous expression is
+           * optional (dots or expression list)*)
+          let map l =
+            (* loop processes the reversed list, leading acc to be in the right
+             * order while having a simpler pattern matching than if the list
+             * was in its orginal order *)
+            let rec loop acc = function
+              | [] -> acc
+              | e::[] -> (false, e)::acc
+              | curr::(prev::_ as l) ->
+                  let prev_is_optional =
+                    match Ast0.unwrap prev with
+                      | Ast0.Edots _
+                      | Ast0.MetaExprList _ -> true
+                      | _ -> false
+                  in
+                  let acc = (prev_is_optional, curr)::acc in
+                  loop acc l in
+            loop [] (List.rev l) in
+          map el in
     dots
-      (function e ->
-	match Ast0.unwrap e with
-	  Ast0.EComma(comma) -> unfavored_mcode comma
-	| _ -> r.VT0.combiner_rec_expression e)
-      k d in
+      (fun (prev_is_optional, e) ->
+        match (prev_is_optional, Ast0.unwrap e) with
+        | (true, Ast0.EComma(comma)) ->
+            (* Only unfavor the comma if the previous expression can be absent
+             * from the match *)
+            unfavored_mcode comma
+        | _ ->
+            r.VT0.combiner_rec_expression e)
+      k (Ast0.rewrap d l) in
   let idots r k d =
     dots
       (function i ->
@@ -208,6 +241,13 @@ bind to that; not good for isomorphisms *)
 	  Ast0.PComma(comma) -> unfavored_mcode comma
 	| _ -> r.VT0.combiner_rec_parameter p)
   k d in
+  let tpdots r k d =
+    dots
+      (function p ->
+	match Ast0.unwrap p with
+	  Ast0.TPComma(comma) -> unfavored_mcode comma
+	| _ -> r.VT0.combiner_rec_template_parameter p)
+  k d in
   let dpdots r k d =
     dots
       (function p ->
@@ -215,6 +255,13 @@ bind to that; not good for isomorphisms *)
 	  Ast0.DPComma(comma) -> unfavored_mcode comma
 	| _ -> r.VT0.combiner_rec_define_param p)
   k d in
+  let enumdots r k d =
+    dots
+      (function p ->
+	match Ast0.unwrap p with
+	  Ast0.EnumComma(comma) -> unfavored_mcode comma
+	| _ -> r.VT0.combiner_rec_enumdecl p)
+  k d in
 
   let sdots r k d = dots r.VT0.combiner_rec_statement k d in
   let ddots r k d = dots r.VT0.combiner_rec_declaration k d in
@@ -256,14 +303,18 @@ bind to that; not good for isomorphisms *)
  (*     Ast0.IfThen(_,_,_,_,_,aft)
     | Ast0.IfThenElse(_,_,_,_,_,_,_,aft)
     | Ast0.While(_,_,_,_,_,aft)
-    | Ast0.For(_,_,_,_,_,_,_,_,aft)
+    | Ast0.For(_,_,_,_,_,aft)
     | Ast0.Iterator(_,_,_,_,_,aft) ->
 	redo_branched (do_nothing r k s) aft*)
-    | Ast0.FunDecl((info,bef),fninfo,name,lp,params,va,rp,lbrace,body,rbrace,
+    | Ast0.FunDecl((info,bef),fninfo,name,lp,params,va,rp,attrs,lbrace,body,rbrace,
 		   (aftinfo,aft)) ->
 	(Toplevel,info,bef)::(k s)@[(Toplevel,aftinfo,aft)]
     | Ast0.Decl((info,bef),decl) ->
-	(Decl,info,bef)::(k s)
+	(match Ast0.unwrap decl with
+	  Ast0.MetaDecl _ ->
+	    (* drop bef, which will be ignored by cocci_vs_c *)
+	    k s
+	| _ -> (Decl,info,bef)::(k s))
     | Ast0.Nest(starter,stmt_dots,ender,whencode,multi) ->
 	mcode starter @ r.VT0.combiner_rec_statement_dots stmt_dots @
 	mcode ender
@@ -275,7 +326,8 @@ bind to that; not good for isomorphisms *)
 
   let forinfo r k s =
     match Ast0.unwrap s with
-      Ast0.ForDecl((info,bef),decl) ->
+      Ast0.ForDecl((info,bef),_,_,_,_)
+    | Ast0.ForRange((info,bef),_,_) ->
 	(Decl,info,bef)::(k s)
     | _ -> k s in
 
@@ -342,14 +394,13 @@ bind to that; not good for isomorphisms *)
 
   let do_top r k (e: Ast0.top_level) = k e in
 
-  V0.flat_combiner bind option_default
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode
-    edots idots pdots sdots ddots fdots cdots dpdots
-    ident expression do_nothing do_nothing
-    typeC initialiser param decl field statement forinfo
-    case_line do_nothing do_top
-
+  V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=do_nothing}
+    ~dotsexpr:edots ~dotsinit:idots ~dotsparam:pdots ~dotstemplateparam:tpdots
+    ~dotsstmt:sdots ~dotsdecl:ddots ~dotsfield:fdots ~dotsenumdecl:enumdots
+    ~dotscase:cdots ~dotsdefpar:dpdots
+    ~ident:ident ~expr:expression
+    ~ty:typeC ~init:initialiser ~param:param ~decl:decl ~field:field
+    ~stmt:statement ~forinfo:forinfo ~case:case_line ~top:do_top ()
 
 let call_collect_minus context_nodes :
     (int * (minus_join_point * Ast0.info * Ast0.mcodekind) list) list =
@@ -365,6 +416,9 @@ let call_collect_minus context_nodes :
       | Ast0.DotsParamTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_parameter_list e)
+      | Ast0.DotsTemplateParamTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_minus_join_points e).VT0.combiner_rec_template_parameter_list e)
       | Ast0.DotsStmtTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_statement_dots e)
@@ -374,6 +428,9 @@ let call_collect_minus context_nodes :
       | Ast0.DotsFieldTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_field_dots e)
+      | Ast0.DotsEnumDeclTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_minus_join_points e).VT0.combiner_rec_enum_decl_dots e)
       | Ast0.DotsCaseTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_case_line_dots e)
@@ -400,6 +457,9 @@ let call_collect_minus context_nodes :
       | Ast0.ParamTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_parameter e)
+      | Ast0.TemplateParamTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_minus_join_points e).VT0.combiner_rec_template_parameter e)
       | Ast0.InitTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_initialiser e)
@@ -409,6 +469,9 @@ let call_collect_minus context_nodes :
       | Ast0.FieldTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_field e)
+      | Ast0.EnumDeclTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_minus_join_points e).VT0.combiner_rec_enumdecl e)
       | Ast0.StmtTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_statement e)
@@ -418,6 +481,12 @@ let call_collect_minus context_nodes :
       | Ast0.StringFragmentTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_string_fragment e)
+      | Ast0.AttributeTag(e) ->
+          (Ast0.get_index e,
+           (collect_minus_join_points e).VT0.combiner_rec_attribute e)
+      | Ast0.AttrArgTag(e) ->
+          (Ast0.get_index e,
+           (collect_minus_join_points e).VT0.combiner_rec_attr_arg e)
       | Ast0.CaseLineTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_minus_join_points e).VT0.combiner_rec_case_line e)
@@ -497,6 +566,7 @@ let mk_arithOp x          = Ast.ArithOpTag x
 let mk_logicalOp x        = Ast.LogicalOpTag x
 let mk_declaration x      = Ast.DeclarationTag (Ast0toast.declaration x)
 let mk_field x            = Ast.FieldTag (Ast0toast.field x)
+let mk_enum_decl x        = Ast.EnumDeclTag (Ast0toast.enum_decl x)
 let mk_topdeclaration x   = Ast.DeclarationTag (Ast0toast.declaration x)
 let mk_storage x          = Ast.StorageTag x
 let mk_inc_file x         = Ast.IncFileTag x
@@ -504,6 +574,8 @@ let mk_statement x        = Ast.StatementTag (Ast0toast.statement x)
 let mk_forinfo x          = Ast.ForInfoTag (Ast0toast.forinfo x)
 let mk_case_line x        = Ast.CaseLineTag (Ast0toast.case_line x)
 let mk_string_fragment x  = Ast.StringFragmentTag (Ast0toast.string_fragment x)
+let mk_attribute x        = Ast.AttributeTag (Ast0toast.attribute x)
+let mk_attr_arg x         = Ast.AttrArgTag (Ast0toast.attr_arg x)
 let mk_const_vol x        = Ast.ConstVolTag x
 let mk_token x info       = Ast.Token (x,Some info)
 let mk_meta (_,x) info    = Ast.Token (x,Some info)
@@ -511,14 +583,17 @@ let mk_code x             = Ast.Code (Ast0toast.top_level x)
 
 let mk_exprdots x  = Ast.ExprDotsTag (Ast0toast.expression_dots x)
 let mk_paramdots x = Ast.ParamDotsTag (Ast0toast.parameter_list x)
+let mk_templateparamdots x = Ast.TemplateParamDotsTag (Ast0toast.template_parameter_list x)
 let mk_stmtdots x  = Ast.StmtDotsTag (Ast0toast.statement_dots x)
 let mk_decldots x  = Ast.AnnDeclDotsTag (Ast0toast.declaration_dots x)
 let mk_fielddots x  = Ast.AnnFieldDotsTag (Ast0toast.field_dots x)
+let mk_enumdecldots x  = Ast.EnumDeclDotsTag (Ast0toast.enum_decl_dots x)
 let mk_casedots x  = failwith "+ case lines not supported"
 let mk_defpardots x= Ast.DefParDotsTag (Ast0toast.define_param_dots x)
 let mk_typeC x     = Ast.FullTypeTag (Ast0toast.typeC false x)
 let mk_init x      = Ast.InitTag (Ast0toast.initialiser x)
 let mk_param x     = Ast.ParamTag (Ast0toast.parameterTypeDef x)
+let mk_template_param x = Ast.TemplateParamTag (Ast0toast.templateParameterTypeDef x)
 
 let collect_plus_nodes root =
   let root_index = Ast0.get_index root in
@@ -581,7 +656,7 @@ let collect_plus_nodes root =
     | Ast0.TopInit(init) -> r.VT0.combiner_rec_initialiser init
     | Ast0.Decl(bef,decl) ->
 	 do_nothing_extra (pre_info bef) [] mk_statement r k e
-    | Ast0.FunDecl(bef,fi,name,lp,params,va,rp,lbrace,body,rbrace,aft) ->
+    | Ast0.FunDecl(bef,fi,name,lp,params,va,rp,attrs,lbrace,body,rbrace,aft) ->
 	do_nothing_extra (pre_info bef) (pre_info aft) mk_statement r k e
     | Ast0.IfThen(iff,lp,exp,rp,branch1,aft) ->
 	do_nothing_extra [] (info aft) mk_statement r k e
@@ -589,8 +664,8 @@ let collect_plus_nodes root =
 	do_nothing_extra [] (info aft) mk_statement r k e
     | Ast0.While(whl,lp,exp,rp,body,aft) ->
 	do_nothing_extra [] (info aft) mk_statement r k e
-    | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,aft) ->
-	do_nothing_extra [] (info aft)mk_statement r k e
+    | Ast0.For(fr,lp,first,rp,body,aft) ->
+	do_nothing_extra [] (info aft) mk_statement r k e
     | Ast0.Iterator(nm,lp,args,rp,body,aft) ->
 	do_nothing_extra [] (info aft) mk_statement r k e
     | _ -> do_nothing mk_statement r k e in
@@ -610,24 +685,27 @@ let collect_plus_nodes root =
 
   let initdots r k e = k e in
 
-  V0.flat_combiner bind option_default
-    (imcode mk_meta) (imcode mk_token) (mcode mk_constant)
-    (mcode mk_simpleAssignOp) (mcode mk_opAssignOp)
-    (mcode mk_fixOp)
-    (mcode mk_unaryOp) (mcode mk_arithOp)
-    (mcode mk_logicalOp) (mcode mk_const_vol)
-    (mcode mk_sign) (mcode mk_structUnion)
-    (mcode mk_storage) (mcode mk_inc_file)
-    (do_nothing mk_exprdots) initdots
-    (do_nothing mk_paramdots) stmt_dots (do_nothing mk_decldots)
-    (do_nothing mk_fielddots)
-    (do_nothing mk_casedots) (do_nothing mk_defpardots)
-    (do_nothing mk_ident) (do_nothing mk_expression)
-    (do_nothing mk_assignOp) (do_nothing mk_binaryOp)
-    (do_nothing mk_typeC) (do_nothing mk_init) (do_nothing mk_param)
-    (do_nothing mk_declaration) (do_nothing mk_field)
-    stmt (do_nothing mk_forinfo) (do_nothing mk_case_line)
-    (do_nothing mk_string_fragment) toplevel
+  V0.combiner_default bind option_default
+    ~meta_mcode:(imcode mk_meta) ~string_mcode:(imcode mk_token) ~const_mcode:(mcode mk_constant)
+    ~simpleAssign_mcode:(mcode mk_simpleAssignOp) ~opAssign_mcode:(mcode mk_opAssignOp)
+    ~fixOp_mcode:(mcode mk_fixOp) ~unaryOp_mcode:(mcode mk_unaryOp) ~arithOp_mcode:(mcode mk_arithOp)
+    ~logicalOp_mcode:(mcode mk_logicalOp) ~cv_mcode:(mcode mk_const_vol)
+    ~sign_mcode:(mcode mk_sign) ~struct_mcode:(mcode mk_structUnion)
+    ~storage_mcode:(mcode mk_storage) ~inc_mcode:(mcode mk_inc_file)
+    ~dotsexpr:(do_nothing mk_exprdots) ~dotsinit:initdots
+    ~dotsparam:(do_nothing mk_paramdots) ~dotstemplateparam:(do_nothing mk_templateparamdots)
+    ~dotsstmt:stmt_dots ~dotsdecl:(do_nothing mk_decldots)
+    ~dotsfield:(do_nothing mk_fielddots) ~dotsenumdecl:(do_nothing mk_enumdecldots)
+    ~dotscase:(do_nothing mk_casedots) ~dotsdefpar:(do_nothing mk_defpardots)
+    ~ident:(do_nothing mk_ident) ~expr:(do_nothing mk_expression)
+    ~assignOp:(do_nothing mk_assignOp) ~binaryOp:(do_nothing mk_binaryOp)
+    ~ty:(do_nothing mk_typeC) ~init:(do_nothing mk_init)
+    ~param:(do_nothing mk_param) ~template_param:(do_nothing mk_template_param)
+    ~decl:(do_nothing mk_declaration) ~field:(do_nothing mk_field)
+    ~enumdecl:(do_nothing mk_enum_decl)
+    ~stmt:stmt ~forinfo:(do_nothing mk_forinfo) ~case:(do_nothing mk_case_line)
+    ~string_fragment:(do_nothing mk_string_fragment) ~attribute:(do_nothing mk_attribute)
+    ~attr_arg:(do_nothing mk_attr_arg) ~top:toplevel ()
 
 let call_collect_plus context_nodes :
     (int * (Ast0.info * Ast.count * Ast.anything) list) list =
@@ -643,6 +721,9 @@ let call_collect_plus context_nodes :
       | Ast0.DotsParamTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_parameter_list e)
+      | Ast0.DotsTemplateParamTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_plus_nodes e).VT0.combiner_rec_template_parameter_list e)
       | Ast0.DotsStmtTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_statement_dots e)
@@ -652,6 +733,9 @@ let call_collect_plus context_nodes :
       | Ast0.DotsFieldTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_field_dots e)
+      | Ast0.DotsEnumDeclTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_plus_nodes e).VT0.combiner_rec_enum_decl_dots e)
       | Ast0.DotsCaseTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_case_line_dots e)
@@ -681,12 +765,18 @@ let call_collect_plus context_nodes :
       | Ast0.ParamTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_parameter e)
+      | Ast0.TemplateParamTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_plus_nodes e).VT0.combiner_rec_template_parameter e)
       | Ast0.DeclTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_declaration e)
       | Ast0.FieldTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_field e)
+      | Ast0.EnumDeclTag(e) ->
+	  (Ast0.get_index e,
+	   (collect_plus_nodes e).VT0.combiner_rec_enumdecl e)
       | Ast0.StmtTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_statement e)
@@ -699,6 +789,12 @@ let call_collect_plus context_nodes :
       | Ast0.StringFragmentTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_string_fragment e)
+      | Ast0.AttributeTag(e) ->
+          (Ast0.get_index e,
+           (collect_plus_nodes e).VT0.combiner_rec_attribute e)
+      | Ast0.AttrArgTag(e) ->
+          (Ast0.get_index e,
+           (collect_plus_nodes e).VT0.combiner_rec_attr_arg e)
       | Ast0.TopTag(e) ->
 	  (Ast0.get_index e,
 	   (collect_plus_nodes e).VT0.combiner_rec_top_level e)
@@ -1156,7 +1252,7 @@ let reevaluate_contextness =
      match Ast0.unwrap e with
        Ast0.Decl(bef,decl) ->
 	 donothing_extra (pre_info bef) r k e
-     | Ast0.FunDecl(bef,fi,name,lp,params,va,rp,lbrace,body,rbrace,aft) ->
+     | Ast0.FunDecl(bef,fi,name,lp,params,va,rp,attrs,lbrace,body,rbrace,aft) ->
 	 donothing_extra ((pre_info bef) @ (pre_info aft)) r k e
      | Ast0.IfThen(iff,lp,exp,rp,branch1,aft) ->
 	 donothing_extra (info aft) r k e
@@ -1164,21 +1260,15 @@ let reevaluate_contextness =
 	 donothing_extra (info aft) r k e
      | Ast0.While(whl,lp,exp,rp,body,aft) ->
 	 donothing_extra (info aft) r k e
-     | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,aft) ->
+     | Ast0.For(fr,lp,first,rp,body,aft) ->
 	 donothing_extra (info aft) r k e
      | Ast0.Iterator(nm,lp,args,rp,body,aft) ->
 	 donothing_extra (info aft) r k e
      | _ -> donothing r k e in
 
   let res =
-    V0.flat_combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing stmt donothing
-      donothing donothing
-      donothing in
+    V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=donothing}
+      ~stmt:stmt () in
   res.VT0.combiner_rec_top_level
 
 (* --------------------------------------------------------------------- *)
@@ -1189,4 +1279,4 @@ let insert_plus minus plus ei =
   let minus_stream = process_minus minus in
   let plus_stream = process_plus plus in
   merge minus_stream plus_stream;
-  List.iter (function x -> let _ =  reevaluate_contextness x in ()) minus
+  List.iter (function x -> ignore(reevaluate_contextness x)) minus
