@@ -67,37 +67,40 @@ let clt2mcode_ext str isSymbol = function
     (Data.MINUS,line,lline,llineend,offset,col,strbef,straft,pos,ws) ->
       (str,Ast0.NONE,
        make_info line lline llineend offset col strbef straft isSymbol ws,
-       Ast0.MINUS(ref(Ast.NOREPLACEMENT,Ast0.default_token_info)),ref pos,-1)
+       Ast0.MINUS(ref(Ast.NOREPLACEMENT,Ast0.default_token_info)),ref pos,
+       {Ast.counter = -1; Ast.ender = false})
   | (Data.OPTMINUS,line,lline,llineend,offset,col,strbef,straft,pos,ws)->
       (str,Ast0.OPT,
        make_info line lline llineend offset col strbef straft isSymbol ws,
-       Ast0.MINUS(ref(Ast.NOREPLACEMENT,Ast0.default_token_info)),ref pos,-1)
+       Ast0.MINUS(ref(Ast.NOREPLACEMENT,Ast0.default_token_info)),ref pos,
+       {Ast.counter = -1; Ast.ender = false})
   | (Data.PLUS,line,lline,llineend,offset,col,strbef,straft,pos,ws)        ->
       (str,Ast0.NONE,
        make_info line lline llineend offset col strbef straft isSymbol ws,
-       Ast0.PLUS(Ast.ONE),ref pos,-1)
+       Ast0.PLUS(Ast.ONE),ref pos,{Ast.counter = -1; Ast.ender = false})
   | (Data.PLUSPLUS,line,lline,llineend,offset,col,strbef,straft,pos,ws)    ->
       (str,Ast0.NONE,
        make_info line lline llineend offset col strbef straft isSymbol ws,
-       Ast0.PLUS(Ast.MANY),ref pos,-1)
+       Ast0.PLUS(Ast.MANY),ref pos,{Ast.counter = -1; Ast.ender = false})
   | (Data.CONTEXT,line,lline,llineend,offset,col,strbef,straft,pos,ws)     ->
       (str,Ast0.NONE,
        make_info line lline llineend offset col strbef straft isSymbol ws,
        Ast0.CONTEXT(ref(Ast.NOTHING,
 			Ast0.default_token_info,Ast0.default_token_info)),
-       ref pos,-1)
+       ref pos,{Ast.counter = -1; Ast.ender = false})
   | (Data.OPT,line,lline,llineend,offset,col,strbef,straft,pos,ws)         ->
       (str,Ast0.OPT,
        make_info line lline llineend offset col strbef straft isSymbol ws,
        Ast0.CONTEXT(ref(Ast.NOTHING,
 			Ast0.default_token_info,Ast0.default_token_info)),
-       ref pos,-1)
+       ref pos,{Ast.counter = -1; Ast.ender = false})
 
 let clt2mcode name clt = clt2mcode_ext name false clt
 let id3name   (name, _, clt) = name
 let id2name   (name, clt) = name
 let id2clt    (name, clt) = clt
 let id2mcode  (name, clt) = clt2mcode name clt
+let tok2mcode (name, clt) = clt2mcode name clt
 let sym2mcode (name, clt) = clt2mcode_ext name true clt
 
 let mkdots str (dot,whencode) =
@@ -149,11 +152,6 @@ let logic_op ast_op left op right =
   let op' = Ast0.wrap (Ast0.Logical (clt2mcode ast_op op)) in  
   Ast0.wrap (Ast0.Binary(left, op', right))
 
-let make_cv cv ty =
-  match cv with
-    None -> ty
-  | Some x -> Ast0.wrap (Ast0.ConstVol(x,ty))
-
 let make_attr attr =
   Ast0.wrap(Ast0.Attribute(attr))
 
@@ -162,6 +160,17 @@ let make_gcc_attr attr_ lp1 lp2 arg rp1 rp2 =
     clt2mcode "(" lp1,clt2mcode "(" lp2, arg,
     clt2mcode ")" rp1,clt2mcode ")" rp2))
 
+let make_cxx_attr lb1 arg rb1 rb2 =
+  Ast0.wrap
+    (Ast0.CxxAttribute
+       (tok2mcode lb1, arg, tok2mcode rb1, tok2mcode rb2))
+
+let make_cxx_attr_using lb1 usng atnm dotdot arg rb1 rb2 =
+  Ast0.wrap
+    (Ast0.CxxAttributeUsing
+       (tok2mcode lb1,clt2mcode "using" usng, atnm,
+	clt2mcode ":" dotdot, arg, tok2mcode rb1, tok2mcode rb2))
+
 let top_dots l = Ast0.wrap l
 
 (* here the offset is that of the first in the sequence of *s, not that of
@@ -170,19 +179,30 @@ let pointerify ty m =
   List.fold_left
     (function inner ->
       function cur ->
-	Ast0.wrap(Ast0.Pointer(inner,clt2mcode "*" cur)))
+	Ast0.wrap(Ast0.Pointer(inner, clt2mcode (fst cur) (snd cur) )))
     ty m
 
 let ty_pointerify ty m =
   List.fold_left
-    (fun inner cur -> Ast0.wrap (Ast0.Pointer (inner, Ast0.make_mcode "")))
+    (fun inner cur -> Ast0.wrap (Ast0.Pointer (inner, Ast0.make_mcode "*")))
     ty m
 
+let make_cv cvbefore ty cvafter =
+  match cvbefore,cvafter with
+    [],[] -> ty
+  | _ -> Ast0.wrap (Ast0.ConstVol(cvbefore,ty,cvafter))
+
+let make_ctype_and_ptr (cv1, ty, cv2, m) =
+  List.fold_left
+    (fun prev (star,cvattrs) ->
+      make_cv [] (pointerify prev [star]) cvattrs)
+    (make_cv cv1 ty cv2) m
+
 let arrayify ty ar =
   List.fold_right
     (function (l,i,r) ->
       function rest ->
-	Ast0.wrap (Ast0.Array(rest,clt2mcode "[" l,i,clt2mcode "]" r)))
+	Ast0.wrap (Ast0.Array(rest,tok2mcode l,i,tok2mcode r)))
     ar ty
 
 (* Left is <=>, Right is =>.  Collect <=>s. *)
@@ -515,7 +535,9 @@ let meta_dparam_list name =
 let exp_stm exp pv =
   Ast0.wrap(Ast0.ExprStatement (exp, clt2mcode ";" pv))
 
-let make_fake_mcode _ = (Ast0.default_info(),Ast0.context_befaft(),-1)
+let make_fake_mcode _ =
+  (Ast0.default_info(),Ast0.context_befaft(),
+   {Ast.counter = -1; Ast.ender = false})
 
 let ifthen iff lp tst rp thn =
   Ast0.wrap(Ast0.IfThen(clt2mcode "if" iff,
@@ -527,17 +549,22 @@ let ifthenelse iff lp tst rp thn e els =
     clt2mcode "else" e,els,make_fake_mcode()))
 
 let forloop fr lp e1 sc1 e2 sc2 e3 rp s =
-  Ast0.wrap(Ast0.For(clt2mcode "for" fr,clt2mcode "(" lp,
-		     Ast0.wrap(Ast0.ForExp(e1,clt2mcode ";" sc1)),e2,
-		     clt2mcode ";" sc2,e3,clt2mcode ")" rp,s,
-		     make_fake_mcode()))
+  let first =
+    Ast0.wrap(Ast0.ForExp(e1,clt2mcode ";" sc1,e2,clt2mcode ";" sc2,e3)) in
+  Ast0.wrap(Ast0.For(clt2mcode "for" fr,clt2mcode "(" lp,first,
+		     clt2mcode ")" rp,s,make_fake_mcode()))
 
 let forloop2 fr lp decl e2 sc2 e3 rp s =
   let bef = (Ast0.default_info(),Ast0.context_befaft()) in
-  Ast0.wrap(Ast0.For(clt2mcode "for" fr,clt2mcode "(" lp,
-		     Ast0.wrap(Ast0.ForDecl (bef,decl)),e2,
-		     clt2mcode ";" sc2,e3,clt2mcode ")" rp,s,
-		     make_fake_mcode()))
+  let first = Ast0.wrap(Ast0.ForDecl (bef,decl,e2,clt2mcode ";" sc2,e3)) in
+  Ast0.wrap(Ast0.For(clt2mcode "for" fr,clt2mcode "(" lp,first,
+		     clt2mcode ")" rp,s,make_fake_mcode()))
+
+let forloop3 fr lp decl i2 rp s =
+  let bef = (Ast0.default_info(),Ast0.context_befaft()) in
+  let first = Ast0.wrap(Ast0.ForRange (bef,decl,i2)) in
+  Ast0.wrap(Ast0.For(clt2mcode "for" fr,clt2mcode "(" lp,first,
+		     clt2mcode ")" rp,s,make_fake_mcode()))
 
 let whileloop w lp e rp s =
   Ast0.wrap(Ast0.While(clt2mcode "while" w,clt2mcode "(" lp,
@@ -559,8 +586,8 @@ let switch s lp e rp lb d c rb =
 	Ast0.wrap(Ast0.Decl((Ast0.default_info(),Ast0.context_befaft()),d)))
       d in
   Ast0.wrap(Ast0.Switch(clt2mcode "switch" s,clt2mcode "(" lp,e,
-			clt2mcode ")" rp,clt2mcode "{" lb,
-			Ast0.wrap d,Ast0.wrap c,clt2mcode "}" rb))
+			clt2mcode ")" rp,tok2mcode lb,
+			Ast0.wrap d,Ast0.wrap c,tok2mcode rb))
 
 let ret_exp r e pv =
   Ast0.wrap(Ast0.ReturnExpr(clt2mcode "return" r,e,clt2mcode ";" pv))
@@ -581,7 +608,7 @@ let goto g i pv =
   Ast0.wrap(Ast0.Goto(clt2mcode "goto" g,i,clt2mcode ";" pv))
 
 let seq lb s rb =
-  Ast0.wrap(Ast0.Seq(clt2mcode "{" lb,s,clt2mcode "}" rb))
+  Ast0.wrap(Ast0.Seq(tok2mcode lb,s,tok2mcode rb))
 
 (* ---------------------------------------------------------------------- *)
 
