@@ -51,7 +51,8 @@ and print_anything_list = function
 	| Ast.Token("if",_) | Ast.Token("while",_) -> true | _ -> false) ||
 	(match aft with
 	  Ast.Rule_elemTag(_) | Ast.AssignOpTag(_) | Ast.BinaryOpTag(_)
-	| Ast.ArithOpTag(_) | Ast.LogicalOpTag(_) | Ast.Token("{",_) -> true
+	| Ast.ArithOpTag(_) | Ast.LogicalOpTag(_) | Ast.Token("{",_)
+	| Ast.Token("<:",_) -> true
 	| _ -> false) in
       if space then print_string " ";
       print_anything_list rest
@@ -242,16 +243,18 @@ let rec expression e =
   | Ast.Paren(lp,exp,rp) ->
       mcode print_string_box lp; expression exp; close_box();
       mcode print_string rp
-  | Ast.ArrayAccess(exp1,lb,exp2,rb) ->
-      expression exp1; mcode print_string_box lb; expression exp2; close_box();
-      mcode print_string rb
+  | Ast.ArrayAccess(fn,lb,args,rb) ->
+      expression fn; mcode print_string_box lb;
+      dots (function _ -> ()) expression args;
+      close_box(); mcode print_string rb
   | Ast.RecordAccess(exp,pt,field) ->
       expression exp; mcode print_string pt; ident field
   | Ast.RecordPtAccess(exp,ar,field) ->
       expression exp; mcode print_string ar; ident field
-  | Ast.Cast(lp,ty,attr,rp,exp) ->
+  | Ast.QualifiedAccess(ty,coloncolon,field) ->
+      print_option fullType ty; mcode print_string coloncolon; ident field
+  | Ast.Cast(lp,ty,rp,exp) ->
       mcode print_string_box lp; fullType ty; close_box();
-      print_attribute_list attr;
       mcode print_string rp; expression exp
   | Ast.SizeOfExpr(sizeof,exp) ->
       mcode print_string sizeof; expression exp
@@ -259,6 +262,25 @@ let rec expression e =
       mcode print_string sizeof;
       mcode print_string_box lp; fullType ty; close_box();
       mcode print_string rp
+  | Ast.Delete(dlt,exp) ->
+      mcode print_string dlt; print_string " "; expression exp
+  | Ast.DeleteArr(dlt,lb,rb,exp) ->
+      mcode print_string dlt; print_string " "; mcode print_string lb; mcode print_string rb; print_string " ";
+      expression exp
+  | Ast.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) ->
+      mcode print_string nw;
+      print_string " ";
+      print_option print_args pp_opt;
+      print_option (function e -> mcode print_string e) lp_opt;
+      fullType ty;
+      print_option (function e -> mcode print_string e) rp_opt;
+      print_string " ";
+      print_option print_args args_opt;
+  | Ast.TemplateInst(name,lp,args,rp) ->
+      expression name; print_string " ";
+      mcode print_string_box lp;
+      dots (function _ -> ()) expression args;
+      close_box(); mcode print_string rp
   | Ast.TypeExp(ty) -> fullType ty
   | Ast.Constructor(lp,ty,rp,init) ->
       mcode print_string_box lp; fullType ty; close_box();
@@ -285,6 +307,12 @@ let rec expression e =
   | Ast.Edots(dots,None) -> mcode print_string dots
   | Ast.OptExp(exp) -> print_string "?"; expression exp
 
+and print_args (lp,args,rp) =
+  mcode print_string lp;
+  dots (function _ -> ()) expression args;
+  mcode print_string rp;
+  print_string " ";
+
 and string_fragment e =
   match Ast.unwrap e with
     Ast.ConstantFragment(str) -> mcode print_string str
@@ -307,19 +335,18 @@ and unaryOp = function
   | Ast.DeRef -> print_string "*"
   | Ast.UnPlus -> print_string "+"
   | Ast.UnMinus -> print_string "-"
-  | Ast.Tilde -> print_string "~"
-  | Ast.Not -> print_string "!"
+  | Ast.Tilde s -> print_string s
+  | Ast.Not s -> print_string s
 
 and assignOp op =
   match Ast.unwrap op with
     Ast.SimpleAssign op -> mcode print_string op
-  | Ast.OpAssign(aop) ->
-      mcode (function op -> arithOp op; print_string "=") aop
+  | Ast.OpAssign(aop) -> mcode opAssignOp aop
   | Ast.MetaAssign(metavar,_,_,_) -> mcode print_meta metavar
 
 and simpleAssignOp op = print_string "="
 
-and opAssignOp aop = arithOp aop; print_string "="
+and opAssignOp op = simple_arithOp (fun s -> print_string s; print_string "=") op
 
 and fixOp = function
     Ast.Dec -> print_string "--"
@@ -331,19 +358,21 @@ and binaryOp op =
   | Ast.Logical(lop) -> mcode logicalOp lop
   | Ast.MetaBinary(metavar,_,_,_) -> mcode print_meta metavar
 
-and arithOp = function
-    Ast.Plus -> print_string "+"
-  | Ast.Minus -> print_string "-"
-  | Ast.Mul -> print_string "*"
-  | Ast.Div -> print_string "/"
-  | Ast.Min -> print_string "<?"
-  | Ast.Max -> print_string ">?"
-  | Ast.Mod -> print_string "%"
-  | Ast.DecLeft -> print_string "<<"
-  | Ast.DecRight -> print_string ">>"
-  | Ast.And -> print_string "&"
-  | Ast.Or -> print_string "|"
-  | Ast.Xor -> print_string "^"
+and simple_arithOp aprint_string = function
+    Ast.Plus -> aprint_string "+"
+  | Ast.Minus -> aprint_string "-"
+  | Ast.Mul -> aprint_string "*"
+  | Ast.Div -> aprint_string "/"
+  | Ast.Min -> aprint_string "<?"
+  | Ast.Max -> aprint_string ">?"
+  | Ast.Mod -> aprint_string "%"
+  | Ast.DecLeft -> aprint_string "<<"
+  | Ast.DecRight -> aprint_string ">>"
+  | Ast.And s -> print_string s
+  | Ast.Or s -> print_string s
+  | Ast.Xor s -> print_string s
+
+and arithOp op = simple_arithOp print_string op
 
 and  logicalOp = function
     Ast.Inf -> print_string "<"
@@ -351,9 +380,9 @@ and  logicalOp = function
   | Ast.InfEq -> print_string "<="
   | Ast.SupEq -> print_string ">="
   | Ast.Eq -> print_string "=="
-  | Ast.NotEq -> print_string "!="
-  | Ast.AndLog -> print_string "&&"
-  | Ast.OrLog -> print_string "||"
+  | Ast.NotEq s -> print_string s
+  | Ast.AndLog s -> print_string s
+  | Ast.OrLog s -> print_string s
 
 and constant = function
     Ast.String(s,sz) ->
@@ -385,14 +414,15 @@ and storage = function
 
 and fullType ft =
   match Ast.unwrap ft with
-    Ast.Type(_,cv,ty) ->
-      (match Ast.unwrap ty with
-	Ast.Pointer(_,_) ->
-	  typeC ty;
-	  print_option (function x -> print_string " "; mcode const_vol x) cv
-      |	_ ->
-	  print_option (function x -> mcode const_vol x; print_string " ") cv;
-	  typeC ty)
+    Ast.Type(_,cvbefore,ty,cvafter) ->
+      let do_cvattr = function
+	  Ast.CV cv -> mcode const_vol cv
+	| Ast.Attr attr -> print_attribute attr in
+      print_between print_space do_cvattr cvbefore;
+      (if cvbefore <> [] then print_string " ");
+      typeC ty;
+      (if cvafter <> [] then print_string " ");
+      print_between print_space do_cvattr cvafter
   | Ast.AsType(ty,asty) -> fullType ty; print_string "@"; fullType asty
   | Ast.DisjType(decls) -> print_disj_list fullType decls "|"
   | Ast.ConjType(decls) -> print_disj_list fullType decls "&"
@@ -401,11 +431,11 @@ and fullType ft =
 and print_parentype (lp,ty,rp) fn =
   let function_pointer ty1 array_dec =
     match Ast.unwrap ty1 with
-     Ast.Type(_,_,fty1) ->
+     Ast.Type(_,_,fty1,_) ->
       (match Ast.unwrap fty1 with
         Ast.Pointer(ty2,star) ->
          (match Ast.unwrap ty2 with
-           Ast.Type(_,_,fty3) ->
+           Ast.Type(_,_,fty3,_) ->
             (match Ast.unwrap fty3 with
               Ast.FunctionType(ty3,lp3,params,rp3) ->
                fullType ty3;
@@ -416,9 +446,9 @@ and print_parentype (lp,ty,rp) fn =
                let _ =
                  match array_dec with
                    Some(lb1,size,rb1) ->
-                      mcode print_string lb1;
-                      print_option expression size;
-                      mcode print_string rb1
+                     mcode print_string lb1;
+                     print_option expression size;
+                     mcode print_string rb1
                  | None -> () in
                mcode print_string rp;
                mcode print_string lp3;
@@ -429,7 +459,7 @@ and print_parentype (lp,ty,rp) fn =
        | _ -> failwith "ParenType Pretty_print_cocci")
      | _ -> failwith "ParenType Pretty_print_cocci" in
   match Ast.unwrap ty with
-    Ast.Type(_,_,fty1) ->
+    Ast.Type(_,_,fty1,_) ->
       (match Ast.unwrap fty1 with
         Ast.Array(ty1,lb1,size,rb1) ->
           function_pointer ty1 (Some(lb1,size,rb1))
@@ -447,25 +477,48 @@ and varargs = function
 
 and print_fninfo = function
     Ast.FStorage(stg) -> mcode storage stg
-  | Ast.FType(ty) -> fullType ty
+  | Ast.FType(ty) -> fullType ty; print_string " "
   | Ast.FInline(inline) -> mcode print_string inline; print_string " "
-  | Ast.FAttr(attr) -> print_attribute attr; print_string " "
 
-and print_attribute_list attrs =
-  if not (attrs = []) then print_string " ";
-  print_between print_space print_attribute attrs
+and print_attribute_list ?(befspace=true) ?(aftspace=false) attrs =
+  if befspace && not (attrs = []) then print_string " ";
+  print_between print_space print_attribute attrs;
+  if aftspace && not (attrs = []) then print_string " "
 
 and print_attribute attr =
   match Ast.unwrap attr with
-    Ast.Attribute(a) -> mcode print_string a
-  | Ast.MetaAttribute(name,_,_,_) -> mcode print_meta name
+    Ast.Attribute(a) -> print_attr_arg a
+  | Ast.GccAttribute(attr_,lp1,lp2,args,rp1,rp2) ->
+      mcode print_string attr_; print_string " ";
+      mcode print_string_box lp1; mcode print_string_box lp2;
+      dots (function _ -> ()) expression args; close_box();
+      mcode print_string_box rp1; close_box();
+      mcode print_string_box rp2
+  | Ast.CxxAttribute(lb1,args,rb1,rb2) ->
+      mcode print_string_box lb1;
+      dots (function _ -> ()) expression args; close_box();
+      mcode print_string_box rb1; mcode print_string_box rb2
+  | Ast.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2) ->
+      mcode print_string_box lb1; mcode print_string usng;
+      ident atnm; mcode print_string dotdot;
+      dots (function _ -> ()) expression args; close_box();
+      mcode print_string_box rb1; mcode print_string_box rb2
+
+and print_attr_arg arg =
+  match Ast.unwrap arg with
+    Ast.MacroAttr(arg) -> mcode print_string arg
+  | Ast.MetaAttr(name,_,_,_) -> mcode print_meta name
+  | Ast.MacroAttrArgs(attr,lp,args,rp) ->
+      mcode print_string attr; mcode print_string_box lp;
+      dots (function _ -> ()) expression args;
+      close_box(); mcode print_string rp
 
 and typeC ty =
   match Ast.unwrap ty with
     Ast.BaseType(ty,strings) ->
-      List.iter (function s -> mcode print_string s; print_string " ") strings
+      print_between print_space (mcode print_string) strings
   | Ast.SignedT(sgn,ty) -> mcode sign sgn; print_option typeC ty
-  | Ast.Pointer(ty,star) -> fullType ty; mcode print_string star
+  | Ast.Pointer(ty,star) -> fullType ty; print_space(); mcode print_string star
   | Ast.ParenType(lp,ty,rp) ->
       print_parentype (lp,ty,rp) (function _ -> ())
   | Ast.FunctionType(ty,lp,params,rp) ->
@@ -483,11 +536,15 @@ and typeC ty =
       print_option (mcode print_string) comma;
       print_option expression precision_opt;
       mcode print_string rp
-  | Ast.EnumName(kind,name) ->
+  | Ast.EnumName(kind,key,name) ->
       mcode print_string kind;
+      print_string " ";
+      print_option (mcode structUnion) key;
       print_option (function x -> ident x; print_string " ") name
-  | Ast.EnumDef(ty,lb,ids,rb) ->
-      fullType ty; mcode print_string lb;
+  | Ast.EnumDef(ty,base,lb,ids,rb) ->
+      fullType ty;
+      print_option enum_base base;
+      mcode print_string lb;
       dots force_newline enum_decl ids;
       mcode print_string rb
   | Ast.StructUnionName(kind,name) ->
@@ -505,10 +562,19 @@ and typeC ty =
       mcode print_string typeof;
       mcode print_string_box lp; fullType ty; close_box();
       mcode print_string rp
+  | Ast.QualifiedType(ty,coloncolon,name) ->
+      print_option fullType ty;
+      mcode print_string coloncolon;
+      ident name
   | Ast.TypeName(name) -> mcode print_string name; print_string " "
   | Ast.AutoType(auto) -> mcode print_string auto; print_string " "
   | Ast.MetaType(name,_,_,_) ->
       mcode print_meta name; print_string " "
+  | Ast.TemplateType(name,lp,args,rp) ->
+      fullType name; print_string " ";
+      mcode print_string_box lp;
+      dots (function _ -> ()) expression args;
+      close_box(); mcode print_string rp
 
 and baseType ty = print_string (Ast.string_of_baseType ty ^ " ")
 
@@ -525,30 +591,34 @@ and const_vol const_vol = print_string (Ast.string_of_const_vol const_vol ^ " ")
 
 and print_named_type ty id =
   match Ast.unwrap ty with
-    Ast.Type(_,None,ty1) ->
+    Ast.Type(_,[],ty1,[]) ->
       (match Ast.unwrap ty1 with
         Ast.Array(ty,lb,size,rb) ->
 	  let rec loop ty k =
 	    match Ast.unwrap ty with
 	      Ast.Array(ty,lb,size,rb) ->
 		(match Ast.unwrap ty with
-		  Ast.Type(_,cv,ty) ->
-		    print_option
-		      (function x -> mcode const_vol x; print_string " ")
-		      cv;
+		  Ast.Type(_,cvbefore,ty,cvafter) ->
+		    let do_cvattr = function
+			Ast.CV cv -> mcode const_vol cv
+		      | Ast.Attr attr -> print_attribute attr in
+		    print_between print_space do_cvattr cvbefore;
+		    (if cvbefore <> [] then print_space());
 		    loop ty
 		      (function _ ->
 			k ();
 			mcode print_string lb;
 			print_option expression size;
-			mcode print_string rb)
+			mcode print_string rb);
+		    (if cvafter <> [] then print_space());
+		    print_between print_space do_cvattr cvafter
 		| _ -> failwith "complex array types not supported")
-	    | _ -> typeC ty; id(); k () in
+	    | _ -> typeC ty; print_space(); id(); k () in
 	  loop ty1 (function _ -> ())
       | Ast.ParenType(lp,ty,rp) ->
           print_parentype (lp,ty,rp) (function _ -> id())
-      | _ -> fullType ty; id())
-  | _ -> fullType ty; id()
+      | _ -> fullType ty; print_space(); id())
+  | _ -> fullType ty; print_space(); id()
 
 and declaration d =
   match Ast.unwrap d with
@@ -556,16 +626,18 @@ and declaration d =
       mcode print_meta name
   | Ast.AsDecl(decl,asdecl) -> declaration decl; print_string "@";
       declaration asdecl
-  | Ast.Init(stg,ty,id,attr,eq,ini,sem) ->
+  | Ast.Init(al,stg,ty,id,endattr,eq,ini,sem) ->
+      print_option alignas al;
       print_option (mcode storage) stg;
       print_named_type ty (fun _ -> ident id);
-      print_attribute_list attr;
+      print_attribute_list endattr;
       print_string " "; mcode print_string eq;
       print_string " "; initialiser ini; mcode print_string sem
-  | Ast.UnInit(stg,ty,id,attr,sem) ->
+  | Ast.UnInit(al,stg,ty,id,endattr,sem) ->
+      print_option alignas al;
       print_option (mcode storage) stg;
       print_named_type ty (fun _ -> ident id);
-      print_attribute_list attr;
+      print_attribute_list endattr;
       mcode print_string sem
   | Ast.FunProto (fninfo,name,lp1,params,va,rp1,sem) ->
       List.iter print_fninfo fninfo;
@@ -573,21 +645,25 @@ and declaration d =
       parameter_list params; varargs va;
       close_box(); mcode print_string rp1;
       mcode print_string sem
-  | Ast.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
-      print_option (mcode storage) stg; ident name; mcode print_string_box lp;
+  | Ast.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
+      print_option (mcode storage) stg;
+      print_attribute_list preattr;
+      ident name; mcode print_string_box lp;
       dots (function _ -> ()) expression args;
       close_box(); mcode print_string rp;
       print_attribute_list attr;
       mcode print_string sem
-  | Ast.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
-      print_option (mcode storage) stg; ident name; mcode print_string_box lp;
+  | Ast.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
+      print_option (mcode storage) stg;
+      print_attribute_list preattr;
+      ident name; mcode print_string_box lp;
       dots (function _ -> ()) expression args;
       close_box(); mcode print_string rp;
+      print_attribute_list attr;
       print_string " "; mcode print_string eq;
       print_string " "; initialiser ini; mcode print_string sem
-  | Ast.TyDecl(ty,attr,sem) ->
+  | Ast.TyDecl(ty,sem) ->
       fullType ty;
-      print_attribute_list attr;
       mcode print_string sem
   | Ast.Typedef(stg,ty,id,sem) ->
       mcode print_string stg; print_string " ";
@@ -597,6 +673,12 @@ and declaration d =
   | Ast.ConjDecl(decls) -> print_disj_list declaration decls "&"
   | Ast.OptDecl(decl) -> print_string "?"; declaration decl
 
+and alignas (Ast.Align(align,lpar,expr,rpar)) =
+  mcode print_string align;
+  mcode print_string lpar;
+  expression expr;
+  mcode print_string rpar
+
 and annotated_decl arity d =
   match Ast.unwrap d with
     Ast.DElem(bef,allminus,decl) ->
@@ -612,17 +694,25 @@ and field d =
     Ast.MetaField(name,_,_,_)
   | Ast.MetaFieldList(name,_,_,_,_) ->
       mcode print_meta name
-  | Ast.Field(ty,id,bf,sem) ->
+  | Ast.Field(ty,id,bf,endattr,sem) ->
       begin
 	match id with
 	  None -> fullType ty
-	| Some id -> print_named_type ty (fun _ -> ident id);
+        | Some id -> print_named_type ty (fun _ -> ident id);
       end;
       let bitfield (c, e) =
 	mcode print_string c;
 	expression e in
       print_option bitfield bf;
+      print_attribute_list endattr;
       mcode print_string sem
+  | Ast.MacroDeclField(name,lp,args,rp,attr,sem) ->
+      ident name; mcode print_string_box lp;
+      dots (function _ -> ()) expression args;
+      close_box(); mcode print_string rp;
+      print_attribute_list attr;
+      mcode print_string sem
+  | Ast.CppField(di) -> directive di
 
 and annotated_field arity d =
   match Ast.unwrap d with
@@ -651,6 +741,10 @@ and enum_decl d =
       mcode print_string dots; print_string "   when != "; enum_decl whencode
   | Ast.EnumDots(dots,None) -> mcode print_string dots
 
+and enum_base (td, ty) =
+  mcode print_string td;
+  print_string " ";
+  fullType ty
 (* --------------------------------------------------------------------- *)
 (* Initialiser *)
 
@@ -701,10 +795,7 @@ and designator = function
 
 and parameterTypeDef p =
   match Ast.unwrap p with
-    Ast.VoidParam(ty,attr) ->
-      fullType ty;
-      print_attribute_list attr
-  | Ast.Param(ty,Some id,attr) ->
+    Ast.Param(ty,Some id,attr) ->
       print_named_type ty (fun _ -> ident id);
       print_attribute_list attr
   | Ast.Param(ty,None,attr) ->
@@ -718,14 +809,28 @@ and parameterTypeDef p =
   | Ast.AsParam(p,asexp) ->
       parameterTypeDef p; print_string "@"; expression asexp
 
+and templateParameterTypeDef p =
+  match Ast.unwrap p with
+    Ast.TypenameOrClassParam(tyorcl,id,eqtyopt) ->
+      mcode print_string tyorcl;
+      ident id;
+      print_option (fun (eq,ty) -> mcode print_string eq; fullType ty) eqtyopt
+  | Ast.VarNameParam(ty,id,eqiniopt) ->
+      print_named_type ty (fun _ -> ident id);
+      print_option (fun (eq,ini) -> mcode print_string eq; initialiser ini) eqiniopt
+  | Ast.TPComma(comma) -> mcode print_string comma; force_newline()
+  | Ast.TPDots(dots) -> mcode print_string dots; force_newline()
+
 and parameter_list l = dots (function _ -> ()) parameterTypeDef l
 
+and template_parameter_list l = dots (function _ -> ()) templateParameterTypeDef l
+
 (* --------------------------------------------------------------------- *)
 (* Top-level code *)
 
 and rule_elem arity re =
   match Ast.unwrap re with
-    Ast.FunHeader(bef,allminus,fninfo,name,lp,params,va,rp) ->
+    Ast.FunHeader(bef,allminus,fninfo,name,lp,params,va,rp,attrs) ->
       mcode (function _ -> ()) ((),Ast.no_info,bef,[]);
       print_string arity; List.iter print_fninfo fninfo;
       ident name; mcode print_string_box lp;
@@ -736,7 +841,15 @@ and rule_elem arity re =
                mcode print_string comma;
               mcode print_string ellipsis
       end; close_box(); mcode print_string rp;
+      print_attribute_list attrs;
       print_string " "
+  | Ast.TemplateDefinitionHeader(tmpkw,lab,params,rab) ->
+      mcode print_string tmpkw;
+      print_string " ";
+      mcode print_string lab;
+      print_string " ";
+      template_parameter_list params;
+      mcode print_string rab
   | Ast.Decl(ann_decl) -> annotated_decl arity ann_decl
   | Ast.SeqStart(brace) ->
       print_string arity; mcode print_string brace;
@@ -763,11 +876,10 @@ and rule_elem arity re =
       mcode print_string whl; print_string " "; mcode print_string_box lp;
       expression exp; close_box(); mcode print_string rp;
       mcode print_string sem
-  | Ast.ForHeader(fr,lp,first,e2,sem2,e3,rp) ->
+  | Ast.ForHeader(fr,lp,first,rp) ->
       print_string arity;
       mcode print_string fr; mcode print_string_box lp; forinfo first;
-      print_option expression e2; mcode print_string sem2;
-      print_option expression e3; close_box();
+      close_box();
       mcode print_string rp; print_string " "
   | Ast.IteratorHeader(nm,lp,args,rp) ->
       print_string arity;
@@ -806,18 +918,12 @@ and rule_elem arity re =
   | Ast.Ty(ty) -> print_string arity; fullType ty
   | Ast.TopId(id) -> print_string arity; ident id
   | Ast.TopInit(init) -> initialiser init
-  | Ast.Include(inc,s) ->
-      mcode print_string inc; print_string " "; mcode inc_file s
-  | Ast.MetaInclude(inc,s) ->
-      mcode print_string inc; print_string " "; expression s
+  | Ast.CppTop(di) -> directive di
   | Ast.Undef(def,id) ->
       mcode print_string def; print_string " "; ident id
   | Ast.DefineHeader(def,id,params) ->
       mcode print_string def; print_string " "; ident id;
       print_define_parameters params
-  | Ast.Pragma(prg,id,body) ->
-      mcode print_string prg; print_string " "; ident id; print_string " ";
-      pragmainfo body
   | Ast.Default(def,colon) ->
       mcode print_string def; mcode print_string colon; print_string " " 
   | Ast.AsRe(re,asre) ->
@@ -835,14 +941,22 @@ and rule_elem arity re =
       force_newline(); print_string ")"
 
 and forinfo = function
-    Ast.ForExp(e1,sem1) ->
-      print_option expression e1; mcode print_string sem1
-  | Ast.ForDecl(ann_decl) -> annotated_decl "" ann_decl
+    Ast.ForExp(e1,sem1,e2,sem2,e3) ->
+      print_option expression e1; mcode print_string sem1;
+      print_option expression e2; mcode print_string sem2;
+      print_option expression e3
+  | Ast.ForDecl(ann_decl,e2,sem2,e3) ->
+      annotated_decl "" ann_decl;
+      print_option expression e2; mcode print_string sem2;
+      print_option expression e3
+  | Ast.ForRange(ann_decl, ini) ->
+      annotated_decl "" ann_decl; print_space(); initialiser ini
 
 and pragmainfo pi =
   match Ast.unwrap pi with
-      Ast.PragmaString(s) -> mcode print_string s
-    | Ast.PragmaDots (dots) -> mcode print_string dots
+    Ast.PragmaString(s) -> mcode print_string s
+  | Ast.PragmaDots (dots) -> mcode print_string dots
+  | Ast.MetaPragmaInfo(metavar,_,_,_) -> mcode print_meta metavar
 
 and print_define_parameters params =
   match Ast.unwrap params with
@@ -854,6 +968,7 @@ and print_define_parameters params =
 and print_define_param param =
   match Ast.unwrap param with
     Ast.DParam(id) -> ident id
+  | Ast.DParamEll(id,dots) -> ident id; mcode print_string dots
   | Ast.MetaDParamList(name,_,_,_,_) -> mcode print_meta name
   | Ast.DPComma(comma) -> mcode print_string comma
   | Ast.DPdots(dots) -> mcode print_string dots
@@ -895,6 +1010,9 @@ and statement arity s =
       dots force_newline (statement arity) body;
       rule_elem arity rbrace;
       mcode (function _ -> ()) ((),Ast.no_info,aft,[])
+  | Ast.TemplateDefinition(header,stmt) ->
+      rule_elem arity header;
+      statement arity stmt
   | Ast.Disj([stmt_dots]) (* useful why? *)
   | Ast.Conj([stmt_dots]) ->
       print_string arity;
@@ -938,6 +1056,32 @@ and statement arity s =
       close_box(); force_newline()
   | Ast.OptStm(s) -> statement "?" s
 
+and directive di =
+  match Ast.unwrap di with
+    Ast.Include(inc,s) ->
+      mcode print_string inc; print_string " "; mcode inc_file s
+  | Ast.MetaInclude(inc,s) ->
+      mcode print_string inc; print_string " "; expression s
+  | Ast.Pragma(prg,id,body) ->
+      mcode print_string prg; print_string " "; ident id; print_string " ";
+      pragmainfo body
+  | Ast.UsingNamespace(usng,nmspc,name,sem) ->
+      mcode print_string usng; print_string " ";
+      mcode print_string nmspc; print_string " ";
+      ident name; print_string " ";
+      mcode print_string sem
+  | Ast.UsingTypename(usng,name,eq,tn,ty,sem) ->
+      mcode print_string usng; print_string " ";
+      ident name; print_string " ";
+      mcode print_string eq  ; print_string " ";
+      print_option (fun x -> mcode print_string x; print_string " ") tn;
+      fullType ty;  print_string " ";
+      mcode print_string sem
+  | Ast.UsingMember(usng,name,sem) ->
+      mcode print_string usng; print_string " ";
+      ident name; print_string " ";
+      mcode print_string sem
+
 and whencode notfn alwaysfn = function
     Ast.WhenNot a ->
       print_string "   WHEN != "; open_box 0; notfn a; close_box()
@@ -1058,6 +1202,9 @@ let unparse_cocci_mv rule = function
   | Ast.MetaAssignmentOperatorDecl(_,(r,n)) -> (* missing constraints *)
       print_string "assignment operator "; print_name rule r n;
       print_string ";"
+  | Ast.MetaPragmaInfoDecl(_,(r,n)) -> (* missing constraints *)
+      print_string "pragmainfo "; print_name rule r n;
+      print_string ";"
   | Ast.MetaParamListDecl(_,(r,n),len) ->
       print_string "parameter list"; print_listlen rule len;
       print_name rule r n; print_string ";"
@@ -1173,6 +1320,7 @@ let _ =
     | Ast.CaseLineTag(x) -> case_line "" x
     | Ast.StringFragmentTag(x) -> string_fragment x
     | Ast.AttributeTag(x) -> print_attribute x
+    | Ast.AttrArgTag(x) -> print_attr_arg x
     | Ast.ConstVolTag(x) -> const_vol x
     | Ast.Token(x,Some info) -> print_string_befaft print_string x info
     | Ast.Token(x,None) -> print_string x
@@ -1183,6 +1331,7 @@ let _ =
     | Ast.Code(x) -> let _ = top_level x in ()
     | Ast.ExprDotsTag(x) -> dots (function _ -> ()) expression x
     | Ast.ParamDotsTag(x) -> parameter_list x
+    | Ast.TemplateParamDotsTag(x) -> dots (function _ -> ()) templateParameterTypeDef x
     | Ast.StmtDotsTag(x) -> dots (function _ -> ()) (statement "") x
     | Ast.AnnDeclDotsTag(x) -> dots (function _ -> ()) (annotated_decl "") x
     | Ast.AnnFieldDotsTag(x) -> dots (function _ -> ()) (annotated_field "") x
@@ -1190,6 +1339,7 @@ let _ =
     | Ast.DefParDotsTag(x) -> dots (function _ -> ()) print_define_param x
     | Ast.TypeCTag(x) -> typeC x
     | Ast.ParamTag(x) -> parameterTypeDef x
+    | Ast.TemplateParamTag(x) -> templateParameterTypeDef x
     | Ast.SgrepStartTag(x) -> print_string x
     | Ast.SgrepEndTag(x) -> print_string x
 
@@ -1293,6 +1443,10 @@ let ident_to_string x =
   print_newlines_disj := true;
   Common.format_to_string (function _ -> ident x)
 
+let expression_to_string x =
+  print_newlines_disj := true;
+  Common.format_to_string (function _ -> expression x)
+
 let unparse_to_string x =
   print_newlines_disj := true;
   Common.format_to_string (function _ -> unparse [] x)
