@@ -36,14 +36,7 @@ let all_removed_recursor =
     | Ast.MINUS(_,_,_,Ast.REPLACEMENT(_,_)) -> Ast.Unsafe
     | Ast.PLUS _ -> failwith "not possible"
     | Ast.CONTEXT(_,info) -> Ast.Unsafe in
-  V.combiner bind option_default
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode mcode mcode mcode
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-    do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
+  V.combiner bind option_default {V.cmcode=mcode} {V.cdonothing=do_nothing} do_nothing
 
 let all_removed_decl =
   all_removed_recursor.V.combiner_declaration
@@ -83,7 +76,8 @@ let contains_modif =
 	bind (mcode r ((),(),bef,[])) res
     | Ast.Decl decl ->
 	bind (mcode r ((),(),annotated_decl decl,[])) res
-    | Ast.ForHeader(fr,lp,Ast.ForDecl(decl),e2,sem2,e3,rp) ->
+    | Ast.ForHeader(fr,lp,Ast.ForDecl(decl,_,_,_),rp)
+    | Ast.ForHeader(fr,lp,Ast.ForRange(decl,_),rp) ->
 	bind (mcode r ((),(),annotated_decl decl,[])) res
     | _ -> res in
   let init r k i =
@@ -92,15 +86,8 @@ let contains_modif =
       Ast.StrInitList(allminus,_,_,_,_) -> allminus || res
     | _ -> res in
   let recursor =
-    V.combiner bind option_default
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode mcode mcode mcode
-      do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-      do_nothing do_nothing do_nothing do_nothing do_nothing do_nothing
-      do_nothing do_nothing do_nothing init do_nothing
-      do_nothing do_nothing do_nothing do_nothing
-      do_nothing do_nothing rule_elem do_nothing do_nothing do_nothing
-      do_nothing do_nothing do_nothing in
+    V.combiner bind option_default {V.cmcode=mcode} {V.cdonothing=do_nothing}
+      ~init:init ~rule:rule_elem do_nothing in
   recursor.V.combiner_fullType
 
 let attr_arg a =
@@ -112,7 +99,12 @@ let attr_arg a =
 let attribute a =
   match Ast.unwrap a with
     Ast.Attribute(attr) -> attr_arg attr
-  | Ast.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) -> attr_arg arg
+  | Ast.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) -> mcode () attr_
+  | Ast.CxxAttribute(lb1,arg,rb1,rb2) -> mcode () lb1
+  | Ast.CxxAttributeUsing(lb1,usng,atnm,dotdot,arg,rb1,rb2) -> mcode () lb1
+
+let alignas (Ast.Align(al,lp,_,rp)) =
+  mcode () al || mcode () lp || mcode () rp
 
 let decl r k e =
   let e = k e in
@@ -121,17 +113,20 @@ let decl r k e =
   | Ast.NoStorage -> {e with Ast.safe_for_multi_decls = Ast.NoStorage}
   | Ast.Unsafe ->
       match Ast.unwrap e with
-	Ast.Init(stg,ty,midattr,_,endattr,_,_,sem)
-      | Ast.UnInit(stg,ty,midattr,_,endattr,sem) ->
+	Ast.Init(al,stg,ty,_,endattr,_,_,sem)
+      | Ast.UnInit(al,stg,ty,_,endattr,sem) ->
+	  let al_modif =
+	    match al with
+	      Some al -> alignas al
+	    | None -> false in
 	  let stg_modif =
 	    match stg with
 	      Some stg -> mcode () stg
 	    | None -> false in
-	  let midattr_modif = List.exists attribute midattr in
 	  let endattr_modif = List.exists attribute endattr in
 	  let ft_modif = contains_modif ty in
 	  let sem_modif = mcode () sem in
-	  if not(stg_modif || midattr_modif || endattr_modif || ft_modif || sem_modif)
+	  if not(al_modif || stg_modif || endattr_modif || ft_modif || sem_modif)
 	  then {e with Ast.safe_for_multi_decls = Ast.Safe}
 	  else e
       | _ -> e
@@ -155,10 +150,11 @@ let field r k e =
   | Ast.NoStorage -> {e with Ast.safe_for_multi_decls = Ast.NoStorage}
   | Ast.Unsafe ->
     match Ast.unwrap e with
-      Ast.Field(ty,_,_bf,sem) ->
+      Ast.Field(ty,_,_bf,endattr,sem) ->
 	let ft_modif = contains_modif ty in
+	let endattr_modif = List.exists attribute endattr in
 	let sem_modif = mcode () sem in
-	if not(ft_modif || sem_modif)
+	if not(ft_modif || endattr_modif || sem_modif)
 	then {e with Ast.safe_for_multi_decls = Ast.Safe}
 	else e
     | _ -> e
@@ -180,13 +176,9 @@ let mcode e = e
 let donothing r k e = k e
 
 let process =
-  let fn = V.rebuilder
-      mcode mcode mcode mcode mcode mcode mcode mcode mcode
-      mcode mcode mcode mcode mcode
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing donothing donothing donothing
-      donothing donothing donothing donothing decl anndecl field annfield
-      donothing donothing donothing donothing donothing donothing donothing
+  let fn =
+    V.rebuilder {V.rmcode=mcode} {V.rdonothing=donothing}
+      ~decl:decl ~annotated_decl:anndecl ~field:field ~annotated_field:annfield
       donothing in
   List.map fn.V.rebuilder_top_level
 
