@@ -21,11 +21,6 @@ let left_dots f l =
     [] -> false
   | x::xs -> f x
 
-let right_dots f l =
-  match List.rev (Ast0.unwrap l) with
-    [] -> false
-  | x::xs -> f x
-
 let modif_before_mcode mc =
   match Ast0.get_mcode_mcodekind mc with
     Ast0.MINUS mc -> true (*conservative; don't want to hunt right for + code*)
@@ -104,18 +99,6 @@ let rec left_ident i =
   | Ast0.OptIdent(id) -> left_ident id
   | Ast0.AsIdent _ -> failwith "not possible"
 
-let rec right_ident i =
-  modif_after i ||
-  match Ast0.unwrap i with
-    Ast0.Id(name) -> modif_after_mcode name
-  | Ast0.MetaId(name,_,_,_) -> modif_after_mcode name
-  | Ast0.MetaFunc(name,_,_) -> modif_after_mcode name
-  | Ast0.MetaLocalFunc(name,_,_) -> modif_after_mcode name
-  | Ast0.DisjId(_,id_list,_,_) -> List.exists right_ident id_list
-  | Ast0.ConjId(_,id_list,_,_) -> List.exists right_ident id_list
-  | Ast0.OptIdent(id) -> right_ident id
-  | Ast0.AsIdent _ -> failwith "not possible"
-
 (* --------------------------------------------------------------------- *)
 (* Expression *)
 
@@ -138,9 +121,15 @@ let rec left_expression e =
   | Ast0.ArrayAccess(exp1,lb,exp2,rb) -> left_expression exp1
   | Ast0.RecordAccess(exp,pt,field) -> left_expression exp
   | Ast0.RecordPtAccess(exp,ar,field) -> left_expression exp
-  | Ast0.Cast(lp,ty,attr,rp,exp) -> modif_before_mcode lp
+  | Ast0.QualifiedAccess(Some ty,coloncolon,field) -> left_typeC ty
+  | Ast0.QualifiedAccess(None,coloncolon,field) -> modif_before_mcode coloncolon
+  | Ast0.Cast(lp,ty,rp,exp) -> modif_before_mcode lp
   | Ast0.SizeOfExpr(szf,exp) -> modif_before_mcode szf
   | Ast0.SizeOfType(szf,lp,ty,rp) -> modif_before_mcode szf
+  | Ast0.Delete(dlt,exp) -> modif_before_mcode dlt
+  | Ast0.DeleteArr(dlt,lb,rb,exp) -> modif_before_mcode dlt
+  | Ast0.TemplateInst(name,lab,args,rab) -> left_expression name
+  | Ast0.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) -> modif_before_mcode nw
   | Ast0.TypeExp(ty) -> left_typeC ty
   | Ast0.Constructor(lp,ty,rp,init) -> modif_before_mcode lp
   | Ast0.MetaErr(name,_,_) -> modif_before_mcode name
@@ -158,10 +147,30 @@ let rec left_expression e =
 (* --------------------------------------------------------------------- *)
 (* Types *)
 
+and left_attr_arg a =
+  modif_before a ||
+  match Ast0.unwrap a with
+    Ast0.MacroAttr(arg) -> modif_before_mcode arg
+  | Ast0.MacroAttrArgs(nm,lp,exps,rp) -> modif_before_mcode nm
+  | Ast0.MetaAttr(nm,x,y) -> modif_before_mcode nm
+
+and left_attr a =
+  modif_before a ||
+  match Ast0.unwrap a with
+    Ast0.Attribute(arg) -> left_attr_arg arg
+  | Ast0.GccAttribute(attr,lp1,lp2,exps,rp2,rp1) ->
+      modif_before_mcode attr
+  | Ast0.CxxAttribute(lb1,exps,rb1,rb2) ->
+      modif_before_mcode lb1
+  | Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,exps,rb1,rb2) ->
+      modif_before_mcode lb1
+
 and left_typeC t =
   modif_before t ||
   match Ast0.unwrap t with
-    Ast0.ConstVol(cv,ty) -> modif_before_mcode (List.hd cv)
+    Ast0.ConstVol([],ty,cvafter) -> left_typeC ty
+  | Ast0.ConstVol((Ast0.CV cv)::_,ty,_) -> modif_before_mcode cv
+  | Ast0.ConstVol((Ast0.Attr attr)::_,ty,_) -> left_attr attr
   | Ast0.BaseType(ty,strings) -> modif_before_mcode (List.hd strings)
   | Ast0.Signed(sgn,ty) -> modif_before_mcode sgn
   | Ast0.Pointer(ty,star) -> left_typeC ty
@@ -170,13 +179,16 @@ and left_typeC t =
   | Ast0.Array(ty,lb,size,rb) -> left_typeC ty
   | Ast0.Decimal(dec,lp,length,comma,precision_opt,rp) ->
       modif_before_mcode dec
-  | Ast0.EnumName(kind,name) -> modif_before_mcode kind
-  | Ast0.EnumDef(ty,lb,ids,rb) -> left_typeC ty
+  | Ast0.EnumName(kind,key,name) -> modif_before_mcode kind
+  | Ast0.EnumDef(ty,base,lb,ids,rb) -> left_typeC ty
   | Ast0.StructUnionName(kind,name) -> modif_before_mcode kind
   | Ast0.StructUnionDef(ty,lb,decls,rb) -> left_typeC ty
   | Ast0.TypeOfExpr(tf,_,_,_) -> modif_before_mcode tf
   | Ast0.TypeOfType(tf,_,_,_) -> modif_before_mcode tf
+  | Ast0.QualifiedType(Some ty,coloncolon,name) -> left_typeC ty
+  | Ast0.QualifiedType(None,coloncolon,name) -> modif_before_mcode coloncolon
   | Ast0.TypeName(name) -> modif_before_mcode name
+  | Ast0.TemplateType(name,_,_,_) -> left_typeC name
   | Ast0.AutoType(auto) -> modif_before_mcode auto
   | Ast0.MetaType(name,_,_) -> modif_before_mcode name
   | Ast0.DisjType(lp,types,mids,rp) -> List.exists left_typeC types
@@ -193,165 +205,36 @@ and left_declaration d =
   modif_before d ||
   match Ast0.unwrap d with
     Ast0.MetaDecl(name,_,_) -> modif_before_mcode name
-  | Ast0.Init(Some stg,ty,midattr,id,endattr,eq,ini,sem) -> modif_before_mcode stg
-  | Ast0.Init(None,ty,midattr,id,endattr,eq,ini,sem) -> left_typeC ty
-  | Ast0.UnInit(Some stg,ty,midattr,id,endattr,sem) -> modif_before_mcode stg
-  | Ast0.UnInit(None,ty,midattr,id,endattr,sem) -> left_typeC ty
-  | Ast0.FunProto(fninfo,attr,name,lp1,params,va,rp1,sem) ->
+  | Ast0.Init(Some(Ast0.Align(a,lp,e,rp)),stg,ty,id,endattr,eq,ini,sem) ->
+      modif_before_mcode a
+  | Ast0.Init(None,Some stg,ty,id,endattr,eq,ini,sem) -> modif_before_mcode stg
+  | Ast0.Init(None,None,ty,id,endattr,eq,ini,sem) -> left_typeC ty
+  | Ast0.UnInit(Some(Ast0.Align(a,lp,e,rp)),stg,ty,id,endattr,sem) ->
+      modif_before_mcode a
+  | Ast0.UnInit(None,Some stg,ty,id,endattr,sem) -> modif_before_mcode stg
+  | Ast0.UnInit(None,None,ty,id,endattr,sem) -> left_typeC ty
+  | Ast0.FunProto(fninfo,name,lp1,params,va,rp1,sem) ->
       (* should not be nested in anything anyway *)
       false
-  | Ast0.MacroDecl(Some stg,name,lp,args,rp,attr,sem) -> modif_before_mcode stg
-  | Ast0.MacroDecl(None,name,lp,args,rp,attr,sem) -> left_ident name
-  | Ast0.MacroDeclInit(Some stg,name,lp,args,rp,eq,ini,sem) ->
+  | Ast0.MacroDecl(Some stg,preattr,name,lp,args,rp,attr,sem) ->
       modif_before_mcode stg
-  | Ast0.MacroDeclInit(None,name,lp,args,rp,eq,ini,sem) -> left_ident name
-  | Ast0.TyDecl(ty,attr,sem) -> left_typeC ty
+  | Ast0.MacroDecl(None,[],name,lp,args,rp,attr,sem) -> left_ident name
+  | Ast0.MacroDecl(None,preattr::_,name,lp,args,rp,attr,sem) ->
+      left_attr preattr
+  | Ast0.MacroDeclInit(Some stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
+      modif_before_mcode stg
+  | Ast0.MacroDeclInit(None,[],name,lp,args,rp,attr,eq,ini,sem) ->
+      left_ident name
+  | Ast0.MacroDeclInit(None,preattr::_,name,lp,args,rp,attr,eq,ini,sem) ->
+      left_attr preattr
+  | Ast0.TyDecl(ty,sem) -> left_typeC ty
   | Ast0.Typedef(stg,ty,id,sem) -> modif_before_mcode stg
   | Ast0.DisjDecl(_,decls,_,_) -> List.exists left_declaration decls
   | Ast0.ConjDecl(_,decls,_,_) -> List.exists left_declaration decls
   | Ast0.OptDecl(decl) -> left_declaration decl
   | Ast0.AsDecl _ -> failwith "not possible"
 
-and right_declaration d =
-  modif_before d ||
-  match Ast0.unwrap d with
-    Ast0.MetaDecl(name,_,_) -> modif_before_mcode name
-  | Ast0.Init(_,ty,midattr,id,eq,ini,endattr,sem) -> modif_after_mcode sem
-  | Ast0.UnInit(_,ty,midattr,id,endattr,sem) -> modif_after_mcode sem
-  | Ast0.FunProto(fninfo,attr,name,lp1,params,va,rp1,sem) -> modif_after_mcode sem
-  | Ast0.MacroDecl(_,name,lp,args,rp,attr,sem) -> modif_after_mcode sem
-  | Ast0.MacroDeclInit(_,name,lp,args,rp,eq,ini,sem) -> modif_after_mcode sem
-  | Ast0.TyDecl(ty,attr,sem) -> modif_after_mcode sem
-  | Ast0.Typedef(stg,ty,id,sem) -> modif_after_mcode sem
-  | Ast0.DisjDecl(_,decls,_,_) -> List.exists right_declaration decls
-  | Ast0.ConjDecl(_,decls,_,_) -> List.exists right_declaration decls
-  | Ast0.OptDecl(decl) -> right_declaration decl
-  | Ast0.AsDecl _ -> failwith "not possible"
-
 (* --------------------------------------------------------------------- *)
-(* Field declaration *)
-
-and left_field d =
-  modif_before d ||
-  match Ast0.unwrap d with
-    Ast0.MetaField(name,_,_)
-  | Ast0.MetaFieldList(name,_,_,_) ->
-      modif_before_mcode name
-  | Ast0.Field(ty,id,_bf,sem) -> left_typeC ty
-  | Ast0.DisjField(_,decls,_,_) -> List.exists left_field decls
-  | Ast0.ConjField(_,decls,_,_) -> List.exists left_field decls
-  | Ast0.OptField(decl) -> left_field decl
-  | Ast0.Fdots(dots,_) -> false
-
-and right_field d =
-  modif_before d ||
-  match Ast0.unwrap d with
-    Ast0.MetaField(name,_,_)
-  | Ast0.MetaFieldList(name,_,_,_) ->
-      modif_before_mcode name
-  | Ast0.Field(ty,id,_bf,sem) -> modif_after_mcode sem
-  | Ast0.DisjField(_,decls,_,_) -> List.exists right_field decls
-  | Ast0.ConjField(_,decls,_,_) -> List.exists right_field decls
-  | Ast0.OptField(decl) -> right_field decl
-  | Ast0.Fdots(dots,_) -> false
-
-(* --------------------------------------------------------------------- *)
-(* Top-level code *)
-(* These functions seem to be never used
-and left_statement s =
-  modif_before s or
-  match Ast0.unwrap s with
-    Ast0.FunDecl(_,fninfo,name,lp,params,rp,lbrace,body,rbrace) ->
-      (* irrelevant *) false
-  | Ast0.Decl(_,decl) -> left_declaration decl
-  | Ast0.Seq(lbrace,body,rbrace) -> modif_before_mcode lbrace
-  | Ast0.ExprStatement(Some exp,sem) -> left_expression exp
-  | Ast0.ExprStatement(None,sem) -> modif_before_mcode sem
-  | Ast0.IfThen(iff,lp,exp,rp,branch1,aft) -> modif_before_mcode iff
-  | Ast0.IfThenElse(iff,lp,exp,rp,branch1,els,branch2,aft) ->
-      modif_before_mcode iff
-  | Ast0.While(whl,lp,exp,rp,body,aft) -> modif_before_mcode whl
-  | Ast0.Do(d,body,whl,lp,exp,rp,sem) -> modif_before_mcode d
-  | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,aft) ->
-      modif_before_mcode fr
-  | Ast0.Iterator(nm,lp,args,rp,body,aft) -> left_ident nm
-  | Ast0.Switch(switch,lp,exp,rp,lb,decls,cases,rb) ->
-      modif_before_mcode switch
-  | Ast0.Break(br,sem) -> modif_before_mcode br
-  | Ast0.Continue(cont,sem) -> modif_before_mcode cont
-  | Ast0.Label(l,dd) -> left_ident l
-  | Ast0.Goto(goto,l,sem) -> modif_before_mcode goto
-  | Ast0.Return(ret,sem) -> modif_before_mcode ret
-  | Ast0.ReturnExpr(ret,exp,sem) -> modif_before_mcode ret
-  | Ast0.Exec(exec,lang,code,sem) -> modif_before_mcode exec
-  | Ast0.MetaStmt(name,pure) -> modif_before_mcode name
-  | Ast0.MetaStmtList(name,_,_) -> modif_before_mcode name
-  | Ast0.Disj(_,statement_dots_list,_,_) ->
-      List.exists (left_dots left_statement) statement_dots_list
-  | Ast0.Nest(starter,stmt_dots,ender,whencode,multi) ->
-      left_dots left_statement stmt_dots
-  | Ast0.Exp(exp) -> false (* can only be replaced by an expression *)
-  | Ast0.TopExp(exp) -> false (* as above *)
-  | Ast0.Ty(ty) -> false (* can only be replaced by a type *)
-  | Ast0.TopId(id) -> false (* can only be replaced by an ident *)
-  | Ast0.TopInit(init) -> false (* can only be replaced by an init *)
-  | Ast0.Dots(d,whn) -> false
-  | Ast0.Include(inc,s) -> modif_before_mcode inc
-  | Ast0.MetaInclude(inc,s) -> modif_before_mcode inc
-  | Ast0.Undef(def,id) -> modif_before_mcode def
-  | Ast0.Define(def,id,params,body) -> modif_before_mcode def
-  | Ast0.Pragma(prg,id,body) -> modif_before_mcode prg
-  | Ast0.OptStm(re) -> left_statement re
-  | Ast0.AsStmt _ -> failwith "not possible"
-
-and right_statement s =
-  modif_after s or
-  match Ast0.unwrap s with
-    Ast0.FunDecl(_,fninfo,name,lp,params,rp,lbrace,body,rbrace) ->
-      (* irrelevant *) false
-  | Ast0.Decl(_,decl) -> right_declaration decl
-  | Ast0.Seq(lbrace,body,rbrace) -> modif_after_mcode rbrace
-  | Ast0.ExprStatement(exp,sem) -> modif_after_mcode sem
-  | Ast0.IfThen(iff,lp,exp,rp,branch1,(_,aft,_)) -> modif_after_mcodekind aft
-  | Ast0.IfThenElse(iff,lp,exp,rp,branch1,els,branch2,(_,aft,_)) ->
-      modif_after_mcodekind aft
-  | Ast0.While(whl,lp,exp,rp,body,(_,aft,_)) -> modif_after_mcodekind aft
-  | Ast0.Do(d,body,whl,lp,exp,rp,sem) -> modif_after_mcode sem
-  | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,(_,aft,_)) ->
-      modif_after_mcodekind aft
-  | Ast0.Iterator(nm,lp,args,rp,body,(_,aft,_)) ->
-      modif_after_mcodekind aft
-  | Ast0.Switch(switch,lp,exp,rp,lb,decls,cases,rb) -> modif_after_mcode rb
-  | Ast0.Break(br,sem) -> modif_after_mcode sem
-  | Ast0.Continue(cont,sem) -> modif_after_mcode sem
-  | Ast0.Label(l,dd) -> modif_after_mcode dd
-  | Ast0.Goto(goto,l,sem) -> modif_after_mcode sem
-  | Ast0.Return(ret,sem) -> modif_after_mcode sem
-  | Ast0.ReturnExpr(ret,exp,sem) -> modif_after_mcode sem
-  | Ast0.Exec(exec,lang,code,sem) -> modif_after_mcode sem
-  | Ast0.MetaStmt(name,pure) -> modif_after_mcode name
-  | Ast0.MetaStmtList(name,_,_) -> modif_after_mcode name
-  | Ast0.Disj(_,statement_dots_list,_,_) ->
-      List.exists (right_dots right_statement) statement_dots_list
-  | Ast0.Nest(starter,stmt_dots,ender,whencode,multi) ->
-      right_dots right_statement stmt_dots
-  | Ast0.Exp(exp) -> false (* can only be replaced by an expression *)
-  | Ast0.TopExp(exp) -> false (* as above *)
-  | Ast0.Ty(ty) -> false (* can only be replaced by a type *)
-  | Ast0.TopId(id) -> false (* can only be replaced by a type *)
-  | Ast0.TopInit(init) -> false (* can only be replaced by an init *)
-  | Ast0.Dots(d,whn) -> false
-  | Ast0.Include(inc,s) -> modif_after_mcode s
-  | Ast0.MetaInclude(inc,s) -> false (* irrelevant, not a substatement *)
-  | Ast0.Undef(def,id) -> right_ident id
-  | Ast0.Define(def,id,params,body) -> right_dots right_statement body
-  | Ast0.Pragma(prg,id,body) -> right_pragma body -- not defined, b/c not used
-  | Ast0.OptStm(re) -> right_statement re
-  | Ast0.AsStmt _ -> failwith "not possible"
-*)
-
-(* --------------------------------------------------------------------- *)
-
 
 (* A very coarse approximation.  We would really only like to return true
 if a new statement is added.  For this it would be best to correlate with the
@@ -372,8 +255,7 @@ let rec adding_something s =
       let (text,tinfo1,tinfo2) = !mc in
       (match text with Ast.NOTHING -> false | _ -> true)
   | Ast0.MIXED(_) ->
-      not(contains_only_minus.VT0.combiner_rec_statement s) (*&&
-      (left_statement s) or (right_statement s)*)
+      not(contains_only_minus.VT0.combiner_rec_statement s)
   | _ -> failwith "unexpected plus code"
 
 (* why do we need this; MINUS should mean the same thing *)
@@ -457,14 +339,12 @@ and contains_only_minus =
 	List.for_all r.VT0.combiner_rec_case_line case_list
     | _ -> k e in
 
-  V0.flat_combiner bind option_default
-    mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode mcode
-    mcode mcode mcode
-    dots dots dots dots dots dots dots dots dots
-    identifier expression donothing donothing typeC donothing donothing
-    declaration field donothing statement donothing case_line donothing
-    donothing donothing donothing
-
+  V0.combiner bind option_default {V0.cmcode=mcode} {V0.cdonothing=donothing}
+    ~dotsexpr:dots ~dotsinit:dots ~dotsparam:dots ~dotstemplateparam:dots
+    ~dotsstmt:dots ~dotsdecl:dots ~dotsfield:dots ~dotsenumdecl:dots
+    ~dotscase:dots ~dotsdefpar:dots
+    ~ident:identifier ~expr:expression ~ty:typeC ~decl:declaration ~field:field
+    ~stmt:statement ~case:case_line ()
 
 (* needs a special case when there is a Disj or an empty DOTS *)
 (* ---------------------------------------------------------------------- *)
@@ -599,6 +479,9 @@ let rec statement dots_before dots_after s =
 	(Ast0.FunDecl(x,fninfo,name,lp,params,va,rp,attrs,lbrace,
 		      statement_dots false true body,
 		      rbrace,y))
+  | Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt) ->
+      Ast0.rewrap s
+        (Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt)) (* FIXME: check *)
   | Ast0.Decl(_,_) -> s
   | Ast0.Seq(lbrace,body,rbrace) ->
       Ast0.rewrap s
@@ -623,11 +506,10 @@ let rec statement dots_before dots_after s =
       do_one
 	(Ast0.rewrap s
 	   (Ast0.Do(d,statement false false body,whl,lp,exp,rp,sem)))
-  | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,x) ->
+  | Ast0.For(fr,lp,first,rp,body,x) ->
       do_one
 	(Ast0.rewrap s
-	   (Ast0.For(fr,lp,first,e2,sem2,e3,rp,
-		     statement false false body,x)))
+	   (Ast0.For(fr,lp,first,rp,statement false false body,x)))
   | Ast0.Iterator(nm,lp,args,rp,body,x) ->
       do_one
 	(Ast0.rewrap s
@@ -677,11 +559,14 @@ let rec statement dots_before dots_after s =
   | Ast0.TopId(id) -> s
   | Ast0.TopInit(init) -> s
   | Ast0.Dots(d,whn) -> s
-  | Ast0.Include(inc,string) -> s (* doesn't affect the need for braces *)
-  | Ast0.MetaInclude(inc,name) -> s (* doesn't affect the need for braces *)
+  | Ast0.CppTop(di) ->
+      (match Ast0.unwrap di with
+	Ast0.Pragma(prg,id,body) -> s (* same as include *)
+      | Ast0.Include(inc,string) -> s (* doesn't affect the need for braces *)
+      | Ast0.MetaInclude(inc,name) -> s (* doesn't affect the need for braces *)
+      | _ -> do_one s)
   | Ast0.Undef(def,id) -> s (* same as include *)
   | Ast0.Define(def,id,params,body) -> s (* same as include *)
-  | Ast0.Pragma(prg,id,body) -> s (* same as include *)
   | Ast0.OptStm(re) ->
       Ast0.rewrap s
 	(Ast0.OptStm(statement dots_before dots_after re))
