@@ -36,12 +36,15 @@ let set_test_poss =
 	Ast0.rewrap e (Ast0.CondExpr(process_exp e1,q,e2,c,e3))
     | Ast0.Binary(e1,op,e2) ->
 	(match Ast0.unwrap op with
-	  Ast0.Logical op' when (let op''=Ast0.unwrap_mcode op' in op''=Ast.AndLog || op''=Ast.OrLog) ->
-	    Ast0.rewrap e (Ast0.Binary(process_exp e1,op,process_exp e2))
+	  Ast0.Logical op' ->
+	    (match Ast0.unwrap_mcode op' with
+	      Ast.AndLog _ | Ast.OrLog _ -> 
+		Ast0.rewrap e (Ast0.Binary(process_exp e1,op,process_exp e2))
+	    | _ -> e)
 	| _ -> e)
     | Ast0.Unary(e1,op) ->
 	(match Ast0.unwrap_mcode op with
-	  Ast.Not -> Ast0.rewrap e (Ast0.Unary(process_exp e1,op))
+	  Ast.Not _ -> Ast0.rewrap e (Ast0.Unary(process_exp e1,op))
 	| _ -> e)
     | _ -> e in
 
@@ -61,18 +64,26 @@ let set_test_poss =
 	Ast0.rewrap s (Ast0.While(i,lp,process_exp e,rp,s1,aft))
     | Ast0.Do(d,s1,w,lp,e,rp,sc) ->
 	Ast0.rewrap s (Ast0.Do(d,s1,w,lp,process_exp e,rp,sc))
-    | Ast0.For(f,lp,first,Some e2,sc2,e3,rp,s1,aft) ->
-	Ast0.rewrap s
-	  (Ast0.For(f,lp,first,Some (process_exp e2),sc2,e3,rp,s1,aft))
+    | Ast0.For(f,lp,first,rp,s1,aft) ->
+	(match Ast0.unwrap first with
+	  Ast0.ForExp(e1a,sc1a,Some e2a,sc2a,e3a) ->
+	    let first =
+	      Ast0.rewrap first
+		(Ast0.ForExp(e1a,sc1a,Some (process_exp e2a),sc2a,e3a)) in
+	    Ast0.rewrap s (Ast0.For(f,lp,first,rp,s1,aft))
+	| Ast0.ForDecl (bef,decla,Some e2a,sc2a,e3a) ->
+	    let first =
+	      Ast0.rewrap first
+		(Ast0.ForDecl(bef,decla,Some (process_exp e2a),sc2a,e3a)) in
+	    Ast0.rewrap s (Ast0.For(f,lp,first,rp,s1,aft))
+	| _ -> s)
     | Ast0.Dots(d,wc) ->
 	Ast0.rewrap s (Ast0.Dots(d,List.map process_wc wc))
     | Ast0.Nest(l,s1,r,wc,m) ->
 	Ast0.rewrap s (Ast0.Nest(l,s1,r,List.map process_wc wc,m))
     | _ -> s in
 
-  V0.rebuilder
-    {V0.rebuilder_functions with
-      VT0.rebuilder_exprfn = expression; VT0.rebuilder_stmtfn = statement}
+  V0.rebuilder_default ~expr:expression ~stmt:statement ()
 
 let process = List.map set_test_poss.VT0.rebuilder_rec_top_level
 
