@@ -214,16 +214,18 @@ let rec expression e =
       | Ast0.Paren(lp,exp,rp) ->
 	  mcode print_string_box lp; expression exp; close_box();
 	  mcode print_string rp
-      | Ast0.ArrayAccess(exp1,lb,exp2,rb) ->
-	  expression exp1; mcode print_string_box lb; expression exp2;
+      | Ast0.ArrayAccess(fn,lb,args,rb) ->
+	  expression fn; mcode print_string_box lb;
+	  let _ = dots (function _ -> ()) expression args in
 	  close_box(); mcode print_string rb
       | Ast0.RecordAccess(exp,pt,field) ->
 	  expression exp; mcode print_string pt; ident field
       | Ast0.RecordPtAccess(exp,ar,field) ->
 	  expression exp; mcode print_string ar; ident field
-      | Ast0.Cast(lp,ty,attr,rp,exp) ->
+      | Ast0.QualifiedAccess(ty,coloncolon,field) ->
+      print_option typeC ty; mcode print_string coloncolon; ident field 
+      | Ast0.Cast(lp,ty,rp,exp) ->
 	  mcode print_string_box lp; typeC ty; close_box();
-          print_attribute_list attr;
 	  mcode print_string rp; expression exp
       | Ast0.SizeOfExpr(szf,exp) ->
 	  mcode print_string szf; expression exp
@@ -231,6 +233,24 @@ let rec expression e =
           mcode print_string szf;
 	  mcode print_string_box lp; typeC ty; close_box();
 	  mcode print_string rp
+      | Ast0.Delete(dlt,exp) ->
+	  mcode print_string dlt; print_string " "; expression exp
+      | Ast0.DeleteArr(dlt,lb,rb,exp) ->
+	  mcode print_string dlt; print_string " "; mcode print_string lb; mcode print_string rb; print_string " ";
+	  expression exp
+      | Ast0.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) ->
+	  mcode print_string nw;
+	  print_string " ";
+	  print_option (print_args true) pp_opt;
+	  print_option (function e -> mcode print_string e) lp_opt;
+	  typeC ty;
+	  print_option (function e -> mcode print_string e) rp_opt;
+	  print_string " ";
+	  print_option (print_args false) args_opt;
+      | Ast0.TemplateInst(tn,lp,args,rp) ->
+	  expression tn; mcode print_string_box lp;
+	  let _ = dots (function _ -> ()) expression args in
+	  close_box(); mcode print_string rp
       | Ast0.TypeExp(ty) -> typeC ty
       | Ast0.Constructor(lp,ty,rp,init) ->
 	  mcode print_string_box lp; typeC ty; close_box();
@@ -269,6 +289,13 @@ let rec expression e =
 
 and expression_dots x = dots (function _ -> ()) expression x
 
+and print_args putspace (lp,args,rp) =
+  mcode print_string lp;
+  expression_dots args;
+  mcode print_string rp;
+  if putspace then
+    print_string " "
+
 and string_fragment e =
   match Ast0.unwrap e with
     Ast0.ConstantFragment(str) -> mcode print_string str
@@ -321,8 +348,13 @@ and typeC t =
   print_context t
     (function _ ->
       match Ast0.unwrap t with
-	Ast0.ConstVol(cv,ty) ->
-	  List.iter (function s -> mcode U.const_vol s) cv; print_string " "; typeC ty
+	Ast0.ConstVol(cvbefore,ty,cvafter) ->
+	  let do_cva = function
+	      Ast0.CV cv -> mcode U.const_vol cv
+	    | Ast0.Attr attr -> print_attribute attr; print_string " " in
+	  List.iter do_cva cvbefore;
+	  print_string " "; typeC ty;
+	  List.iter do_cva cvafter
       |	Ast0.BaseType(ty,strings) ->
 	  List.iter (function s -> mcode print_string s; print_string " ")
 	    strings
@@ -345,11 +377,20 @@ and typeC t =
 	  print_option (mcode print_string) comma;
 	  print_option expression precision_opt;
 	  mcode print_string rp
-      | Ast0.EnumName(kind,name) ->
+      | Ast0.EnumName(kind,key,name) ->
 	  mcode print_string kind;
+	  print_string " ";
+	  print_option (mcode U.structUnion) key;
 	  print_option (function x -> ident x; print_string " ") name
-      | Ast0.EnumDef(ty,lb,ids,rb) ->
-	  typeC ty; mcode print_string lb;
+      | Ast0.EnumDef(ty,base,lb,ids,rb) ->
+	  typeC ty;
+	  print_option
+	    (function (td, ty) ->
+	      mcode print_string td;
+	      print_string " ";
+	      typeC ty)
+	    base;
+	  mcode print_string lb;
 	  dots force_newline enum_decl ids;
 	  mcode print_string rb
       | Ast0.StructUnionName(kind,name) ->
@@ -367,12 +408,20 @@ and typeC t =
           mcode print_string tf;
 	  mcode print_string_box lp; typeC ty; close_box();
 	  mcode print_string rp
+      | Ast0.QualifiedType(ty,coloncolon,name) ->
+          print_option typeC ty;
+          mcode print_string coloncolon;
+          ident name;
       | Ast0.TypeName(name)-> mcode print_string name; print_string " "
       | Ast0.AutoType(auto) -> mcode print_string auto; print_string " "
       | Ast0.MetaType(name,_,_)-> mcode print_meta name; print_string " "
       | Ast0.DisjType(_,types,_,_) -> do_disj types typeC "|"
       | Ast0.ConjType(_,types,_,_) -> do_disj types typeC "&"
       | Ast0.OptType(ty) -> print_string "?"; typeC ty
+      | Ast0.TemplateType(tn,lp,args,rp) ->
+	  typeC tn; mcode print_string_box lp;
+	  let _ = dots (function _ -> ()) expression args in
+	  close_box(); mcode print_string rp
       | Ast0.AsType(ty,asty) -> typeC ty; print_string "@"; typeC asty)
 
 (* --------------------------------------------------------------------- *)
@@ -380,7 +429,7 @@ and typeC t =
 (* Even if the Cocci program specifies a list of declarations, they are
    split out into multiple declarations of a single variable each. *)
 
-and print_named_type ty midattr id =
+and print_named_type ty id =
   match Ast0.unwrap ty with
     Ast0.Array(ty,lb,size,rb) ->
       let rec loop ty k =
@@ -392,12 +441,17 @@ and print_named_type ty midattr id =
 		mcode print_string lb;
 		print_option expression size;
 		mcode print_string rb)
-        | _ -> typeC ty; print_attribute_list midattr ~befspace:false ~aftspace:true; ident id; k () in
+        | _ -> typeC ty; ident id; k () in
       loop ty (function _ -> ())
   | Ast0.ParenType(lp,ty,rp) ->
       print_parentype (lp,ty,rp) (function _ -> ident id)
-  | _ -> typeC ty; print_attribute_list midattr ~befspace:false ~aftspace:true; ident id
+  | _ -> typeC ty; ident id
 
+and alignas (Ast0.Align(align,lpar,expr,rpar)) =
+  mcode print_string align;
+  mcode print_string lpar;
+  expression expr;
+  mcode print_string rpar
 
 and declaration d =
   print_context d
@@ -405,42 +459,45 @@ and declaration d =
       match Ast0.unwrap d with
 	Ast0.MetaDecl(name,_,_) ->
 	  mcode print_meta name
-      |	Ast0.Init(stg,ty,midattr,id,endattr,eq,ini,sem) ->
+      |	Ast0.Init(al,stg,ty,id,endattr,eq,ini,sem) ->
+	  print_option alignas al;
 	  print_option (mcode U.storage) stg;
-	  print_named_type ty midattr id;
+	  print_named_type ty id;
           print_attribute_list endattr;
 	  print_string " ";
 	  mcode print_string eq; print_string " "; initialiser ini;
 	  mcode print_string sem
-      | Ast0.UnInit(stg,ty,midattr,id,endattr,sem) ->
-	  print_option (mcode U.storage) stg; print_named_type ty midattr id;
+      | Ast0.UnInit(al,stg,ty,id,endattr,sem) ->
+	  print_option alignas al;
+	  print_option (mcode U.storage) stg; print_named_type ty id;
           print_attribute_list endattr;
 	  mcode print_string sem
-      | Ast0.FunProto(fninfo,attr,name,lp1,params,va,rp1,sem) ->
+      | Ast0.FunProto(fninfo,name,lp1,params,va,rp1,sem) ->
 	  List.iter print_fninfo fninfo;
-          print_attribute_list attr ~befspace:false ~aftspace:true;
 	  ident name; mcode print_string_box lp1;
 	  parameter_list params; varargs va;
 	  close_box(); mcode print_string rp1;
 	  mcode print_string sem
-      | Ast0.MacroDecl(stg,name,lp,args,rp,attr,sem) ->
+      | Ast0.MacroDecl(stg,preattr,name,lp,args,rp,attr,sem) ->
 	  print_option (mcode U.storage) stg;
+          print_attribute_list preattr;
 	  ident name; mcode print_string_box lp;
 	  let _ = dots (function _ -> ()) expression args in
 	  close_box(); mcode print_string rp;
           print_attribute_list attr;
 	  mcode print_string sem
-      | Ast0.MacroDeclInit(stg,name,lp,args,rp,eq,ini,sem) ->
+      | Ast0.MacroDeclInit(stg,preattr,name,lp,args,rp,attr,eq,ini,sem) ->
 	  print_option (mcode U.storage) stg;
+          print_attribute_list preattr;
 	  ident name; mcode print_string_box lp;
 	  let _ = dots (function _ -> ()) expression args in
 	  close_box(); mcode print_string rp;
+          print_attribute_list attr;
           print_string " ";
           mcode print_string eq; print_string " "; initialiser ini;
 	  mcode print_string sem
-      | Ast0.TyDecl(ty,attr,sem) ->
+      | Ast0.TyDecl(ty,sem) ->
           typeC ty;
-          print_attribute_list attr;
           mcode print_string sem
       | Ast0.Typedef(stg,ty,id,sem) ->
 	  mcode print_string stg; typeC ty; typeC id;
@@ -465,17 +522,25 @@ and field d =
 	Ast0.MetaField(name,_,_)
       | Ast0.MetaFieldList(name,_,_,_) ->
 	  mcode print_meta name
-      | Ast0.Field(ty,id,bf,sem) ->
+      | Ast0.Field(ty,id,bf,endattr,sem) ->
 	  begin
 	    match id with
 	      None -> typeC ty
-            | Some id -> print_named_type ty [] id
+            | Some id -> print_named_type ty id
 	  end;
 	  let bitfield (c, e) =
 	    mcode print_string c;
 	    expression e in
 	  Common.do_option bitfield bf;
+          print_attribute_list endattr;
 	  mcode print_string sem
+      | Ast0.MacroDeclField(name,lp,args,rp,attr,sem) ->
+	  ident name; mcode print_string_box lp;
+	  let _ = dots (function _ -> ()) expression args in
+	  close_box(); mcode print_string rp;
+          print_attribute_list attr;
+	  mcode print_string sem
+      | Ast0.CppField(di) -> directive di
       | Ast0.DisjField(_,fields,_,_) ->
 	  do_disj fields field "|"
       | Ast0.ConjField(_,fields,_,_) ->
@@ -552,15 +617,11 @@ and parameterTypeDef p =
   print_context p
     (function _ ->
       match Ast0.unwrap p with
-        Ast0.VoidParam(ty,attr) ->
-          typeC ty;
+        Ast0.Param(ty,Some id,attr) ->
+          print_named_type ty id;
           print_attribute_list attr;
-      | Ast0.Param(ty,midattr,Some id,attr) ->
-          print_named_type ty midattr id;
-          print_attribute_list attr;
-      | Ast0.Param(ty,midattr,None,attr) ->
+      | Ast0.Param(ty,None,attr) ->
           typeC ty;
-          assert (midattr = []);
           print_attribute_list attr;
       | Ast0.MetaParam(name,_,_) -> mcode print_meta name
       | Ast0.MetaParamList(name,_,_,_) -> mcode print_meta name
@@ -570,7 +631,25 @@ and parameterTypeDef p =
       |	Ast0.AsParam(p,asexp) -> parameterTypeDef p; print_string "@";
 	  expression asexp)
 
+and templateParameterTypeDef p =
+  print_context p
+    (function _ ->
+      match Ast0.unwrap p with
+        Ast0.TypenameOrClassParam(tyorcl,id,eqtyopt) ->
+          mcode print_string tyorcl;
+          ident id;
+          print_option (fun (eq,ty) -> mcode print_string eq; typeC ty) eqtyopt
+      | Ast0.VarNameParam(ty,id,eqiniopt) ->
+          print_named_type ty id;
+          print_option (fun (eq,ini) -> mcode print_string eq; initialiser ini)
+	    eqiniopt
+      | Ast0.TPComma(cm) ->
+          mcode print_string cm
+      | Ast0.TPDots(dots) ->
+          mcode print_string dots)
+
 and parameter_list l = dots (function _ -> ()) parameterTypeDef l
+and template_parameter_list l = dots (function _ -> ()) templateParameterTypeDef l
 and varargs va = match va with
   | None -> ()
   | Some (comma, ellipsis) ->
@@ -591,6 +670,12 @@ and statement arity s =
 	  print_string arity; mcode print_string lbrace; start_block();
 	  dots force_newline (statement arity) body;
 	  end_block(); print_string arity; mcode print_string rbrace
+      | Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt) ->
+	  print_string arity;
+          mcode print_string_box tmpkw;
+          mcode print_string_box lab;
+	  template_parameter_list params; close_box(); mcode print_string rab;
+	  statement arity stmt
       | Ast0.Decl(_,decl) ->
 	  print_string arity; declaration decl
       | Ast0.Seq(lbrace,body,rbrace) ->
@@ -627,15 +712,21 @@ and statement arity s =
 	  mcode print_string whl; print_string " "; mcode print_string_box lp;
 	  expression exp; close_box(); mcode print_string rp;
 	  mcode print_string sem
-      | Ast0.For(fr,lp,first,e2,sem2,e3,rp,body,(info,aft,adj)) ->
+      | Ast0.For(fr,lp,first,rp,body,(info,aft,adj)) ->
 	  print_string arity;
 	  mcode print_string fr; mcode print_string_box lp;
 	  (match Ast0.unwrap first with
-	    Ast0.ForExp(e1,sem1) ->
-	      print_option expression e1; mcode print_string sem1
-	  | Ast0.ForDecl (_,decl) -> declaration decl);
-	  print_option expression e2; mcode print_string sem2;
-	  print_option expression e3; close_box();
+	    Ast0.ForExp(e1,sem1,e2,sem2,e3) ->
+	      print_option expression e1; mcode print_string sem1;
+	      print_option expression e2; mcode print_string sem2;
+	      print_option expression e3
+	  | Ast0.ForDecl (_,decl,e2,sem2,e3) ->
+	      declaration decl;
+	      print_option expression e2; mcode print_string sem2;
+	      print_option expression e3
+	  | Ast0.ForRange (_,decl,ini) ->
+	      declaration decl; print_string " "; initialiser ini);
+	  close_box();
 	  mcode print_string rp; print_string " "; statement arity body;
 	  mcode (function _ -> ()) ((),(),info,aft,ref [],adj)
       | Ast0.Iterator(nm,lp,args,rp,body,(info,aft,adj)) ->
@@ -718,10 +809,7 @@ and statement arity s =
 	  List.iter
 	    (whencode (dots force_newline (statement "")) (statement ""))
 	    whn
-      | Ast0.Include(inc,s) ->
-	  mcode print_string inc; print_string " "; mcode U.inc_file s
-      | Ast0.MetaInclude(inc,s) ->
-	  mcode print_string inc; print_string " "; expression s
+      | Ast0.CppTop(di) -> directive di
       | Ast0.Undef(def,id) ->
 	  mcode print_string def; print_string " "; ident id
       | Ast0.Define(def,id,params,body) ->
@@ -729,18 +817,10 @@ and statement arity s =
 	  print_define_parameters params;
 	  print_string " ";
 	  dots force_newline (statement arity) body
-      | Ast0.Pragma(prg,id,body) ->
-	  mcode print_string prg; print_string " "; ident id;
-	  print_string " "; pragmainfo body
       | Ast0.OptStm(re) -> statement "?" re
       | Ast0.AsStmt(stm,asstm) -> statement arity stm; print_string "@";
 	  statement arity asstm)
 
-and pragmainfo pi =
-  match Ast0.unwrap pi with
-      Ast0.PragmaString(s) -> mcode print_string s
-    | Ast0.PragmaDots(dots) -> mcode print_string dots
-
 and print_define_parameters params =
   match Ast0.unwrap params with
     Ast0.NoParams -> ()
@@ -751,6 +831,7 @@ and print_define_parameters params =
 and print_define_param param =
   match Ast0.unwrap param with
     Ast0.DParam(id) -> ident id
+  | Ast0.DParamEll(id,dots) -> ident id; mcode print_string dots
   | Ast0.MetaDParamList(name,_,_,_) -> mcode print_meta name
   | Ast0.DPComma(comma) -> mcode print_string comma
   | Ast0.DPdots(dots) -> mcode print_string dots
@@ -762,7 +843,6 @@ and print_fninfo = function
     Ast0.FStorage(stg) -> mcode U.storage stg
   | Ast0.FType(ty) -> typeC ty
   | Ast0.FInline(inline) -> mcode print_string inline
-  | Ast0.FAttr(attr) -> print_attribute attr
 
 and print_attribute_list ?(befspace=true) ?(aftspace=false) attrs =
   if befspace && not (attrs = []) then print_string " ";
@@ -772,12 +852,25 @@ and print_attribute_list ?(befspace=true) ?(aftspace=false) attrs =
 and print_attribute a =
   match Ast0.unwrap a with
     Ast0.Attribute(attr) -> print_attr_arg attr
-  | Ast0.GccAttribute(attr_,lp1,lp2,arg,rp1,rp2) ->
+  | Ast0.GccAttribute(attr_,lp1,lp2,args,rp1,rp2) ->
       mcode print_string attr_; print_string " ";
       mcode print_string_box lp1; mcode print_string_box lp2;
-      print_attr_arg arg; close_box();
-      mcode print_string_box rp1; close_box();
-      mcode print_string_box rp2;
+      let _ = dots (function _ -> ()) expression args in
+      close_box(); mcode print_string_box rp1;
+      close_box(); mcode print_string_box rp2
+  | Ast0.CxxAttribute(lb1,args,rb1,rb2) ->
+      mcode print_string_box lb1;
+      let _ = dots (function _ -> ()) expression args in
+      close_box();
+      mcode print_string_box rb1; mcode print_string_box rb2
+  | Ast0.CxxAttributeUsing(lb1,usng,atnm,dotdot,args,rb1,rb2) ->
+      mcode print_string_box lb1;
+      mcode print_string usng;
+      ident atnm;
+      mcode print_string dotdot;
+      let _ = dots (function _ -> ()) expression args in
+      close_box();
+      mcode print_string_box rb1; mcode print_string_box rb2
 
 and print_attr_arg a =
   match Ast0.unwrap a with
@@ -788,6 +881,38 @@ and print_attr_arg a =
       let _ = dots (function _ -> ()) expression args in
       close_box(); mcode print_string rp
 
+and pragmainfo pi =
+  match Ast0.unwrap pi with
+    Ast0.PragmaString(s) -> mcode print_string s
+  | Ast0.PragmaDots(dots) -> mcode print_string dots
+  | Ast0.MetaPragmaInfo(name,_,_) -> mcode print_meta name
+
+and directive d =
+  match Ast0.unwrap d with
+    Ast0.Include(inc,s) ->
+      mcode print_string inc; print_string " "; mcode U.inc_file s
+  | Ast0.MetaInclude(inc,s) ->
+      mcode print_string inc; print_string " "; expression s
+  | Ast0.Pragma(prg,id,body) ->
+      mcode print_string prg; print_string " "; ident id;
+      print_string " "; pragmainfo body
+  | Ast0.UsingNamespace(usng,nmspc,s,sem) ->
+      mcode print_string usng; print_string " ";
+      mcode print_string nmspc; print_string " ";
+      ident s; print_string " ";
+      mcode print_string sem
+  | Ast0.UsingTypename(usng,name,eq,tn,ty,sem) ->
+      mcode print_string usng; print_string " ";
+      ident name; print_string " ";
+      mcode print_string eq; print_string " ";
+      print_option (mcode print_string) tn; print_string " ";
+      typeC ty; print_string " ";
+      mcode print_string sem
+  | Ast0.UsingMember(usng,name,sem) ->
+      mcode print_string usng; print_string " ";
+      ident name; print_string " ";
+      mcode print_string sem
+
 and whencode notfn alwaysfn = function
     Ast0.WhenNot (_,_,a) ->
       print_string "   WHEN != "; open_box 0; notfn a; close_box()
@@ -857,6 +982,8 @@ let rec unparse_anything x =
       expression_dots d
   | Ast0.DotsParamTag(d) ->
       parameter_list d
+  | Ast0.DotsTemplateParamTag(d) ->
+      template_parameter_list d
   | Ast0.DotsInitTag(d) ->
       initialiser_list d
   | Ast0.DotsStmtTag(d) ->
@@ -877,6 +1004,7 @@ let rec unparse_anything x =
       print_string ("BinaryOp: " ^ (Ast0.string_of_binaryOp d)); force_newline();
   | Ast0.TypeCTag(d) -> typeC d
   | Ast0.ParamTag(d) -> parameterTypeDef d
+  | Ast0.TemplateParamTag(d) -> templateParameterTypeDef d
   | Ast0.InitTag(d)  -> initialiser d
   | Ast0.DeclTag(d)  -> declaration d
   | Ast0.FieldTag(d)  -> field d
@@ -887,9 +1015,16 @@ let rec unparse_anything x =
   | Ast0.ForInfoTag(fi)  ->
       print_string "ForInfo:"; force_newline();
       (match Ast0.unwrap fi with
-	Ast0.ForExp(e1,sem1) ->
-	  print_option expression e1; mcode print_string sem1
-      | Ast0.ForDecl (_,decl) -> declaration decl)
+	Ast0.ForExp(e1,sem1,e2,sem2,e3) ->
+	  print_option expression e1; mcode print_string sem1;
+	  print_option expression e2; mcode print_string sem2;
+	  print_option expression e3
+      | Ast0.ForDecl (_,decl,e2,sem2,e3) ->
+	  declaration decl;
+	  print_option expression e2; mcode print_string sem2;
+	  print_option expression e3
+      | Ast0.ForRange (_,decl,ini) ->
+	  declaration decl; initialiser ini)
   | Ast0.CaseLineTag(d)  -> case_line "" d
   | Ast0.StringFragmentTag(d)  -> string_fragment d
   | Ast0.AttributeTag(d) -> print_attribute d
