@@ -186,7 +186,7 @@ let add_file_to_score score res correct diffxs =
 (* If extra test is provided, then all failing tests with the standard
    comparison are considered ok, and only the correct result are subjected to
    the extra test *)
-let testall_bis extra_test expected_score_file update_score_file =
+let testall_bis setup extra_test expected_score_file update_score_file =
 
   let score  = empty_score () in
 
@@ -204,6 +204,8 @@ let testall_bis extra_test expected_score_file update_score_file =
 	then matched1 res
 	else raise (Impossible 164) in
       let base = if x =~ "\\(.*\\)_ver[0-9]+" then matched1 x else x in
+      let cocci_file = "tests/" ^ base ^ ".cocci" in
+      setup cocci_file;
       let cfile      = "tests/" ^ x ^ ".c" in
       let cfile      =
         let cppfile = "tests/" ^ x ^ ".cpp" in
@@ -218,7 +220,6 @@ let testall_bis extra_test expected_score_file update_score_file =
             Flag.c_plus_plus := Flag.Off;
             cfile
           end in
-      let cocci_file = "tests/" ^ base ^ ".cocci" in
       let expected = "tests/" ^ res in
       let out = base ^ out_suffix in
       let expected_out = "tests/" ^ out in
@@ -227,7 +228,6 @@ let testall_bis extra_test expected_score_file update_score_file =
 
       try (
         Common.timeout_function "testing" timeout_testall  (fun () ->
-
 	  pr2 res;
 
 	  let (xs, current_out) =
@@ -269,7 +269,13 @@ let testall_bis extra_test expected_score_file update_score_file =
       )
       with exn ->
         Common.reset_pr_indent();
+	Iteration.reset();
+	Flag.reset_virt();
         let s = "PROBLEM\n" ^ ("   exn = " ^ Printexc.to_string exn ^ "\n") in
+	(* clean up state *)
+	Flag.defined_virtual_rules := [];
+	Flag.defined_virtual_env := [];
+	Iteration.clear_pending_instance();
         Hashtbl.add score res (Common.Pb s)
     );
 
@@ -366,8 +372,8 @@ let testall_bis extra_test expected_score_file update_score_file =
 
   end
 
-let testall = testall_bis None
-let test_spacing = testall_bis (Some Compare_c.exact_compare)
+let testall setup = testall_bis setup None
+let test_spacing setup = testall_bis setup (Some Compare_c.exact_compare)
 
 (* ------------------------------------------------------------------------ *)
 
@@ -611,7 +617,8 @@ let test_parse_cocci file =
       Prepare_ocamlcocci.test ());
   flush stdout; flush stderr;
   match (!Flag.scanner,query) with
-    ((Flag.CocciGrep|Flag.NoScanner|Flag.GitGrep),(query,_,_,_)) ->
+    ((Flag.CocciGrep|Flag.NoScanner|Flag.GitGrep
+     |Flag.PatchDiff|Flag.PatchDiffRange _),(query,_,_,_)) ->
       (match query with
 	None -> pr "No grep query"
       | Some x ->
