@@ -11,8 +11,8 @@ open Common
 (*****************************************************************************)
 val testone : string (*prefix*) -> string (*test*) ->
   string option (*compare_expected*) -> unit
-val testall : string -> bool -> unit
-val test_spacing : string -> bool -> unit
+val testall : (string -> unit) -> string -> bool -> unit
+val test_spacing : (string -> unit) -> string -> bool -> unit
 
 (*****************************************************************************)
 (* works with tests-big/. The .res, .ok, .spatch_ok, .failed, .var *)
