@@ -183,8 +183,8 @@ let rec type_pos t snp
         "in position generator."
       )
   | Ast0.ConjType(lp,tlist,pipelist,rp) -> None (* not sure *)
-  | Ast0.ConstVol(const,t) ->
-      let constructor ~item = Ast0.ConstVol(const,item) in
+  | Ast0.ConstVol(cv1,t,cv2) ->
+      let constructor ~item = Ast0.ConstVol(cv1,item,cv2) in
       item_wrap ~item:t ~item_posfn:type_pos ~constructor snp
   | Ast0.Signed(sign,t) ->
       let constructor ~mc = Ast0.Signed(mc,t) in
@@ -204,14 +204,14 @@ let rec type_pos t snp
   | Ast0.Decimal(dec,lp,expr,comma,expr2,rp) ->
       let constructor ~mc = Ast0.Decimal(dec,lp,expr,comma,expr2,mc) in
       mcode_wrap ~mc:rp ~constructor snp
-  | Ast0.EnumName(enum,Some nm) ->
-      let constructor ~id = Ast0.EnumName(enum,Some id) in
+  | Ast0.EnumName(enum, key, Some nm) ->
+      let constructor ~id = Ast0.EnumName(enum, key, Some id) in
       id_wrap ~id:nm ~constructor snp
-  | Ast0.EnumName(enum,None) ->
-      let constructor ~mc = Ast0.EnumName(mc,None) in
+  | Ast0.EnumName(enum, key, None) ->
+      let constructor ~mc = Ast0.EnumName(mc, key, None) in
       mcode_wrap ~mc:enum ~constructor snp
-  | Ast0.EnumDef(t,lcb,exprdots,rcb) ->
-      let c ~item ~mc = Ast0.EnumDef(item,lcb,exprdots,mc) in
+  | Ast0.EnumDef(t,base,lcb,exprdots,rcb) ->
+      let c ~item ~mc = Ast0.EnumDef(item,base,lcb,exprdots,mc) in
       let alt() = mcode_wrap ~mc:rcb ~constructor:(c ~item:t) snp in
       item_wrap ~item:t ~item_posfn:type_pos ~constructor:(c ~mc:rcb) ~alt snp
   | Ast0.StructUnionName(sumc,Some nm) ->
@@ -234,6 +234,13 @@ let rec type_pos t snp
   | Ast0.TypeName(nm) ->
       let constructor ~mc = Ast0.TypeName(mc) in
       mcode_wrap ~mc:nm ~constructor snp
+  | Ast0.QualifiedType(Some typec,coloncolon,name) ->
+      let _ = type_pos typec snp in
+      let constructor ~id = Ast0.QualifiedType(Some typec, coloncolon, id) in
+      id_wrap ~id:name ~constructor snp
+  | Ast0.QualifiedType(None,coloncolon,name) ->
+      let constructor ~id = Ast0.QualifiedType(None, coloncolon, id) in
+      id_wrap ~id:name ~constructor snp
   | Ast0.AutoType(auto) ->
       let constructor ~mc = Ast0.AutoType(mc) in
       mcode_wrap ~mc:auto ~constructor snp
@@ -243,6 +250,9 @@ let rec type_pos t snp
   | Ast0.OptType(t) ->
       let constructor ~item = Ast0.OptType(item) in
       item_wrap ~item:t ~item_posfn:type_pos ~constructor snp
+  | Ast0.TemplateType(name, lab, dots, rab) ->
+      let constructor ~mc = Ast0.TemplateType(name, mc, dots, rab) in
+      mcode_wrap ~mc:lab ~constructor snp
 
 (* NB: if implementing disj generation, make sure that the statement dots in
  * the clist are generated in no_gen mode...
@@ -347,10 +357,10 @@ let rec expression_pos exp snp
       let c ~exp ~mc = Ast0.Paren (mc, exp, rp) in
       let alt() = mcode_wrap ~mc:lp ~constructor:(c ~exp) snp in
       exp_wrap ~exp ~constructor:(c ~mc:lp) ~alt snp
-  | Ast0.ArrayAccess(exp1, lb, exp2, rb) ->
-      let c ~exp1 ~exp2 ~mc = Ast0.ArrayAccess(exp1, mc, exp2, rb) in
-      let alt() = mcode_wrap ~mc:lb ~constructor:(c ~exp1 ~exp2) snp in
-      exp_wrap2 ~exp1 ~exp2 ~constructor:(c ~mc:lb) ~alt snp
+  | Ast0.ArrayAccess(exp, lb, expdots, rb) ->
+      let c ~exp ~mc = Ast0.ArrayAccess(exp, mc, expdots, rb) in
+      let alt() = mcode_wrap ~mc:lb ~constructor:(c ~exp) snp in
+      exp_wrap ~exp ~constructor:(c ~mc:lb) ~alt snp
   | Ast0.RecordAccess(exp, stop, id) ->
       let c ~exp ~id = Ast0.RecordAccess(exp, stop, id) in
       let alt() = id_wrap ~id ~constructor:(c ~exp) snp in
@@ -359,9 +369,16 @@ let rec expression_pos exp snp
       let c ~exp ~id = Ast0.RecordPtAccess(exp, arrow, id) in
       let alt() = id_wrap ~id ~constructor:(c ~exp) snp in
       exp_wrap ~exp ~constructor:(c ~id) ~alt snp
-  | Ast0.Cast(lp, typec, attr, rp, exp) ->
+  | Ast0.QualifiedAccess(Some typec, coloncolon, id) ->
+      let _ = type_pos typec snp in
+      let constructor ~id = Ast0.QualifiedAccess(Some typec, coloncolon, id) in
+      id_wrap ~id:id ~constructor snp
+  | Ast0.QualifiedAccess(None, coloncolon, id) ->
+      let constructor ~id = Ast0.QualifiedAccess(None, coloncolon, id) in
+      id_wrap ~id:id ~constructor snp
+  | Ast0.Cast(lp, typec, rp, exp) ->
       let _ = type_pos typec snp in (* sanity check for disj *)
-      let c ~exp ~mc = Ast0.Cast(lp, typec, attr, mc, exp) in
+      let c ~exp ~mc = Ast0.Cast(lp, typec, mc, exp) in
       let alt() = mcode_wrap ~mc:rp ~constructor:(c ~exp) snp in
       exp_wrap ~exp ~constructor:(c ~mc:rp) ~alt snp
   | Ast0.SizeOfExpr(sizeofmc, exp) ->
@@ -371,6 +388,18 @@ let rec expression_pos exp snp
       let _ = type_pos typec snp in (* sanity check for disj *)
       let constructor ~mc = Ast0.SizeOfType(mc, lp, typec, rp) in
       mcode_wrap ~mc:sizeofmc ~constructor snp
+  | Ast0.New(nw,pp_opt,lp_opt,ty,rp_opt,args_opt) ->
+      let constructor ~mc = Ast0.New(mc,pp_opt,lp_opt,ty,rp_opt,args_opt) in
+      mcode_wrap ~mc:nw ~constructor snp
+  | Ast0.Delete(deletemc, exp) ->
+      let constructor ~mc = Ast0.Delete(mc, exp) in
+      mcode_wrap ~mc:deletemc ~constructor snp
+  | Ast0.DeleteArr(deletemc,lb,rb,exp) ->
+      let constructor ~mc = Ast0.DeleteArr(mc,lb,rb,exp) in
+      mcode_wrap ~mc:deletemc ~constructor snp
+  | Ast0.TemplateInst(name, lab, dots, rab) ->
+      let constructor ~mc = Ast0.TemplateInst(name, mc, dots, rab) in
+      mcode_wrap ~mc:lab ~constructor snp
   | Ast0.TypeExp(typec) ->
       let constructor ~item = Ast0.TypeExp(item) in
       item_wrap ~item:typec ~item_posfn:type_pos ~constructor snp
@@ -404,32 +433,32 @@ let rec declaration_pos decl snp
   | Ast0.MetaDecl _
   | Ast0.AsDecl _ ->
       None
-  | Ast0.Init(st, ty, midattr, id, endattr, eq, ini, sem) ->
+  | Ast0.Init(al, st, ty, id, endattr, eq, ini, sem) ->
       let _ = type_pos ty snp in (* sanity check *)
-      let constructor ~id = Ast0.Init(st, ty, midattr, id, endattr, eq, ini, sem) in
+      let constructor ~id = Ast0.Init(al, st, ty, id, endattr, eq, ini, sem) in
       id_wrap ~id ~constructor snp
-  | Ast0.UnInit(st, ty, midattr, id, endattr, sem) ->
+  | Ast0.UnInit(al, st, ty, id, endattr, sem) ->
       let _ = type_pos ty snp in (* sanity check *)
-      let constructor ~id = Ast0.UnInit(st, ty, midattr, id, endattr, sem) in
+      let constructor ~id = Ast0.UnInit(al, st, ty, id, endattr, sem) in
       id_wrap ~id ~constructor snp
-  | Ast0.TyDecl (t, attr, sem) ->
-      let c ~item ~mc = Ast0.TyDecl(item, attr, mc) in
+  | Ast0.TyDecl (t, sem) ->
+      let c ~item ~mc = Ast0.TyDecl(item, mc) in
       let alt() = mcode_wrap ~mc:sem ~constructor:(c ~item:t) snp in
       item_wrap ~item:t ~item_posfn:type_pos ~constructor:(c ~mc:sem) ~alt snp
   | Ast0.Typedef (tm, tc, tc2, sem) ->
       let constructor ~mc = Ast0.Typedef (mc, tc, tc2, sem) in
       mcode_wrap ~mc:tm ~constructor snp
-  | Ast0.MacroDecl (st,id,lp,ed,rp,attr,sem) ->
-      let constructor ~id = Ast0.MacroDecl (st, id, lp, ed, rp, attr, sem) in
+  | Ast0.MacroDecl (st,pa,id,lp,ed,rp,attr,sem) ->
+      let constructor ~id = Ast0.MacroDecl (st, pa, id, lp, ed, rp, attr, sem) in
       id_wrap ~id ~constructor snp
-  | Ast0.MacroDeclInit (st,id,lp,ed,rp,eq,init,sem) ->
-      let constructor ~id = Ast0.MacroDeclInit (st,id,lp,ed,rp,eq,init,sem) in
+  | Ast0.MacroDeclInit (st,pa,id,lp,ed,rp,attr,eq,init,sem) ->
+      let constructor ~id = Ast0.MacroDeclInit (st,pa,id,lp,ed,rp,attr,eq,init,sem) in
       id_wrap ~id ~constructor snp
   | Ast0.OptDecl(dec) ->
       let constructor ~item = Ast0.OptDecl item in
       item_wrap ~item:dec ~item_posfn:declaration_pos ~constructor snp
-  | Ast0.FunProto(fninfo,attr,id,lp1,params,va,rp1,sem) ->
-      let constructor ~id = Ast0.FunProto(fninfo,attr,id,lp1,params,va,rp1,sem) in
+  | Ast0.FunProto(fninfo,id,lp1,params,va,rp1,sem) ->
+      let constructor ~id = Ast0.FunProto(fninfo,id,lp1,params,va,rp1,sem) in
       id_wrap ~id ~constructor snp
 
 let rec field_pos decl snp
@@ -440,27 +469,58 @@ let rec field_pos decl snp
   | Ast0.Fdots _
   | Ast0.MetaField _
   | Ast0.MetaFieldList _
-  | Ast0.Field(_, None, _, _) -> None
-  | Ast0.Field(ty, Some id, bf, sem) ->
+  | Ast0.Field(_, None, _, _, _) -> None
+  | Ast0.Field(ty, Some id, bf, endattr, sem) ->
       let _ = type_pos ty snp in (* sanity check *)
-      let constructor ~id = Ast0.Field(ty, Some id, bf, sem) in
+      let constructor ~id = Ast0.Field(ty, Some id, bf, endattr, sem) in
       id_wrap ~id ~constructor snp
+  | Ast0.MacroDeclField (id,lp,ed,rp,attr,sem) ->
+      let constructor ~id = Ast0.MacroDeclField (id, lp, ed, rp, attr, sem) in
+      id_wrap ~id ~constructor snp
+  | Ast0.CppField di ->
+      let constructor ~item = Ast0.CppField item in
+      item_wrap ~item:di ~item_posfn:directive_pos ~constructor snp
   | Ast0.OptField(dec) ->
       let constructor ~item = Ast0.OptField item in
       item_wrap ~item:dec ~item_posfn:field_pos ~constructor snp
 
+and directive_pos di snp
+: (Ast0.base_directive Ast0.wrap * Snap.t) option =
+  match Ast0.unwrap di with
+    Ast0.Include (incmc,filemc) ->
+      let constructor ~mc = Ast0.Include(incmc, mc) in
+      mcode_wrap ~mc:filemc ~constructor snp
+  | Ast0.MetaInclude (incmc,filemc) ->
+      let constructor ~mc = Ast0.MetaInclude(mc, filemc) in
+      mcode_wrap ~mc:incmc ~constructor snp
+  | Ast0.Pragma (pragmc, id, praginfo) ->
+      let constructor ~id = Ast0.Pragma(pragmc, id, praginfo) in
+      id_wrap ~id ~constructor snp
+  | Ast0.UsingNamespace (usngmc,nmspcmc,namemc,semmc) ->
+      let constructor ~mc = Ast0.UsingNamespace(usngmc, mc, namemc, semmc) in
+      mcode_wrap ~mc:nmspcmc ~constructor snp
+  | Ast0.UsingTypename (usngmc,namemc,eqmc,tnmc,tymc,semmc) ->
+      let constructor ~mc = Ast0.UsingTypename(usngmc, namemc, mc, tnmc, tymc, semmc) in
+      mcode_wrap ~mc:eqmc ~constructor snp
+  | Ast0.UsingMember(usngmc,namemc,semmc) ->
+      let constructor ~mc = Ast0.UsingMember(usngmc, namemc, mc) in
+      mcode_wrap ~mc:semmc ~constructor snp
+
 let forinfo_pos f snp
 : (Ast0.base_forinfo Ast0.wrap * Snap.t) option =
   match Ast0.unwrap f with
-  | Ast0.ForExp (Some exp, sem) ->
-      let c ~exp ~mc = Ast0.ForExp(Some exp, mc) in
-      let alt() = mcode_wrap ~mc:sem ~constructor:(c ~exp) snp in
-      exp_wrap ~exp ~constructor:(c ~mc:sem) ~alt snp
-  | Ast0.ForExp (None, sem) ->
-      let constructor ~mc = Ast0.ForExp (None, mc) in
-      mcode_wrap ~mc:sem ~constructor snp
-  | Ast0.ForDecl (bef, decl) ->
-      let constructor ~item = Ast0.ForDecl(bef, item) in
+  | Ast0.ForExp (Some exp,sem1,expo1,sem2,expo2) ->
+      let c ~exp ~mc = Ast0.ForExp(Some exp,mc,expo1,sem2,expo2) in
+      let alt() = mcode_wrap ~mc:sem1 ~constructor:(c ~exp) snp in
+      exp_wrap ~exp ~constructor:(c ~mc:sem1) ~alt snp
+  | Ast0.ForExp (None,sem1,expo1,sem2,expo2) ->
+      let constructor ~mc = Ast0.ForExp (None,mc,expo1,sem2,expo2) in
+      mcode_wrap ~mc:sem1 ~constructor snp
+  | Ast0.ForDecl (bef,decl,expo1,sem,expo2) ->
+      let constructor ~item = Ast0.ForDecl(bef,item,expo1,sem,expo2) in
+      item_wrap ~item:decl ~item_posfn:declaration_pos ~constructor snp
+  | Ast0.ForRange (bef,decl,inio) ->
+      let constructor ~item = Ast0.ForRange(bef,item,inio) in
       item_wrap ~item:decl ~item_posfn:declaration_pos ~constructor snp
 
 let rec statement_pos s snp
@@ -483,12 +543,10 @@ let rec statement_pos s snp
   | Ast0.Ty _ -> None
   | Ast0.TopInit _ -> None
 
-  | Ast0.Include (incmc,filemc) ->
-      let constructor ~mc = Ast0.Include(incmc, mc) in
-      mcode_wrap ~mc:filemc ~constructor snp
-  | Ast0.MetaInclude (incmc,filemc) ->
-      let constructor ~mc = Ast0.MetaInclude(mc, filemc) in
-      mcode_wrap ~mc:incmc ~constructor snp
+  | Ast0.CppTop di -> 
+      let c ~item = Ast0.CppTop item in
+      item_wrap ~item:di ~item_posfn:directive_pos ~constructor:c snp
+
   | Ast0.Undef (defmc, id) ->
       let constructor ~id = Ast0.Undef(defmc, id) in
       id_wrap ~id ~constructor snp
@@ -506,9 +564,6 @@ let rec statement_pos s snp
        let defparam = Ast0.wrap (Ast0.DParams(lp, defp, newrp)) in
        wrap (c ~id ~defparam) snp
     )
-  | Ast0.Pragma (pragmc, id, praginfo) ->
-      let constructor ~id = Ast0.Pragma(pragmc, id, praginfo) in
-      id_wrap ~id ~constructor snp
   | Ast0.OptStm stm ->
       let c ~item = Ast0.OptStm item in
       item_wrap ~item:stm ~item_posfn:statement_pos ~constructor:c snp
@@ -543,8 +598,8 @@ let rec statement_pos s snp
       let c ~exp ~mc = Ast0.Do(mc, s, whmc, l, exp, r, sem) in
       let alt() = mcode_wrap ~mc:d ~constructor:(c ~exp) snp in
       exp_wrap ~exp ~constructor:(c ~mc:d) ~alt snp
-  | Ast0.For (fo,lp, fi,expo1,sem,expo2,rp,stmt,a) ->
-      let c ~item ~mc = Ast0.For (mc,lp,item,expo1,sem,expo2,rp,stmt,a) in
+  | Ast0.For (fo,lp, fi,rp,stmt,a) ->
+      let c ~item ~mc = Ast0.For (mc,lp,item,rp,stmt,a) in
       let alt() = mcode_wrap ~mc:fo ~constructor:(c ~item:fi) snp in
       item_wrap
         ~item:fi ~item_posfn:forinfo_pos ~constructor:(c ~mc:fo) ~alt snp
@@ -578,3 +633,6 @@ let rec statement_pos s snp
   | Ast0.FunDecl (b, f, id, lp, ps, op, rp, ea, lb, sd, rb, a) ->
       let constructor ~id = Ast0.FunDecl(b,f,id,lp,ps,op,rp,ea,lb,sd,rb,a) in
       id_wrap ~id ~constructor snp
+  | Ast0.TemplateDefinition(tmpkw,lab,params,rab,stmt) ->
+      let constructor ~mc = Ast0.TemplateDefinition(mc,lab,params,rab,stmt) in
+      mcode_wrap ~mc:tmpkw ~constructor snp
